#include "TableEditor3D.h"

#include <wx/dcbuffer.h>
#include <wx/filename.h>
#include <wx/arrimpl.cpp> // this is a magic incantation which must be done!
#include <algorithm>


#include "TableEditor3DConfigHelper.h"
#include "TableEditor3DSQLHelper.h"

#include "../Math.h"
#include "Basic3DMath.h"

#include "../DrawHelper.h"

#include "../vemsgauge/GaugeGroupManager.h"
#include "../gaugedesc/GaugeProperties.h"

#include "../hotkeys/wxHotKeyManager.h"
#include "../common/VemsMatrixHelper.h"

#include "../WindowHelper.h"
#include "../agg/wxAgg.h"

#include "../logger.h"
#include "../wxPointEx.h"
#include "../wxVariantHelper.h"

#include "../vemsgauge/TableEditorPanel.h"
#include "../bcWidgets/bcLabel.h"

#include "../vemsgauge/FocusParentSinkerBase.h"


#define PI 3.14159265
#define SURFACE_HEIGHT_RATE 0.9

#define MAX_CAMERA_EYE_X 0.5
#define MAX_CAMERA_EYE_Y 1.0
#define CAMERA_STEP 0.1

#define ptrdel(ptr) \
    if( (ptr) != NULL ) { delete (ptr); (ptr) = NULL; }

WX_DEFINE_OBJARRAY(ArrayOfColours);

/// Event ids of TableEditor3D popup context-sensitive menu
enum
{
    TABLE_EDITOR_3D_PM_INC_CELL_LARGE = 2000,
    TABLE_EDITOR_3D_PM_DEC_CELL_LARGE,
    TABLE_EDITOR_3D_PM_INC_CELL_SMALL,
    TABLE_EDITOR_3D_PM_DEC_CELL_SMALL,
    TABLE_EDITOR_3D_PM_ROTATE_LEFT,	                    // Rotate table left
    TABLE_EDITOR_3D_PM_ROTATE_RIGHT,	                // Rotate table right
    TABLE_EDITOR_3D_PM_ROTATE_UP,	                    // Rotate table up
    TABLE_EDITOR_3D_PM_ROTATE_DOWN, 	                // Rotate table down
    TABLE_EDITOR_3D_PM_INTERPOLATE, 	                // Interpolate selected
    TABLE_EDITOR_3D_PM_SET_Y_RANGE_MIN,                 // set y-axis min value
    TABLE_EDITOR_3D_PM_SET_Y_RANGE_MAX,                 // set y-axis max value
    TABLE_EDITOR_3D_PM_IMPORT,
    TABLE_EDITOR_3D_PM_REMAP,

};



BEGIN_EVENT_TABLE( TableEditor3D, wxControl )
    EVT_MENU_RANGE( TABLE_EDITOR_3D_PM_INC_CELL_LARGE, TABLE_EDITOR_3D_PM_REMAP, TableEditor3D::OnPopupMenu )
    EVT_SIZE ( TableEditor3D::OnSize )
    EVT_PAINT( TableEditor3D::OnPaint )
    EVT_KEY_DOWN( TableEditor3D::OnKeyDown )
    EVT_RIGHT_DOWN( TableEditor3D::OnMouseRightClick )


    EVT_LEFT_DOWN( TableEditor3D::OnMouseLeftClick )
    EVT_LEFT_UP(TableEditor3D::OnMouseLeftUp)
	EVT_MOTION(TableEditor3D::OnMouseMove)


	EVT_SET_FOCUS(TableEditor3D::OnSetFocus)
	EVT_KILL_FOCUS(TableEditor3D::OnKillFocus)


END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS(TableEditor3D, wxControl);

TableEditor3D::TableEditor3D( wxWindow* parent, GaugeGroupManager * ggm, GaugeProperties &prop)
    : GroupMember( parent->GetParent(), prop, ggm, parent->GetParent()->GetId() )
{
    m_areLabelsReady = false;

    CreateBcParent(this);

    m_delta = wxDefaultPosition;

    m_helpPanelBmp = NULL;
#if defined(__WXMSW__)
    SetBackgroundStyle( wxBG_STYLE_CUSTOM );
#elif defined(__UNIX__)
    SetBackgroundStyle( wxBG_STYLE_SYSTEM );
#endif
    m_tableName = prop.symbolName;
    Create( parent, wxID_ANY );

    m_hotkeyMgr = new wxHotKeyManager(this);

    setDefault();
    loadParams();
    m_pSurface = new Surface3D();
    m_pBackBuffer = NULL;
    setNeedUpdateBackBuffer();
    setRangeY();

    m_isResized = false;

    m_padding = 5;
    m_labelTextPadding = 2;
    m_maxLabelWidth = 0;

    m_rowNumberDigits = 0;
    m_colNumberDigits = 0;
    m_tableNumberDigits = 0;

    m_isAutoFollowLoad = false;

//    initLabels();
}

TableEditor3D::~TableEditor3D()
{
    saveParams();

    delete m_pBackBuffer;
    m_pBackBuffer = NULL;

    delete m_pSurface;
    m_pSurface = NULL;

//    delete m_selectedTextDesc;
//    delete m_markedTextDesc;
}

void TableEditor3D::setRowNumberDigits( size_t digits )
{
    m_rowNumberDigits = digits;
    m_rowNumberFormat = wxString::Format( _T("%%5.%df"), m_rowNumberDigits ) ;
}

void TableEditor3D::setColNumberDigits( size_t digits )
{
    m_colNumberDigits = digits;
    m_colNumberFormat = wxString::Format( _T("%%5.%df"), m_colNumberDigits ) ;;
}

void TableEditor3D::setTableNumberDigits( size_t digits )
{
    m_tableNumberDigits = digits;
    m_tableNumberFormat = wxString::Format( _T("%%5.%df"), m_tableNumberDigits ) ;;;
}

void TableEditor3D::initLabels()
{
    /// selected label
    m_selectedLabel = new bcLabel( this );

    m_selectedLabel->SetProperty( BCBS_BACKGROUND_COLOR, wxColor( 0, 0, 0, 0) );

    m_selectedLabel->SetProperty( BCBS_BORDER_COLOR, wxColor( 255, 0, 0, 255 ) );
    m_selectedLabel->SetProperty( BCBS_BORDER_WIDTH, 1.0 );
    m_selectedLabel->SetPosition( 5.0, 5.0 );
    m_selectedLabel->SetProperty( BCWB_ISSHOWN, false );
    m_selectedLabel->SetProperty( BCBS_PADDDING, (double)m_labelTextPadding );

    m_selectedTextDesc = new TextDesc();
    m_selectedTextDesc->Color      = wxColor( 255, 0, 0, 255 );
    m_selectedTextDesc->Height     = m_labelFontHeight;
    m_selectedTextDesc->Fontfile   = _T("VeraMono.ttf");

    /// marked label
    m_markedLabel = new bcLabel( this );

    m_markedLabel->SetProperty( BCBS_BACKGROUND_COLOR, wxColor( 0, 0, 0, 0) );

    m_markedLabel->SetProperty( BCBS_BORDER_COLOR, wxColor( 0, 255, 0, 255 ) );
    m_markedLabel->SetProperty( BCBS_BORDER_WIDTH, 1.0 );
    m_markedLabel->SetPosition( 5.0, 5.0 );
    m_markedLabel->SetProperty( BCWB_ISSHOWN, false );
    m_markedLabel->SetProperty( BCBS_PADDDING, (double)m_labelTextPadding );

    m_markedTextDesc =  new TextDesc();
    m_markedTextDesc->Color      = wxColor( 0, 255, 0, 255 );
    m_markedTextDesc->Height     = m_labelFontHeight;
    m_markedTextDesc->Fontfile   = _T("VeraMono.ttf");

    /// marked load
    m_markedLoadLabel = new bcLabel( this );
    m_markedLoadLabel->SetProperty( BCBS_BACKGROUND_COLOR, wxColor( 0, 0, 0, 0) );

    m_markedLoadLabel->SetProperty( BCBS_BORDER_COLOR, wxColor( 0, 255, 0, 255 ) );
    m_markedLoadLabel->SetProperty( BCBS_BORDER_WIDTH, 1.0 );
    m_markedLoadLabel->SetPosition( 5.0, 5.0 );
    m_markedLoadLabel->SetProperty( BCWB_ISSHOWN, false );
    m_markedLoadLabel->SetProperty( BCBS_PADDDING, (double)m_labelTextPadding );

    m_markedLoadTextDesc = new TextDesc();
    m_markedLoadTextDesc->Color      = wxColor( 255, 255, 255, 255 );
    m_markedLoadTextDesc->Height     = m_labelFontHeight;
    m_markedLoadTextDesc->Fontfile   = _T("VeraMono.ttf");

    m_areLabelsReady = true;

}

void TableEditor3D::drawLabels()
{
//    drawSelectedLabel();
//    drawMarkedLabel();
//
//    if ( m_bUseTargetMarker ) {
//        drawLoadLabel();
//    }
}

void TableEditor3D::drawSelectedLabel()
{
    double xValue = m_columnHeaders[ m_selectedCol ];
    double yValue = m_rowHeaders[ m_rows - m_selectedRow - 1 ];
    double zValue = m_table[ index( m_rows - m_selectedRow - 1, m_selectedCol ) ];

    wxString numberFormat = _T("[") + m_colNumberFormat + _T(":") + m_rowNumberFormat + _T("] = ") + m_tableNumberFormat;
    wxString text = wxString::Format( numberFormat, xValue, yValue, zValue );
    m_selectedTextDesc->Text = text;


    m_selectedLabel->WriteLabel( *m_selectedTextDesc );
    m_selectedLabel->SetProperty( BCWB_ISSHOWN, true );
}

void TableEditor3D::drawMarkedLabel()
{
    wxString numberFormat = _T("[") + m_colNumberFormat + _T(":") + m_rowNumberFormat + _T("] = ") + m_tableNumberFormat;
    wxString text = wxString::Format( numberFormat, m_markedX, m_markedY, m_markedPoint.y );

    m_markedTextDesc->Text       = text;

    m_markedLabel->WriteLabel( *m_markedTextDesc );
    m_markedLabel->SetProperty( BCWB_ISSHOWN, true );
}

void TableEditor3D::drawLoadLabel()
{
    wxString text = wxString::Format( m_tableNumberFormat, m_markedZ );

    m_markedLoadTextDesc->Text = text;

    m_markedLoadLabel->WriteLabel( *m_markedLoadTextDesc );
    m_markedLoadLabel->SetProperty( BCWB_ISSHOWN, true );
}

void TableEditor3D::calculateMaxLabelWidth()
{
    wxString numberFormat = _T("[") + m_colNumberFormat + _T(":") + m_rowNumberFormat + _T("] = ") + m_tableNumberFormat;
    wxString text = wxString::Format( numberFormat, 0.0, 0.0, 0.0 );

    m_selectedTextDesc->Text = text;
    wxRect selectedRect = m_selectedLabel->GetTextBox( *m_selectedTextDesc );
    selectedRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    m_markedTextDesc->Text = text;
    wxRect markedRect = m_markedLabel->GetTextBox( *m_markedTextDesc );
    markedRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    text = wxString::Format( m_tableNumberFormat, 0.0 );
    m_markedLoadTextDesc->Text = text;
    wxRect loadRect = m_markedLoadLabel->GetTextBox( *m_markedLoadTextDesc );
    loadRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    int selectedWidth = selectedRect.GetWidth();
    int markedWidth = markedRect.GetWidth();

    if ( m_bUseTargetMarker ) {
        markedWidth += loadRect.GetWidth() + m_padding;
    }

    m_maxLabelWidth = wxMax( selectedWidth, markedWidth );
}

void TableEditor3D::alignAllLabels()
{
    wxRect selectedRect = m_selectedLabel->GetTextBox( *m_selectedTextDesc );
    wxRect markedRect   = m_markedLabel->GetTextBox( *m_markedTextDesc );
    wxRect loadRect     = m_markedLoadLabel->GetTextBox( *m_markedLoadTextDesc );

    selectedRect.Inflate( m_labelTextPadding, m_labelTextPadding );
    markedRect.Inflate( m_labelTextPadding, m_labelTextPadding );
    loadRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    wxSize parentSize = GetClientSize();

    wxPoint selectedPos;
    wxPoint markedPos;
    wxPoint loadPos;

    if ( m_selectedLabelAlign & wxALIGN_RIGHT ) {
        selectedPos.x   = parentSize.GetWidth() - m_padding - m_maxLabelWidth;
        markedPos.x     = selectedPos.x;
        loadPos.x       = parentSize.GetWidth() - loadRect.GetWidth() - m_padding;
    } else {
        selectedPos.x   = m_padding;
        markedPos.x     = selectedPos.x;
        loadPos.x       = markedPos.x  + markedRect.GetWidth() + m_padding;
    }

    if ( m_selectedLabelAlign & wxALIGN_BOTTOM ) {
        selectedPos.y   = parentSize.GetHeight() - m_padding - markedRect.GetHeight() - m_padding - selectedRect.GetHeight();
        markedPos.y     = parentSize.GetHeight() - m_padding - markedRect.GetHeight();
        loadPos.y       = markedPos.y;
    } else {
        selectedPos.y   = m_padding;
        markedPos.y     = m_padding + selectedRect.GetHeight() + m_padding;
        loadPos.y       = markedPos.y;
    }

    m_selectedLabel->SetPosition( selectedPos );
    m_markedLabel->SetPosition( markedPos );
    m_markedLoadLabel->SetPosition( loadPos );
}

void TableEditor3D::alignLabels()
{
//    calculateMaxLabelWidth();
//
//    if ( m_selectedLabelAlign == m_markedLabelAlign ) {
//        alignAllLabels();
//    } else {
//        alignSingleSelected();
//        alignSingleMarked();
//    }
}


void TableEditor3D::alignSingleSelected()
{
    wxRect selectedRect = m_selectedLabel->GetTextBox( *m_selectedTextDesc );
    selectedRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    wxSize parentSize   = GetClientSize();

    wxPoint labelPosition;

    if ( m_selectedLabelAlign & wxALIGN_RIGHT ) {
        labelPosition.x = parentSize.GetWidth() - m_padding - selectedRect.GetWidth();
    } else {
        labelPosition.x = m_padding;
    }

    if ( m_selectedLabelAlign & wxALIGN_BOTTOM ) {
        labelPosition.y = parentSize.GetHeight() - m_padding - selectedRect.GetHeight();
    } else {
        labelPosition.y = m_padding;
    }

    m_selectedLabel->SetPosition( labelPosition );
}

void TableEditor3D::alignSingleMarked()
{
    wxRect markedRect    = m_markedLabel->GetTextBox( *m_markedTextDesc );
    wxRect loadRect    = m_markedLoadLabel->GetTextBox( *m_markedLoadTextDesc );

    markedRect.Inflate( m_labelTextPadding, m_labelTextPadding );
    loadRect.Inflate( m_labelTextPadding, m_labelTextPadding );

    wxSize parentSize   = GetClientSize();

    wxPoint markedPos;
    wxPoint loadPos;

    if ( m_markedLabelAlign & wxALIGN_RIGHT ) {
        markedPos.x = parentSize.GetWidth() - m_padding - m_maxLabelWidth;
        loadPos.x = parentSize.GetWidth() - m_padding - loadRect.GetWidth();;
    } else {
        markedPos.x = m_padding;
        loadPos.x = m_padding + markedRect.GetWidth() + m_padding;
    }

    if ( m_markedLabelAlign & wxALIGN_BOTTOM ) {
        markedPos.y = parentSize.GetHeight() - m_padding - markedRect.GetHeight();
        loadPos.y = markedPos.y;
    } else {
        markedPos.y = m_padding;
        loadPos.y = markedPos.y;
    }

    m_markedLabel->SetPosition( markedPos );
    m_markedLoadLabel->SetPosition( loadPos );
}



SelectableWidgetBase* TableEditor3D::GetSelectableWidgetBase()
{
    wxWindow* win = GetGrandParent();

    if(!win) return NULL;

    TableEditorPanel* tep = wxDynamicCast(win,TableEditorPanel);

    if(!tep) return NULL;

    return tep;
}

wxWindow* TableEditor3D::GetSelectFaceParent()
{
    wxWindow* gp = GetGrandParent();

    if(!gp) return NULL;

    wxWindow* gpp = gp->GetParent();

    return gpp;
}

void TableEditor3D::OnMouseLeftUp(wxMouseEvent& evt)
{
    m_delta = wxDefaultPosition;

    if(!m_groupManager->IsEditMode()) return;

    wxWindow* gpp = GetSelectFaceParent();
    if(gpp) gpp->AddPendingEvent(evt);

    if(HasCapture())
        ReleaseMouse();

    evt.Skip();
}

void TableEditor3D::OnMouseMove(wxMouseEvent& evt)
{
    if(!m_groupManager->IsEditMode()) return;

    if(m_delta == wxDefaultPosition) {
        wxWindow* gpp = GetSelectFaceParent();
        if(gpp) gpp->AddPendingEvent(evt);
        evt.Skip();
        return;
    }

    if (evt.Dragging())
    {

        wxPoint pos = ClientToScreen(evt.GetPosition());

        wxPoint cpos = GetGrandParent()->GetPosition();

        GetGrandParent()->Move(wxPoint(pos.x - m_delta.x, pos.y - m_delta.y));

        //if(m_winTip) m_winTip->ShowMsg(wxString::Format(_T("xpos=%i\nypos=%i"),GetPosition().x,GetPosition().y));

        wxPointEx rdelta = GetGrandParent()->GetPosition() - cpos;

        wxVariantHash hash;
        hash[_T("groupname")] = m_props.groupName;
        hash[_T("cmd")] =  _T("gauge_move");
        hash[_T("sender_id")] = GetGrandParent() == NULL ? 0L : (long)GetGrandParent()->GetId();
        hash[_T("target_id")] = _T("0");
        hash[_T("deltapos")] << rdelta;

        m_groupManager->ProcessWidgetCommand(hash);

    }



    evt.Skip();
}

void TableEditor3D::OnMouseLeftClick(wxMouseEvent& event)
{
    event.Skip();

    if(m_groupManager->IsEditMode()) {


        CaptureMouse();

        wxPoint pos = ClientToScreen(event.GetPosition());

        wxPoint origin = GetGrandParent()->GetPosition();
        int dx =  pos.x - origin.x;
        int dy = pos.y - origin.y;
        m_delta = wxPoint(dx, dy);

        Refresh(false);


        SelectableWidgetBase* selw = GetSelectableWidgetBase();

        if( !selw->IsSelected() ) {

            if( !event.ShiftDown() ) {

                wxVariantHash hash;
                long id = GetGrandParent()->GetId();
                hash[_T("groupname")] = m_props.groupName;
                hash[_T("cmd")] =  _T("gauge_click");
                hash[_T("sender_id")] = id;
                hash[_T("target_id")] = _T("0");

                m_groupManager->ProcessWidgetCommand(hash);

            }

            selw->SetSelected(true);
            selw->DrawSelectionBitmap(this);
            Refresh(false);

        } else {
            if( event.ShiftDown() ) {
                selw->SetSelected(false);
                Refresh(false);
        }
        }

        return;
    }

}

void TableEditor3D::setDefault()
{
    m_xAngle = 0.0;
    m_yAngle = 0.0;
    m_cols = 0;
    m_rows = 0;
    m_selectedCol = 0;
    m_selectedRow = 0;

    //selected point params
    m_selectedColor     = *wxRED;
    m_highlightedColor  = wxColor( 0xFF, 0xFF, 0x00 );
    m_selectedCol   = 0;
    m_selectedRow   = 0;
    m_highlightOriginRow = -1;
    m_highlightOriginCol = -1;
    m_bHighlighted  = false;
    m_bShowHelp     = false;

    //marked point params
    m_markedColor = *wxGREEN;
    m_markedX = 0.0;
    m_markedY = 0.0;
    m_markedZ = 0.0;
    m_markedPoint = Point3D( 0, 0, 0 );
    m_bUseTargetMarker = false;

    m_smallStep     = 1.0;
    m_largeStep     = 10.0;

    m_min               = Point3D();
    m_max               = Point3D( 1.0, 1.0, 1.0 );
    m_scale             = Point3D( 1.0, 1.0 , 1.0 );
    m_zoom              = 1.0;

    m_tableMinValue     = 0.0;
    m_tableMaxValue     = 1.0;
    m_rowHeaderMinValue = 0.0;
    m_rowHeaderMaxValue = 1.0;
    m_colHeaderMinValue = 0.0;
    m_colHeaderMaxValue = 1.0;

    m_userMinValue      = m_tableMinValue;
    m_userMaxValue      = m_tableMaxValue;

    /// matrices
    Basic3DMath::translate( &m_mTranslateToCenter, -0.5, -0.5, -0.5 );

    m_axis.setAxisColour( Axis3D::X_AXIS, wxColor( 0x00, 0x80, 0xFF ) );
    m_axis.setAxisColour( Axis3D::Y_AXIS, wxColor( 0xFF, 0xFF, 0x00 ) );
    m_axis.setAxisColour( Axis3D::Z_AXIS, wxColor( 0xFF, 0x80, 0x80 ) );

    m_cameraEye = Vector3D( 0.0, 0.0, 1.5 );
 }


/** \purpose    set table dimension
    \require    nothing
    \promise    ...
*/
void TableEditor3D::setDimension( size_t rows, size_t cols )
{
    m_cols = cols;
    m_rows = rows;

    m_columnHeaders.setSize( 1, m_cols );
    m_rowHeaders.setSize( 1, m_rows );
    m_table.setSize( m_rows, m_cols );
    m_highlights.setSize( m_rows, m_cols );

    m_pSurface->setSize( m_rows, m_cols );
}


bool TableEditor3D::calculateMarkedPosition()
{
    if ( ( m_markedX < m_columnHeaders[0] || m_markedX > m_columnHeaders[ m_cols - 1 ] ) ||
         ( m_markedY > m_rowHeaders[0] || m_markedY < m_rowHeaders[ m_rows - 1 ] ) )
    {
        return false;
    }
    else
    {
        interpolate( m_markedX, m_markedY );
        return true;
    }
}

void TableEditor3D::interpolate( double x, double y )
{
	int colRight = findFirstGreater( x , m_columnHeaders );
    int rowBottom = findFirstGreater( y , m_rowHeaders, true );

    normalizeIndex(colRight, m_columnHeaders.size() );
    normalizeIndex(rowBottom, m_rowHeaders.size() );

    int colLeft = colRight -1;
    int rowTop = rowBottom +1;

    normalizeIndex(colLeft, m_columnHeaders.size() );
    normalizeIndex(rowTop, m_rowHeaders.size() );

/**
p0(x0,y0,z0) | p1(x1,y1,z1)
p2(x2,y2,z2) | p3(x3,y3,z3)

assume: x0 = x2, x1 = x3
assume: y0 = y1, y2 = y3
**/

	double leftTop[3]       = { m_columnHeaders[ colLeft ],
                                m_rowHeaders[ rowBottom ],
                                m_table[ index( rowBottom, colLeft ) ] };
    double rightTop[3]      = { m_columnHeaders[ colRight ],
                                m_rowHeaders[ rowBottom],
                                m_table[ index( rowBottom, colRight ) ] };

	double leftBottom[3]    = { m_columnHeaders[ colLeft ],
                                m_rowHeaders[ rowTop ],
                                m_table[ index( rowTop, colLeft ) ] };
	double rightBottom[3]   = { m_columnHeaders[ colRight ],
                                m_rowHeaders[ rowTop ],
                                m_table[ index( rowTop, colRight ) ] };
    double marked[3]        = { x, y, 0.0 };

	Math::linearInterpolation3D( leftTop, rightTop, leftBottom, rightBottom, marked );

    int markedColIdx =  m_cols - 1;

    normalizeIndex(markedColIdx, m_columnHeaders.size() );

	m_markedPoint = Point3D( x / m_columnHeaders[ markedColIdx ], marked[2], y / m_rowHeaders[ 0 ]  );
}


size_t TableEditor3D::index( size_t row, size_t col )
{
    return m_cols * row + col;
}


int TableEditor3D::findFirstGreater( double d, VemsMatrix<double>& a, bool searchFromEnd )
{
    for ( size_t i = 0; i < a.size(); ++i )
    {
        int idx = searchFromEnd ? a.size() - i - 1  : i;

        if ( a[ idx ] > d ) { return idx; }
    }

    return wxNOT_FOUND;
}


void TableEditor3D::updateSurface( wxDC& dc, const wxSize& size )
{
    m_camera.setViewPort( Basic3D::Rectangle( 0, 0, size.GetWidth(), size.GetHeight() ) );
    m_camera.setZoomLevel( m_zoom );
    m_camera.setEye( m_cameraEye );
    Matrix Mv = m_camera.createViewMatrix();

    Matrix Mtr1;
    Basic3DMath::translate( &Mtr1, -m_min.x, -m_min.y, -m_min.z );

    Matrix Ms;
    Basic3DMath::scale( &Ms, m_scale );

    Matrix Mrx;
    Basic3DMath::rotateX( &Mrx, m_xAngle );

    Matrix Mry;
    Basic3DMath::rotateY( &Mry, m_yAngle );

    m_transformationMatrix = Mtr1 * Ms * m_mTranslateToCenter * Mry * Mrx * Mv;

    dc.SetBackground( *wxBLACK );
    dc.Clear();

    if ( m_bShowHelp ) { m_pSurface->setColor( wxColour( 0x80, 0x80, 0x80 ) ); }
    else { m_pSurface->setColor( wxColour( 0xFF, 0xFF, 0xFF ) ); }

    m_pSurface->render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
}


void TableEditor3D::updateHighlightedPoints( wxDC& dc, const wxSize& size )
{
    if ( m_highlights.size() )
    {
        for ( size_t i = 0; i < m_table.size(); ++i )
        {
            if ( !m_highlights[ i ] ) { continue; }

            Marker3D selected( m_pSurface->getPoint( i ), m_min, 3, MARKER_CROSS, m_highlightedColor, false );
            selected.render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
        }
    }


    if ( m_selectedCol >= m_cols )
    {
        Logger::logf( LOG_MSG, _T("-----TableEditor3D::updateHighlightedPoints() - boundary check failed") );
        return;
    }

    if ( !( m_rows - m_selectedRow ) )
    {
        Logger::logf( LOG_MSG, _T("-----TableEditor3D::updateHighlightedPoints() - boundary check failed") );
        return;
    }

    Point4D selectedPoint = m_pSurface->getPoint( m_rows - m_selectedRow - 1, m_selectedCol );
    Marker3D selected( selectedPoint, m_min, 5, MARKER_CROSS, m_selectedColor );
    selected.render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
}


void TableEditor3D::updateLoadMarker( wxDC& dc, const wxSize& size )
{
    setMarkedPosition( m_markedX, m_markedY );
    calculateMarkedPosition();

    if ( m_bUseTargetMarker )
    {
        m_targetPoint = Point3D( m_markedPoint.x, m_markedZ, m_markedPoint.z );
        Marker3D target( m_targetPoint, m_min, 5, MARKER_CIRCLE, *wxWHITE );
        target.render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
    }


    Marker3D marked( m_markedPoint, m_min, 5, MARKER_CIRCLE, m_markedColor );
    marked.render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
}


void TableEditor3D::updateAxes( wxDC& dc, const wxSize& size )
{
    m_axis.setCenter( m_min );
    m_axis.setAxisLength( Axis3D::X_AXIS, m_max.x );
    m_axis.setAxisLength( Axis3D::Y_AXIS, m_max.y );
    m_axis.setAxisLength( Axis3D::Z_AXIS, m_max.z );

    m_axis.render( dc, size.GetWidth(), size.GetHeight(), &m_transformationMatrix );
}


void TableEditor3D::updateBackBuffer( wxDC& dc, const wxSize& size )
{
    if ( !m_needUpdateBackBuffer ) { return; }

    delete m_pBackBuffer;

    m_pBackBuffer = new wxBitmap( size.GetWidth(), size.GetHeight() );
    //m_pBackBuffer = new wxBitmap( dc.GetSize().GetWidth(), dc.GetSize().GetHeight() );

    wxMemoryDC memoryDC( *m_pBackBuffer );

    updateSurface( memoryDC, size );

    resetNeedUpdateBackBuffer();
}


void TableEditor3D::setNeedUpdateBackBuffer()
{
    m_needUpdateBackBuffer = true;
}


void TableEditor3D::resetNeedUpdateBackBuffer()
{
    m_needUpdateBackBuffer = false;
}


void TableEditor3D::OnSize( wxSizeEvent& event )
{
    setNeedUpdateBackBuffer();
    m_isResized = true;

    SelectableWidgetBase* selw = GetSelectableWidgetBase();
    if(selw) selw->DrawSelectionBitmap(this);

    if ( m_areLabelsReady ) {
        alignLabels();
        drawLabels();
    }

    Refresh();
}

void TableEditor3D::OnKillFocus(wxFocusEvent &evt)
{
    if(m_focusSinker) m_focusSinker->OnFocusMsg( FPST_KILLFOCUS );
}

void TableEditor3D::OnSetFocus(wxFocusEvent &evt)
{
    if(m_focusSinker) m_focusSinker->OnFocusMsg( FPST_SETFOCUS );
}

void TableEditor3D::OnPaint( wxPaintEvent& event )
{
    wxBufferedPaintDC dc( this );

    wxSize size = GetClientSize();
	//dc.SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
    //size.SetHeight( size.GetHeight() * SURFACE_HEIGHT_RATE );

    updateBackBuffer( dc, size );

    dc.DrawBitmap( *m_pBackBuffer, 0, 0, true );

    updateHighlightedPoints( dc, size );
    updateLoadMarker( dc, size );
    updateAxes( dc, size );
    renderText( dc );
    renderHelp( dc );

    m_isResized = false;

    SelectableWidgetBase* selw = GetSelectableWidgetBase();

    if(selw && selw->IsSelected()) {
        wxBitmap* selbmp = selw->GetSelectionBitmap();
        if(selbmp) dc.DrawBitmap( *selbmp, 0,0 );
    }

    PaintWidgets(dc);
}


void TableEditor3D::displayValues( wxDC& dc, const wxPoint& pos, const wxString& name, const DoubleArray& values, const ArrayOfColours& colors )
{
    const size_t VALUE_COUNT = 3;
    const size_t COLOR_COUNT = 4;

    if ( values.GetCount() < VALUE_COUNT ) { return; }
    if ( colors.GetCount() < COLOR_COUNT ) { return; }

    size_t length = 0;

//    wxFont font = dc.GetFont();
//    int pointSize = font.GetPointSize();
//    font.SetPointSize( 14 );
//    dc.SetFont( font );

    dc.SetTextForeground( colors[0] );
    dc.DrawText( name, pos.x, pos.y );
    length  += dc.GetTextExtent( name ).GetWidth();

    wxString valueStr = wxString::Format( _T(" %.2f, "), values[0] );

    dc.SetTextForeground( colors[1] );
    dc.DrawText( valueStr, pos.x + length, pos.y );
    length  += dc.GetTextExtent( valueStr ).GetWidth();

    valueStr = wxString::Format( _T(" %.2f"), values[1] );

    dc.SetTextForeground( colors[2] );
    dc.DrawText( valueStr, pos.x + length, pos.y );
    length  += dc.GetTextExtent( valueStr ).GetWidth();

    valueStr = wxString::Format( _T(" ) = ") );

    dc.SetTextForeground( colors[0] );
    dc.DrawText( valueStr, pos.x + length, pos.y );
    length  += dc.GetTextExtent( valueStr ).GetWidth();

    valueStr = wxString::Format( _T(" %.2f"), values[2] );

    dc.SetTextForeground( colors[3] );
    dc.DrawText( valueStr, pos.x + length, pos.y );
    length  += dc.GetTextExtent( valueStr ).GetWidth();

    if ( values.GetCount() < (VALUE_COUNT + 1) ) { return; }
    if ( colors.GetCount() < (COLOR_COUNT + 1) ) { return; }

    valueStr = wxString::Format( _T(" ( %.2f )"), values[3] );

    dc.SetTextForeground( colors[4] );
    dc.DrawText( valueStr, pos.x + length, pos.y );

//    font.SetPointSize( pointSize );
//    dc.SetFont( font );
}


void TableEditor3D::renderText( wxDC& dc )
{
    const int textHeight = dc.GetTextExtent( _T("(TEST)") ).GetHeight();
    // const size_t lineCount = 2;
    // const int linesHeight = lineCount * textHeight;

    const int textPosY = 5;// GetSize().GetHeight() - linesHeight;
    const int textPosX = 10;

    /// draw selected values
    DoubleArray values;
    values.Add( m_columnHeaders[ m_selectedCol ] );
    values.Add( m_rowHeaders[ m_rows - m_selectedRow - 1 ] );
    values.Add( m_table[ index( m_rows - m_selectedRow - 1, m_selectedCol ) ] );

    ArrayOfColours colors;
    colors.Add( m_selectedColor );
    //colors.Add( m_xColor );
    //colors.Add( m_zColor );
    //colors.Add( m_yColor );
    colors.Add( m_axis.getAxisColour( Axis3D::X_AXIS ) );
    colors.Add( m_axis.getAxisColour( Axis3D::Z_AXIS ) );
    colors.Add( m_axis.getAxisColour( Axis3D::Y_AXIS ) );


    wxPoint textPos( textPosX, textPosY );
    displayValues( dc, textPos, _T("selected ("), values, colors );

    /// draw marked values
    values[0] = m_markedX;
    values[1] = m_markedY;
    values[2] = m_markedPoint.y;

    colors[0] = m_markedColor;

    if ( m_bUseTargetMarker )
    {
        values.Add( m_markedZ );
        colors.Add( *wxWHITE );
    }

    textPos.y += textHeight;;
    displayValues( dc, textPos, _T("marked ("), values, colors );

    //Logger::logf( LOG_MSG, _T("TableEditor3D::renderText() - textPosY: %d"), textPos.y  );
}

/*
    min font size
    max font size
*/

int TableEditor3D::getHelpAlign()
{
    bool aligns[4]; //TOP_LEFT, TOP_RIGHT, BOTTOM_LEFT, BOTTOM_RIGHT
    int wxAligns[4] = { (wxALIGN_LEFT | wxALIGN_TOP),
                        (wxALIGN_RIGHT | wxALIGN_TOP),
                        (wxALIGN_LEFT | wxALIGN_BOTTOM),
                        (wxALIGN_RIGHT | wxALIGN_BOTTOM) };

    for ( size_t i = 0; i < 4; ++i ) {
        aligns[i] = true;
    }

    if ( m_selectedLabelAlign & wxALIGN_RIGHT ) {
        if ( m_selectedLabelAlign & wxALIGN_BOTTOM ) {
            aligns[3] = false;
        } else {
            aligns[1] = false;
        }
    } else {
        if ( m_selectedLabelAlign & wxALIGN_BOTTOM ) {
            aligns[2] = false;
        } else {
            aligns[0] = false;
        }
    }

    if ( m_markedLabelAlign & wxALIGN_RIGHT ) {
        if ( m_markedLabelAlign & wxALIGN_BOTTOM ) {
            aligns[3] = false;
        } else {
            aligns[1] = false;
        }
    } else {
        if ( m_markedLabelAlign & wxALIGN_BOTTOM ) {
            aligns[2] = false;
        } else {
            aligns[0] = false;
        }
    }

    if ( aligns[1] ) {
        return wxAligns[1];
    }

    for ( size_t i = 0; i < 4; ++i ) {
        if ( aligns[i] ) {
            return wxAligns[i];
        }
    }

    return wxAligns[1];
}

void TableEditor3D::renderHelp( wxDC& dc )
{
    dc.SetTextForeground( *wxWHITE );

    wxString  hotkeyStr = m_hotkeyMgr->GetHotKeyStringBySymbol( _T("TABLE_EDITOR_3D_KEY_DISPLAY_HELP") );
    wxString  titleStr = m_hotkeyMgr->GetHotKeyTitleBySymbol( _T("TABLE_EDITOR_3D_KEY_DISPLAY_HELP") );
    wxString helpString = titleStr + _T(": ") + hotkeyStr;

    wxRect clientRect = GetClientRect();
    clientRect.Deflate( 5, 5 );

    dc.DrawLabel( helpString, clientRect, getHelpAlign() );

    if ( !m_bShowHelp ) { return; }
    drawHelpPanel( dc );


}


void TableEditor3D::OnKeyDown( wxKeyEvent& event )
{
    if( event.GetKeyCode() == WXK_F12 ){
        if(m_groupManager)
            m_groupManager->ResetGaugesMinMax( m_props.groupName );

    }

    if( m_groupManager->IsEditMode() ) {

        WindowHelper whe(GetGrandParent());
        int res = whe.Action(event);

        whe.GroupAction(m_groupManager, res, m_props.groupName, true);

        event.Skip();

        return;
    }

    wxString evtstr = m_hotkeyMgr->GetEvent(event);

    if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_INC_SMALL" ) ) )
    {
        increaseSelectedCells( m_smallStep );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_INC_LARGE") ) )
    {
        increaseSelectedCells( m_largeStep );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_DEC_SMALL") ) )
    {
        increaseSelectedCells( -m_smallStep );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_DEC_LARGE") ) )
    {
        increaseSelectedCells( -m_largeStep );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_LEFT") ) )
    {
        moveAndUnhighlight( CURSOR_MOVE_LEFT );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_RIGHT") ) )
    {
        moveAndUnhighlight( CURSOR_MOVE_RIGTH );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_UP") ) )
    {
        moveAndUnhighlight( CURSOR_MOVE_UP );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_DOWN") ) )
    {
        moveAndUnhighlight( CURSOR_MOVE_DOWN );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_LEFT") ) )
    {
        moveAndHighlightSelected( CURSOR_MOVE_LEFT );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_RIGHT") ) )
    {
        moveAndHighlightSelected( CURSOR_MOVE_RIGTH );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_UP") ) )
    {
       moveAndHighlightSelected( CURSOR_MOVE_UP );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_DOWN") ) )
    {
        moveAndHighlightSelected( CURSOR_MOVE_DOWN );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_HIGHLIGHT_ALL") ) )
    {
        highlightAll();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_INTERPOLATE_HIGHLITED") ) )
    {
        interpolateHighlighted();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ROTATE_LEFT") ) )
    {
        m_yAngle += ( PI / 180.0 * 1.0 );
        if ( m_yAngle > 2 * PI ) { m_yAngle -= 2 *PI; }
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ROTATE_RIGHT") ) )
    {
        m_yAngle -= ( PI / 180.0 * 1.0 );
        if ( m_yAngle < 0.0 ) { m_yAngle += 2 * PI; }
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ROTATE_UP") ) )
    {
        m_xAngle += ( PI / 180.0 * 1.0 );
//        if ( m_xAngle > PI / 4 ) { m_xAngle = PI / 4; }
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ROTATE_DOWN") ) )
    {
        m_xAngle -= ( PI / 180.0 * 1.0 );
//        if ( m_xAngle < -PI / 4 ) { m_xAngle = -PI / 4; }
        setNeedUpdateBackBuffer();
    }
#ifdef TABLE_EDITOR_3D_TEST
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_MARKER_LEFT") ) )
    {
        setMarkedPosition( m_markedX - m_markerStepX, m_markedY );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_MARKER_RIGHT") ) )
    {
        setMarkedPosition( m_markedX + m_markerStepX, m_markedY );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_MARKER_UP") ) )
    {
        setMarkedPosition( m_markedX, m_markedY + m_markerStepY );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_MARKER_DOWN") ) )
    {
        setMarkedPosition( m_markedX, m_markedY - m_markerStepY );
    }
#endif
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_SET_Y_RANGE_MIN") ) )
    {
        showDialogRangeYMin();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_SET_Y_RANGE_MAX") ) )
    {
        showDialogRangeYMax();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_SAVE_PARAMS") ) )
    {
        saveParams();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_DISPLAY_HELP") ) )
    {
        m_bShowHelp = !m_bShowHelp;
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_SHOW_HIDE_AXES") ) )
    {
        //m_bShowAxes = !m_bShowAxes;
        m_axis.setVisible( !m_axis.getVisible() );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_SELECT_LOAD") ) )
    {
        SelectLoad();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_UNDO_ALL_CELLS_MODIFICATION") ) )
    {
        Logger::logf( LOG_MSG, _T("TableEditor3D::OnKeyDown() - TABLE_EDITOR_3D_KEY_UNDO_ALL_CELLS_MODIFICATION") );
        sendVEMSTableEditorEvent( TableEditorEvent::UNDO_ALL_CELLS_MODIFICATION );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ZOOM_IN") ) )
    {
        m_zoom -= 0.1;
        if ( m_zoom <= 0.5 ) m_zoom = 0.5;
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_ZOOM_OUT") ) )
    {
        m_zoom += 0.1;
        if ( m_zoom >= 1.5 ) m_zoom = 1.5;
        setNeedUpdateBackBuffer();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_WRITE_FW_CONFIG") ) )
    {
        sendVEMSTableEditorEvent( TableEditorEvent::WRITE_FW_CONFIG );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_BURN_FW_CONFIG") ) )
    {
        sendVEMSTableEditorEvent( TableEditorEvent::BURN_FW_CONFIG );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_REFRESH_FW_CONFIG") ) )
    {
        sendVEMSTableEditorEvent( TableEditorEvent::REFRESH_FW_CONFIG );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_MOVE_CAMERA_LEFT") ) )
    {
        moveCameraLeft();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_MOVE_CAMERA_RIGHT") ) )
    {
        moveCameraRight();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_MOVE_CAMERA_UP") ) )
    {
        moveCameraUp();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_MOVE_CAMERA_DOWN") ) )
    {
        moveCameraDown();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_RESET_VIEW") ) )
    {
        resetView();
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_AUTO_FOLLOW_LOAD") ) )
    {
        m_isAutoFollowLoad = !m_isAutoFollowLoad;
    }

    Refresh();
}

/*
else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_LEFT") ) )
    {
        moveAndHighlightSelected( PI );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_RIGHT") ) )
    {
        moveAndHighlightSelected( 0 );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_UP") ) )
    {
       moveAndHighlightSelected( PI / 2 );
    }
    else if ( evtstr.IsSameAs( _T("TABLE_EDITOR_3D_KEY_NAVIGATE_AND_HIGHLIGHT_DOWN") ) )
    {
        moveAndHighlightSelected( PI / 2 * 3 );
    }
*/

void TableEditor3D::moveSelectedAbsolute( CursorMove dir )
{
    switch ( dir )
    {
        case CURSOR_MOVE_LEFT:
            if ( m_selectedCol > 0 ) --m_selectedCol;
            break;
        case CURSOR_MOVE_RIGTH:
            if ( m_selectedCol < ( m_cols - 1 ) ) ++m_selectedCol;
            break;
        case CURSOR_MOVE_UP:
            if ( m_selectedRow < ( m_rows -1 ) ) ++m_selectedRow;
            break;
        case CURSOR_MOVE_DOWN:
            if ( m_selectedRow > 0 ) --m_selectedRow;
            break;
    }
}

void TableEditor3D::moveSelectedRelative( CursorMove dir )
{
    double dirAngle = m_yAngle;

    switch ( dir )
    {
        case CURSOR_MOVE_LEFT:
            dirAngle += PI;
            break;
        case CURSOR_MOVE_RIGTH:
            break;
        case CURSOR_MOVE_UP:
            dirAngle += ( PI / 2 );
            break;
        case CURSOR_MOVE_DOWN:
            dirAngle += ( PI / 2 * 3 );
            break;
    }

    if ( dirAngle > 2 * PI ) { dirAngle -= 2 *PI; }

    if ( ( ( dirAngle >= 0 && dirAngle < PI / 4 ) || ( dirAngle > PI / 4 * 7 ) ) && m_selectedCol < ( m_cols - 1 ) )
    {
        ++m_selectedCol;
    }
    else if ( dirAngle >= PI / 4 && dirAngle < PI / 4  * 3  && m_selectedRow < ( m_rows -1 ) )
    {
        ++m_selectedRow;
    }
    else if ( dirAngle >= PI / 4 * 3 && dirAngle < PI / 4 * 5  && m_selectedCol > 0 )
    {
        --m_selectedCol;
    }
    else if ( dirAngle >= PI / 4 * 5 && dirAngle <= PI / 4 * 7 && m_selectedRow > 0 )
    {
        --m_selectedRow;
    }
}

void TableEditor3D::moveSelected( CursorMove dir )
{
    switch ( m_cursorMovingMode )
    {
        case TE3D_CursorMovingMode::CMM_RELATIVE:
            moveSelectedRelative( dir );
            break;
        case TE3D_CursorMovingMode::CMM_ABSOLUTE:
            moveSelectedAbsolute( dir );
            break;
    }
    drawLabels();
}

void TableEditor3D::moveAndUnhighlight( CursorMove dir )
{
    resetHighlightOrigin();
    unhighlightAllCells();

    int prevoiusRow = m_selectedRow;
    int previousCol = m_selectedCol;

    moveSelected( dir );

    if ( prevoiusRow == (int)m_selectedRow && previousCol == (int)m_selectedCol ) return;

    sendVEMSTableEditorEvent( TableEditorEvent::CURSOR_MOVED, m_rows - m_selectedRow - 1, m_selectedCol );
}

void TableEditor3D::moveAndHighlightSelected( CursorMove dir )
{
    //Logger::logf( LOG_MSG, _T("TableEditor3D::moveAndHighlightSelected()") );

    const int prevSelectedRow = m_selectedRow;
    const int prevSelectedCol = m_selectedCol;

    moveSelected( dir );

    if ( m_highlightOriginRow < 0 || m_highlightOriginRow >= getNumberRows() ||
         m_highlightOriginCol < 0 || m_highlightOriginCol >= getNumberCols() )
    {
        m_highlightOriginRow = prevSelectedRow;
        m_highlightOriginCol = prevSelectedCol;
    }

    unhighlightAllCells();
    highlightArea( m_highlightOriginRow, m_highlightOriginCol, m_selectedRow, m_selectedCol );
}


void TableEditor3D::increaseSelectedCells( double inc )
{
    if ( !m_bHighlighted )
    {

        increaseCellValue( m_selectedRow, m_selectedCol, inc );
//        Logger::logf( LOG_MSG, _T("TableEditor3D::increaseSelectedCells() - row: %d, col: %d, value: %f"),
//            m_selectedRow, m_selectedCol, getCellValue( m_rows - m_selectedRow - 1, m_selectedCol ) );
        sendSingleCellChangedEvent( m_rows - m_selectedRow - 1 , m_selectedCol,
                                    getCellValue( m_rows - m_selectedRow - 1 , m_selectedCol ) );
        return;
    }

    int left    = wxMin( (int)(m_selectedCol), m_highlightOriginCol );
    int right   = wxMax( (int)(m_selectedCol), m_highlightOriginCol );
    int top     = wxMax( (int)(m_selectedRow), m_highlightOriginRow );
    int bottom  = wxMin( (int)(m_selectedRow), m_highlightOriginRow );
    //int top     = m_rows - wxMax( m_selectedRow, m_highlightOriginRow ) - 1;
    //int bottom  = m_rows - wxMin( m_selectedRow, m_highlightOriginRow ) - 1;

    for ( int i = bottom; i <= top; ++i )
    {
        for ( int j = left; j <= right; ++j )
        {
            increaseCellValue( i, j, inc );
        }
    }

    sendVEMSTableEditorEvent( TableEditorEvent::CELLS_CHANGED );
    setNeedUpdateBackBuffer();
}

void TableEditor3D::increaseCellValue( int row, int col, double inc )
{
    double& d = m_table[ index( m_rows - row - 1, col ) ];
    d += inc;
    if ( d > m_max.y ) { d = m_max.y; }
    if ( d < m_min.y ) { d = m_min.y; }
}



//under dev
void TableEditor3D::setRowLabelValue( int row, double value )
{
    m_rowHeaders[ row ] = value;

    for ( size_t col = 0; col < m_cols; ++col )
    {
        m_pSurface->setPoint( row, col, createSurfacePoint( row, col ) );
    }
    setNeedUpdateBackBuffer();
}

void TableEditor3D::setColLabelValue( int col, double value )
{
    m_columnHeaders[ col ] = value;

    for ( size_t row = 0; row < m_rows; ++row )
    {
        m_pSurface->setPoint( row, col, createSurfacePoint( row, col ) );
    }
    setNeedUpdateBackBuffer();
}

void TableEditor3D::setCellValue( int row, int col, double value )
{
    m_table[ index( row, col ) ] = value;
    m_pSurface->setPoint( row, col, createSurfacePoint( row, col ) );
    setNeedUpdateBackBuffer();
}

double TableEditor3D::getCellValue( int row, int col )
{
    return m_table[ index( row, col ) ];
}

int TableEditor3D::getNumberCols()
{
    return m_cols;
}

int TableEditor3D::getNumberRows()
{
    return m_rows;
}

void TableEditor3D::setMarkedPosition( double x, double y )
{
    if ( x < m_columnHeaders[0] ) { m_markedX = m_columnHeaders[0]; }
    else if ( x > m_columnHeaders[ m_cols - 1 ] ) { m_markedX = m_columnHeaders[ m_cols - 1 ] - 1; }
    else { m_markedX = x; }

    if ( y < m_rowHeaders[ m_rows - 1 ] ) { m_markedY = m_rowHeaders[ m_rows - 1 ]; }
    else if ( y > m_rowHeaders[0] ) { m_markedY = m_rowHeaders[0]; }
    else { m_markedY = y; }
}


void TableEditor3D::setTargetPosition( double z )
{
    m_markedZ = z;
    drawLabels();
}


void TableEditor3D::resetHighlightOrigin()
{
    m_highlightOriginRow = -1;
    m_highlightOriginCol = -1;
}


void TableEditor3D::unhighlightAllCells()
{
	//m_highlights.Clear();
    //m_highlights.Add( 0, m_rows * m_cols );
    m_highlights.fill( 0 );
    m_bHighlighted = false;
}


void TableEditor3D::highlightArea( int startRow, int startCol, int endRow, int endCol )
{
    // swap values
	if ( startRow > endRow )
	{
		int tmp = startRow;
		startRow = endRow;
		endRow = tmp;
	}

	if ( startCol > endCol )
	{
		int tmp = startCol;
		startCol = endCol;
		endCol = tmp;
	}

	// highlight area
	for ( int i = startRow; i <= endRow; ++i )
	{
		for ( int j = startCol; j <= endCol; ++j )
		{
            m_highlights[ index( m_rows - i - 1, j ) ] = 1;
            //Logger::logf( LOG_MSG, _T("\m_highlights[ %d ] = %u"),
            //    index( m_rows - i - 1, j ), m_highlights.get( index( m_rows - i - 1, j ) )  );
		}
	}

	m_bHighlighted = true;
}


void TableEditor3D::highlightAll()
{
    highlightArea( 0, 0, m_rows-1, m_cols-1 );

    m_highlightOriginRow = 0;
    m_highlightOriginCol = 0;
    m_selectedRow = m_rows - 1;
    m_selectedCol = m_cols - 1;
}


void TableEditor3D::interpolateHighlighted()
{
    if ( !m_bHighlighted ) { return; }

    MatrixSelection s;

    s.left      = wxMin( (int)m_selectedCol, m_highlightOriginCol );
    s.right     = wxMax( (int)m_selectedCol, m_highlightOriginCol );
    s.top       = (int)m_rows - wxMax( (int)m_selectedRow, m_highlightOriginRow ) - 1;
    s.bottom    = (int)m_rows - wxMin( (int)m_selectedRow, m_highlightOriginRow ) - 1;

    VemsMatrixHelper::interpolate( &m_table, m_rowHeaders, m_columnHeaders, s );
    sendVEMSTableEditorEvent( TableEditorEvent::CELLS_CHANGED );
}


void TableEditor3D::OnMouseRightClick( wxMouseEvent& event )
{
    wxMenu menu;

    //menu.Append( TABLE_EDITOR_3D_PM_ROTATE_LEFT, _T("Rotate left \t") + getKeyTextualForFunction(TABLE_EDITOR_3D_KEY_ROTATE_LEFT ) );
    //menu.Append( TABLE_EDITOR_3D_PM_ROTATE_RIGHT, _T("Rotate right \t") + getKeyTextualForFunction(TABLE_EDITOR_3D_KEY_ROTATE_RIGHT ) );
    //menu.Append( TABLE_EDITOR_3D_PM_ROTATE_UP, _T("Rotate up \t") + getKeyTextualForFunction(TABLE_EDITOR_3D_KEY_ROTATE_UP ) );
    //menu.Append( TABLE_EDITOR_3D_PM_ROTATE_DOWN, _T("Rotate down \t") + getKeyTextualForFunction(TABLE_EDITOR_3D_KEY_ROTATE_DOWN ) );
    //menu.Append( TABLE_EDITOR_3D_PM_IMPORT, _T("Import") );
    //menu.Append( TABLE_EDITOR_3D_PM_REMAP, _T("Remap") );
    //menu.AppendSeparator();

    if( GetItemID() != -1 )
    {

        // tables from ini file
        wxMenu* tables = new wxMenu();
        for ( unsigned int i = 0; i < m_tables.GetCount(); ++i )
        {
            int id = wxNewId();
            tables->Append( id, m_tables[i] );
            m_menuIDtoTableName[id] = m_tables[i];
            Connect( id, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler(TableEditor3D::OnPopupMenuChangeTable) );
        }

        menu.SetTitle( m_props.groupName );
        //menu.AppendSeparator();
        menu.AppendSubMenu( tables, _T("Change Table To") );
        menu.AppendSeparator();
        menu.AppendSubMenu( createAddSubMenu(), _T("add") );
        menu.Append( createRemoveMenu(&menu) );
        menu.Append( createSaveMenu(&menu) );
    }

	const wxPoint& pos = event.GetPosition();
    PopupMenu( &menu, pos.x, pos.y);
}


void TableEditor3D::OnPopupMenuChangeTable( wxCommandEvent& event )
{
    TableEditorEvent e( TableEditorEvent::TABLE_NAME_CHANGED );

    e.name = m_menuIDtoTableName[ event.GetId() ];

    if ( GetParent() ) { GetParent()->AddPendingEvent( e ); }
}


void TableEditor3D::setMinRangeY( double y )
{
    if ( y >= m_userMaxValue ) { return; }
#ifndef __WINCE_STANDARDSDK__
	m_userMinValue = std::max( m_tableMinValue, y );
#else
	m_userMinValue = max( m_tableMinValue, y  );
#endif
    setRangeY();
}


void TableEditor3D::setMaxRangeY( double y )
{
    if ( y <= m_userMinValue ) { return; }
#ifndef __WINCE_STANDARDSDK__
	m_userMaxValue = std::min( m_tableMaxValue, y );
#else
	m_userMaxValue = min( m_tableMaxValue, y );
#endif
    setRangeY();
}


void TableEditor3D::showDialogRangeYMin()
{
    wxString title = wxString::Format(  _T("Set y-axis min value( %f, %f )"),
                                        m_userMinValue, m_userMaxValue );
                                        //m_min.y, m_max.y );

    wxString str = ::wxGetTextFromUser( _T(""),
                                        title,
                                        _T(""),
                                        this,
                                        wxDefaultCoord,
                                        wxDefaultCoord,
                                        true );

    double d;
    if ( str.ToDouble( &d ) )
    {
        setMinRangeY( d );
    }
}


void TableEditor3D::showDialogRangeYMax()
{
    wxString title = wxString::Format(  _T("Set y-axis max value( %f, %f )"),
                                        //m_min.y, m_max.y );
                                        m_userMinValue, m_userMaxValue );

    wxString str = ::wxGetTextFromUser( _T(""),
                                        title,
                                        _T(""),
                                        this,
                                        wxDefaultCoord,
                                        wxDefaultCoord,
                                        true );

    double d;
    if ( str.ToDouble( &d ) )
    {
        setMaxRangeY( d );
    }
}


int TableEditor3D::GetRowLabelValue( int row )
{
    return (int)(m_rowHeaders[ row ]);
}


int TableEditor3D::GetColLabelValue( int col )
{
    return (int)(m_columnHeaders[ col ]);
}


void TableEditor3D::OnPopupMenu( wxCommandEvent& event )
{
    switch ( event.GetId() )
    {
        case TABLE_EDITOR_3D_PM_ROTATE_LEFT     :
            m_yAngle += ( PI / 180.0 * 1.0 );
            if ( m_yAngle > 2 * PI ) { m_yAngle -= 2 *PI; }
            break;
        case TABLE_EDITOR_3D_PM_ROTATE_RIGHT    :
            m_yAngle -= ( PI / 180.0 * 1.0 );
            if ( m_yAngle < 0.0 ) { m_yAngle += 2 * PI; }
            break;
        case TABLE_EDITOR_3D_PM_ROTATE_UP       :
            m_xAngle += ( PI / 180.0 * 1.0 );
//            if ( m_xAngle > PI / 4 ) { m_xAngle = PI / 4; }
            break;
        case TABLE_EDITOR_3D_PM_ROTATE_DOWN     :
            m_xAngle -= ( PI / 180.0 * 1.0 );
//            if ( m_xAngle < -PI / 4 ) { m_xAngle = -PI / 4; }
            break;
        case TABLE_EDITOR_3D_PM_IMPORT:
        {
            sendVEMSTableEditorEvent( TableEditorEvent::IMPORT );
        } break;
        case TABLE_EDITOR_3D_PM_REMAP:
        {
            sendVEMSTableEditorEvent( TableEditorEvent::REMAP );
        } break;

    }

    Refresh();
}


void TableEditor3D::loadParams()
{
    TableEditor3DParams params;

    if ( isGroupMember() )
    {
        TableEditor3DSQLHelper::loadParams( m_props.itemID, m_groupManager, params );
    }
    else
    {
        TableEditor3DConfigHelper::loadParams( m_tableName, params );
    }

    TableEditor3DConfigHelper::loadGlobalParams( params );

    m_xAngle            = params.rotationAngleX;
    m_yAngle            = params.rotationAngleY;
    m_cameraEye.x       = params.cameraEyeX;
    m_cameraEye.y       = params.cameraEyeY;
    m_zoom              = params.zoom;
    m_cursorMovingMode  = params.cursorMovingMode;
    m_selectedLabelAlign = TE3D_Align::getWxAlign( params.selectedLabelAlign );
    m_markedLabelAlign   = TE3D_Align::getWxAlign( params.markedLabelAlign );
    m_axis.setVisible( params.showAxes );
    m_labelFontHeight   = params.labelFontSize;
}


void TableEditor3D::saveParams()
{
    TableEditor3DParams params;

    params.rotationAngleX   = m_xAngle;
    params.rotationAngleY   = m_yAngle;
    params.cameraEyeX       = m_cameraEye.x;
    params.cameraEyeY       = m_cameraEye.y;
    params.zoom             = m_zoom;
    params.showAxes         = m_axis.getVisible();

    //params.labelFontSize    = m_labelFontHeight;


    if ( isGroupMember() )
    {
        TableEditor3DSQLHelper::saveParams( m_props.itemID, m_groupManager, params );
    }
    else
    {
        TableEditor3DConfigHelper::saveParams( m_tableName, params );
    }
}


void TableEditor3D::sendVEMSTableEditorEvent( TableEditorEvent::EventType type, int row, int col, double d )
{
    TableEditorEvent e( type, row, col, d );
    if ( GetParent() ) { GetParent()->AddPendingEvent( e ); }
}


void TableEditor3D::sendSingleCellChangedEvent( int row, int col, double d )
{
    TableEditorEvent e( TableEditorEvent::CELL_CHANGED );
    e.row = row;
    e.col = col;
    e.value = d;
    //e.readAfterWrite = readAfterWrite;

    if ( GetParent() )
    {
        GetParent()->AddPendingEvent( e );
    }
}


void TableEditor3D::SelectLoad()
{
    int colRight = findFirstGreater( m_markedX , m_columnHeaders );
    int rowBottom = findFirstGreater( m_markedY , m_rowHeaders, true );

    normalizeIndex(colRight, m_columnHeaders.size() );
    normalizeIndex(rowBottom, m_rowHeaders.size() );

    int colLeft = colRight -1;
    int rowTop = rowBottom +1;

    normalizeIndex(colLeft, m_columnHeaders.size() );
    normalizeIndex(rowTop, m_rowHeaders.size() );

    double cx = ( m_columnHeaders[colRight] + m_columnHeaders[colLeft] ) / 2.0;
    double cy = ( m_rowHeaders[rowBottom] + m_rowHeaders[rowTop] ) / 2.0;

    int row = rowTop;
    int col = colLeft;

    if ( m_markedX > cx ) { col = colRight; }
    if ( m_markedY > cy ) { row = rowBottom; }

    resetHighlightOrigin();
    unhighlightAllCells();

    m_selectedCol = col;
    m_selectedRow = m_rows - row - 1;
}


void TableEditor3D::setCellsMinValue( double min )
{
    m_tableMinValue = min;
    m_userMinValue  = min;
    setRangeY();
}


void TableEditor3D::setCellsMaxValue( double max )
{
    m_tableMaxValue = max;
    m_userMaxValue  = max;
    setRangeY();
}


Point4D TableEditor3D::createSurfacePoint( size_t row, size_t col )
{
    if ( m_cols < 2 || m_rows < 2 ) { return Point4D(); }

    double z = m_rowHeaders[ 0 ] ? m_rowHeaders[ row ] / m_rowHeaders[ 0 ] : 0.0;
    double x = m_columnHeaders[ m_cols - 1 ] ? m_columnHeaders[ col ] / m_columnHeaders[ m_cols - 1 ] : 0.0;
    double y = m_table.get( row, col );

#ifdef TABLE_EDITOR_3D_TEST
    if ( m_cols < 1 || m_rows < 1 ) { return; }

    setMarkedPosition( m_markedX, m_markedY );

    m_markerStepX = ( m_columnHeaders[ m_cols - 1 ] - m_columnHeaders[0] ) / 10.0;
    m_markerStepY = ( m_rowHeaders[0] - m_rowHeaders[ m_rows - 1 ] ) / 10.0;
#endif

    double xRange   = m_columnHeaders[m_cols-1] - m_columnHeaders[0];
    m_scale.x       = xRange ? m_columnHeaders[m_cols-1] / xRange : 1.0;
    m_min.x         = m_columnHeaders[m_cols-1] ? m_columnHeaders[0] / m_columnHeaders[m_cols-1] : 0.0;

    double zRange   = m_rowHeaders[0] - m_rowHeaders[m_rows-1];
    m_scale.z       = zRange  ? m_rowHeaders[0] / zRange : 1.0;
    m_min.z         = m_rowHeaders[0] ? m_rowHeaders[m_rows-1] / m_rowHeaders[0] : 0.0;

    Point4D p( x, y, z );
    //Logger::logf( LOG_MSG, _T("TableEditor3D::createSurfacePoint()  - ") + p.toString() );
    return p;
}


void TableEditor3D::setRangeY()
{
    double yRange   = m_userMaxValue - m_userMinValue;
    m_scale.y       = yRange ? 1.5 / yRange : 1.0;
    m_min.y         = m_userMinValue;
    m_max.y         = m_userMaxValue;

    m_pSurface->setMinY( m_min.y );

    //Logger::logf( LOG_MSG, _T("TableEditor3D::setRangeY() - m_min( %f, %f, %f)"), m_min.x, m_min.y, m_min.z );
    //Logger::logf( LOG_MSG, _T("TableEditor3D::setRangeY() - m_max( %f, %f, %f)"), m_max.x, m_max.y, m_max.z );
    //ogger::logf( LOG_MSG, _T("TableEditor3D::setRangeY() - m_scale( %f, %f, %f)"), m_scale.x, m_scale.y, m_scale.z );
}


void TableEditor3D::normalizeIndex( int& idx, int arraySize )
{
    if(idx < 0){
         idx = 0;
         return;
    }
    int maxSize = arraySize -1;
    if(idx > maxSize) idx = maxSize;
}


void TableEditor3D::moveCursorToPosition( const wxPoint& position )
{
    int row = m_rows - position.y - 1;
    int col = position.x;

    if ( row < 0 || row > ( (int)m_rows - 1 ) || col < 0 || col > ((int)m_cols - 1 ) ) return;
    if ( row == (int)m_selectedRow && col == (int)m_selectedCol ) return;

    m_selectedRow = row;
    m_selectedCol = col;

    resetHighlightOrigin();
    unhighlightAllCells();

    Refresh();
}

void TableEditor3D::loadFromVemsMatrix( const VemsMatrix<double>& m )
{
    for ( size_t i = 0; i < m.rows(); ++i )
    {
        for ( size_t j = 0; j < m.cols(); ++j )
        {
            setCellValue( i, j, m.get( i, j ) );
        }
    }
}

void TableEditor3D::loadRowHeaderFromVemsMatrix( const VemsMatrix<double>& m )
{
    for ( size_t i = 0; i < m.cols(); ++i )
    {
        setRowLabelValue( i, m.get( i ) );
    }
}

void TableEditor3D::loadColHeaderFromVemsMatrix( const VemsMatrix<double>& m )
{
    for ( size_t i = 0; i < m.cols(); ++i )
    {
        setColLabelValue( i, m.get( i ) );
    }
}


void TableEditor3D::drawHelpPanel( wxDC& dc )
{
    wxSize size = GetClientSize();

    if ( !m_helpPanelBmp || m_isResized ) { createHelpPanel( size.GetHeight() ); }

    int helpBmpWidth    = m_helpPanelBmp->GetWidth();
    int helpBmpHeight   = m_helpPanelBmp->GetHeight();

    int helpBmtLeft = ( size.GetWidth() - helpBmpWidth ) / 2;
    int helpBmtTop = ( size.GetHeight() - helpBmpHeight ) / 2;

    dc.DrawBitmap( *m_helpPanelBmp, helpBmtLeft, helpBmtTop );
}


double TableEditor3D::getHelpTextHeight( int clientHeight, size_t lineCount, double margin, double rowScale )
{
    const double MAX_TEXT_HEIGHT = 9.5;
    const double MIN_TEXT_HEIGHT = 5.0;

    double textAreaHeight = (double)clientHeight - ( margin * 2 ) - 2;

    double rowHeight = textAreaHeight / (double)lineCount;

    double textHeight = rowHeight / rowScale;

    if ( textHeight > MAX_TEXT_HEIGHT ) { textHeight = MAX_TEXT_HEIGHT; }
    else if ( textHeight < MIN_TEXT_HEIGHT ) { textHeight = MIN_TEXT_HEIGHT ;}

    return textHeight;
}

void TableEditor3D::createHelpPanel( int clientHeight )
{
    HotKeyDescArray arr;

    double m = 10.0;
    double rowScale = 1.6;

    m_hotkeyMgr->GetHotKeyList(arr);

    double maxTitleWidth = 0.0, maxHKWidth = 0.0;

    size_t arrc = arr.GetCount();

    double th = getHelpTextHeight( clientHeight, arrc, m, rowScale );

    for( size_t a=0; a<arrc; a++ ) {
        double tw = wxAgg::getTextWidth( arr[a].Title, th );
        double hw = wxAgg::getTextWidth( arr[a].HotKey, th );

        if( maxTitleWidth < tw ) maxTitleWidth = tw;
        if( maxHKWidth < hw ) maxHKWidth = hw;
    }

    double rowheight = th * rowScale;

    double pw = 2*m + maxTitleWidth + 2*m + maxHKWidth;
    double ph = arrc * rowheight + 2*m + 2;

    wxAgg aggCanvas( (int)pw, (int)ph );

    aggCanvas.clear(wxColor(0,0,0,0));

    aggCanvas.setPen( wxColor(0,0,0,255) );

    aggCanvas.setBrush(wxColor( 255,255,255,200 ));

    double bw = 2;

    aggCanvas.drawFilledRoundedRectangle(bw/2, bw/2, pw-2*bw, ph-2*bw, m, bw);

    aggCanvas.setPen( wxColor(0,0,0,255) );

    for( size_t a=0; a<arrc; a++ ) {

        double ty = (a+1) * rowheight + m;

        aggCanvas.drawText( arr[a].Title, m, ty, -1, th, 1.0);
        aggCanvas.drawText( arr[a].HotKey, 2*m + maxTitleWidth, ty, -1, th, 1.0);

    }

    if(m_helpPanelBmp)
    {
        ptrdel(m_helpPanelBmp);
    }

    m_helpPanelBmp = aggCanvas.renderToBitmap();
}

void TableEditor3D::moveCameraLeft()
{
    m_cameraEye.x -= CAMERA_STEP;

    if ( m_cameraEye.x < -MAX_CAMERA_EYE_X ) { m_cameraEye.x = -MAX_CAMERA_EYE_X; }

    setNeedUpdateBackBuffer();
}

void TableEditor3D::moveCameraRight()
{
    m_cameraEye.x += CAMERA_STEP;

    if ( m_cameraEye.x > MAX_CAMERA_EYE_X ) { m_cameraEye.x = MAX_CAMERA_EYE_X; }

    setNeedUpdateBackBuffer();
}

void TableEditor3D::moveCameraUp()
{
    m_cameraEye.y += CAMERA_STEP;

    if ( m_cameraEye.y > MAX_CAMERA_EYE_Y ) { m_cameraEye.y = MAX_CAMERA_EYE_Y; }

    setNeedUpdateBackBuffer();
}

void TableEditor3D::moveCameraDown()
{
    m_cameraEye.y -= CAMERA_STEP;

    if ( m_cameraEye.y < -MAX_CAMERA_EYE_Y ) { m_cameraEye.y = -MAX_CAMERA_EYE_Y; }

    setNeedUpdateBackBuffer();
}

void TableEditor3D::resetView()
{
    m_cameraEye = Vector3D( 0.0, 0.0, 1.5 );
    m_zoom      = 1.0;

    setNeedUpdateBackBuffer();
}

bool TableEditor3D::isGroupMember()
{
    return ( GetItemID() != -1 );
}


void TableEditor3D::AutoFollowLoad()
{
    if ( m_isAutoFollowLoad ) {
        SelectLoad();
    }
}
