#include "ChannelCollision.h"

#include "Symbol.h"
#include "Math.h"

#include "common/types/IntArray.h"
#include "common/types/String2IntHash.h"
#include "common/VemsMatrix.h"
#include "gaugedesc/OutputChannelSQL.h"
#include "stdatamanager/STDataModelBase.h"

wxArrayString ChannelCollision::detectInvalidChannels( STDataModelBase * stdata_model, SymbolTable * symbols )
{
    wxArrayString retVal;

    wxObjectArray arr = OutputChannelSQL::getSymbols();

    // get valid symbolnames from DB (output channels)
    for ( size_t i = 0; i < arr.GetCount(); ++i )
    {
        wxString symname = ((dbsymbol*)arr[i])->name;
        SymbolTable::const_iterator it = symbols->find(symname);

        if ( it != symbols->end() )
        {
            uint8_t val = ((uint8_t)stdata_model->getViewerData(*(*symbols)[symname]));

            if ( val > 128 ) val -= 128;

            if ( OutputChannelSQL::isValidType( val ) )
            {
                if ( (val & 15) == 0x0f ) // disabled
                    continue;
            }
            else
            {
                wxString ret = ((dbsymbol*)arr[i])->fullname + _T("|") + ((dbsymbol*)arr[i])->name;
                retVal.Add( ret );
            }
        }
    }

    return retVal;
}

wxArrayString ChannelCollision::detectCollisions( STDataModelBase * stdata_model, SymbolTable * symbols )
{
    wxArrayString collisions;
    collisions.Clear();

    wxArrayString symnames;
    symnames.Clear();

    wxArrayString nice_name;
    nice_name.Clear();

    wxObjectArray arr = OutputChannelSQL::getSymbols();

    String2IntHash oc_values;
    oc_values.clear();

    // get valid symbolnames from DB (output channels)
    for ( size_t i = 0; i < arr.GetCount(); ++i )
    {
        wxString symname = ((dbsymbol*)arr[i])->name;
        SymbolTable::const_iterator it = symbols->find(symname);

        if ( it != symbols->end() )
        {
            // get value from model

            uint8_t val = ((uint8_t)stdata_model->getViewerData(*(*symbols)[symname]));
            // ignore inverted
            if ( val > 128 ) val -= 128;

            if ( OutputChannelSQL::isValidType( val ) )
            {
                if ( (val & 15) == 0x0f ) // disabled
                    continue;
            }
            else // INVALID
                continue;

            oc_values[symname] = val;
            symnames.Add( symname );
            nice_name.Add( ((dbsymbol*)arr[i])->fullname );
        }
    }

    int symcount = symnames.GetCount();

    //
    int alt_cyl = -1;
    SymbolTable::const_iterator stit = symbols->find(_T("alternateCyl"));

    if ( stit != symbols->end() )
        alt_cyl = (uint8_t) stdata_model->getViewerData( *(*symbols)[_T("alternateCyl")] );

    uint8_t ignoutputs= (uint8_t) stdata_model->getViewerData(*(*symbols)[_T("ignchmax")]);

    int mat_size = symcount + (ignoutputs + 1);
    mat_size += ( alt_cyl + 1 );

    VemsMatrix<bool> collision_mat( mat_size, mat_size, false );
    collision_mat.fill( false );

    for ( String2IntHash::const_iterator it = oc_values.begin(); it != oc_values.end(); ++it )
    {
        int val = it->second;
        for ( String2IntHash::const_iterator it2 = oc_values.begin(); it2 != oc_values.end(); ++it2 )
        {
            int val2 = it2->second;
            if ( (val == val2) && ( it->first != it2->first ) )
            {
                int idx1 = symnames.Index( it->first ), idx2 = symnames.Index( it2->first );
                collision_mat.set( idx1, idx2, true );
                collision_mat.set( idx2, idx1, true );
            }
        }
    }

    // check injector outputs
    for ( int j = 0; j <= alt_cyl; ++j )
    {
        wxString hexid = wxString::Format( _T("%d0"),j );
        int channel = Math::hex2int( hexid );
        symnames.Add( wxString::Format( _T("Injector output %d"), j ) );

        for ( String2IntHash::const_iterator it2 = oc_values.begin(); it2 != oc_values.end(); ++it2 )
        {
            int val = it2->second;
            if ( val == channel )
            {
                int idx1 = symcount + j, idx2 = symnames.Index( it2->first );
                collision_mat.set( idx1, idx2, true );
                collision_mat.set( idx2, idx1, true );
            }
        }
    }

    // ignition outputs
    bool isDual = ( (uint8_t)stdata_model->getViewerData(*(*symbols)[_T("ign_out_dual")]) )!=0?true:false;

    for ( size_t j = 0; j <= ignoutputs; ++j )
    {
        wxString h2str = ( isDual ) ? wxString::Format(_T("h2d%d"), j) : wxString::Format(_T("h2%d"), j);

        symnames.Add( wxString::Format( _T("Ignition output %d"), j ) );

        uint8_t pinval = (uint8_t) stdata_model->getViewerData(*(*symbols)[h2str]);

        uint8_t chId, otype;

        if ( OutputChannelSQL::getIgnitionMapping( isDual, pinval, otype, chId) )
        {
            if ( isDual )
            {
                int chval1 = ( 16* ( chId -1 ) + otype ), chval2 = (16 * chId + otype);
                for ( String2IntHash::const_iterator it = oc_values.begin(); it != oc_values.end(); ++it )
                {
                    int val = it->second;
                    if ( val == chval1 || val == chval2 )
                    {
                        int idx1 = symcount + alt_cyl + 1 + j, idx2 = symnames.Index( it->first );
                        collision_mat.set( idx1, idx2, true );
                        collision_mat.set( idx2, idx1, true );
                    }
                }
            }
            else
            {
                int chval = 16 * chId + otype;
                for ( String2IntHash::const_iterator it = oc_values.begin(); it != oc_values.end(); ++it )
                {
                    int val = it->second;
                    if ( val == chval )
                    {
                        int idx1 = symcount + alt_cyl + 1 + j, idx2 = symnames.Index( it->first );
                        collision_mat.set( idx1, idx2, true );
                        collision_mat.set( idx2, idx1, true );
                    }
                }
            }
        }
    }

    //
    for ( int i = 0; i < mat_size; ++i )
    {
        for ( int j = i; j < mat_size; ++j )
        {
            if ( collision_mat.get( i, j ) )
            {
                wxString left = ( i < symcount ) ? nice_name[i] : symnames[i];
                wxString right = ( j < symcount ) ? nice_name[j] : symnames[j];
                wxString s = left + _T("|") + right;
                s += _T("|") + symnames[i] + _T("|") + symnames[j];
                collisions.Add(s);
            }
        }
    }

    // iac_stepper
    SymbolTable::const_iterator it = symbols->find( _T("iac_type") );
    if ( it != symbols->end() )
    {
        IntArray iacvals; iacvals.Add( -1, 5 );
        uint8_t val = (uint8_t) stdata_model->getViewerData(*(*symbols)[_T("iac_type")]);
        if ( val == 3 )
        {
            for ( size_t i = 1; i <= 4; ++i )
            {
                wxString symn = wxString::Format( _T("iac_step_seq_%d"),i );
                it = symbols->find( symn );
                if ( it != symbols->end() )
                {
                    int val = (int) stdata_model->getViewerData(*(*symbols)[symn]);
                    val = ( val + 3 ) * 16 + 6;
                    iacvals[i] = val;
                }
            }

            for ( size_t i = 0; i < symcount; ++i )
            {
                const wxString& sn = symnames[i];
                int val = oc_values[sn];

                wxString conflicts1 = _T(""), conflicts2 = _T("");

                for ( size_t j = 1; j <= 4; ++j )
                {
                    if ( val == iacvals[j] )
                    {
                        conflicts1 += wxString::Format( _T("iac_step_seq_%d "),j );
                        if ( conflicts2 == _T("") )
                            conflicts2 = wxString::Format( _T("iac_step_seq_%d"), j );
                    }
                }

                if ( conflicts2 != _T("") )
                {
                    wxString tmpS = nice_name[i] + _T("|") + conflicts1 + _T("|") + symnames[i] + _T("|") + conflicts2;
                    collisions.Add( tmpS );
                }
            }
        }
    }
    return collisions;
}
