#include "SDImageHandler.h"

#include "logger.h"
#include "events/SDCardEvent.h"
#include "communication/clients/SDCardFileHelper.h"
#include "common/types/wxVariantHash.h"
#include "common/VemsContext.h"
#include "FirmwareInfo.h"

SDImageHandler::SDImageHandler( const wxString& imgname )
{
    m_imageFile = new wxFile( imgname );
    m_imageSize = m_imageFile->SeekEnd( 0 );
    m_evtHandler = NULL;
}

SDImageHandler::~SDImageHandler()
{
    delete m_imageFile;
}

void SDImageHandler::setEvtHandler( wxEvtHandler* evtHandler )
{
    m_evtHandler = evtHandler;
}

void SDImageHandler::searchFiles()
{
    m_stopIdx = getStopIdx();

    //m_size = getDataSize();
    m_size = 69;
    Logger::logf(LOG_MSG, _T("stopidx=%d,ds=%d"), m_stopIdx,m_size);

    SDCardEvent event;

    FirmwareInfo* fwinfo = dynamic_cast<FirmwareInfo*>( VemsContext::getObject( _T("FirmwareInfo") ) );

    FirmwareVersion fv;
    unsigned int fwver_int;
    uint8_t inodeVersion = 0;

    for ( uint32_t blockIdx = 0; ; ++blockIdx )
    {
        uint8_t data[514];

        getMMCBlock( blockIdx, data, 514 );

        if ( SDCardFile::checkSectorIsEmpty( data, 512 ) )
            break;

        if ( blockIdx == 0 )
        {
            if ( !SDCardFile::isVEMSFilesystem( data ) )
            {
                event.SetType( SDCardEventTypes::SEARCHING_FINISHED );
                event.SetInt( SDCardFile::SDFS_NOTVALIDFS );
                m_evtHandler->AddPendingEvent( event );
                return;
            }


            fv = SDCardFile::getFwVersion( data );
            fwver_int = SDCardFile::getFWVersionAsInt( data );
            inodeVersion = SDCardFile::getInodeVersion( data );

            if ( fwinfo )
            {
                int prohibit;
                fwinfo->getProhibit( prohibit );

                if ( fwinfo->getFirmwareVersionAsString() >= _T("1.1.87") )
                {
                    if ( prohibit != fwver_int )
                    {
                        SDCardEvent event;
                        event.SetType( SDCardEventTypes::FIRMWARE_VERSION );
                        event.SetString( fv.toString() );
                        if ( m_evtHandler )
                        {
                            m_evtHandler->AddPendingEvent( event );
                        }
                    }
                }
            }
        }

        switch( inodeVersion )
        {
            case 1:// inode version 1
            {
                for( uint16_t i = 8; i < 512; i += 3 )
                {
                    uint32_t ad = data[i] << 15;
                    ad += data[i+1] << 7;
                    ad += data[i+2] >> 1;

                    Logger::logf(LOG_MSG, _T("address=%d, raw"), ad);

                    if ( ad == 0 )
                        break;

                    m_addrarr.Add( ad );
                    m_rttypearr.Add( SDCardFile::SDLOG_RAW );
                }
            }
            break;
            case 2:// inode version 2
            {
                for( uint16_t i = 8; i < 512; i += 4 )
                {
                    uint32_t ad = data[i] << 15;
                    ad += data[i+1] << 7;
                    ad += data[i+2] >> 1;

                    if ( ad == 0 )
                        break;

                    m_addrarr.Add( ad );

                    m_rttypearr.Add( data[i+3] );//data type : 1: old raw rt data 2: bytestuffed
                }
            }
            break;

            // card is empty
            default:
            {
                LOG
                event.SetType( SDCardEventTypes::SEARCHING_FINISHED );
                event.SetInt( SDCardFile::SDFS_EMPTY );
                m_evtHandler->AddPendingEvent( event );
                return;
            }
        }
    }

    if ( m_addrarr.GetCount() == 0 )
    {
        event.SetType( SDCardEventTypes::SEARCHING_FINISHED );
        event.SetInt( SDCardFile::SDFS_EMPTY );
        m_evtHandler->AddPendingEvent( event );
        return;
    }

    m_mmcdescarr.Clear();

    for ( size_t i = 0; i < m_addrarr.GetCount(); ++i )
    {
        MMCFilesDesc desc;

        desc.DataType = m_rttypearr[i];

        desc.StartAddress = m_addrarr[i];

        if ( i < ( m_addrarr.GetCount() - 1 ) )
        {
            desc.EndAddress = m_addrarr[i + 1] - 1;
        }
        else
        {
            desc.EndAddress = m_stopIdx;
        }

        Logger::logf(LOG_MSG, _T("start ad=%d, end_adr=%d, type=%d"), desc.StartAddress, desc.EndAddress, desc.DataType );

        desc.StopIndex = m_stopIdx;

        switch(desc.DataType)
        {
            case SDCardFile::SDLOG_RAW:
                LOG
                desc.Time = SDCardFile::CalcLogTimeLength( ( desc.EndAddress - desc.StartAddress ) + 1, m_size );
                desc.DataSize = m_size;
                desc.isApprox = false;
                break;

            case SDCardFile::SDLOG_TR_FRAME:
                    //CalcLogTimeLength( desc.EndAddress - desc.StartAddress, 69 );
                    // TODO: fetch the first TR FRAME block and calculate the size from it.
                    //m_state = GET_FIRST_TR_BLOCK;
                // desc.Time = SDCardFile::CalcLogTimeLength( ( desc.EndAddress - desc.StartAddress ) + 1 );
                desc.Time = (size_t)((double)(SDCardFile::CalcLogTimeLength( ( desc.EndAddress - desc.StartAddress ) + 1 )) / 1.166667f);
                desc.isApprox = true;
                break;
        }

        // add to the array
        if ( desc.StartAddress < desc.EndAddress )
        {
            m_mmcdescarr.Add( desc );
        }
    }

    //SDStaticInfo::setDirty( false );

    event.SetType( SDCardEventTypes::SEARCHING_FINISHED );
    event.SetInt( SDCardFile::SDFS_OK );
    event.SetFiles( m_mmcdescarr );

    m_evtHandler->AddPendingEvent( event );
}

void SDImageHandler::downloadSDFile( const MMCFilesDesc& desc )
{
    uint32_t blockIdx = desc.StartAddress;
    uint32_t endIdx = desc.EndAddress;
    uint32_t stopIdx = m_stopIdx;

    if ( endIdx == 0 )
    {
        --endIdx;
    }

    if ( endIdx > stopIdx )
    {
        endIdx = stopIdx;
    }

    uint8_t* data = new uint8_t[( endIdx - blockIdx + 2 ) * 514];
    uint8_t* buffer = new uint8_t[514];
    size_t datalen = 0;

    m_lastSecCountInBlock = -255;

    do {
        double percent = (double)( blockIdx + 1 - desc.StartAddress ) * 100.0 / (double)( desc.EndAddress - desc.StartAddress );

        SDCardEvent event;
        event.SetType( SDCardEventTypes::DOWNLOADING_PROGRESS );
        wxVariant var = percent;
        event.SetVariant( var );
        m_evtHandler->ProcessEvent( event );

        getMMCBlock( blockIdx, buffer, 514 );

        uint8_t* dataPtr = &buffer[0];

        if ( SDCardFile::isDataValid( dataPtr, 514 - 18 ) )
        {
            if ( isNextLog( dataPtr ) )
            {
                if ( datalen > 0 )
                {
                    ///...
                }
                delete[] data;
                data = new uint8_t[( endIdx - blockIdx + 2) * 514];
                datalen = 0;
            }
            memcpy( data + datalen, buffer, 512 );
            datalen += 512;
        }
        ++blockIdx;

        if ( blockIdx > endIdx )
        {
            break;
        }
    } while(1 && !m_aborted);

    delete[] buffer;

    SDCardEvent event;

    if ( m_aborted )
    {
        event.SetType( SDCardEventTypes::INTERRUPTED );
    }
    else
    {
        event.SetType( SDCardEventTypes::DOWNLOADING_FINISHED );
        wxVariant var = (void*)data;
        event.SetVariant( var );
        event.SetInt( datalen );
    }

    m_evtHandler->AddPendingEvent( event );
}

void SDImageHandler::createSDSummary( const MMCFilesDesc& desc )
{
    unsigned char buffer[514+2];

    wxVariantHash* hash = new wxVariantHash;

    uint32_t len = desc.EndAddress - desc.StartAddress;

    uint32_t inc = len / desc.SampleCnt;

    if ( inc == 0 )
        inc = 1;

    (*hash)[_T("Max RPM")] = (long) 0;
    (*hash)[_T("AVG RPM")] = (long) 0;
    (*hash)[_T("Max MAP")] = (long) 0;
    (*hash)[_T("AVG MAP")] = (long) 0;
    (*hash)[_T("ALS On")] = (long) 0;
    (*hash)[_T("Launch On")] = (long) 0;

    unsigned int realSampleCnt = 0;

    for ( uint32_t idx = desc.StartAddress; idx < desc.EndAddress; idx += inc )
    {
        ++realSampleCnt;

        getMMCBlock( idx, buffer, 514 );

        double percent = (double)( idx - desc.StartAddress ) * 100.0 / (double)len;
        wxVariant v = percent;

        SDCardEvent event;
        event.SetType( SDCardEventTypes::SUMMARY_PROGRESS );
        event.SetVariant( v );
        m_evtHandler->ProcessEvent( event );

        long rpm = ( buffer[14] * 256) + buffer[15];
        long map = ( buffer[4] * 256 + buffer[5] ) / 4;

        if ( (*hash)[_T("Max RPM")].GetLong() < rpm )
            (*hash)[_T("Max RPM")] = (long) rpm;

        if ( (*hash)[_T("Max MAP")].GetLong() < map )
            (*hash)[_T("Max MAP")] = (long) map;

        (*hash)[_T("AVG RPM")] = (*hash)[_T("AVG RPM")].GetLong() + rpm;
        (*hash)[_T("AVG MAP")] = (*hash)[_T("AVG MAP")].GetLong() + map;

        if ( (buffer[47] & 128) == 128 )
            (*hash)[_T("ALS On")] = (*hash)[_T("ALS On")].GetLong() + 1;

        if ( (buffer[47] & 64) == 64 )
            (*hash)[_T("Launch On")] = (*hash)[_T("Launch On")].GetLong() + 1;
    }

    (*hash)[_T("Sample count")] = (long)realSampleCnt;

    (*hash)[_T("AVG RPM")] = (double)( (*hash)[_T("AVG RPM")].GetLong() ) / (double) realSampleCnt;
    (*hash)[_T("AVG MAP")] = (double)( (*hash)[_T("AVG MAP")].GetLong() ) / (double) realSampleCnt;
    (*hash)[_T("ALS On")] = (double)( (*hash)[_T("ALS On")].GetLong() ) * 100.0 / (double) realSampleCnt;
    (*hash)[_T("Launch On")] = (double) ( (*hash)[_T("Launch On")].GetLong() ) * 100.0 / (double) realSampleCnt;

    SDCardEvent event;
    event.SetType( SDCardEventTypes::SUMMARY_FINISHED );

    wxVariant var = (void*)hash;
    event.SetVariant( var );

    m_evtHandler->AddPendingEvent( event );
}

void SDImageHandler::Abort()
{
    m_aborted = true;
}

// private methods

uint32_t SDImageHandler::getStopIdx()
{
    uint32_t blockIdx = 0x100;
    uint32_t minIdx = 0x100;

    short count = 0;

    uint32_t mask = 0x01;

    do {
        blockIdx |= mask;
        mask <<= 1;
        ++count;
    } while ( ! CheckSectorIsEmpty( blockIdx ) );

    ++blockIdx;

    mask <<= 1;

    ++count;

    do {
        if ( CheckSectorIsEmpty( blockIdx ) )
            blockIdx &= ~mask;

        mask >>= 1;

        blockIdx |= mask;

        if ( blockIdx < minIdx )
        {
            blockIdx = minIdx;
        }

        --count;
    } while ( count >= 0 );

    return blockIdx;
}

bool SDImageHandler::getMMCBlock( uint32_t blockIdx, uint8_t* data, size_t len )
{
    unsigned int blocknumber = len / 514;
    bool isok = false;

    for ( unsigned int i = 0; i < blocknumber; ++i )
    {
        if ( __getMMCBlock( blockIdx++, data + ( 514 * i ), 512 ) == true )
            isok = true;

        data[512 + (514*i)] = 0xff;
        data[513 + (514*i)] = 0xff;
    }

    return isok;
}

bool SDImageHandler::__getMMCBlock( uint32_t blockIdx, uint8_t* data, size_t len )
{
    uint32_t filepos = blockIdx * 512;

    if ( ( filepos + len ) > m_imageSize )
    {
        memset( data, 0, len );
        return false;
    }

    m_imageFile->Seek( filepos );

    int readB = m_imageFile->Read( data, len );

    if ( readB == (int)len )
    {
        return true;
    }
    else
        return false;
}

short SDImageHandler::getDataSize()
{
    uint8_t data[514+2];
    if ( getMMCBlock( m_stopIdx, data, 514 ) )
    {
        short dataSize = data[510];

        if ( dataSize <= 0 || dataSize == 255 )
            return -2;
        else
            return dataSize;
    }

    return -1;
}

bool SDImageHandler::CheckSectorIsEmpty( uint32_t blockIdx )
{
    uint8_t data[514 + 2];
    size_t sum = 0;

    if ( getMMCBlock( blockIdx, data, 514 ) )
    {
        for ( unsigned i = 0; i < 512; ++i )
            sum += data[i];

        return ( ( sum == 0 ) || ( sum == 512*255 ) );
    }

    return true;
}

bool SDImageHandler::isNextLog( uint8_t* data )
{
    int fscInBlock = data[0];

    if ( ( fscInBlock == m_lastSecCountInBlock ) ||
         ( ( fscInBlock - m_lastSecCountInBlock ) == 1 ) ||
         ( ( fscInBlock == 0 ) && ( m_lastSecCountInBlock == 255 ) ) )
    {
        m_lastSecCountInBlock = fscInBlock;
        return false;
    }

    m_lastSecCountInBlock = fscInBlock;

    return true;
}
