#ifndef WXAGG2_H
#define WXAGG2_H

#include "../common/typedef.h"
#include "../common/types/RealPointArray.h"
#include "../common/types/wxRealRectArray.h"
#include "../wxRealRect.h"

#include "agg_pixfmt_rgb.h"
#include "agg_pixfmt_rgba.h"
#include "agg_pixfmt_gray.h"
#include "agg_pixfmt_amask_adaptor.h"
#include "agg_alpha_mask_u8.h"
#include "agg_rendering_buffer.h"
#include "agg_renderer_base.h"
#include "agg_renderer_scanline.h"
#include "agg_rasterizer_scanline_aa.h"
#include "agg_scanline_p.h"
#include "agg_scanline_u.h"
#include "agg_scanline_bin.h"
#include "agg_math_stroke.h"
#include "agg_conv_contour.h"
#include "agg_pixfmt_rgb_packed.h"
#include "agg_path_storage.h"

#include "wxAggPen.h"
#include "wxAggBrush.h"
#include "wxAggFont.h"
#include "wxAggBlur.h"

typedef agg::pixfmt_rgba32_plain                    pixfmtT;
typedef agg::renderer_base<pixfmtT>                 ren_base;
typedef agg::pod_auto_array< agg::rgba8, 256 >      color_array_type;

class BitmapCanvasWidgetBase;
class wxAggObjectStack;

class wxAgg2
{
    public:
        // initialization
        wxAgg2( uint32_t width, uint32_t height, const wxColor& clearColor = wxColor(0,0,0,0) );
        wxAgg2( const wxSize& size, const wxColor& clearColor = wxColor(0,0,0,0) );
        wxAgg2( wxWindow* win, const wxColor& clearColor = wxColor(0,0,0,0) );
        wxAgg2( BitmapCanvasWidgetBase* bcWidget, const wxColor& clearColor = wxColor(0,0,0,0) );
        wxAgg2( wxBitmap* bitmap );

        virtual ~wxAgg2();

        void        clear( const wxColor& color );

        bool        isOk() const { return m_initialized; }

        // set/get pen, brush
        void        setPen( const wxColour& pencolor );
        void        setPen( const wxPen& pen );
        void        setPen( const wxAggPen& pen );
        void        setTransparentPen();

        void        setPenWidth( double width );
        double      getPenWidth() const;

        void        setBrush( const wxColour& brushcolor);
        void        setBrush( const wxBrush& brush );
        void        setBrush( const wxAggBrush& brush );
        void        setTransparentBrush();

        wxAggPen*   getPen();
        wxAggBrush* getBrush();

        // save and restore brush&pen&font
        void        saveState();
        void        restoreState();

        // blurring - shadow/glow
        void        setBlur( const wxAggBlur& blur );
        wxAggBlur*  getBlur() { return m_blur; }
        void        disableBlur();
        void        setDropShadow( double radius );

        // clipping
        void        setClippingRect( const wxRect& rect ) { m_clippingRect = rect; }
        void        clearClipping() { m_clippingRect = wxDefaultRect; }

        // drawing line
        void        drawLine( double x1, double y1, double x2, double y2 );
        void        drawLine( wxRealPoint point1, wxRealPoint point2 );
        void        drawLine( const RealPointArray& coords, bool closeIt = false  );

        // drawing rectangles
        void        drawRectangle( double x, double y, double width, double height );
        void        drawRectangle( const wxRealRect& rect ) { drawRectangle(rect.x,rect.y,rect.w,rect.h); };
        void        drawRoundedRectangle( double x, double y, double width, double height, double radius );
        void        drawRoundedRectangle( const wxRealRect& rect, double radius );
        void        drawRotatedRectangle( double x, double y, double width, double height, double rot_degree );

        // drawing polygons
        void        drawPolygon( const RealPointArray& points, bool alignPatternToLeft = false );

        // drawing arc
        void        drawArc( double centerx, double centery, double radius, double startAngleRad, double deltaAngleRad );
        void        drawArc( wxRealPoint center, double radius, double startAngleRad, double deltaAngleRad );
        void        drawArc( wxPoint center, double radius, double startAngleRad, double deltaAngleRad );

        // drawing circle, ellipse
        /// TODO
        void        drawEllipse( double centerx, double centery, double r1, double r2 );
        void        drawEllipse( double centerx, double centery, double r1, double r2, double startAngleRad, double deltaAngleRad );

        void        drawCircle( double centerx, double centery, double radius );
        void        drawCircle( wxRealPoint center, double radius );

        void        drawFilledCircleSection( double centerx, double centery, double radius, double startAngleRad, double deltaAngleRad );
        void        drawGradientFilledCircle( const wxColour& begincolor, const wxColour& endcolor, double deg, double centerx, double centery, double radius );

        // font handling
        bool        loadTTFont( const wxString& fontfilename, double height, double weight, double width = 0 );
        void        setFontHeight( double h );
        double      getFontHeight() const;
        void        setFontWidth( double w );
        double      getFontWidth() const;

        // drawing text
        void        drawText( const wxString& text, double x, double y, double w, double h, double s );
        static
            double  getTextWidth( const wxString& text, double h );

        wxRealRect  drawTTFText( double x, double y, const wxString& text, double angleDeg = 0 );
        wxRealRect  drawTTFText( const wxString& text, double x, double y, double angleDeg = 0 );
        wxRealRect  getTTFTextBound( const wxString& text, double angleDeg = 0, double x = 0, double y = 0 );
        wxRealRect  getTTFTextBound( const wxString& text, wxRealRectArray& charRects, double angleDeg = 0, double x = 0, double y = 0 );

        // image drawing
        void        drawBitmap(wxBitmap* bmp, const wxPoint& pos);
        void        drawImage(wxImage* img, const wxPoint& pos);

        // other
        static      wxColor gradient( const wxColor& bc, const wxColor& ec, double k );
        void        drawInnerRectShadow( int width, int alpha, const wxRect& rect, const wxColor& color = wxColor(*wxBLACK) );
        void        drawInnerRectShadow( int width, int alpha, const wxRealRect& rect, const wxColor& color = wxColor(*wxBLACK) );

        // render to DC, image, file
        void        render( wxDC& dc, unsigned int xpos = 0, unsigned int ypos = 0 );
        wxBitmap*   renderToBitmap(bool isSupportAlphaChannel = true, const wxRect& renderBox = wxDefaultRect);
        wxImage*    renderToImage(bool isSupportAlphaChannel = true);
        wxImage*    renderCopyToImage(bool isSupportAlphaChannel = true);
        bool        renderToFile( const wxString& imgname, long bitmaptype = wxBITMAP_TYPE_PNG );


        wxSize      getSize() { return wxSize(m_width,m_height); };

        // testing
        void        aggTest1();
        void        aggTest2();

    protected:

    private:
        bool                            m_initialized;

        wxAggObjectStack*               m_states;

        uint32_t                        m_width;
        uint32_t                        m_height;
        unsigned char*                  m_buffer;

        unsigned char*                  m_imgBuffer;
        unsigned char*                  m_alphaBuffer;

        agg::rasterizer_scanline_aa<>   m_ras;
        agg::scanline_p8                m_sl;
        agg::rendering_buffer*          m_rb;
        ren_base*                       m_renderer;
        pixfmtT*                        m_pixf;

        wxAggPen*                       m_pen;
        wxAggBrush*                     m_brush;
        wxAggFont*                      m_font;
        wxAggBlur*                      m_blur;

        wxRealRect                      m_clippingRect;

        // pattern filling
        agg::int8u*                     m_patternBuffer;
        agg::rendering_buffer           m_pattern_rbuf;
        unsigned int                    m_pattern_w, m_pattern_h;

        wxBitmap*                       m_ownBitmap;

    private:
        void        Create( uint32_t width, uint32_t height, const wxColor& clearColor = wxColor(0,0,0,0) );
        void        getBuffers(const wxRect& renderBox, bool isSupportAlphaChannel = true);
        void        getBuffers(unsigned char* rgb_buffer, unsigned char* a_buffer = NULL);
        void        applyClipping();
        void        generatePattern();
        void        fillColorArray( color_array_type& array, agg::rgba8 begin, agg::rgba8 end );

        template<class T>
        void        drawOutlines(T path);

        void        gradient_rendering( double centerx, double centery, double r1, double r2 );
};

#endif // WXAGG2_H
