#include "LEDDisplay.h"

#include "../agg/wxAgg2.h"
#include "../logger.h"

String2IntHash LEDDisplay::m_charTable;

LEDDisplay::LEDDisplay()
{
    m_digitHeight = 24;
    m_activecolor = wxColor(  226, 73, 61, 224  );
    m_inactivecolor = wxColor( 20, 20, 20, 224 );
    m_inclination = 8;
    init();
}

LEDDisplay::LEDDisplay( double height, int ledcolor )
{
    m_digitHeight = height;
    m_inclination = 8;

    switch ( ledcolor )
    {
        case GREEN_LED:
            m_activecolor = wxColor( 173, 252, 56, 224 );
            m_inactivecolor = wxColor( 35, 35, 35, 224 );
            break;

        case RED_LED:
        default:
            m_activecolor = wxColor( 226, 73, 61, 224 );
            m_inactivecolor = wxColor( 20, 20, 20,224 );
            break;
    }

    init();
}

LEDDisplay::LEDDisplay( double height, const wxColor& activesegmentcolor, double inclinationDeg, const wxColor& inactivesegmentcolor ) :
    m_digitHeight( height ), m_activecolor( activesegmentcolor ), m_inactivecolor( inactivesegmentcolor ), m_inclination( inclinationDeg )
{
    init();
}

void LEDDisplay::init()
{
    m_canvas = NULL;
    m_xpos = m_ypos = 0;

    if ( m_inclination < 0 )
        m_inclination = 8;

    m_regions = new wxPoint*[8];

    int segmentsizes[] = { 4, 5, 5, 4, 5, 5, 6, 4 };
    for ( size_t i = 0; i < 8; ++i )
    {
        m_regions[i] = new wxPoint[segmentsizes[i]];
    }

    if ( m_charTable.size() == 0 )
        fillCharTable();

    calcCoordinates();
}

LEDDisplay::~LEDDisplay()
{
    for ( size_t i = 0; i < 8; ++i )
        delete[] m_regions[i];

    delete[] m_regions;
}

void LEDDisplay::setCanvas( wxAgg2* canvas )
{
    m_canvas = canvas;
}

wxRealRect LEDDisplay::getLEDSize() const
{
    wxRealRect r;
    r.x = m_xs; r.y = m_ys;
    r.w = m_xpos - m_xs;
    r.h = m_digitHeight;

    return r;
}

void LEDDisplay::draw( unsigned char segments,  double x, double y )
{
    m_xs = m_xpos = x;
    m_ys = m_ypos = y;

    drawSegments( segments, true );

    m_xpos += m_digitwidth;
}


void LEDDisplay::drawText( const wxString& str, double x, double y )
{
    m_canvas->saveState();

    m_xs = m_xpos = x;
    m_ys = m_ypos = y;

    for ( size_t i = 0; i < str.Length(); ++i )
    {
        wxString ch = str[i];

        String2IntHash::iterator it = m_charTable.find( ch );

        if ( it == m_charTable.end() )
            continue;

        int segments = it->second;

        // check if dot should be lightened out

        if ( i < str.Length() - 1 )
        {
            wxChar c2 = str[i+1];
            if ( c2 == '.' )
                segments |= 128;
        }

        drawSegments( segments, true );

        m_xpos += m_digitwidth;
    }

    m_canvas->restoreState();
}

void LEDDisplay::drawText( const wxChar& ch, double x, double y )
{
    m_xs = m_xpos = x;
    m_ys = m_ypos = y;

    String2IntHash::iterator it = m_charTable.find( ch );

    if ( it == m_charTable.end() )
        return;

    int segments = it->second;

    drawSegments( segments, true );

    m_xpos += m_digitwidth;
}

void LEDDisplay::drawSegments( unsigned char segments, bool showDot )
{
    if ( m_canvas == NULL )
        return;

    m_canvas->setBrush( wxDefaultAggBrush );
    m_canvas->setPen( wxAGGPEN_TRANSPARENT );
    RealPointArray pts;

    wxRealPoint tr( m_xpos, m_ypos );

    // draw 'segment a'
    wxColor c = ( segments & 1 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmenta, 4, tr );
    convertAndAddRegion( pts, 0 );
    m_canvas->drawPolygon( pts );

    // draw 'segment b'
    c = ( segments & 2 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmentb, 5, tr );
    convertAndAddRegion( pts, 1 );
    m_canvas->drawPolygon( pts );

    // draw 'segment c'
    c = ( segments & 4 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmentc, 5, tr );
    convertAndAddRegion( pts, 2 );
    m_canvas->drawPolygon( pts );

    // draw 'segment d'
    c = ( segments & 8 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmentd, 4, tr );
    convertAndAddRegion( pts, 3 );
    m_canvas->drawPolygon( pts );

    // draw 'segment e'
    c = ( segments & 16 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmente, 5, tr );
    convertAndAddRegion( pts, 4 );
    m_canvas->drawPolygon( pts );

    // draw 'segment f'
    c = ( segments & 32 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmentf, 5, tr );
    convertAndAddRegion( pts, 5 );
    m_canvas->drawPolygon( pts );

    // draw 'segment g'
    c = ( segments & 64 ) ? m_activecolor : m_inactivecolor;
    m_canvas->setBrush( c );
    translateSegment( pts, m_segmentg, 6, tr );
    convertAndAddRegion( pts, 6 );
    m_canvas->drawPolygon( pts );

    // draw 'segment dot'
    if ( showDot )
    {
        c = ( segments & 128 ) ? m_activecolor : m_inactivecolor;
        m_canvas->setBrush( c );
        m_canvas->drawCircle( m_segmentdot + tr, m_dotradius );

        m_regions[7][0].x = (int)(m_segmentdot.x + tr.x - m_dotradius ); m_regions[7][0].y = (int)( m_segmentdot.y + tr.y - m_dotradius );
        m_regions[7][1].x = (int)(m_segmentdot.x + tr.x + m_dotradius ); m_regions[7][1].y = (int)( m_segmentdot.y + tr.y - m_dotradius );
        m_regions[7][2].x = (int)(m_segmentdot.x + tr.x + m_dotradius ); m_regions[7][2].y = (int)( m_segmentdot.y + tr.y + m_dotradius );
        m_regions[7][3].x = (int)(m_segmentdot.x + tr.x - m_dotradius ); m_regions[7][3].y = (int)( m_segmentdot.y + tr.y + m_dotradius );
    }
}

void LEDDisplay::convertAndAddRegion( RealPointArray& pts, unsigned int segment )
{
    size_t count = pts.GetCount();

    for ( size_t i = 0; i < count; ++i )
    {
        m_regions[segment][i].x = (int)pts[i].x;
        m_regions[segment][i].y = (int)pts[i].y;
    }
}

void LEDDisplay::translateSegment( RealPointArray& pts, wxRealPoint* segment, int pointnum, wxRealPoint& transl )
{
    pts.Clear();
    for ( int i = 0; i < pointnum; ++i )
    {
        wxRealPoint pt; pt.x = segment[i].x + transl.x; pt.y = segment[i].y + transl.y;
        pts.Add( pt );
    }
}

void LEDDisplay::calcCoordinates()
{
    double gap = m_digitHeight * 0.07;
    double width = m_digitHeight * 0.6;

    double segmentw = m_digitHeight * 0.1;

    double rad = m_inclination / 180.0 * M_PI;

    // segment "a"
    m_segmenta[0].x = segmentw/2; m_segmenta[0].y = 0;
    m_segmenta[1].x = width - segmentw/2; m_segmenta[1].y = 0;
    m_segmenta[2].x = width - 2 * gap; m_segmenta[2].y = segmentw;
    m_segmenta[3].x = 2 * gap; m_segmenta[3].y = segmentw;

    for ( size_t i = 0; i < 4; ++i ) m_segmenta[i].x += inclineTr( rad, m_digitHeight - m_segmenta[i].y );

    // segment "b"
    m_segmentb[0].x = width; m_segmentb[0].y = gap;
    m_segmentb[1].x = width; m_segmentb[1].y  = m_digitHeight / 2 - segmentw / 2;
    m_segmentb[2].x = width - segmentw / 2; m_segmentb[2].y = m_digitHeight / 2 - gap/6;
    m_segmentb[3].x = width - segmentw; m_segmentb[3].y = m_digitHeight / 2 - segmentw / 2;
    m_segmentb[4].x = width - segmentw; m_segmentb[4].y = segmentw + gap;

    for ( size_t i = 0; i < 5; ++i ) m_segmentb[i].x += inclineTr( rad, m_digitHeight - m_segmentb[i].y );

    // segment "f"
    m_segmentf[0].x = 0; m_segmentf[0].y = gap;
    m_segmentf[1].x = segmentw; m_segmentf[1].y  = segmentw + gap;
    m_segmentf[2].x = segmentw; m_segmentf[2].y = m_digitHeight / 2 - segmentw / 2;
    m_segmentf[3].x = segmentw / 2; m_segmentf[3].y = m_digitHeight / 2 - gap / 6;
    m_segmentf[4].x = 0; m_segmentf[4].y = m_digitHeight / 2 - segmentw / 2;

    for ( size_t i = 0; i < 5; ++i ) m_segmentf[i].x += inclineTr( rad, m_digitHeight - m_segmentf[i].y );

    // segment "c"
    m_segmentc[0].x = width - segmentw; m_segmentc[0].y = m_digitHeight / 2 + segmentw / 2;
    m_segmentc[1].x = width - segmentw / 2; m_segmentc[1].y = m_digitHeight / 2 + gap / 6;
    m_segmentc[2].x = width; m_segmentc[2].y = m_digitHeight / 2 + segmentw / 2;
    m_segmentc[3].x = width; m_segmentc[3].y  = m_digitHeight - gap;
    m_segmentc[4].x = width - segmentw; m_segmentc[4].y = m_digitHeight - segmentw - gap;

    for ( size_t i = 0; i < 5; ++i ) m_segmentc[i].x += inclineTr( rad, m_digitHeight - m_segmentc[i].y );

    // segment "e"
    m_segmente[0].x = 0; m_segmente[0].y = m_digitHeight / 2 + segmentw / 2;
    m_segmente[1].x = segmentw / 2; m_segmente[1].y = m_digitHeight / 2 + gap / 6;
    m_segmente[2].x = segmentw; m_segmente[2].y = m_digitHeight / 2 + segmentw / 2;
    m_segmente[3].x = segmentw; m_segmente[3].y  = m_digitHeight - segmentw - gap;
    m_segmente[4].x = 0; m_segmente[4].y = m_digitHeight - gap;

    for ( size_t i = 0; i < 5; ++i ) m_segmente[i].x += inclineTr( rad, m_digitHeight - m_segmente[i].y );

    // segment "d"
    m_segmentd[0].x = 2 * gap; m_segmentd[0].y = m_digitHeight - segmentw;
    m_segmentd[1].x = width - 2* gap; m_segmentd[1].y = m_digitHeight - segmentw;
    m_segmentd[2].x = width - segmentw / 2; m_segmentd[2].y = m_digitHeight;
    m_segmentd[3].x = segmentw/2; m_segmentd[3].y = m_digitHeight;

    for ( size_t i = 0; i < 4; ++i ) m_segmentd[i].x += inclineTr( rad, m_digitHeight - m_segmentd[i].y );

    // segment "g"
    m_segmentg[0].x = segmentw + gap; m_segmentg[0].y = m_digitHeight / 2 - segmentw / 2;
    m_segmentg[1].x = width - segmentw - gap; m_segmentg[1].y = m_digitHeight / 2 - segmentw / 2;
    m_segmentg[2].x = width - segmentw / 2 - gap; m_segmentg[2].y = m_digitHeight / 2;
    m_segmentg[3].x = width - segmentw - gap; m_segmentg[3].y = m_digitHeight / 2 + segmentw / 2;
    m_segmentg[4].x = segmentw + gap; m_segmentg[4].y = m_digitHeight / 2 + segmentw / 2;
    m_segmentg[5].x = segmentw / 2 + gap; m_segmentg[5].y = m_digitHeight / 2;

    for ( size_t i = 0; i < 6; ++i ) m_segmentg[i].x += inclineTr( rad, m_digitHeight - m_segmentg[i].y );

    m_dotradius = segmentw / 2.0;
    m_segmentdot.x = width + m_dotradius * 3;
    m_segmentdot.y = m_digitHeight - m_dotradius;

    m_digitwidth = m_segmenta[1].x + 4 * m_dotradius;
}

void LEDDisplay::fillCharTable()
{
    int a = 1, b = 2, c = 4, d = 8, e = 16, f = 32, g = 64;
    m_charTable[_T(" ")] = 0;
    m_charTable[_T("0")] = a + b + c + d + e + f;
    m_charTable[_T("1")] = b + c;
    m_charTable[_T("2")] = a + b + g + e + d;
    m_charTable[_T("3")] = a + b + g + c + d;
    m_charTable[_T("4")] = f + g + b + c;
    m_charTable[_T("5")] = a + f + g + c + d;
    m_charTable[_T("6")] = a + f + g + e + c + d;
    m_charTable[_T("7")] = a + b + c;
    m_charTable[_T("8")] = a + b + c + d + e + f + g;
    m_charTable[_T("9")] = a + b + c + d + f + g;
    m_charTable[_T("-")] = g;

    m_charTable[_T("a")] = m_charTable[_T("A")] = a + f + b + e + g + c;
    m_charTable[_T("b")] = m_charTable[_T("B")] = f + g + e + d + c;
    m_charTable[_T("c")] = g + e + d;
    m_charTable[_T("C")] = a + f + e + d;
    m_charTable[_T("d")] = m_charTable[_T("D")] = b + g + e + c + d;
    m_charTable[_T("e")] = m_charTable[_T("E")] = a + f + g + e + d;
    m_charTable[_T("f")] = m_charTable[_T("F")] = a + g + e;
    m_charTable[_T("U")] = f + e + d + c + b;
    m_charTable[_T("u")] = e + d + c;
}

double LEDDisplay::inclineTr( double angleRad, double h )
{
    return h * cos ( M_PI / 2 - angleRad );
}
