#include "RoundDemoModeSettings.h"

#include <wx/dcbuffer.h>
#include <wx/statline.h>

#include "PolygonHelper.h"

#include "../agg/wxAgg2.h"
#include "../DirHelper.h"
#include "../logger.h"
#include "../vemsdialog/ConfigletPanel.h"

/// LED CLICK EVENT
DEFINE_EVENT_TYPE( wxEVT_LEDCLICK )
IMPLEMENT_DYNAMIC_CLASS( LedClickEvent, wxEvent )
/// END LED CLICK EVENT


/// GLOBAL EVENT IDs
int idLEDClick          = wxNewId();
int idDisableDemoMode   = wxNewId();


/// ----------------------------------------------------------------
/// ROUND DEMO MODE LED DISPLAY

BEGIN_EVENT_TABLE( RoundDemoModeLedDisplay, wxPanel )
    EVT_PAINT( RoundDemoModeLedDisplay::OnPaint )
    EVT_LEFT_DOWN( RoundDemoModeLedDisplay::OnClick )
END_EVENT_TABLE()

/********************************************

RoundDemoModeLedDisplay class

Displays a round display ( 3 + 4 leds )

 @parent: the parent window AND the event handler (click events will be sent to)

 @topcolor:  RED_LED, GREEN_LED etc (see LEDDisplay.h)
 @bottomcolor: RED_LED, GREEN_LED etc (see LEDDisplay.h)

*********************************************/

RoundDemoModeLedDisplay::RoundDemoModeLedDisplay( wxWindow* parent, int id, const wxPoint& pos, const wxSize& size, int topcolor, int bottomcolor ) :
    wxPanel( parent, id, pos, size )
{
    m_topcolor = topcolor;
    m_bottomcolor = bottomcolor;

    for ( size_t i = 0; i < 7; ++i )
    {
        m_ledvalues[i] = 0;
        m_leds[i] = NULL;
    }

    m_evtHandler = (wxEvtHandler*) parent;

    Refresh();
}

RoundDemoModeLedDisplay::~RoundDemoModeLedDisplay()
{
    for ( size_t i = 0; i < 7; ++i )
    {
        if ( m_leds[i] != NULL )
        {
            delete m_leds[i];
            m_leds[i] = NULL;

        }
    }
}

/// redraws display panel
void RoundDemoModeLedDisplay::redraw()
{
    Refresh( false );
}

/// set segments for each LED
/// @display: LEDs
///   top row :  0 1 2
///   bottom  : 3 4 5 6
/// @value: segment value (see LEDDisplay.h for details)
void RoundDemoModeLedDisplay::setSegments( int display, int value )
{
    m_ledvalues[display] = value;
}

/// drawing to dc with agg
void RoundDemoModeLedDisplay::OnPaint( wxPaintEvent& event )
{
    for ( size_t i = 0; i < 7; ++i )
        if ( m_leds[i] != NULL )
            delete m_leds[i];

    wxBufferedPaintDC dc( this );
    dc.SetBackground( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    dc.Clear();

    wxAgg2 canvas( this );

    wxRealSize r = GetSize();

    double h = r.h;

    //Logger::logf(LOG_MSG, _T("OnPaint w=%.1f h=%.1f"),r.w, r.h);

    double cy = h / 2, cx = r.w / 2;

    double roundsize = h * 0.86;

    canvas.setBrush( wxDefaultAggBrush );
    canvas.setBrush( wxColor( 130, 130, 130 ) );
    canvas.setPen( wxColor( 130, 130, 130 ) );
    canvas.drawCircle( cx, cy, roundsize / 2 );

    canvas.setBrush( wxColor( 0,0,0 ) );
    canvas.setPen( wxColor( 130,130,130 ) );
    canvas.drawCircle( cx, cy, cy * 0.83 );

    // top
    double x = cx - 80 * 1.2;
    for ( size_t i = 0; i < 3; ++i )
    {
        if ( m_ledvalues[i] < 0 )
        {
            // hide it ( forced output )
            m_leds[i] = new LEDDisplay( 80, wxColor(0,0,0,0), 8, wxColor(0,0,0,0) );
        }
        else
        {
            m_leds[i] = new LEDDisplay( 80, m_topcolor );
        }

        if ( canvas.isOk() )
            m_leds[i]->setCanvas( &canvas );

        m_leds[i]->draw( m_ledvalues[i], x, cy - 100 );

        x += m_leds[i]->getLEDSize().w;

        m_regions[i] = m_leds[i]->getSegmentRegions();
    }

    // bottom
    x = cx - 65 * 1.67;
    for ( size_t i = 3; i < 7; ++i )
    {
        if ( m_ledvalues[i] < 0 )
        {
            // hide it ( forced output )
            m_leds[i] = new LEDDisplay( 65, wxColor(0,0,0,0), 8, wxColor(0,0,0,0) );
        }
        else
        {
            m_leds[i] = new LEDDisplay( 65, m_bottomcolor );
        }

        m_leds[i]->setCanvas( &canvas );

        m_leds[i]->draw( m_ledvalues[i], x, cy + 5 );

        x += m_leds[i]->getLEDSize().w;

        m_regions[i] = m_leds[i]->getSegmentRegions();
    }

    // button
    wxAggBrush* brush = canvas.getBrush();
    brush->SetStyle( wxAGGBRUSH_GRADIENT );
    brush->SetEndColor( wxColor( 30, 30, 60 ) );
    brush->SetColor( wxColor( 255, 255, 255 ) );
    brush->SetGradientOrientation( 45 );
    canvas.setTransparentPen();

    canvas.drawCircle( cx, cy + 105, 16 );

    // finally render it to dc
    canvas.render( dc );
}

/// send LEDClickEvent when click occured on a segment
void RoundDemoModeLedDisplay::OnClick( wxMouseEvent& event )
{
    for ( size_t i = 0; i < 7; ++i )
    {
        // it is invalid because the output of this LED is forced
        if ( m_ledvalues[i] < 0 )
            continue;

        // check which led, which segment was clicked on
        wxPoint** segments = m_regions[i];

        for ( size_t j = 0; j < 8; ++j )
        {
            int segsize[] = { 4, 5, 5, 4, 5, 5, 6, 4 };
            wxPoint* region = segments[j];
            if ( PolygonHelper::PtInPolygon( region, segsize[j], event.GetPosition() ) )
            {
                LedClickEvent evt( idLEDClick );

                // set the lednumber, and segmentnumber (see LEDDisplay.h)
                evt.lednumber = i;
                evt.segmentnumber = j;

                // send event
                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent( evt );
            }
        }
    }
}

///---------------------------------------------------------------------------

BEGIN_EVENT_TABLE( RoundDemoModeSettings, wxPanel )
    EVT_LEDCLICK( wxID_ANY, RoundDemoModeSettings::OnLEDClick )
    EVT_BUTTON( idDisableDemoMode, RoundDemoModeSettings::OnButton )
    EVT_CLOSE( RoundDemoModeSettings::OnClose )
    //EVT_SIZE( RoundDemoModeSettings::OnSize )
END_EVENT_TABLE()

RoundDemoModeSettings::RoundDemoModeSettings( wxWindow* parent, STDataMgrBase* stMgr, SymbolTable* symbols, SymbolMgr* symbolMgr ) : wxPanel( parent, wxID_ANY )
{
    m_stDataMgr = stMgr;
    m_symbols = symbols;
    m_symbolMgr = symbolMgr;

    m_statvcontroller = new StaticViewersController( parent, _T("Display Demo") );
    m_statvcontroller->setController( m_stDataMgr );
    m_statvcontroller->addViewer( this );

    buildGUI( parent );

    SetDataMgr( m_stDataMgr );
    m_stDataMgr->RegisterViewer( this );

    update();
}

RoundDemoModeSettings::~RoundDemoModeSettings()
{
    unregisterViewer();


    if ( m_statvcontroller )
    {
        m_statvcontroller->removeViewer(this);
        delete m_statvcontroller;
    }

    if ( m_stDataMgr )
        m_stDataMgr->UnRegisterViewer( this );
}


void RoundDemoModeSettings::buildGUI( wxWindow* parent )
{
    SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    ClearBackground();

    wxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );

    wxSizer* cbsizer = new wxBoxSizer( wxHORIZONTAL );

    wxStaticText* txt = new wxStaticText( this, wxID_ANY, _T("Force outputs to:") );

    cbsizer->Add( txt, 0, wxALL, 3 );

    for ( size_t i = 0; i < 7; ++i )
    {
        int id = wxNewId();

        wxString label;
        if ( i < 3 )
            label = wxString::Format( _T("Top %d"), i );
        else
            label = wxString::Format( _T("Bottom %d"), i - 3 );

        m_forceOutput[i] = new wxCheckBox( this, id, label );
        m_forceOutput[i]->Connect( id, wxEVT_COMMAND_CHECKBOX_CLICKED, wxCommandEventHandler( RoundDemoModeSettings::OnCheckBox ), 0, this );

        m_foIds[i] = id;

        cbsizer->Add( m_forceOutput[i], 0, wxALL, 3 );
    }



    wxButton* resetButton = new wxButton( this, idDisableDemoMode, _T("Disable demo mode (reset values)") );

    m_displayPanel = new RoundDemoModeLedDisplay( this, -1, wxDefaultPosition, wxSize( -1, 400 ), LEDDisplay::RED_LED, LEDDisplay::GREEN_LED );

    mainsizer->Add( cbsizer, 0, wxEXPAND | wxALL | wxALIGN_TOP, 3 );

    txt = new wxStaticText( this, wxID_ANY, _T("Note: you can change the settings of an output only when the output is forced! (the LED is visible only in that case)") );

    mainsizer->Add( txt, 0, wxEXPAND | wxALL, 5 );

    mainsizer->Add( new wxStaticLine( this, -1 ), 0, wxALL | wxEXPAND, 4 );

    mainsizer->Add( resetButton, 0, wxALL, 4 );

    mainsizer->Add( m_displayPanel, 0, wxEXPAND | wxALL | wxALIGN_TOP, 3 );

    addConfigletPanel( mainsizer );

    m_statcpanel = new StaticControlPanel( this, m_statvcontroller );

    mainsizer->Add( m_statcpanel, 0, wxEXPAND | wxALL, 10 );

    SetSizer( mainsizer );
    mainsizer->Fit( this );
    mainsizer->SetSizeHints( this );
}

void RoundDemoModeSettings::addConfigletPanel( wxSizer* sizer )
{
    wxString configletDir = DirHelper::GetPath( dCONFIGLETS ) + _T("ledDisplayDemo") + DirHelper::Separ();

    if ( !::wxDirExists( configletDir ) ) return;

    ConfigletPanel* configletPanel = new ConfigletPanel( this, configletDir, m_symbolMgr );
    m_statvcontroller->addViewer( configletPanel );

    sizer->Add( configletPanel, 0, wxEXPAND );
}

int RoundDemoModeSettings::segmentMapper( int value, bool isTranslationA )
{
    int retval = 0;

    if ( isTranslationA )
    {
        if ( value & 1 ) retval += 64;
        if ( value & 2 ) retval += 32;
        if ( value & 4 ) retval += 1;
        if ( value & 8 ) retval += 2;
        if ( value & 16 ) retval += 128;
        if ( value & 32 ) retval += 4;
        if ( value & 64 ) retval += 8;
        if ( value & 128 ) retval += 16;
    }
    else
    {
        if ( value & 1 ) retval += 2;
        if ( value & 2 ) retval += 1;
        if ( value & 4 ) retval += 32;
        if ( value & 8 ) retval += 64;
        if ( value & 16 ) retval += 16;
        if ( value & 32 ) retval += 8;
        if ( value & 64 ) retval += 4;
        if ( value & 128 ) retval += 128;
    }

    return retval;
}

int RoundDemoModeSettings::reverseSegmentMapper( int value, bool isTranslationA )
{
    int segmentTrA[8] = { 4, 8, 32, 64, 128, 2, 1, 16 };
    int segmentTrB[8] = { 2, 1, 64, 32, 16, 4, 8, 128 };

    if ( isTranslationA )
        return segmentTrA[value];
    else
        return segmentTrB[value];
}

void RoundDemoModeSettings::update()
{
    wxString symName = _T("display_mask");

    Symbol* sym = NULL;

    SymbolTable::iterator it = m_symbols->find( symName );

    if ( it != m_symbols->end() )
    {
        sym = it->second;

        if ( sym )
        {
            m_demomodeVal = (int)m_stDataMgr->getDataFromViewerModel( *sym, 0, 0 );
        }

        if ( m_demomodeVal > 127 )
            m_demomodeVal -= 128;

        // check which output is forced
        for ( size_t i = 0; i < 7; ++i )
        {
            int out = 1 << i;
            if ( m_demomodeVal & out )
            {
                m_forceOutput[i]->SetValue( false );
                m_displayPanel->setSegments( i, -1 );
            }
            else
            {
                symName = wxString::Format( _T("segment%d"), i );

                it = m_symbols->find( symName );

                int val = 0;

                if ( it != m_symbols->end() )
                {
                    sym = it->second;

                    if ( sym )
                        val = (int)m_stDataMgr->getDataFromViewerModel( *sym, 0, 0 );

                    // "A" translation
                    if ( i % 2 == 0)
                        val = segmentMapper( val, true );
                    // "B" translation
                    else
                        val = segmentMapper( val, false );
                }

                m_forceOutput[i]->SetValue( true );
                m_displayPanel->setSegments( i, val );
            }
        }

        m_displayPanel->redraw();
    }
}

void RoundDemoModeSettings::updateModel()
{
    // nop
}

void RoundDemoModeSettings::requestWrite()
{
    // nop
}

void RoundDemoModeSettings::OnLEDClick( LedClickEvent& event )
{
    if ( event.GetId() == idLEDClick )
    {
        int lednumber = event.lednumber;
        int segment = event.segmentnumber;

        int segmentvalue = reverseSegmentMapper( segment, ( lednumber % 2 == 0 ) );

        wxString symName = wxString::Format( _T("segment%d"), lednumber );

        SymbolTable::iterator it = m_symbols->find( symName );

        if ( it != m_symbols->end() )
        {
            Symbol* sym = it->second;

            if ( sym )
            {
                int val = (int)m_stDataMgr->getDataFromViewerModel( *sym, 0, 0 );
                val ^= segmentvalue;

                m_stDataMgr->setModelData( *sym, 0, 0, (double)val );

                m_stDataMgr->requestWrite( *sym );

                m_stDataMgr->updateRegisteredViewers();
            }
        }
    }
}

void RoundDemoModeSettings::OnCheckBox( wxCommandEvent& event )
{
    int id = event.GetId();

    for ( size_t i = 0; i < 7; ++i )
    {
        if ( m_foIds[i] == id )
        {
            int val = 1 << i;
            m_demomodeVal ^= val;

            wxString symName = _T("display_mask");

            Symbol* sym = NULL;

            SymbolTable::iterator it = m_symbols->find( symName );

            if ( it != m_symbols->end() )
            {
                sym = it->second;
            }
            if ( sym )
            {
                m_stDataMgr->setModelData( *sym, 0, 0, m_demomodeVal );

                m_stDataMgr->requestWrite( *sym );

                m_stDataMgr->updateRegisteredViewers();

                return;
            }
        }
    }
}

void RoundDemoModeSettings::OnButton( wxCommandEvent& event )
{
    int id = event.GetId();

    if ( id == idDisableDemoMode )
    {
        if ( m_stDataMgr == NULL )
        {
            event.Skip();
            return;
        }

        wxString symName = _T("display_mask");

        Symbol* sym = NULL;

        SymbolTable::iterator it = m_symbols->find( symName );

        if ( it != m_symbols->end() )
        {
            sym = it->second;
        }
        if ( sym )
        {
            m_stDataMgr->setModelData( *sym, 0, 0, 255.0 );
            m_stDataMgr->requestWrite( *sym );
        }

        for ( size_t i = 0; i < 7; ++i )
        {
            symName = wxString::Format( _T("segment%d"), i );

            it = m_symbols->find( symName );

            sym = NULL;

            if ( it != m_symbols->end() )
                sym = it->second;

            if ( sym )
            {
                m_stDataMgr->setModelData( *sym, 0,0, 255.0 );
                m_stDataMgr->requestWrite( *sym );
            }

            m_stDataMgr->updateRegisteredViewers();
        }

    }
    else
        event.Skip();
}

void RoundDemoModeSettings::OnClose( wxCloseEvent& event )
{
    if ( GetParent() )
        GetParent()->Close();
}
