#include "ProcessBase.h"

#include "../ringbuffer/CommunicationThread.h"
#include "../logger.h"

/**
More detail:
http://docs.wxwidgets.org/trunk/classwx_thread.html
*/

int ProcessBase::s_threadCount = 0;

ProcessBase::ProcessBase()
{
    //ctor
    m_isDisabled = false;
    m_pThread = NULL;
}

ProcessBase::~ProcessBase()
{
    //dtor
    if(isRunning())
        kill();
}

int ProcessBase::GetPID()
{
    wxCriticalSectionLocker enter(m_pThreadCS);
    if(!m_pThread) return -1;
    return m_pThread->GetCurrentId();
}

void ProcessBase::start(int priority)
{
    if(m_pThread) return;
    if( isDisabled() ) return;
    reset();
    //request( new EnableReadRealTimeCommand( true ) );
    m_pThread = new CommunicationThread( this );

    if( m_pThread->Create() != wxTHREAD_NO_ERROR )
    {
        Logger::logf(LOG_MSG, _T("ProcessBase::start() Can't create thread!"));
        /// wx doc cannot tell how to handle this case

        ///delete m_pThread;   /// it sets m_pThread to NULL as well but dunno it is safe or not!!

        /// safe but cause little memory leak
        m_pThread = NULL;

        return;
    }

    m_pThread->SetPriority(priority);
    m_pThread->Run();
    //Logger::logf(LOG_MSG, _T("ProcessBase::start() FINISHED!"));

}

void ProcessBase::kill()
{
    {
        wxCriticalSectionLocker enter(m_pThreadCS);
        if ( !m_pThread ){
            Logger::logf(LOG_MSG,_T("ProcessBase::kill() thread already stopped!"));
            return; /// already stopped
        }
        if(m_pThread->Kill() != wxTHREAD_NO_ERROR) {
            Logger::logf(LOG_MSG,_T("ProcessBase::kill() Can't kill the thread!"));
            m_pThread = NULL;
            return;
        }
    }

            Logger::logf(LOG_MSG,_T("ProcessBase::kill() thread killed"));

            while (1) {
                Logger::logf(LOG_MSG, _T("ProcessBase::kill() WAIT TO thread destroy"));
        {
                wxCriticalSectionLocker enter(m_pThreadCS);
                if(!m_pThread) {
                    Logger::logf(LOG_MSG,_T("ProcessBase::kill() thread destroyed"));
                    break;
                }
        }
                wxThread::This()->Sleep(1);
            }
    reset();
}

void ProcessBase::stop()
{
    Logger::logf(LOG_MSG,_T("ProcessBase::stop() STARTED!"));
    {
        wxCriticalSectionLocker enter(m_pThreadCS);
        if ( !m_pThread ){
            Logger::logf(LOG_MSG,_T("ProcessBase::stop() thread already stopped!"));
            return; /// already stopped
        }
//        m_pThread->stop();   /// release onWait()
         Logger::logf(LOG_MSG,_T("ProcessBase::stop() TRY to delete!"));
        if (m_pThread->Delete() != wxTHREAD_NO_ERROR ){ /// try to stop;
            Logger::logf(LOG_MSG,_T("ProcessBase::stop() Can't delete the thread!"));
        }
        Logger::logf(LOG_MSG,_T("ProcessBase::stop() DELETED!"));
    }

    int waitMs = 0;
    while (1)
    {
        { // was the ~MyThread() function executed?
            wxCriticalSectionLocker enter(m_pThreadCS);
            if (!m_pThread) break;
        }

        // wait for thread completion
        Logger::logf(LOG_MSG, _T("ProcessBase::stop() WAIT TO thread destroy"));

        if(++waitMs > 100 ) {
            Logger::logf(LOG_MSG, _T("ProcessBase::stop() KILL the thread!"));
            kill();
            break;
        }else{
             wxThread::This()->Sleep(1);
        }

    }
    Logger::logf(LOG_MSG, _T("ProcessBase::stop() FINISH"));
    reset();
}

void ProcessBase::pause()
{
    // anytime we access the m_pThread pointer we must ensure that it won't
    // be modified in the meanwhile; since only a single thread may be
    // inside a given critical section at a given time, the following code
    // is safe:
    wxCriticalSectionLocker enter(m_pThreadCS);

    if (m_pThread)         // does the thread still exist?
    {
        // without a critical section, once reached this point it may happen
        // that the OS scheduler gives control to the MyThread::Entry() function,
        // which in turn may return (because it completes its work) making
        // invalid the m_pThread pointer

        if (m_pThread->Pause() != wxTHREAD_NO_ERROR )
             Logger::logf(LOG_MSG,_T("ProcessBase::Pause() an't pause the thread!"));
    }
}

void ProcessBase::resume()
{
    wxCriticalSectionLocker enter(m_pThreadCS);
    if (m_pThread)         // does the thread still exist?
    {
        if (m_pThread->Resume() != wxTHREAD_NO_ERROR )
             Logger::logf(LOG_MSG,_T("ProcessBase::Pause() an't pause the thread!"));
    }
}

bool ProcessBase::isRunning()
{
    //return m_pThread ? m_pThread->IsRunning() : false;
    wxCriticalSectionLocker enter(m_pThreadCS);

    return m_pThread ? true : false;
}



class CommunicationLocker
{
    public:
    CommunicationLocker(ProcessBase* commCtr){
        m_commCtr = commCtr;
        m_commCtr->stop();
    };
    ~CommunicationLocker(){
        m_commCtr->start();
    };

    private:
        ProcessBase* m_commCtr;

};
