#include "CommunicationController2.h"
#include "CommunicationControllerClient.h"
#include "CommunicationControllerEvent.h"
#include "../events/CommunicationEvent.h"
#include "../CommunicationBase.h"
#include "../SerialException.h"
#include "../logger.h"

CommunicationController2::CommunicationController2()
{
    m_client        = NULL;
    m_commBase      = NULL;
    m_eventHandler  = NULL;

    m_errorHandler  = new CommunicationControllerErrorHandler( this );

    m_errorCode = OK;

    m_sleepBeforeClearing = 0;
    m_clearingReadCount = 0;

    m_isStopped     = false;
}

CommunicationController2::CommunicationController2( wxEvtHandler* eventHandler, CommunicationControllerClient* client, CommunicationBase* commBase )
{
    m_client        = client;
    m_commBase      = commBase;
    m_eventHandler  = eventHandler;

    m_errorHandler  = new CommunicationControllerErrorHandler( this );

    m_errorCode = OK;

    m_sleepBeforeClearing = 0;
    m_clearingReadCount = 0;

    m_isStopped = false;
}

CommunicationController2::~CommunicationController2()
{
    delete m_errorHandler;
}

void CommunicationController2::registerClient( CommunicationControllerClient* client )
{
    m_client = client;
}

void CommunicationController2::setCommunicationBase( CommunicationBase* commBase )
{
    m_commBase = commBase;
}

void CommunicationController2::setEventHandler( wxEvtHandler* eventHandler )
{
    m_eventHandler  = eventHandler;
}

bool CommunicationController2::process()
{
    int rxCount = 1;
    int txCount = 0;

    try {
        m_client-> startProcess();

        if ( !m_commBase->IsOpen() ) {
        	m_commBase->open();
        }

        if ( m_clearingReadCount > 0 ) {
            if ( m_sleepBeforeClearing > 0 ) {
                LOGP( wxString::Format( _T("m_clearingReadCount: %d, m_sleepBeforeClearing: %d"),
                    m_clearingReadCount, m_sleepBeforeClearing ) );
                wxThread::Sleep( m_sleepBeforeClearing );
                m_sleepBeforeClearing = 0;
            }
            --m_clearingReadCount;
            m_commBase->Read( m_rxBuff, rxCount );
            return true;
        }

        m_client->write( m_txBuff, txCount );

        if ( txCount > 0 ) {
            m_commBase->SendString( m_txBuff, txCount );

            size_t interWriteDelay = m_client->getInterWriteDelay();
            if ( interWriteDelay > 0 ) {
                wxThread::Sleep(interWriteDelay);
            }
        }

        // because GPS receivedBytes = 0 with some BlueTooth driver
        m_rxBuff[0] = 0;

        int receivedBytes = m_commBase->Read( m_rxBuff, rxCount );

        if ( receivedBytes >= 0 ) {
            m_client->read( m_rxBuff, receivedBytes );
        } else {
            m_errorCode = READ_ERROR;
            //m_commBase->close();
            return false;
        }
        return  !m_client->isFinished();
    } catch( SerialException& se ) {
        LOGP( wxString::Format( _T("SERIAL EXCEPTION: ") + se.getErrorCodeAsString() ) );
        if ( se.getCode() == SP_CANNOT_OPEN ) {
            m_errorCode = CANNOT_OPEN_PORT;
            LOG
        } else {
            m_errorCode = UNKNOWN_ERROR;
        }
        return false;
    } catch (...) {
        LOGP( _T("UNKNOWN EXCEPTION") );
        m_errorCode = UNKNOWN_ERROR;
        return false;
    }
}

void CommunicationController2::start( int priority )
{
    m_isStopped = false;

    m_errorCode = OK;

    this->ProcessBase::start();
}

void CommunicationController2::stop()
{
    m_isStopped = true;

    this->ProcessBase::stop();
}

void CommunicationController2::reset()
{
}

/// thread stopped by user(stop()) or client( isFinished() == true ) or error occured
void CommunicationController2::finished()
{
    if ( error() ) {
        m_errorHandler->handleError();
        return;
    }

    if ( m_isStopped ) {
        return;
    }

    m_client->finish();
}

bool CommunicationController2::error() const
{
    return m_errorCode != OK;
}

CommunicationController2::ErrorCode CommunicationController2::lastError() const
{
    return m_errorCode;
}

void CommunicationController2::clear( size_t readCount, size_t sleepBeforeClear )
{
    m_clearingReadCount     = readCount;
    m_sleepBeforeClearing   = sleepBeforeClear;
}


/// CommunicationControllerErrorHandler
CommunicationControllerErrorHandler::CommunicationControllerErrorHandler( CommunicationController2* parent )
    : m_errorTimer( this, wxID_NO )
{
    m_parent        = parent;
    m_isPortClosed  = false;

    Connect( COMM_CTRL_EVENT_CLOSING_PORT, wxCommandEventHandler( CommunicationControllerErrorHandler::OnClosingPort ), NULL, this );
    Connect( COMM_CTRL_EVENT_PORT_CLOSED, wxCommandEventHandler( CommunicationControllerErrorHandler::OnPortClosed ), NULL, this );
    Connect( wxID_NO, wxEVT_TIMER, wxTimerEventHandler( CommunicationControllerErrorHandler::OnErrorTimer ), NULL, this );
}

CommunicationControllerErrorHandler::~CommunicationControllerErrorHandler()
{
    Disconnect( COMM_CTRL_EVENT_CLOSING_PORT, wxCommandEventHandler( CommunicationControllerErrorHandler::OnClosingPort ), NULL, this );
    Disconnect( COMM_CTRL_EVENT_PORT_CLOSED, wxCommandEventHandler( CommunicationControllerErrorHandler::OnPortClosed ), NULL, this );
    Disconnect( wxID_NO, wxEVT_TIMER, wxTimerEventHandler( CommunicationControllerErrorHandler::OnErrorTimer ), NULL, this );
}


/// handle serial exceptions
/// handled in CommunicationController2 thread
void CommunicationControllerErrorHandler::handleError()
{
    m_isPortClosed = false;

    if ( m_parent->m_errorCode == CommunicationController2::CANNOT_OPEN_PORT ) {
        if ( m_parent->m_eventHandler ) {
            CommunicationEvent event( CommunicationEvent::CANNOT_OPEN_PORT );
            m_parent->m_eventHandler->AddPendingEvent( event );
        }
        return;
    }

    /// send an event before try to close comm port
    wxCommandEvent closingPortEvent( COMM_CTRL_EVENT_CLOSING_PORT, wxID_ANY );
    AddPendingEvent( closingPortEvent );

    /// try to close comm port
    /// if USB-Serial converter is wrong, block CommunicationController2 thread
    m_parent->m_commBase->close();

    m_isPortClosed = true;

    /// send an event after comm port closed
    wxCommandEvent portClosedEvent( COMM_CTRL_EVENT_PORT_CLOSED, wxID_ANY );
    AddPendingEvent( portClosedEvent );
}

/// CommunicationController2 try to close comm port
/// handled in the MainThread
void CommunicationControllerErrorHandler::OnClosingPort( wxCommandEvent& event )
{
    m_errorTimer.Start( 1000, true );
}

/// comm port succesfully closed
/// handled in the MainThread
void CommunicationControllerErrorHandler::OnPortClosed( wxCommandEvent& event )
{
    m_errorTimer.Stop();

    if ( m_parent->m_eventHandler ) {
        CommunicationEvent event( CommunicationEvent::PORT_CLOSED );
        m_parent->m_eventHandler->AddPendingEvent( event );
    }
}

/// probably USB-Serial converter failure, send REMOVE_USB event
/// handled in the MainThread
void CommunicationControllerErrorHandler::OnErrorTimer( wxTimerEvent& event )
{
    if ( !m_isPortClosed && m_parent->m_eventHandler ) {
        CommunicationEvent usbEvent( CommunicationEvent::REMOVE_USB );
        m_parent->m_eventHandler->AddPendingEvent( usbEvent );
    }
}
