#include "FirmwareVerifier.h"
#include "FirmwareVerifierListener.h"
#include "TimeoutListener.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../Math.h"
#include "../../vemsprotocol/FirmwareData.h"
#include "../../events/FirmwareVerifyEvent.h"
#include "../CommunicationClientEventHandler.h"


FirmwareVerifier::FirmwareVerifier( FirmwareVerifierListener* verifierListener,
                                    TimeoutListener* timeoutListener,
                                    FirmwareData* fwData )
{
    m_verifierListener  = verifierListener;
    m_timeoutListener   = timeoutListener;
    m_firmwareData      = fwData;

    m_state = SEND_PAGE_ADDRESS;
    m_pageNum = m_firmwareData->getPages();
    m_currentPage = 0;
//    m_timeoutCount = 0;
    m_rxDataLength = 0;
    m_pageErrorCount = 0;
    m_result = TIMEOUT;

    // read-in patch files
    if ( !m_ntcfinder.isOk )
    {
        m_ntcfinder.getFileArrray( NTC_COOLANT );
        m_ntcfinder.getFileArrray( NTC_MAT );
        m_ntcfinder.toString();
    }

    /** send the data to a verification file*/
    //m_firmwareData->writeToIhex(); //write all page data to iHexfile
}

FirmwareVerifier::~FirmwareVerifier()
{
    delete m_firmwareData;
}

void FirmwareVerifier::write( uint8_t* buffer, int& size )
{
    switch ( m_state )
    {
        case SEND_PAGE_ADDRESS:
            sendPageAddress( buffer, size );
            break;

        case SEND_GET_CHECKSUM_CMD:
            sendGetChecksumCmd( buffer, size );
            break;

        default:
            break;
    }
}

void FirmwareVerifier::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case WAIT_PAGE_ADDRESS_OK:
            waitPageAddressOk( buffer, size );
            break;

        case WAIT_CHECKSUM:
            waitChecksum( buffer, size );
            break;

        default:
            break;
    }
}

void FirmwareVerifier::finish()
{
    switch ( m_result )
    {
        case FIRMWARE_IDENTICAL:
            sendFirmwareVerifierEvent( FirmwareVerifyEvent::SUCCEEDED );
            break;

        case FIRMWARE_DIFFERENT:
            sendFirmwareVerifierEvent( FirmwareVerifyEvent::FAILED );
            break;

        case TIMEOUT:
            m_timeoutListener->OnTimeout();
            break;
    }
}

bool FirmwareVerifier::isFinished()
{
    return ( m_state == FINISHED );
}

void FirmwareVerifier::sendPageAddress( uint8_t* buffer, int& size )
{
//    m_timeoutCount = 0;

    uint32_t pageIdx = m_firmwareData->getPageIndexbyOrderIdx(m_currentPage);
    uint32_t addressW = m_firmwareData->getPageAddressW(pageIdx);
    /// addresW is actually 2 bytes

    ///TODO - handle command size > 1

    uint8_t str[3];
    str[0] = 'A';
    str[1] = (addressW >> 8) & 0xFF;
    str[2] = addressW & 0xFF;

    TxData command = TxData( str, 3 );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_timeoutTimer.start( m_timeout );

    m_state = WAIT_PAGE_ADDRESS_OK;
}


void FirmwareVerifier::waitPageAddressOk( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        if( buffer[0] == '\r' ) {
            m_state = SEND_GET_CHECKSUM_CMD;
        } else {
            m_result = TIMEOUT;
            m_state = FINISHED;
        }
    }
    else
    {
        wxThread::Sleep( m_sleep );

        if ( m_timeoutTimer.isTimedout() ) {
            m_result = TIMEOUT;
            m_state = FINISHED;
        }
    }
}

void FirmwareVerifier::sendGetChecksumCmd( uint8_t* buffer, int& size )
{
    TxData command;


    if ( m_firmwareData->getType() == FirmwareData::HEX ) {
        //V3
        command = TxData( _T("Dll") );
    } else {
        //ROUND
        command = TxData( _T("DCC") );
    }

    m_rxDataLength = 0;
//    m_timeoutCount = 0;
    m_timeoutTimer.start( m_timeout );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = WAIT_CHECKSUM;
}

void FirmwareVerifier::waitChecksum( uint8_t* buffer, int size )
{
    if ( size > 0 )
    {
        m_rxBuffer[ m_rxDataLength++ ] = buffer[0];

        if ( m_rxDataLength == CHECKSUM_LENGTH ) {
            uint32_t pageIdx = m_firmwareData->getPageIndexbyOrderIdx(m_currentPage);
            uint32_t filePageCheckSum = m_firmwareData->getPageCheckSum(pageIdx);
            uint32_t ecuPageCheckSum = Math::getNumber( m_rxBuffer );

            bool crcOK = false;

            if ( m_firmwareData->getType() == FirmwareData::HEX ) {
                //V3
                crcOK = ( ecuPageCheckSum == filePageCheckSum );
            } else {
                //ROUND
                uint16_t received_chk_high = ecuPageCheckSum >> 16;
                uint16_t chk_hexfile_high = filePageCheckSum >> 16;

                crcOK = ( received_chk_high == chk_hexfile_high );
            }

            bool patched = false;
            wxString fname = wxEmptyString;

            if( !crcOK ) {
                /// firmware may has been patched
                fname = m_ntcfinder.getFile( ecuPageCheckSum );
                if ( fname != wxEmptyString )
                {
                    IntSet pages = m_ntcfinder.getPageSet( fname );
                    IntSet::const_iterator it = pages.find( m_currentPage );
                    if ( it != pages.end() )
                    {
                        patched = true;
                    }
                }

                if ( !patched )
                {
                    ///TODO error code PAGE_VERIFY_CHECKSUM_ERROR
                    Logger::logf(LOG_MSG, _T("FirmwareVerifier::waitChecksum different checksum! idx: %i,%08X,%08X"), pageIdx, filePageCheckSum, ecuPageCheckSum);
                    if ( ++m_pageErrorCount >= MAX_PAGE_ERROR_COUNT){
                        m_result = FIRMWARE_DIFFERENT;
                        m_state = FINISHED;
                        return;
                    }
                }
            }

            if ( patched )
                sendFirmwareVerifierEvent( FirmwareVerifyEvent::PAGE_VERIFIED_AND_PATCHED, fname );
            else
                sendFirmwareVerifierEvent( FirmwareVerifyEvent::PAGE_VERIFIED );

            if(m_currentPage < ( m_pageNum - 1 ) ){
                m_currentPage++;
                m_state = SEND_PAGE_ADDRESS;
            }else{
                m_result = FIRMWARE_IDENTICAL;
                m_state = FINISHED;
            }
        }
    }
    else
    {
        wxThread::Sleep( m_sleep );

        if ( m_timeoutTimer.isTimedout() ) {
            m_result = TIMEOUT;
            m_state = FINISHED;
        }
    }
}

void FirmwareVerifier::sendFirmwareVerifierEvent( int eventType, const wxString& patchfilename )
{
    FirmwareVerifyEvent event( eventType );
    event.setFirmwarePageCount( m_firmwareData->getPages() );
    event.setFirmwarePageIdx( m_currentPage );

    if ( patchfilename != wxEmptyString )
        event.setPatchFileName( patchfilename );

    m_verifierListener->OnFirmwareVerifyEvent( event );
}

