#include "SDCardFileDownloader.h"

#include "../../Math.h"
#include "../../logger.h"
#include "../../events/SDCardEvent.h"
#include "../../common/TxData.h"

#include "SDCardFileHelper.h"

using namespace SDCardFile;

SDCardFileDownloader::SDCardFileDownloader( const MMCFilesDesc& desc, wxEvtHandler* evthandler, unsigned int blockCount )
{
    m_nextBlock = true;

    if ( blockCount != 1 && blockCount != 60 )
        m_blockCount = 1;
    else
        m_blockCount = blockCount;

    m_buffer = new unsigned char[m_blockCount * ( 514 + 2 )];
    memset( m_buffer, 0, m_blockCount* (514+2) );
    m_buflen = 0;

    m_evthandler = evthandler;

    m_blockIdx = desc.StartAddress;
    m_endIdx = desc.EndAddress;

    uint32_t stopidx = desc.StopIndex;

    if ( m_endIdx == 0 )
        m_endIdx--;

    if ( m_endIdx > stopidx )
        m_endIdx = stopidx;

    m_desc = desc;

    // LOGP( wxString::Format(_T("Start=%d, end=%d"), m_blockIdx, m_endIdx) )

    m_timeoutCounter = 0;

    m_tryCount = 0;

    m_data = new unsigned char[ ( m_endIdx - m_blockIdx + 2 ) * 514 ];
    m_datalen = 0;

    m_lastSecCountInBlock = -255;

    m_msscounter = 0;

    // m_timeout += 1000; // 25ms

    m_state = GET_STATUS;
    Logger::logf(LOG_MSG, _T("startIdx=%d, endIdx =%d"), m_blockIdx, m_endIdx);
}

SDCardFileDownloader::~SDCardFileDownloader()
{
    delete[] m_buffer;

}

void SDCardFileDownloader::Abort()
{
    m_state = ABORTED;
    m_aborted = true;
}

void SDCardFileDownloader::write( uint8_t* buffer, int& size )
{
    m_timeout = TIMEOUT_VAL;
    switch ( m_state )
    {
        case GET_STATUS:
            getStatus( buffer, size );
            break;

        case GET_SECTOR:
        {
            if ( m_nextBlock )
            {
                memset( m_buffer, 0, ( 514+2 ) * m_blockCount );
                m_buflen = 0;

                if ( m_blockCount == 1 )
                {
                    double percent = ( ( m_blockIdx + 1 - m_desc.StartAddress ) * 100.0 / (double)( m_desc.EndAddress - m_desc.StartAddress ) );

                    wxVariant var = percent;

                    SDCardEvent event;
                    event.SetType( SDCardEventTypes::DOWNLOADING_PROGRESS );
                    event.SetVariant( var );
                    m_evthandler->AddPendingEvent( event );
                }

                uint32_t blockIdx;
                blockIdx = m_blockIdx & 0xffffff;

                wxString hexIdx = wxString::Format( _T("%06X"), blockIdx );
                wxString cmd = ( m_blockCount == 1 ) ? _T("msg") : _T("msG");
                cmd += hexIdx;

                LOGP( cmd )

                TxData command = TxData( cmd );

                size = command.size;
                memcpy( buffer, command.data, command.size );

                m_nextBlock = false;

                m_state = CHECK_SECTOR;

            }
            break;
        } // case get_sector

        default: break;
    }
}

void SDCardFileDownloader::read( uint8_t* buffer, int size )
{
    // Logger::logf(LOG_MSG, _T("m_sleep=%d"), m_sleep);
    m_sleep = 2;

    switch ( m_state )
    {
        case CHECK_STATUS:
            checkStatus( buffer, size );
            break;

        case CHECK_SECTOR:
            readSector( buffer, size );
            break;

        default: break;
    }
}

void SDCardFileDownloader::finish()
{
    SDCardEvent event;

    if ( m_state == TIMEOUT )
    {
        event.SetType( SDCardEventTypes::TIMEOUT );
        LOGP( _T("Timeout") );
        delete[] m_data;
    }
    else if ( m_state == ABORTED )
    {
        event.SetType( SDCardEventTypes::INTERRUPTED );
        LOGP( _T("Aborted") )
        delete[] m_data;
    }
    else if ( m_state == STATUS_FAILED )
    {
        event.SetType( SDCardEventTypes::MSx_FAILED );
        LOGP( _T("mss failed") )
        delete[] m_data;
    }
    else
    {
        event.SetType( SDCardEventTypes::DOWNLOADING_FINISHED );
        wxVariant var; var = (void*)m_data;
        event.SetVariant( var );
        event.SetInt( m_datalen );
        event.SetDataType( m_desc.DataType );

        LOGP( wxString::Format(_T("Data=%p, datasize=%d"), m_data, m_datalen) )
    }
    m_evthandler->AddPendingEvent( event );
}

bool SDCardFileDownloader::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEOUT || m_state == ABORTED || m_state == STATUS_FAILED );
}

///

void SDCardFileDownloader::getStatus( uint8_t* buffer, int& size )
{
    TxData cmd( _T("mss") );

    size = cmd.size;
    memcpy( buffer, cmd.data, cmd.size );

    m_state = CHECK_STATUS;
    m_timeoutCounter = 0;
}

void SDCardFileDownloader::checkStatus( uint8_t* buffer, int size )
{
    if ( size == 0 )
    {
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        if ( m_timeoutCounter > m_timeout )
        {
            m_state = TIMEOUT;
        }
        return;
    }
    else m_timeoutCounter = 0;

    if ( size > 0 )
    {
        char c = buffer[0];

        if ( c == '0' )
        {
            m_state = GET_SECTOR;
            m_msscounter = 0;
        }

        if ( c == 'u' )
        {
            ++m_msscounter;
            // check status again after 100ms
            m_state = GET_STATUS;

            wxThread::Sleep( 100 );

            if ( m_msscounter > 30 )
            {
                LOGP(_T("ERROR: mss still responding 'u' after 30 tries") )
                m_state = STATUS_FAILED;
            }
        }
    }
}

void SDCardFileDownloader::readSector( uint8_t* buffer, int size )
{
    // timeout?
    if ( size == 0 )
    {
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        if ( m_timeoutCounter > m_timeout )
        {
            if ( ++m_tryCount < MAX_TRY_COUNT )
            {
                m_state = GET_STATUS;
                m_nextBlock = true;
                m_timeoutCounter = 0;
            }
            else
            {
                m_state = TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCounter = 0;

    memcpy( m_buffer + m_buflen, buffer, size );
    m_buflen += size;

    // send progress event (only in case of 60 block/read)
    if ( m_blockCount == 60 && ( m_buflen % 514 == 0 ) )
    {
        int sBlocks = (int)ceil( ( ( m_desc.EndAddress - m_desc.StartAddress ) / 60.0 ) );
        double percent = ( m_blockIdx + 1 + ( m_buflen / 514 ) - m_desc.StartAddress ) * 100 / (double)( sBlocks * 60 );
        wxVariant var = percent;

        SDCardEvent event;
        event.SetType( SDCardEventTypes::DOWNLOADING_PROGRESS );
        event.SetVariant( var );
        m_evthandler->AddPendingEvent( event );
    }

    if ( m_buflen >= (m_blockCount * 514) )
    {
        uint8_t* dataPtr = &m_buffer[0];

        for ( size_t i = 0; i < m_blockCount; ++i )
        {
            if ( ( m_blockIdx + i ) > m_endIdx )
                break;

            bool isvalid = isDataValid( dataPtr, 512 );

            if ( isvalid )
            {
                /*
                if ( isNextLog( dataPtr ) )
                {
                    if ( m_datalen > 0 )
                    {
                        // send event
                    }
                    delete[] m_data;
                    m_data = new unsigned char[ ( m_endIdx - m_blockIdx + 2 ) * 514 ];
                    m_datalen = 0;
                }
                */
                memcpy( m_data + m_datalen, m_buffer + i * 514, 512 );
                m_datalen += 512;
            }

            dataPtr = dataPtr + 514;
        }

        m_blockIdx += m_blockCount;
//        LOGP( wxString::Format(_T("Idx=%d, end=%d"), m_blockIdx, m_endIdx) )

        if ( m_blockIdx > m_endIdx )
        {
            m_state = FINISHED;
        }
        else
        {
            m_state = GET_SECTOR;
            m_nextBlock = true;
        }
    }
}

/// PRIVATE 'HELPERS'
bool SDCardFileDownloader::isNextLog( uint8_t* data )
{
    int offset = 0;
    int fscInBlock = data[offset];

    if ( ( fscInBlock == m_lastSecCountInBlock ) ||
         ( ( fscInBlock - m_lastSecCountInBlock ) == 1 ) ||
         ( ( fscInBlock == 0 ) && ( m_lastSecCountInBlock == 255 ) ) )
    {
        m_lastSecCountInBlock = fscInBlock;
        return false;
    }

    m_lastSecCountInBlock = fscInBlock;
    return true;
}
