#include "SDCardFileSearch.h"
#include "SDCardFileHelper.h"
#include "../../events/SDCardEvent.h"
#include "../../common/TxData.h"
#include "../../logger.h"
#include "../../Math.h"
#include "../../FirmwareInfo.h"
#include "../../common/VemsContext.h"


/// initialize SdStaticInfo

MMCFilesDescA SDStaticInfo::m_mmcdescarr ;
bool SDStaticInfo::m_dirty = true;

///
using namespace SDCardFile;

/**
    SDCardFileSearch Constructor
    Params:
        deviceMode: current devicemode (will be restored after closing dialog)
        evthandler: a wxEvtHandler which will handle SDCardEvents
    Task:
        initialization
        check if rescan needed
        go to man mode if neccessary
        alloc memory for readbuffer (1 sector)

    Return value:
        none
**/
SDCardFileSearch::SDCardFileSearch( wxEvtHandler * evthandler )
{
    // initialization
    m_msscounter = 0;
    m_nextstate = INVALID;

    m_evthandler = evthandler;
    m_mmcdescarr = &(SDStaticInfo::m_mmcdescarr);

    // for searching files
    m_address = 0x100;  // Inode end address
    m_minIdx = 0x100;   // min address
    m_count = 0;
    m_mask = 0x01;

    m_upward = true;
    m_nextBlock = true;

    // create a little bit more buffer
    m_buffer = new unsigned char[514+2];
    memset( m_buffer, 0, 514+2 );
    m_bufLen = 0;

    // needs rescan?
    if ( SDStaticInfo::isDirty() )
    {
        // we need a status check before sending msg
        m_state = GET_STATUS_BGET;
        m_nextstate = GET_STOPIDX;
    }
    // then searching is alread done
    else
    {
        m_state = FINISHED;
        m_cardState = SDFS_OK;
        return;
    }

    m_timeoutCounter = 0;
    m_tryCount = 0;
}

/**
    SDCardFileSearch destructor
    Params:
        none

    Task:
        frees up memory

    Return value:
        none
**/
SDCardFileSearch::~SDCardFileSearch()
{
    // free memory
    delete[] m_buffer;
}

/**
    write() - CommunicationControllerClient() virtual method, Tx, outgoing commands

    Params:
        buffer  : character buffer (a command) which will be sent to comm.port
        size    : size of the buffer

    Task:
        sets m_timeout
        creates the appropriate command (according to the current state)

    Return value:
        none
**/
void SDCardFileSearch::write( uint8_t* buffer, int& size )
{
    /// TODO
    m_timeout = TIMEOUT_VAL;

    switch ( m_state )
    {
        case GET_STOPIDX:
            getStopIdx( buffer, size );
            break;

        case GET_DATASIZE:
            getDataSize( buffer, size );
            break;

        case GET_INODE:
            getInode( buffer, size );
            break;

        case FINISH_INODE:
            finishInode();
            break;

        case CARD_STATE_FINISHED:
            checkCardState();
            break;

        case GET_STATUS_BGET:
            getStatusBeforeGet( buffer, size );
            break;

        case SEND_MSD:
            sendDisable( buffer, size );
            break;

        default:
            break;
    }
}

/**
    read()  - CommunicationControllerClient() virtual method, incoming data

    Params:
        buffer  : the incoming data
        size    : the size of the data in bytes

    Task:
        interprets the data according to the current state

    Return value:
        none
**/
void SDCardFileSearch::read( uint8_t* buffer, int size )
{
    m_sleep = 2;

    switch ( m_state )
    {
        case CHECK_STOPIDX:
            checkStopIdx( buffer, size );
            break;

        case CHECK_DATASIZE:
            checkDataSize( buffer, size );
            break;

        case CHECK_INODE:
            checkInode( buffer, size );
            break;

        case CHECK_STATUS_BGET:
            checkStatusBeforeGet( buffer, size );
            break;

        case CHECK_MSD:
            checkDisable( buffer, size );
            break;

        default:
            break;
    }
}

/**
    finish() - CommunicationControllerClient() virtual method, will be called if isFinished() is true

    Params:
        none

    Tasks:
        sending event ( timeout, searching finished - send mmcfiles desc too, ms(x) command failed ) to m_evthandler

    Return value:
        none
**/
void SDCardFileSearch::finish()
{
    LOG
    SDCardEvent event;

    /// timeout
    if ( m_cardState == SDFS_TIMEOUT )
    {
        LOGP( _T("Timeout"))
        event.SetType( SDCardEventTypes::TIMEOUT );
    }
    else if ( m_cardState == SDFS_MSSFAIL || m_cardState == SDFS_MSDFAIL ) // msd or mss failed (max # of msd/mss reached)
    {
        LOGP( _T("msd/mss failed)") )
        event.SetType( SDCardEventTypes::MSx_FAILED );
    }
    else // searching is done ( not valid card / empty card / ok )
    {
        event.SetType( SDCardEventTypes::SEARCHING_FINISHED );
        event.SetInt( m_cardState );
        event.SetFiles( *m_mmcdescarr );
    }
    m_evthandler->AddPendingEvent( event );
}

/**
    isFinished() - CommunicationControllerClient() virtual method

    Params:
        none

    Tasks:
        check whether searching is finished

    Return value:
        bool: if true => thread should be stopped
**/
bool SDCardFileSearch::isFinished()
{
    return ( m_state == FINISHED || m_state == TIMEOUT );
}

/**
    getStatusBeforeGet()

    Params:
        buffer  : will contain mss
        size    : size of 'mss'

    Tasks:
        status checking of the ecu - sends out mss to the communication port
        next state is checking the response

    Return value:
        none
**/
void SDCardFileSearch::getStatusBeforeGet( uint8_t* buffer, int & size )
{
    wxString cmd = _T("msdmss");
    TxData command( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    m_state = CHECK_STATUS_BGET;
}

/**
    checkStatusBeforeGet()

    Params:
        buffer  : the response of the mss command
        size    : size of the response

    Tasks:
        check if the response is
            -> empty: timeout?
            -> 0?   : mss is OK, logging is disabled
            -> u?   : undefined, waiting, and check status ( resend mss )
            -> 1?   : we must disable logging before sending and msg command

    Return value:
        none
**/
void SDCardFileSearch::checkStatusBeforeGet( uint8_t* buffer, int size )
{
    // no incoming data
    if ( size == 0 )
    {
        // increment timeout counter&sleep m_sleep ms
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        // oops, timeout -> searching finished
        if ( m_timeoutCounter > m_timeout )
        {
            m_state = TIMEOUT;
            m_cardState = SDFS_TIMEOUT;
        }
        return;
    }

    m_timeoutCounter = 0;

    if ( size > 0 )
    {
        // response is the first charachter
        char response = buffer[0];

        // disabled status
        if ( response == '0' )
        {
            LOGP( _T("mss OK"))
            // ok lets go to the next state
            m_state = m_nextstate;
            m_msscounter = 0;
        }

        // undefined status
        if ( response == 'u' )
        {
            ++m_msscounter;

            // we should wait a little bit, check the status again -> send mss
            wxThread::Sleep( 100 );
            m_state = CHECK_STATUS_BGET;

            // something went wrong...
            if ( m_msscounter > MAX_MSS_TRIES )
            {
                LOGP( wxString::Format( _T("ERROR: mss still responding 'u' after %d tries"), MAX_MSS_TRIES ) )

                m_state = FINISHED;
                m_cardState = SDFS_MSSFAIL;
            }
        }

        // .. -> we have to disable logging first, then check status again
        if ( response == '1' )
        {
            LOGP( _T("mss returned 1, sending msd"))
            m_msscounter = 0;
            m_state = SEND_MSD;
        }
    }
}

/**
    sendDisable()

    Params:
        buffer  : will contain msd
        size    : size of 'msd'

    Tasks:
        disable logging, send msd to the commport

    Return value:
        none
**/
void SDCardFileSearch::sendDisable( uint8_t* buffer, int& size )
{
    wxString cmd = _T("msd");

    m_isWaiting = false;

    TxData command( cmd );

    size = command.size;

    memcpy( buffer, command.data, command.size );

    m_state = CHECK_MSD;
}

/**
    checkDisable()

    Params:
        buffer  : unused
        size    : unused

    Tasks:
        set state to GET_STATUS_BGET after a while

    Return value:
        none
**/
void SDCardFileSearch::checkDisable( uint8_t* buffer, int size )
{
    if ( m_isWaiting )
    {
        return;
    }

    m_isWaiting = true;
    wxThread::Sleep( 1000 );
    m_state = GET_STATUS_BGET;
}


/**
    getStopIdx()

    Params:
        buffer  : will be filled with msg(#address) command
        size    : size of the command

    Task:
        sending out command to commport, searching for the stop index

    Return value:
        none
**/
void SDCardFileSearch::getStopIdx( uint8_t* buffer, int& size )
{
    // if it is a new block
    if ( m_nextBlock )
    {
        memset( m_buffer, 0, 514+2 );
        m_bufLen = 0;

        // counting upward: addr = addr * 2
        if ( m_upward )
        {
            m_address |= m_mask;
            m_mask <<= 1;
            ++m_count;
        }

        uint32_t blockIdx = m_address;
        blockIdx &= 0xffffff;

        // create the command
        wxString hexIdx = wxString::Format( _T("%06X"), blockIdx );
        wxString cmd = _T("msg") + hexIdx;
        Logger::logf(LOG_MSG, _T("Command=") + cmd );
        TxData command = TxData( cmd );

        size = command.size;
        memcpy( buffer, command.data, command.size );

        m_nextBlock = false;

        // set state to CHECK_STOPIDX
        m_state = CHECK_STOPIDX;
    }
    m_state = CHECK_STOPIDX;
}

/**
    checkStopIdx()

    Params:
        buffer  : incoming byte
        size    : size of the incoming data

    Task:
        ->fills the m_buffer with 514 bytes of data
        ->check whether it is empty
        searching for the stop index

    Return value:
        none
**/
void SDCardFileSearch::checkStopIdx( uint8_t* buffer, int size )
{
    // no incoming data
    if ( size == 0 )
    {
        // inc. timeout counter&sleep for m_sleep ms
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        // timeout?
        if ( m_timeoutCounter > m_timeout )
        {
            //LOGP( wxString::Format( _T("m_timeoutCounter: %d, m_timeout: %d"), m_timeoutCounter, m_timeout ) );

            // check state before msg (MAX_TRY_COUNT)
            if ( ++m_tryCount < MAX_TRY_COUNT )
            {
                m_state = GET_STATUS_BGET;
                m_nextstate = GET_STOPIDX;
                m_timeoutCounter = 0;
                m_nextBlock = true;
                if ( m_upward )
                {
                    m_mask >>= 1;
                }
                else
                {
                    m_mask <<= 1;
                }
            }
            else
            {
                // that's timeout
                m_state = TIMEOUT;
                m_cardState = SDFS_TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCounter = 0;

    // append incoming byte to the buffer
    memcpy( m_buffer + m_bufLen, buffer, size );
    m_bufLen += size;

    // counting upward
    if ( m_upward )
    {
        // a full block is read in
        if ( m_bufLen == 514 )
        {
            // it block is empty->count downward
            if ( checkSectorIsEmpty( m_buffer, 512 ) )
            {
                m_address++;
                m_upward = false;
                m_mask <<= 1;
                m_count++;
                m_state = GET_STOPIDX;
            }

            // send a searching event
            SDCardEvent event;
            event.SetType( SDCardEventTypes::SEARCHING_PROGRESS );
            m_evthandler->AddPendingEvent( event );

            //Logger::logf(LOG_MSG,_T("MMCLogger getstopidx%d()up blockidx=%i maszk=%d"),
            //    m_count, m_address, m_mask );
            //wxString message = wxString::Format(_T("Searching: %d"), m_address );
            //Logger::logf(LOG_MSG, message);

            // set state to GET_STOPIDX to read another block
            m_state = GET_STOPIDX;

            m_nextBlock = true;
            m_tryCount = 0;

            //Logger::logf(LOG_MSG, _T("m_count = %d"), m_count);
        }
    }
    // downward
    else
    {
        // full block is read in
        if ( m_bufLen == 514 )
        {
            if ( checkSectorIsEmpty( m_buffer, 512 ) )
            {
                m_address &= ~m_mask;
            }

            m_mask >>= 1;

            m_address |= m_mask;

            if ( m_address < m_minIdx )
            {
                m_address = m_minIdx;
            }

            m_count--;
            m_state = GET_STOPIDX;

            //Logger::logf(LOG_MSG,_T("MMCLogger getstopidx%d() down blockidx=%i maszk=%d"),
            //    m_count, m_address, m_mask );
            //wxString message = wxString::Format(_T("Searching: %d"), m_address );
            //Logger::logf(LOG_MSG, message);

            // send a searching event
            SDCardEvent event;
            event.SetType( SDCardEventTypes::SEARCHING_PROGRESS );
            m_evthandler->AddPendingEvent( event );

            m_nextBlock = true;
            m_tryCount = 0;
            //Logger::logf(LOG_MSG, _T("m_count = %d"), m_count);
        }
    }

    // stop idx is found, now lets get the data size
    if ( m_count < 0 )
    {
        m_state = GET_DATASIZE;
        m_nextstate = GET_DATASIZE;
        m_stopIdx = m_address;
        m_tryCount = 0;
    }
}

/**
    getDataSize()

    Params:
        buffer  : msg(#addr) command
        size    : size of the command

    Tasks:
        send command to commport

    Return value:
        none
**/
void SDCardFileSearch::getDataSize( uint8_t* buffer, int& size )
{
    memset( m_buffer, 0, 514 + 2 );
    m_bufLen = 0;

    uint32_t bIdx = m_stopIdx & 0xffffff;

    wxString cmd = wxString::Format( _T("msg%06x"), bIdx );
    TxData command = TxData( cmd );

    size = command.size;
    memcpy( buffer, command.data, command.size );

    // set new state
    m_state = CHECK_DATASIZE;
}

/**
    checkDataSize()

    Params:
        buffer  : incoming data
        size    : size of data

    Tasks:
        fills the m_buffer with 514 bytes(2 crc)
        gets the m_datasize

    Return value:
        none
**/
void SDCardFileSearch::checkDataSize( uint8_t* buffer, int size )
{
    // timeout?
    if ( size == 0 )
    {
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        // if counter>timeout
        if ( m_timeoutCounter > m_timeout )
        {
            // try mss (MAX_TRY_COUNT)
            if ( ++m_tryCount < MAX_TRY_COUNT )
            {
                m_state = GET_STATUS_BGET;
                // get datasize again
                m_nextstate = GET_DATASIZE;
            }
            else
            {
                // sorry, timeout
                m_state = TIMEOUT;
                m_cardState = SDFS_TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCounter = 0;

    memcpy( m_buffer + m_bufLen, buffer, size );

    m_bufLen += size;

    // buffer filled
    if ( m_bufLen == 514 )
    {
        // check for valid crc
        bool crc = isCRCValid( m_buffer, 514 );

        if ( crc == false ) m_datasize = -1;

        // get datasize
        m_datasize = ( m_buffer[510] == 0 || m_buffer[510] == 255 ) ? -2 : m_buffer[510];

        // set next state -> get the inodes
        m_state = m_nextstate = GET_INODE;
        m_nextBlock = true;

        m_address = 0x0;

        m_tryCount = 0;
    }
}

/**
    getInode()

    Params:
        buffer  : msg(#addr) command
        size    : size of the command

    Tasks:
        send command to comm.port

    Return value:
        none
**/
void SDCardFileSearch::getInode( uint8_t* buffer, int& size )
{
    // Logger::logf(LOG_MSG, _T("m_trycount=%d"),m_tryCount);
    if ( m_nextBlock )
    {
        memset( m_buffer, 0, 514 + 2 );
        m_bufLen = 0;

        m_timeoutCounter = 0;

        uint32_t bIdx = m_address & 0xffffff;

        // create command
        wxString cmd = wxString::Format( _T("msg%06x"), bIdx );
        LOGP( cmd )

        TxData command(cmd);

        size = command.size;
        memcpy( buffer, command.data, command.size );

        // set states
        m_state = CHECK_INODE;
        m_nextstate = GET_INODE;
        m_nextBlock = false;

        // create searching event & send it
        SDCardEvent event;
        event.SetType( SDCardEventTypes::SEARCHING_PROGRESS );
        m_evthandler->AddPendingEvent( event );
    }
}

/**
    checkInod()

    Params:
        buffer  : incoming byte
        size    : size of data

    Tasks:
        1. fill the m_buffer with 514 bytes
        2. check whether it's empty -> inode checked
        3. ->inc.address and get that sector -> 1

    Return value:
        none
**/
void SDCardFileSearch::checkInode( uint8_t* buffer, int size )
{
    // timeout?
    // no incoming data
    if ( size == 0 )
    {
        // inc counter & sleep m_sleep ms
        ++m_timeoutCounter;
        wxThread::Sleep( m_sleep );

        // it reached the limit
        if ( m_timeoutCounter > m_timeout )
        {
            // send an mss command and get inode block again (while count<max)
            if ( ++m_tryCount < MAX_TRY_COUNT )
            {
                m_state = GET_STATUS_BGET;
                m_nextstate = GET_INODE;
                m_nextBlock = true;
            }
            // else timeout
            else
            {
                m_state = TIMEOUT;
                m_cardState = SDFS_TIMEOUT;
            }
        }
        return;
    }
    else m_timeoutCounter = 0;

    // append data to the m_buffer
    memcpy( m_buffer + m_bufLen, buffer, size );
    m_bufLen += size;

    // a full block read in
    if ( m_bufLen >= 514 )
    {
        // if it is empty -> inode table checking is finished
        if ( checkSectorIsEmpty( m_buffer, 512 ) )
        {
            m_state = FINISH_INODE;
            return;
        }

        // if block does not begin with 'VEMS' -> invalid block -> searching finished
        if ( !SDCardFile::isVEMSFilesystem(m_buffer) )
        {
            m_cardState = SDFS_NOTVALIDFS;
            m_state = CARD_STATE_FINISHED;
            SDStaticInfo::setDirty( false );
            return;
        }

        FirmwareVersion fwVersion = SDCardFile::getFwVersion( m_buffer );
        // get integer fw version
        unsigned int fwVer_int = SDCardFile::getFWVersionAsInt( m_buffer );

        FirmwareInfo* fwInfo = dynamic_cast<FirmwareInfo*>( VemsContext::getObject( _T("FirmwareInfo") ) );

        if ( fwInfo )
        {
            wxString ecuFw = fwInfo->getFirmwareVersionAsString();

            if ( ecuFw >= _T("1.1.87") )
            {
                int prohibit;
                fwInfo->getProhibit( prohibit );
                wxString sdFw = fwVersion.toString();

                if ( fwVer_int != prohibit )
                {
                    SDCardEvent event;
                    event.SetType( SDCardEventTypes::FIRMWARE_VERSION );
                    event.SetString( sdFw );
                    if ( m_evthandler )
                    {
                        m_evthandler->AddPendingEvent( event );
                    }
                }
            }
        }

        m_buffer[512] = 0;
        m_buffer[513] = 0;

        uint8_t inodeVersion = SDCardFile::getInodeVersion(m_buffer);
        // save addresses
        switch( inodeVersion )
        {
            case 1:// inode version 1
            {
                for( uint16_t i = 8; i < 512; i += 3 )
                {
                    uint32_t ad = m_buffer[i] << 15;
                    ad += m_buffer[i+1] << 7;
                    ad += m_buffer[i+2] >> 1;

                    if ( ad == 0 )
                        break;

                    m_addrarr.Add( ad );
                    m_rttypearr.Add(SDLOG_RAW);
                }
            }
            break;
            case 2:// inode version 2
            {
                for( uint16_t i = 8; i < 512; i += 4 )
                {
                    uint32_t ad = m_buffer[i] << 15;
                    ad += m_buffer[i+1] << 7;
                    ad += m_buffer[i+2] >> 1;

                    if ( ad == 0 )
                        break;

                    m_addrarr.Add( ad );

                    m_rttypearr.Add(m_buffer[i+3]);//data type : 1: old raw rt data 2: bytestuffed
                }
            }
            break;

            // card is empty
            default:
            {
                m_cardState = SDFS_EMPTY;
                // mark info as not dirty
                SDStaticInfo::setDirty( false );

                // searching finished
                m_state = CARD_STATE_FINISHED;
                return;
            }
        }

        // increment address, get next block
        ++m_address;
        m_nextBlock = true;
        m_state = GET_INODE;
        m_tryCount = 0;
    }
}

/**
    finishInode()

    Params:
        none

    Tasks:
        fill the static descriptor array

    Return value:
        none
**/
void SDCardFileSearch::finishInode()
{
    // if no address found -> card is empty
    if ( m_addrarr.GetCount() == 0 )
    {
        m_cardState = SDFS_EMPTY;
        m_state = CARD_STATE_FINISHED;
        return;
    }

    // create a descriptor
    MMCFilesDesc desc;

    // clear the array
    m_mmcdescarr->Clear();

    // set cardstate to OK
    m_cardState = SDFS_OK;

    // fill the descriptor
    for ( size_t i = 0; i < m_addrarr.GetCount(); ++i )
    {
        desc.DataType = m_rttypearr[i];
        desc.StartAddress = m_addrarr[i];

        if ( i < ( m_addrarr.GetCount() - 1 ) )
        {
            desc.EndAddress = m_addrarr[i + 1] - 1;
        }
        else
        {
            desc.EndAddress = m_stopIdx;
        }

        if ( m_datasize == 0x7e || m_datasize < 64 || m_datasize > 75 )
        {
            m_datasize = 69;
        }

        desc.StopIndex = m_stopIdx;
        desc.DataSize = m_datasize;

        switch(desc.DataType)
        {
            case SDCardFile::SDLOG_RAW:
                //Logger::logf(LOG_MSG, _T("Datasize=%d"), m_datasize);
                desc.Time = CalcLogTimeLen( ( desc.EndAddress - desc.StartAddress ) + 1, m_datasize );
                desc.isApprox = false;
                break;

            case SDCardFile::SDLOG_TR_FRAME:
                    //CalcLogTimeLen( desc.EndAddress - desc.StartAddress, 69 );
                    // TODO: fetch the first TR FRAME block and calculate the size from it.
                    //m_state = GET_FIRST_TR_BLOCK;
                    // desc.Time = 0;
                /// NOT EXACT!!
                desc.Time = (size_t)((double)(CalcLogTimeLen( ( desc.EndAddress - desc.StartAddress ) + 1 )) / 1.166667f);
                desc.isApprox = true;

                break;
        }

        // add to the array
        if ( desc.StartAddress < desc.EndAddress )
        {
            m_mmcdescarr->Add( desc );
        }
    }
    // array is filled
    m_state = CARD_STATE_FINISHED;

    SDStaticInfo::setDirty( false );
}

/**
    checkCardState()

    Params:
        none

    Tasks:
        set state to finished

    Return value:
        none
**/
void SDCardFileSearch::checkCardState()
{
    m_state = FINISHED;
}

/// PRIVATE 'HELPERS'

/**
    CalcLogTimeLen()

    Params:
        len     :
        datalen :

    Tasks:
        calculates an approx. time length from descriptor

    Return value:
        size_t: length in seconds
**/
size_t SDCardFileSearch::CalcLogTimeLen( size_t len, size_t datalen )
{
    return SDCardFile::CalcLogTimeLength( len, datalen );
    // return (size_t)( ((int) (512.0 / (double)datalen) ) * 0.1 * len );
    //return (size_t)( (double)len * 51.2f / (double)datalen );
}

/// we assume that 7 RT in a frame -> so result is not exact
size_t SDCardFileSearch::CalcLogTimeLen( size_t len )
{
    return SDCardFile::CalcLogTimeLength( len );
    //return (size_t)( (double)len * 0.1f * 7.0 );
}
