#include "V3iniParser.h"
#include "../mtconfigdata.h"
#include "../vemsdialog/UserMenu.h"
#include "../vemsdialog/UserDialog.h"
#include "../vemsgauge/VemsTableDesc.h"
#include "../condtree.h"
#include "../logger.h"
#include "../vemsprotocol/SIPRTypes.h"
#include "../DirHelper.h"
#include "../UserLog.h"

#include <wx/regex.h>
#include <wx/string.h>
#include <wx/tokenzr.h>

V3iniParserPriv::V3iniParserPriv(MTConfigData* mtConfigData, BoolHash* conditionTable)
{
    m_MTConfigData = mtConfigData;
    if(conditionTable == NULL){
		m_hasOwnConditionTable = true;
        m_conditionTable = new BoolHash();

        (*m_conditionTable)[_T("12X12")]           = true; //todo - must be set outside
        (*m_conditionTable)[_T("16X14")]           = false; //todo - must be set outside
        (*m_conditionTable)[_T("HIGHBOOST")]       = true;
        (*m_conditionTable)[_T("CELSIUS")]         = true;
        (*m_conditionTable)[_T("NARROW_BAND_EGO")] = false;
        (*m_conditionTable)[_T("VER2")]            = false;
    }else{
		m_hasOwnConditionTable = false;
        m_conditionTable = conditionTable;
    }
    wxString key =_T("12X12");
    Logger::logf(LOG_MSG, _T("V3iniParserPriv() ")+ key +_T(" = %d"), (*m_conditionTable)[key]);

    m_currentPage = new long();
    *m_currentPage       = -1;

    m_currentSection = new wxString(_T(""));
    m_currentCT = NULL;

    m_currentCondValue = new bool();
    *m_currentCondValue  = false;

    m_logMsg    = new wxString();
    m_errorLog = new wxArrayString();

};

V3iniParserPriv::~V3iniParserPriv()
{
    delete m_currentPage;
    delete m_currentSection;
    delete m_currentCT;
    delete m_currentCondValue;
    delete m_logMsg;
    delete m_errorLog;
	if ( m_hasOwnConditionTable ) { delete m_conditionTable; }
};

V3iniParser::V3iniParser(MTConfigData* data, BoolHash* conditionTable)
{
    m_priv = new  V3iniParserPriv(data, conditionTable);
    m_del = true;
}

V3iniParser::V3iniParser(V3iniParserPriv* priv)
{
    m_priv = priv;
    m_del = false;
}

V3iniParser::~V3iniParser()
{
    //dtor
    if(m_del){
        delete m_priv;
    }
}


void V3iniParser::parseLine()
{
    wxString line = m_line;
    //Logger::logf(LOG_MSG, _T("V3iniParser() parseLine")+ line );
    if(line.Trim(false) == _T("")) return;
    if(line.Trim(false).GetChar(0) == _T(';')) return;

    Tokenizer t;
    t.tokenize(line);
    //Logger::logf(LOG_MSG, _T("V3iniParser() parseLine2") );

    CondTree* ct        = NULL;
    CondTree* parent    = NULL;

    //TEST
    getLogMsg().Clear();
    getLogMsg().Append( wxString::Format(  _T("Line : %4d -> "),
                                        (int)m_lineNumber+1));

    if(t[1] == _T("include")){
        wxString baseDir = DirHelper::GetPath(dCONFIG);
        wxString includeFile  = t[2];
        includeFile = includeFile.AfterFirst('"');
        includeFile = includeFile.BeforeLast('"');
        wxString incFile = baseDir + includeFile;
        if( wxFileExists(incFile) ){
            Logger::logf( LOG_MSG, _T("V3iniParser::parseLine() Found include: ")+ incFile);
            V3iniParser parser(m_priv);
            parser.parse(incFile);
        }else{
            Logger::logf( LOG_MSG, _T("V3iniParser::parseLine() Not found include: ")+ incFile);
            UserLog::uLog(ULOG_ERROR, _T("Include not found: ") +incFile );
        }
    }

    if(t[1] == _T("if"))
    {
        parent = getCondTree(); // it can be null!!
        ct = new  CondTree();

        wxString condStr = t[2].Upper();
        if(getCondTable()->find( condStr ) != getCondTable()->end())
        {
            getCondValue() = (*getCondTable() )[condStr];
        }
        else
        {
            getCondValue() = false;
            getErrorLog().Add(wxString::Format( condStr + _T(" does not match condition!  - line: %d"), m_lineNumber ) );
        }
        ct->setCondValue(getCondValue()); // true if the if expr is true!!
        ct->setParent(parent);
        ct->setState(CondTree::IN_IF);
        setCondTree(ct);
        getLogMsg().Append(_T("if - ") + condStr);
        //wxLogMessage(getLogMsg());
        return;
    }

    if(t[1] == _T("else"))
    {
        ct = getCondTree();
        if(ct == NULL){ throw ParserException(_T("IF statement missing!"),m_lineNumber); }
        ct->setState(CondTree::IN_ELSE);
        getLogMsg().Append(_T("else"));
        //wxLogMessage(getLogMsg());
        return;
    }

    if(t[1] == _T("endif"))
    {
        ct = getCondTree();
        if(ct == NULL) { throw ParserException(_T("IF statement missing!"),m_lineNumber); }
        else{
            setCondTree(ct->getParent() );
            delete ct;
        }
        //wxLogMessage(getLogMsg());
        getLogMsg().Append(_T("endif"));
        return;
    }

    //ParseSection
    if(t.getType(0) == T_BLK)
    {
        getSection() = t[0];

        if( getSection() == MT_GPSDATA )
        {
            getMTConfig().addGroup( MT_GPSDATA );
            Logger::logf( LOG_MSG, _T("V3iniParser::parseLine() add gps group") );
        }
        //TEST
        getLogMsg().Append(_T("section : ") + getSection());
        //wxLogMessage(getLogMsg());
        return;
    }

    if( getCondTree() == NULL || getCondTree()->isActive() )
    {
        try
        {
            if(getSection() == MT_PARAMS ) { parseParams(t); }
            if(getSection() == MT_MEGATUNE) { parseMegatune(t); }
            if(getSection() == MT_CONSTANTS) { parseConstants(t); }
            if(getSection() == MT_MENU) { parseMenu(t); }
            if( getSection() == MT_TABLEMAPPER ){ parseTableMapper( t ); }
            if(getSection() == MT_TABLEEDITOR){ parseTableEditor( t ); }
            if(getSection() == MT_CURVEEDITOR){ parseCurveEditor( t ); }
            if(getSection() == MT_SUPER_CURVE_EDITOR){ parseSuperCurveEditor( t ); }
//            if(getSection() == MT_GAUGECONFIG) { parseGaugeConfig(t); }
            //if(getSection() == MT_FRONTPAGE){ parseFrontPage(); }
            //if(getSection() == MT_RUNTIME){ parseRunTime(); }
            //if(getSection() == MT_TUNING){ parseTuning(); }
            if(getSection() == MT_USERDEFINED){ parseUserDefined(t); }
            if( getSection() == MT_GPSDATA ) { parseGPSData( t ); }
            if( getSection() == MT_AIMDATA ) { parseAIMData( t ); }
            if( getSection() == MT_SIPR_DATA ) { parseSIPRData( t ); }
            if( getSection() == MT_SIPR_DESCRIPTOR ) { parseSIPRDescriptor( t ); }
            if(getSection() == MT_OUTPUTCHANNELS) { parseOutputChannels(t); }
            if(getSection() == MT_DATALOG){ parseDatalog(t); }
            if( getSection() == MT_SYMBOL_ALIAS ){ parseSymbolAlias( t ); }
            if( getSection() == MT_SAVING_ORDER ) { parseConfigTablesSavingOrder( t ); }
            if( getSection() == MT_USER_DATA ){ parseUserData( t ); }
            if( getSection() == MT_EXPRESSION_GRID ){ parseExpressionGrid( t ); }
            if( getSection() == MT_INTERPOLATION_WIDGET ){ parseInterpolationWidget( t ); }
        }
        catch(ParserException e)
        {
            getErrorLog().Add(e.toString());
        }
    }
    //wxLogMessage(getLogMsg());
}//parseLine

void V3iniParser::parseParams( const Tokenizer& t )
{
    getMTConfig().addParam( t[0], t.getInt(1)!=0?true:false );
    Logger::logf( LOG_MSG, _T("V3iniParser::parseParams() - ") + t[0] + _T(" = %d"), t.getInt(1) );
    Logger::logf( LOG_MSG, _T("V3iniParser::parseParams() - addr of getCondTable(): %p"), getCondTable()  );
    (*getCondTable())[t[0]] = t.getInt(1)!=0?true:false;
}

void V3iniParser::parseMegatune(const Tokenizer& t)
{
    if ( t[0] == _T("signature") )
    {
        getMTConfig().setSignature( t[1] );
        Logger::logf( LOG_MSG, _T("V3iniParser::parseMegatune() sign : ") + getMTConfig().getSignature() );
    }
}

long  V3iniParser::lookupOffset(const wxString& str)
{
    long offset = -1;
    if( !str.ToLong(&offset) ){
        /// it is a lookup string to the header files
         MTConfigData& mtconfigdata = getMTConfig();
        SymbolTable& symbolH = mtconfigdata.getSymbolTable();
        SymbolTable::const_iterator it = symbolH.find(str);
        if(it == symbolH.end() ){
            /// error;
    //         Logger::logf( LOG_MSG, _T("Logger::logf() symbol not found: ")+ str );
        }else{
            Symbol* lookupSymbol = it->second;
            offset = lookupSymbol->getOffset();
  //           Logger::logf( LOG_MSG, _T("Logger::logf() lookup symbol found: ")+ str+_T(" offset: %i"), offset);
        }
    }else{
//        Logger::logf( LOG_MSG, _T("Logger::logf() offset found: %i"), offset);
    }
    return offset;
}

void V3iniParser::parseConstants(const Tokenizer& t)
{
    if(t[0] == _T("page")){
        getPage() = t.getInt(1);
        return;
    }

    if(t[1] == _T("scalar") || t[1] == _T("array") || t[1] == _T("bits") || t.getType(1) == T_EXP ){

        Symbol* symbol = NULL;
        symbol = new Symbol();

        try{
            if(t[1] == _T("scalar")){

                long offset = lookupOffset(t[3]);
                symbol->setBasic(t[0],t[4], true, getPage(), offset );
                symbol->setInputType(  SIPR_GET_TABLEDATA );
                parseTypeSpec(symbol,t[2]);
                symbol->setFormat(t.getDouble(5), t.getDouble(6), t.getDouble(7), t.getDouble(8), t.getInt(9));
            }

            if(t[1] == _T("array")){
                long offset = lookupOffset(t[3]);
                symbol->setBasic(t[0],t[5], true, getPage(), offset );
                symbol->setInputType( SIPR_GET_TABLEDATA );
                parseTypeSpec(symbol,t[2]);
                parseArraySpec(symbol,t[4]);
                symbol->setFormat(t.getDouble(6), t.getDouble(7), t.getDouble(8), t.getDouble(9), t.getInt(10));

                if ( ( t.getType(6) == T_EXP ) && ( t.getType(7) == T_EXP ) )
                {
                    Logger::logf( LOG_MSG, _T("-----V3iniParser::parseConstants() - ") + t[0] +
                        _T(", fromRaw: ") + t[6] + _T(", toRaw: ") + t[7] );
                    symbol->setExpressionsForTransformations( t[6], t[7] );
                }
            }

            if(t[1] == _T("bits")){
                long offset = lookupOffset(t[3]);
                symbol->setBasic(t[0],_T(""), true, getPage(), offset );
                symbol->setInputType( SIPR_GET_TABLEDATA );
                parseTypeSpec(symbol,t[2]);
                parseBitSpec(symbol,t);
            }

            if(t.getType(1) == T_EXP){
//                Logger::logf(   LOG_MSG, _T("-----V3iniParser::parseConstants() - const expression found: ")
//                                + t[0] + _T(" = ") + t[1] );
                symbol->setExpr(t[0],t[1] );
                symbol->setExprReverse( t[2] );
                symbol->setTarget( t[3] );
                symbol->setConst( true );
            }

            if(getMTConfig().getSymbol(symbol->getName()) != NULL){
                getErrorLog().Add( _T("Duplicate symbol! : ") + symbol->getName()
                                + wxString::Format(_T(" - line : %d"),m_lineNumber));
            }

            getMTConfig().addSymbol(symbol);

            getLogMsg().Append(_T("add ") + t[1] + _T(" : ") + symbol->toString()); //TEST

        }catch(ParserException e){
            if ( symbol != NULL ) delete symbol;
            getErrorLog().Add(e.toString());
        }
    }
}//parseConstant

/** @brief parseMenu
  *
  * @todo: document this function
  */
void V3iniParser::parseMenu(const Tokenizer& t)
{
    static wxString currentMenu = _T("default");

    UserMenuList* uml = getMTConfig().getUserMenuList();

    if(t[0] == _T("menu")){
        uml->addUserMenu(t[1]);
        currentMenu = t[1];
    }

    if(t[0] == _T("subMenu")){
        uml->addUserMenuItem(currentMenu,t[1],t[2],t.getInt(3),t[4]);
        getMTConfig().fillMenuAndDialogNamesByDialogId(t[1], currentMenu +_T(":")+ t[2] );
    }
}

void V3iniParser::parseTableMapper( const Tokenizer& t )
{
    static wxString currentMapper = wxEmptyString;

    TblMapDescHash& mappers = getMTConfig().m_tableMappers;

    if( t[0].IsSameAs( _T("name") ) )
    {
//        Logger::logf( LOG_MSG, _T("V3iniParser::parseTableMapper(): ") + t[1] );
        currentMapper = t[1];
        mappers[currentMapper] = new TblMapDesc();
        mappers[currentMapper]->m_name = t[1];
    }
    else if( t[0].IsSameAs( _T("xBins") ) ) //col header
    {
        mappers[currentMapper]->m_colHeaderTbl      = t[1];
        mappers[currentMapper]->m_colHeaderTitle    = t[2];
    }
    else if( t[0].IsSameAs( _T("yBins") ) ) //row header
    {
        mappers[currentMapper]->m_rowHeaderTbl      = t[1];
        mappers[currentMapper]->m_rowHeaderTitle    = t[2];
    }
    else if( t[0].IsSameAs( _T("zBins") ) ) //tables need to remap
    {
        for ( size_t i = 1; i < t.getSize(); ++i )
        {
            mappers[currentMapper]->m_tables.Add( t[i] );
//            Logger::logf( LOG_MSG, _T("V3iniParser::parseTableMapper() : ") + t[i] );
        }


    }
}

void V3iniParser::parseTableEditor( const Tokenizer& t )
{
    static wxString currentTable = wxEmptyString;
    static wxString currentTable3D = wxEmptyString;

    VemsTableDescH* tables = getMTConfig().getTables();

    if( t[0] == _T("table") )
    {
        currentTable = t[1];
        (*tables)[currentTable] = new VemsTableDesc();
        (*tables)[currentTable]->name = t[1];
        (*tables)[currentTable]->coupleName = t[2];
        (*tables)[currentTable]->title = t[3];
        (*tables)[currentTable]->is3D = false;

        if ( !t[2].IsEmpty() )
        {
            currentTable3D = t[2];
            (*tables)[currentTable3D] = new VemsTableDesc();
            (*tables)[currentTable3D]->name = t[2];
            (*tables)[currentTable3D]->coupleName = t[1];
            (*tables)[currentTable3D]->title = t[3];
            (*tables)[currentTable3D]->is3D = true;
        }
    }
    else if( t[0] == _T("xBins") )
    {
        (*tables)[currentTable]->xSTSymbol = t[1];
        (*tables)[currentTable]->xRTSymbol = t[2];
        (*tables)[currentTable3D]->xSTSymbol = t[1];
        (*tables)[currentTable3D]->xRTSymbol = t[2];

    }
    else if( t[0] == _T("yBins") )
    {
        (*tables)[currentTable]->ySTSymbol = t[1];
        (*tables)[currentTable]->yRTSymbol = t[2];
        (*tables)[currentTable3D]->ySTSymbol = t[1];
        (*tables)[currentTable3D]->yRTSymbol = t[2];
    }

    else if( t[0] == _T("zBins") )
    {
        (*tables)[currentTable]->zSTSymbol = t[1];
        (*tables)[currentTable]->zRTSymbol = t[2];
        (*tables)[currentTable3D]->zSTSymbol = t[1];
        (*tables)[currentTable3D]->zRTSymbol = t[2];
    }
    else if( t[0] == _T("numberWidth") )
    {
        (*tables)[currentTable]->numberWidth =  t.getInt( 1 );
        (*tables)[currentTable3D]->numberWidth =  t.getInt( 1 );
    }
    else if( t[0] == _T("numberPrec") )
    {
        (*tables)[currentTable]->numberPrec = t.getInt( 1 );
        (*tables)[currentTable3D]->numberPrec = t.getInt( 1 );
    }
    else if( t[0] == _T("largeCellStep") )
    {
        (*tables)[currentTable]->largeCellStep = t.getDouble( 1 );
        (*tables)[currentTable3D]->largeCellStep = t.getDouble( 1 );
    }
    else if( t[0] == _T("smallCellStep") )
    {
        (*tables)[currentTable]->smallCellStep = t.getDouble( 1 );
        (*tables)[currentTable3D]->smallCellStep = t.getDouble( 1 );
    }
    else if ( t[0] == _T("useTableMapper") )
    {
        (*tables)[currentTable]->useTableMapper = t.getInt( 1 )!=0?true:false;
    }
}


void V3iniParser::parseCurveEditor( const Tokenizer& t )
{
    static wxString currentCurve = wxEmptyString;
    VemsCurveDialogDescH& curves = getMTConfig().getCurveDialogs();

    if( t[0] == _T("curve") )
    {
        currentCurve = t[1];
        curves[currentCurve] = new VemsCurveDialogDesc();
//        Logger::logf( LOG_MSG, _T(" V3iniParser::parseCurveEditor(): ") + t[1] );
        curves[currentCurve]->name = t[1];
        curves[currentCurve]->title = t[2];
        curves[currentCurve]->showHorizontalHelpLine = false;
        curves[currentCurve]->showVerticalHelpLine   = false;
    }
    else if( t[0] == _T("topicHelp") )
    {
        curves[ currentCurve ]->topicHelp = t[1];
    }else if( t[0] == _T("columnLabel") ){
        curves[ currentCurve ]->columnLabelX = t[1];
        curves[ currentCurve ]->columnLabelY = t[2];
    }else if( t[0] == _T("xAxis") ){
        curves[ currentCurve ]->minX = t.getDouble( 1 );
        curves[ currentCurve ]->maxX = t.getDouble( 2 );
        curves[ currentCurve ]->divisionsX = t.getInt( 3 );
        curves[ currentCurve ]->labelFormatX = t[4];
    }else if( t[0] == _T("yAxis") ){
        curves[ currentCurve ]->minY = t.getDouble( 1 );
        curves[ currentCurve ]->maxY = t.getDouble( 2 );
        curves[ currentCurve ]->divisionsY = t.getInt( 3 );
        curves[ currentCurve ]->labelFormatY = t[4];
    }else if( t[0] == _T("xBins") ){
        curves[ currentCurve ]->xBins       = t[1];
        curves[ currentCurve ]->xRealTime   = t[2];
        curves[ currentCurve ]->isXBinsReadOnly = ( t[3] == _T("readonly") );
    }else if( t[0] == _T("yBins") ){
        curves[ currentCurve ]->yBins = t[1];
        curves[ currentCurve ]->yRealTime   = t[2];
    }else if( t[0] == _T("gauge") ){
        curves[ currentCurve ]->gauge = t[1];
    }else if ( t[0].IsSameAs( _T("showHorizontalHelpLine"), false ) ) {
        curves[currentCurve]->showHorizontalHelpLine = true;
    }else if ( t[0].IsSameAs( _T("showVerticalHelpLine"), false ) ) {
        curves[currentCurve]->showVerticalHelpLine = true;
    }else if ( t[0].IsSameAs( _T("xBinsSteps"), false ) ) {
        curves[currentCurve]->xBinsSmallStep = t.getDouble( 1 );
        curves[currentCurve]->xBinsLargeStep = t.getDouble( 2 );
    }else if ( t[0].IsSameAs( _T("yBinsSteps"), false ) ) {
        curves[currentCurve]->yBinsSmallStep = t.getDouble( 1 );
        curves[currentCurve]->yBinsLargeStep = t.getDouble( 2 );
    }
}

void V3iniParser::parseSuperCurveEditor( const Tokenizer& t )
{
#ifndef COMM_MODEL_WITH_GUI
    static wxString currentCurve = wxEmptyString;
    SuperCurveDescriptorH& curves = getMTConfig().getSuperCurveDialogs();

    if( t[0] == _T("supercurve") )
    {
        LOGP( t[1] );
        currentCurve = t[1];
        curves[currentCurve] = new SuperCurveDescriptor();
        curves[currentCurve]->name = t[1];
//        curves[currentCurve]->title = t[2];
    } else if ( t[0].IsSameAs( _T("curve"), false ) ) {
        curves[currentCurve]->records.Add( SuperCurveRecord( t[1], t[2] ) );
    }
#endif
}

void V3iniParser::parseFrontPage(){}
void V3iniParser::parseRunTime(){}
void V3iniParser::parseTuning(){}


void V3iniParser::parseUserDefined(const Tokenizer& t)
{
    static wxString currentDialog = _T("default");

    UserDialogList* udl = getMTConfig().getUserDialogList();

    if(t[0] == _T("dialog")){
        currentDialog = t[1];
        udl->add(t[1], t[2], t.getInt(3));
    }

    if(t[0] == _T("field")){
        wxString label = t[1].AfterFirst( _T('#') );
        if ( label.IsEmpty() ) { label = t[1]; }
        udl->add(currentDialog, label, t[2], t[3]);
        getMTConfig().fillDialogNamesBySymbol(currentDialog, t[2]);
    }
    else if( t[0].IsSameAs( _T("topicHelp"), false ) )
    {
        udl->setHelpID( currentDialog, t[1] );
    }

    // found help
    static wxString currentHelpDialog = _T("default");

    if( t[0].IsSameAs( _T("help"), false ) )
    {
        currentHelpDialog = t[1];
    }
    else if ( t[0].IsSameAs( _T("text"), false ) )
    {
        getMTConfig().m_iniHelps[ currentHelpDialog ] += t[1];
    }
}

void V3iniParser::parseGPSData( const Tokenizer& t )
{
	//Logger::logf( LOG_MSG, _T("V3iniParser::parseGPSData()") );
    if( t[1] == _T("size") )
    {
        //todo set gpsdata size
        return;
    }

    parseOutputChannels( t );
}

void V3iniParser::parseAIMData( const Tokenizer& t )
{
    if( t[1] == _T("size") )
    {
        //todo set aimdata size
        return;
    }
    parseOutputChannels( t );
}

void V3iniParser::parseSIPRDescriptor( const Tokenizer& t )
{
    SIPRDescriptor* pDesc = getMTConfig().getSIPRDescriptor();

    if ( t[0].IsSameAs( _T("realtime_page"), false ) )
    {
        pDesc->addRealTimePage( t[1], t[2] );
    }
    else if( t[0].IsSameAs( _T("burnCommand"), false ) )
    {
        pDesc->setBurnCommand( t[1] );
    }
    else if ( t[0].IsSameAs( _T("page"), false ) )
    {
        pDesc->addPage( t.getInt(1), t[2], t[3], t[4] );

        for ( size_t i = 5; i < t.getSize(); ++i )
        {
            pDesc->addSIPRTableToPage( t.getInt(1), t[i] );
        }
    }
    else if ( t[0].IsSameAs( _T("sipr_table"), false ) )
    {
        pDesc->addSIPRTable( t[1], t.getInt(2), t.getInt(3), t[4] );
    }
    else if ( t[0].IsSameAs( _T("sipr_config"), false ) )
    {
        pDesc->addSIPRConfig( t[1], t.getInt(2), t.getInt(3), t[4] );
    }


}

void V3iniParser::parseSIPRData( const Tokenizer& t )
{
    if( t[1] == _T("size") )
    {
        //todo set aimdata size
        return;
    }
    parseOutputChannels( t );
}



void V3iniParser::parseOutputChannels(const Tokenizer& t)
{

    if ( t[0].IsSameAs( _T("ochGetCommand") ) )
    {
        getMTConfig().getSIPRDescriptor()->setOchGetCommand( t[1] );
        return;
    }

    if(t.getType(1) == T_EXP || t[1] == _T("scalar")){
        Symbol* symbol = NULL;
        symbol = new Symbol;

        try{
            if(t.getType(1) == T_EXP){
                symbol->setExpr(t[0],t[1]);
                getLogMsg().Append(_T("add expr : ") + symbol->toString()); //TEST
                symbol->setGaugeType(t[2]);
                symbol->setTitle(t[3]);
            }

            if(t[1] == _T("scalar")){
                symbol->setBasic(t[0],t[4], false, getPage(), t.getInt(3));
                symbol->setInputType(  SiprSensorType::SIPR_RT_MT_NORMAL );
                parseTypeSpec(symbol,t[2]);
                symbol->setFormat(t.getDouble(5), t.getDouble(6), t.getDouble(7), t.getDouble(8), t.getInt(9));

                symbol->setGaugeType(t[7]);
                symbol->setTitle(t[8]);

                getLogMsg().Append(_T("add ") + t[1] + _T(" : ") + symbol->toString()); //TEST
            }

            if(getMTConfig().getSymbol(t[0]) != NULL){
                getErrorLog().Add( _T("Duplicate symbol! : ") + symbol->getName()
                                + wxString::Format(_T(" - line : %d"),m_lineNumber));
            }

            if( getSection() == MT_GPSDATA )
            {
                symbol->setInputType(  SiprSensorType::GPS );
                symbol->setGroup( getMTConfig().getGroup( MT_GPSDATA ) );
                //Logger::logf( LOG_MSG, _T("V3iniParser::parseOutputChannels() - group addres : %p, symbol : ") +  symbol->getName(),
                //getSection().getGroup( MT_GPSDATA ) );
            }

            if( getSection() == MT_AIMDATA )
            {
                symbol->setInputType(  SiprSensorType::SIPR_AIM );
                symbol->setGroup( getMTConfig().getGroup( MT_AIMDATA ) );
                symbol->setChannel( t.getInt( 7 ) );
            }

            if ( getSection() == MT_SIPR_DATA )
            {
                symbol->setInputType( SiprSensorType::PAGE_0 );
            }

            getMTConfig().addSymbol(symbol);

        }catch(ParserException e){
            if ( symbol != NULL ) delete symbol;
            getErrorLog().Add(e.toString());
        }
    }
}

void V3iniParser::parseDatalog(const Tokenizer& t)
{
    if(t[0] != _T("entry")) return;

    getMTConfig().addDataLogEntry(t[1], t[2], t[3], t[4]);
}

void V3iniParser::parseTypeSpec(Symbol* symbol, const wxString& typeField)
{
    const  wxString RE_TYPE  = _T("(U|S)(08|16|24|32)");
    wxRegEx reType;
    reType.Compile( RE_TYPE, wxRE_ADVANCED);

    if(!reType.Matches(typeField)) { throw ParserException(_T("Type field error!"), m_lineNumber); }

    bool isSigned;

    if(reType.GetMatch(typeField, 1) == _T("S") ) { isSigned = true; }
    else if(reType.GetMatch(typeField, 1) == _T("U") ) { isSigned = false; }
    else{ throw ParserException(_T("Type sign error!"), m_lineNumber); }

    int sizeOf;

    if(reType.GetMatch(typeField, 2) == _T("08") ){ sizeOf = 1; }
    else if( reType.GetMatch(typeField, 2) == _T("16") ){ sizeOf = 2; }
    else if( reType.GetMatch(typeField, 2) == _T("24") ){ sizeOf = 3; }
    else if( reType.GetMatch(typeField, 2) == _T("32") ){ sizeOf = 4; }
    else{ throw ParserException(_T("Type size error!"), m_lineNumber); }

    symbol->setType( isSigned, sizeOf);
}

void V3iniParser::parseArraySpec(Symbol* symbol, const wxString& as )
{
    wxString arrayspec = as;

    arrayspec.Replace(_T(" "), _T(""));
    //arrayspec.Replace(_T("\t"), _T(""));

    const wxString RE_ARRAY_SPEC = _T("^(Y:|X:|)(?:(\\d+)|(\\d+)x(\\d+))");

    bool _xOrder = true;
    int  _order  = 0; //0 - scalar, 1,2 - 1,2 dimension array
    long _dim1 = 1;
    long _dim2 = 1;


    wxRegEx reArraySpec;
    reArraySpec.Compile( RE_ARRAY_SPEC,  wxRE_ICASE | wxRE_ADVANCED );

    if(!reArraySpec.Matches(arrayspec)) { throw ParserException(_T("Array Shape error!"), m_lineNumber ); }

    if((reArraySpec.GetMatch(arrayspec ,1) == _T("X:")) || (reArraySpec.GetMatch(arrayspec ,1 )).IsEmpty())
    {
        _xOrder = true;
    }
    else if(reArraySpec.GetMatch(arrayspec ,1) == _T("Y:"))
    {
        _xOrder = false;
    }
    else { throw ParserException(_T("Array xOrder error!"), m_lineNumber ); }


    if(!reArraySpec.GetMatch(arrayspec, 2).IsEmpty() )
    {
        if( reArraySpec.GetMatch(arrayspec, 2).ToLong(&_dim1))
        {
            _order = 1;
        }
        else { throw ParserException(_T("Array size error!"), m_lineNumber); }
    }

    if(!reArraySpec.GetMatch(arrayspec, 3).IsEmpty() )
    {
        if(reArraySpec.GetMatch(arrayspec, 3).ToLong(&_dim1))
        {
            _order = 2;
        }
        else{ throw ParserException(_T("Array size 1 error!"), m_lineNumber); }

        if(!reArraySpec.GetMatch(arrayspec, 4).IsEmpty() )
        {
            if(!reArraySpec.GetMatch(arrayspec, 4).ToLong(&_dim2) )
            {
                throw ParserException(_T("Array size 2 error!"), m_lineNumber);
            }
        }
        else{ throw ParserException(_T("Missing size 2 error!"), m_lineNumber); }
    }

    symbol->setArray( _xOrder, _order, _dim1, _dim2);
}


void V3iniParser::parseBitSpec(Symbol* symbol, const Tokenizer& t)
{
    wxString bitspec = t[4];

    const wxString RE_BITSPEC = _T("^(\\d+):(\\d+)\\+?(\\d*)");

    long _bitLo = -1;
    long _bitHi = -1;
    long _bitOfs = 0;
    wxArrayString _fields;

    wxRegEx reBitSpec;
    reBitSpec.Compile( RE_BITSPEC, wxRE_ADVANCED );

    if(!reBitSpec.Matches(bitspec)) { throw ParserException(_T("BitField error!"), m_lineNumber); }

    if(!((reBitSpec.GetMatch(bitspec, 1)).ToLong(&_bitLo))) { throw ParserException(_T("loBit value error!"), m_lineNumber ); }
    if(!((reBitSpec.GetMatch(bitspec, 2)).ToLong(&_bitHi))) { throw ParserException(_T("hiBit value error!"), m_lineNumber ); }

    if(!reBitSpec.GetMatch(bitspec, 3).IsEmpty())
    {
        if(!((reBitSpec.GetMatch(bitspec, 3)).ToLong(&_bitOfs))) { throw ParserException( _T("offsBit value error!"), m_lineNumber); }
    }

    int cnt = reBitSpec.GetMatchCount();
    wxString  _bitData;
    if(cnt > 4){
       _bitData = reBitSpec.GetMatch(bitspec, 4);
    }

    for(size_t i=5; i<t.getSize(); i++)
    {
        if ( t.getType(i) == T_EXP ) {
            symbol->setValidationExpr( t[i] );
            break;
        }

        _fields.Add(t[i]);
    }

    symbol->setBit(_bitLo, _bitHi, _bitOfs, _fields );
}


void V3iniParser::parseSymbolAlias( const Tokenizer& t )
{
    Symbol* s = getMTConfig().getSymbol( t[0] );

    if ( s )
    {
        Symbol* alias = new Symbol( *s );
        alias->setName( t[1] );
        getMTConfig().addSymbol( alias );
        s->setAlias( t[1] );
    }
}


void V3iniParser::parseConfigTablesSavingOrder( const Tokenizer& t )
{
    //getMTConfig().addSymbol( alias );
}


void V3iniParser::parseUserData( const Tokenizer& t )
{
    UserDataDesc* desc = new UserDataDesc( t[0], t[1] );
    getMTConfig().m_userDataDescA.Add( desc );
}

void V3iniParser::parseExpressionGrid( const Tokenizer& t )
{
    static ExpressionGridDesc* descriptor = NULL;

    if ( t[0] == _T("name") )
    {
        descriptor = new ExpressionGridDesc();
        descriptor->name    = t[1];
        descriptor->title   = t[2];
        descriptor->disableCol = -1;
        getMTConfig().m_expressionGridDescriptors.Add( descriptor );
        return;
    }

    if ( !descriptor )
    {
        Logger::logf( LOG_MSG, _T("V3iniParser::parseExpressionGrid() - descriptor is null!") );
        return;
    }

    size_t rows = (size_t)descriptor->cells.shape()[0];
    size_t cols = (size_t)descriptor->cells.shape()[1];

    if ( t[0] == _T("row") )
    {
        size_t newRows = rows + 1;
#ifndef __WINCE_STANDARDSDK__
		size_t newCols = std::max( cols, ( t.getSize() - 1 ) );
#else
        size_t newCols = max( cols, ( t.getSize() - 1 ) );
#endif
        ExpressionCellDescriptorMatrix::extent_gen extents;
        descriptor->cells.resize(extents[newRows][newCols]);

        for ( size_t col = 0; col < newCols; ++col )
        {
            ExpressionCellDescriptor& cell = descriptor->cells[newRows-1][col];
            cell.text = t[col+1];
            cell.width = -1;
            cell.precision = -1;
            cell.align = -1;
        }
    }
    else if ( t[0] == _T("col_format") )
    {
        int col = t.getInt(1);

        if ( col >= (int)cols ) return;

        for ( size_t row = 0; row < rows; ++row )
        {
            ExpressionCellDescriptor& cell = descriptor->cells[row][col];
            cell.precision = t.getInt(2);
            cell.align = ExpressionGridDescHelper::parseAlign( t[3] );
        }
    }
    else if ( t[0] == _T("cell_format") )
    {
        int row = t.getInt(1);
        int col = t.getInt(2);

        if ( col >= (int)cols || row >= (int)rows ) return;

        ExpressionCellDescriptor& cell = descriptor->cells[row][col];

        cell.precision = t.getInt(3);
        cell.align = ExpressionGridDescHelper::parseAlign( t[4] );
    }
    else if ( t[0] == _T("link") )
    {
        int row = t.getInt(1);
        int col = t.getInt(2);

        if ( col >= (int)cols || row >= (int)rows ) return;

        ExpressionCellDescriptor& cell = descriptor->cells[row][col];

        ExpressionCellLink link;

        link.dialogId = t[3];

        for ( size_t i = 4; i < t.getSize(); ++i )
        {
            if ( t.getType(i) == T_EXP ) {
                link.enableExpr = t[i];
                if ( i < ( t.getSize() - 1 ) ) {
                    LOGP( wxString::Format( _T("TOO MANY PARAMETERS! - m_lineNumber: %d"), m_lineNumber ) );
                }
                break;
            }

            if ( t.getType(i) == T_ID )
            {
                link.variables.Add( t[i] );
                continue;
            }

            link.fieldIds.Add( t[i] );
        }

        cell.links.Add( link );
    }
    else if ( t[0] == _T("disable_col") )
    {
        int disableCol = t.getInt(1);

        if ( disableCol < -1 || disableCol > (int)(cols - 1) )
        {
            wxString msg = wxString::Format( _T("Invalid disableCol! [%d]"), disableCol );
            LOGP( msg );
            return;
        }

        descriptor->disableCol = t.getInt(1);
    }
    else if ( t[0] == _T("help") )
    {
        int row = t.getInt(1);
        int col = t.getInt(2);

        if ( col >= (int)cols || row >= (int)rows ) return;

        ExpressionCellDescriptor& cell = descriptor->cells[row][col];
        cell.help = t[3];
    } else if ( t[0] == _T("line_enable") ) {
        int line = t.getInt(1);
        wxString expr = t[2];
        descriptor->linesEnableExpr[line] = expr;
    } else if ( t[0] == _T("cell_text") ) {
        int row = t.getInt(1);
        int col = t.getInt(2);

        if ( col >= (int)cols || row >= (int)rows ) return;

        ExpressionCellDescriptor& cell = descriptor->cells[row][col];

        ExprCellText cellText;
        cellText.enableExpr = t[3];
        cellText.text = t[4];
        cell.texts.Add( cellText );
    } else if ( t[0] == _T("cell_color") ) {
        int row = t.getInt(1);
        int col = t.getInt(2);

        if ( col >= (int)cols || row >= (int)rows ) return;

        ExpressionCellDescriptor& cell = descriptor->cells[row][col];

        ExprCellColor cellColor;
        cellColor.enableExpr = t[3];
        cellColor.bgColor = t[4];
        cellColor.fgColor = t[5];
        cell.colors.Add( cellColor );
    }
}

void V3iniParser::parseInterpolationWidget( const Tokenizer& t )
{
    static wxString currentWidget = wxEmptyString;
    InterpolationWidgetDescH& widgets = getMTConfig().m_interpolationWidgets;

    if( t[0] == _T("interpolationWidget") )
    {
        currentWidget = t[1];
        widgets[currentWidget] = new InterpolationWidgetDescriptor();
        widgets[currentWidget]->id = t[1];
        widgets[currentWidget]->title = t[2];
    }
    else if( t[0] == _T("realtime") )
    {
        widgets[ currentWidget ]->realtime = t[1];
    }
    else if( t[0] == _T("xAxis") )
    {
        widgets[ currentWidget ]->xAxis.min         = t.getDouble(1);
        widgets[ currentWidget ]->xAxis.max         = t.getDouble(2);
        widgets[ currentWidget ]->xAxis.divisions   = t.getInt(3);
        widgets[ currentWidget ]->xAxis.format      = t[4];
        widgets[ currentWidget ]->xAxis.title       = t[5];
        widgets[ currentWidget ]->xAxis.unit        = t[6];
    }
    else if( t[0] == _T("yAxis") )
    {
        widgets[ currentWidget ]->yAxis.min         = t.getDouble(1);
        widgets[ currentWidget ]->yAxis.max         = t.getDouble(2);
        widgets[ currentWidget ]->yAxis.divisions   = t.getInt(3);
        widgets[ currentWidget ]->yAxis.format      = t[4];
        widgets[ currentWidget ]->yAxis.title       = t[5];
        widgets[ currentWidget ]->yAxis.unit        = t[6];
    }
    else if( t[0] == _T("x0") )
    {
        int type = convertTokenTypeToInterpolationWidgetFieldType( t.getType(2) );

        widgets[ currentWidget ]->x0.type         = type;
        widgets[ currentWidget ]->x0.label        = t[1];
        widgets[ currentWidget ]->x0.value        = t[2];
        widgets[ currentWidget ]->x0.expression   = t[3];
    }
    else if( t[0] == _T("x1") )
    {
        int type = convertTokenTypeToInterpolationWidgetFieldType( t.getType(2) );

        widgets[ currentWidget ]->x1.type         = type;
        widgets[ currentWidget ]->x1.label        = t[1];
        widgets[ currentWidget ]->x1.value        = t[2];
        widgets[ currentWidget ]->x1.expression   = t[3];
    }
    else if( t[0] == _T("y0") )
    {
        int type = convertTokenTypeToInterpolationWidgetFieldType( t.getType(2) );

        widgets[ currentWidget ]->y0.type         = type;
        widgets[ currentWidget ]->y0.label        = t[1];
        widgets[ currentWidget ]->y0.value        = t[2];
        widgets[ currentWidget ]->y0.expression   = t[3];
    }
    else if( t[0] == _T("y1") )
    {
        int type = convertTokenTypeToInterpolationWidgetFieldType( t.getType(2) );

        widgets[ currentWidget ]->y1.type         = type;
        widgets[ currentWidget ]->y1.label        = t[1];
        widgets[ currentWidget ]->y1.value        = t[2];
        widgets[ currentWidget ]->y1.expression   = t[3];

        Logger::logf( LOG_MSG, widgets[ currentWidget ]->toString() );
    }
}

int V3iniParser::convertTokenTypeToInterpolationWidgetFieldType( type tokenType )
{
    switch ( tokenType )
    {
        case T_EXP:
            return InterpolationWidgetInputField::EXPRESSION;
        case T_ID:
            return InterpolationWidgetInputField::VARIABLE;
        case T_NUM:
            return InterpolationWidgetInputField::NUMBER;
        default:
            return InterpolationWidgetInputField::INVALID;
    }
}

int V3iniParser::tokenizeToArray(wxString& text, const wxString& token, wxArrayString& a)
{
    wxStringTokenizer tkz(text, token);
    int countToken = 0;

    while ( tkz.HasMoreTokens() )
    {
        wxString tokenStr = tkz.GetNextToken();
        a.Add(tokenStr );
        countToken++;
    }

   return countToken;
}

