/*
correction for lambda (l) table:

fuelcalc.c :

lambda=0.9
corr.lambda= 256 / 0.9 - 200 = 84 = 0x54

#define LAMBDACORR_OFFSET (200)

res = mult16_16(res, (uint16_t)(corr.lambda) + LAMBDACORR_OFFSET) >> 8; // effectively * 200..455 / 256

Megatune nem tud reciprokot szmolni, ezrt:

comm.h:
#define MT_LAMBDA_OFFSET 73
#define LAMBDA_INT_OFFSET 200

mt_val = mult16_8((corr.lambda + LAMBDA_INT_OFFSET), 72) >> 7;
                mt_val = (mult16_8(div_65535_x(mt_val), 72) >> 7) - MT_LAMBDA_OFFSET;
                return mt_val & 0xFF;

65535/((84+200) * 72 / 128)      * 72  / 128 - 73  = 157.76 (MT parancsban ezt kldeni).

egyszerusitve:
65535/(( 84 +200))  - 73
*/
#include "FirmwareManager.h"

#include <wx/filename.h>

#include "ConfigCriteria.h"
#include "ConfigCriteriaData.h"
#include "ConfigReplace.h"
#include "ConfigReplacerMgr.h"
#include "ConfigValidatorMgr.h"

#include "../logger.h"
#include "../MTConfigHelper.h"
#include "../VemsMsgListCtrl.h"
#include "../FirmwareInfo.h"
#include "../DirHelper.h"

#include "../common/MenuLinkHelper.h"
#include "../common/VemsContext.h"
#include "../common/wxArrayHelper.h"

#include "../fileparser/parseheader.h"
#include "../fileparser/FirmwareConfigParser.h"
#include "../fileparser/ConfigCriteriaParser.h"

#include "../stdatamanager/STDataModelBase.h"
#include "../stdatamanager/STDataIOVemsCfg.h"
#include "../vemsprotocol/CommandRequestBase.h"
#include "../vemsprotocol/VemsCommand.h"
#include "../vemsdialog/UserDialogBase.h"
#include "../UserLog.h"

#include "FirmwareConfigLoader.h"
#include "../vemsprotocol/ECUInfo.h"

#include "../common/ConfigSwitchTargetTypes.h"


FWConfigUploadManager::FWConfigUploadManager(UserDialogBase* userDialog, FirmwareInfo* ecuFwInfo) :
    //m_fcDataFile    ( _T("File") ),
    m_fcDataECU     ( _T("ECU") ),
    m_model         ( m_cfgSymbolTable )
{
    m_userDialog = userDialog;
    m_pConfigParser     = NULL;
    m_iState            = UMS_WAIT;

    m_ecuFwInfo = ecuFwInfo;

    m_firmwareConfigLoader = new FirmwareConfigLoader();
    FirmwareConfigData& fcDataFile = m_firmwareConfigLoader->getFirmwareConfigData();
    m_fcDataFile = &fcDataFile;

    m_uploadCount = 0;
}


FWConfigUploadManager::~FWConfigUploadManager()
{
    delete m_firmwareConfigLoader;
}


void FWConfigUploadManager::clear()
{
    WX_CLEAR_HASH_MAP( SymbolTable, m_cfgSymbolTable );
    m_fcDataFile->clear();
    m_fcDataECU.clear();
    m_model.clear();

    configsVsHeader.clear();
    parsedTables.clear();
    menuLinks.clear();

    if ( m_pConfigParser )
    {
        delete m_pConfigParser;
        m_pConfigParser = NULL;
    }

    m_uploadCount = 0;
    m_uploadData.clear();

    m_iState = UMS_WAIT;
}


bool FWConfigUploadManager::createUploadConfig( const wxArrayString& configFiles )
{

    if( !parseConfigFromFiles( configFiles ) ){
        ///
        return false;
    }

    switch( m_ecuFwInfo->getDevice() )
    {
        case ECUInfo::DEVICE_V3:
            return validateV3Config();
        case ECUInfo::DEVICE_ROUND:
            return validateRoundConfig();
        default:
            return false;
    }
}

bool FWConfigUploadManager::validateRoundConfig()
{
    ConfigCriteriaData criteriaData;
    ConfigCriteriaParser::parse( criteriaData, _T("Round-ConfigCriteria.xml") );

    bool hasNoError = true;
    const SymbolTable& targetSymbolTable = m_firmwareConfigLoader->getSymbolTable();
    FirmwareConfigDifferences* fcDiff = m_fcDataFile->compare( targetSymbolTable );

    configsVsHeader = *fcDiff;

     const wxArrayString& missingScalar = fcDiff->missingScalars;
    menuLinks += MenuLinkHelper::getHtmlTable(missingScalar, *MTConfigHelper::getMTConfigData() );

    if ( !fcDiff->getSummary().IsEmpty() ) { hasNoError = false; }

    VemsMsgListCtrl* messageViewer = new VemsMsgListCtrl(m_userDialog);
    messageViewer->Show();

    VariableDataDH vd;
    setVariableData(vd, *m_fcDataFile);

    ///replace( m_fcDataFile, fcDiff->getMissingScalarA() );
    ConfigReplacerMgr cfReplacer(&criteriaData);
    cfReplacer.registerVemsMessageViewer(messageViewer);

    /// remove replaced from missing symbols
    wxArrayString remainedA = fcDiff->missingScalars;
    cfReplacer.replace( *m_fcDataFile, fcDiff->missingScalars,  remainedA);

     ///Load ramained missing scalars and tables from default config;
    loadDefaultConfig(remainedA, fcDiff->missingTables);

    ///Validator
    ConfigValidatorMgr cfValidator(&criteriaData, MTConfigHelper::getMTConfigData() );
    cfValidator.registerVemsMessageViewer(messageViewer);
    cfValidator.validate(vd);

    return true;
}

bool FWConfigUploadManager::validateV3Config()
{
    ConfigCriteriaData criteria;
    ConfigCriteriaParser::parse( criteria, _T("ConfigCriteria.xml") );

    bool hasNoError = true;
    const SymbolTable& targetSymbolTable = m_firmwareConfigLoader->getSymbolTable();
    FirmwareConfigDifferences* fcDiff = m_fcDataFile->compare( targetSymbolTable );

    configsVsHeader = *fcDiff;
    m_fcDataFile->tableDataToArrayString( parsedTables );

    const wxArrayString& missingScalar = fcDiff->missingScalars;
    menuLinks += MenuLinkHelper::getHtmlTable(missingScalar, *MTConfigHelper::getMTConfigData() );

    if ( !fcDiff->getSummary().IsEmpty() ) { hasNoError = false; }

    //wxObject* obj = VemsContext::getObject(_T("ConfigCriteriaData"));
    //if(! obj) return hasNoError;

    VemsMsgListCtrl* messageViewer = new VemsMsgListCtrl(m_userDialog);
    messageViewer->Show();

    //ConfigCriteriaData* criteriaData = (ConfigCriteriaData*) obj;
    VariableDataDH vd;
    setVariableData(vd, *m_fcDataFile);

    ///replace( m_fcDataFile, fcDiff->getMissingScalarA() );
    ConfigReplacerMgr cfReplacer(&criteria);
    cfReplacer.registerVemsMessageViewer(messageViewer);

    /// remove replaced from missing symbols
    wxArrayString remainedA = fcDiff->missingScalars;
    cfReplacer.replace( *m_fcDataFile, fcDiff->missingScalars,  remainedA);

    ///Confswitch hack remove confswitch
    wxString filterFile = DirHelper::GetPath(dDEFAULTCONFIGS) + _T("v3_configfilter.vemscfg");
    Logger::logf(LOG_MSG, _T("FWConfigUploadManager::validateV3Config() filter file: ")+filterFile);
    filter(filterFile);

    /// isPartial cfg3_
    if ( m_fcDataFile->isPartialCfg() ){
        Logger::logf(LOG_MSG, _T("FWConfigUploadManager::validateV3Config() cfg is PARTIAL!"));
        UserLog::uLog(ULOG_INFO, _T("Detected: partial config!"));
        return hasNoError;
    }else{
        Logger::logf(LOG_MSG, _T("FWConfigUploadManager::validateV3Config() cfg is FULL!"));
        UserLog::uLog(ULOG_INFO, _T("Detected: full config!"));
    }

    /// isSerialnum different
    if( m_firmwareConfigLoader->isAlienConfig() ){
         /// Load calibrations vemscfg file
        wxString calibrationsFile = DirHelper::GetPath(dDEFAULTCONFIGS) + _T("calibrations.vemscfg");
        Logger::logf(LOG_MSG, _T("FWConfigUploadManager::validateV3Config() calibrations file: ")+calibrationsFile);
        filter(calibrationsFile);
    }

    ///Load ramained missing scalars and tables from default config;
    loadDefaultConfig(remainedA, fcDiff->missingTables);

    safetyConfigSetup();

    ///Validator
    ConfigValidatorMgr cfValidator(&criteria, MTConfigHelper::getMTConfigData() );
    cfValidator.registerVemsMessageViewer(messageViewer);
    cfValidator.validate(vd);


    //try to calculate missing keys
   // Logger::logf(LOG_MSG, _T("FirmwareManager::createUploadConfig() size:%i"), m_replaceHash.size() );
    //replace( m_fcDataFile, fcDiff->getMissingScalarA() );
    //validate keys
    //validate( m_fcDataFile );
    return  hasNoError;
}

void FWConfigUploadManager::filter(const wxString& file)
{
    if( wxFileName::FileExists(file) ){
        Logger::logf(LOG_MSG, _T("FWConfigUploadManager::filter file: ")+file);
        //UserLog::uLog(ULOG_INFO, _T("!"));
        FirmwareConfigData filterData;
        FirmwareConfigParser fwConfParser(&filterData );
        fwConfParser.parse(file);
        m_fcDataFile->filter(filterData);
    }else{
        Logger::logf(LOG_MSG, _T("FWConfigUploadManager::filter file NOT FOUND: ")+file);
    }
    //m_fcDataFile->filterScalar( _T("confswitch") );
}

void FWConfigUploadManager::loadDefaultConfig(const wxArrayString& missingScalars, const wxArrayString& missingTables)
{
    /// Checks missings
    if( missingScalars.IsEmpty() && missingTables.IsEmpty() ) return;

    /// Find default config;
    wxString defaultConfigFile = MTConfigHelper::getDefaultConfigFile();
    if(defaultConfigFile == wxEmptyString){
        Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() default config NOT FOUND!"));
        return;
    }else{
        Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() File: ") + defaultConfigFile );
    }

    /// show dialog
    size_t missingCount = missingScalars.GetCount() + missingTables.GetCount();

    wxString message = wxString::Format( _T("%d missing symbols have been found.\nUse default config to load missing symbols?"), missingCount );
    int answer = wxMessageBox( message, _T("Config upload"), wxYES_NO );

    if ( answer == wxNO ) return;

    Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() defaultConfigFile found: ")+ defaultConfigFile );
    UserLog::uLog(ULOG_INFO, _T("Default config file: ") + defaultConfigFile);

    /// Load default vemscfg file
    FirmwareConfigData defaultConfigData;
    FirmwareConfigParser fwConfParser(&defaultConfigData );
    fwConfParser.parse(defaultConfigFile);

    /// Scalars
    if( !missingScalars.IsEmpty() ){
        /// fill missing keys from default config;
        for(size_t i = 0; i < missingScalars.GetCount(); ++i){
            wxString key = missingScalars.Item(i);
            int value;
            bool isFound = defaultConfigData.getScalar(key, value);
            if(isFound){
                m_fcDataFile->addScalar(key, value);
                wxString msg =_T("Loaded Default key: ") + key+ _T(" value: ") + wxString::Format(_T("%i"), value);
                //UserLog::uLog(ULOG_INFO, msg);
            }
        }
    }

    /// Tables
    if( !missingTables.IsEmpty() ){
        Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() missing tables: %i"), missingTables.GetCount() );

        for(size_t i = 0; i < missingTables.GetCount(); ++i){
            wxString key = missingTables.Item(i);
            VemsMatrix<uint8_t>* defTable = defaultConfigData.getTable(key);
            if(defTable){
                wxString msg =_T("Loaded default table: ")+ key;
                m_fcDataFile->addTable(key, defTable);
                Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() ")+msg);
                UserLog::uLog(ULOG_INFO, msg);
            }else{
                wxString msg =_T("NOT FOUND default table: ")+ key;
                Logger::logf(LOG_MSG, _T("FirmwareManager::loadDefaultConfig() ")+msg);
                UserLog::uLog(ULOG_INFO, msg);
            }
        }
    }
}

void FWConfigUploadManager::safetyConfigSetup()
{
    int prohibit_conf = 0;
    bool isFound = m_fcDataFile->getScalar(_T("prohibit_config"), prohibit_conf);
    if(!isFound) return; // old ECU fw

    bool isIButton = ( prohibit_conf & (1 << 7)) != 0;
    Logger::logf( LOG_MSG, _T("FWConfigUploadManager::safetyConfigSetup() prohibit_conf=%d , isIbutton=%d"),prohibit_conf, isIButton );

    if(!isIButton){
        LOG
        int prohibit = 0;
        m_ecuFwInfo->getProhibit(prohibit);
        m_fcDataFile->addScalar(_T("prohibit_h"), (prohibit >> 8) & 255);
        m_fcDataFile->addScalar(_T("prohibit_l"), prohibit & 255);
    }
}

bool FWConfigUploadManager::parseConfigFromFiles(const wxArrayString& configFiles )
{
    m_firmwareConfigLoader->setDestinationFwInfo(m_ecuFwInfo);
    return m_firmwareConfigLoader->parse(configFiles);

/*
    ParseHeader headerParser( m_cfgSymbolTable );
    headerParser.parse( headerFile ) ;

    FirmwareConfigParser configParser( &m_fcDataFile );

    for ( size_t i = 0; i < configFiles.Count(); ++i )
    {
        configParser.parse( configFiles[ i ] );
    }
*/
    //setTablesOffsetAndSize();
}

/*
void FWConfigUploadManager::setTablesOffsetAndSize()
{
    //set table size
    const String2IntHash& tableSizeH = m_fcDataFile.getTableSizeH();
    const String2IntHash& tableColCountH = m_fcDataFile.getTableColCountH();

    String2IntHash::const_iterator IT = tableSizeH.begin();
    for ( ; IT != tableSizeH.end(); ++IT )
    {
        wxString symbolName = IT->first;
        size_t colSize = tableColCountH.find( symbolName )->second;
        size_t rowSize = tableSizeH.find( symbolName )->second / colSize;
        m_cfgSymbolTable[symbolName] = Symbol::createArray( symbolName, colSize, rowSize, -1, -1);
    }
}
*/
void FWConfigUploadManager::updateModel()
{
    m_firmwareConfigLoader->load(&m_model);
}

void FWConfigUploadManager::uploadConfig( CommandRequestBase* controller, int currentConfigType, int uploadConfigType )
{
    m_firmwareConfigLoader->load(&m_model);
    updateECU( controller, currentConfigType, uploadConfigType );
}

void FWConfigUploadManager::updateECU( CommandRequestBase* controller, int currentConfigType, int uploadConfigType )
{
    //writeAllDirty( m_model, controller );
    writeUploadData( controller );

    int device = m_ecuFwInfo->getDevice();

    if ( device == ECUInfo::DEVICE_V3 ) {
        requestAdvancedBurn( controller, currentConfigType, uploadConfigType );
    } else {
        controller->request( new BurnCommand() );
    }
}

void FWConfigUploadManager::requestAdvancedBurn( CommandRequestBase* controller, int currentConfigType, int uploadConfigType )
{
    size_t sleepAfterWrite = 0;

    switch ( uploadConfigType )
    {
        case CFG_BOTH:
            controller->request( new StringCommand( _T("ManmcS0mbye"), 10000, true ) );
            controller->request( new StringCommand( _T("ManmcS1mbye"), sleepAfterWrite, true ) );
            break;

        case CFG_A:
            switch ( currentConfigType )
            {
                case CFG_A:
                    controller->request( new StringCommand( _T("ManmcS0mbye"), sleepAfterWrite, true ) );
                    break;

                case CFG_B:
                    controller->request( new StringCommand( _T("ManmcS0mbye"), sleepAfterWrite, true  ) );
                    //controller->request( new StringCommand( _T("ManmcL1mbye"), sleepAfterWrite, true ) );
                    break;

                default:
                    LOGP( wxString::Format( _T("Invalid currentConfigType! - %d"), currentConfigType, true ) );
            }
            break;

        case CFG_B:
            switch ( currentConfigType )
            {
                case CFG_A:
                    controller->request( new StringCommand( _T("ManmcS1mbye"), sleepAfterWrite, true ) );
                    //controller->request( new StringCommand( _T("ManmcL0mbye"), sleepAfterWrite, true ) );
                    break;

                case CFG_B:
                    controller->request( new StringCommand( _T("ManmcS1mbye"), sleepAfterWrite, true ) );
                    break;

                default:
                    LOGP( wxString::Format( _T("Invalid currentConfigType! - %d"), currentConfigType ) );
            }
            break;

        default:
            LOGP( wxString::Format( _T("Invalid uploadConfigType! - %d"), uploadConfigType ) );
    }
}

/**
 * \details Update raw ecu model from parsed files
 */
void FWConfigUploadManager::updateModel( const SymbolTable& iniSymbolTable )
{

/*
    String2CharRefHash& tableDataH    = m_fcDataFile.getTableDataHNoConst();
    const String2IntHash& scalarValueH      = m_fcDataFile.getScalarValuesH();

    createHeaderSymbol(iniSymbolTable, _T("r"));
    createHeaderSymbol(iniSymbolTable, _T("k"));

    SymbolTable& neededTables = MTConfigHelper::getMTConfigData()->getSymbolMgr().getConfigTableSymbols();
{
    /// New concept for tables
    SymbolTable::const_iterator IT = neededTables.begin();
    for ( ; IT != neededTables.end(); ++IT ){
        wxString name = IT->first;
        Symbol* neededSymbol = IT->second;
        const int page      = neededSymbol->getPage();
        const int offset    = neededSymbol->getOffset();

        Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel()name: ")+ name +_T(" (p:%i,o:%i)"),page, offset);

    }
}
    SymbolTable::const_iterator IT = m_cfgSymbolTable.begin();
    for ( ; IT != m_cfgSymbolTable.end(); ++IT )
    {
        Symbol* ptrSymbol = IT->second;
        const wxString symbolName = ptrSymbol->getName();
        if ( ptrSymbol->isScalar() )
        {
            const int page      = ptrSymbol->getPage();
            const int offset    = ptrSymbol->getOffset();
            String2IntHash::const_iterator scalarValuesIT = scalarValueH.find( symbolName );

            if( scalarValuesIT == scalarValueH.end() ) {
                ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel() scalarValueH.find not found: ")+ symbolName );
                continue;
            }

            ptrSymbol->setValid( true );
            uint8_t value = scalarValuesIT->second;
            m_model.setRawData( value, page, offset, STDataModelBase::USER_FORCE_WRITE );
        }
        else if ( ptrSymbol->isArray() )
        {
            ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): ")+ symbolName );
            if(iniSymbolTable.find(symbolName) == iniSymbolTable.end() ){ continue;}

            String2CharRefHash::const_iterator tableDataIT = tableDataH.find( symbolName );
            if( tableDataIT == tableDataH.end() ) { continue; }
            const uint8_t* tableData = (uint8_t*)tableDataIT->second;
            /// Symbol from parsed ini file
            Symbol* iniSymbol = iniSymbolTable.find(symbolName)->second;
            const int page      = iniSymbol->getPage();
            const int offset    = iniSymbol->getOffset();

            ptrSymbol->setPage(page);
            ptrSymbol->setOffset(offset);
            ptrSymbol->setValid(true);

            int oldColSize = ptrSymbol->getArrayColSize();
            int oldRowSize = ptrSymbol->getArrayRowSize();
            int oldSize = ptrSymbol->getArraySize();

            int newColSize = iniSymbol->getArrayColSize();
            int newRowSize = iniSymbol->getArrayRowSize();
            int newSize = iniSymbol->getArraySize();

            ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): ")+ symbolName+_T(" %i,%i"),oldSize, newSize );
            if( oldSize == newSize ){
//                Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel():")+ ptrSymbol->getName()+_T(" p=%i,o=%i"),page,offset );
                for( int i = 0; i < oldSize; ++i ){
                    setRawData(iniSymbol, tableData[i], page, offset + i );
                    if ( ptrSymbol->getArrayDimension() == 1 ){
//                        Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): [%i]=%i"),i, tableData[i]);
                    }
                }
            }else if(oldSize < newSize ){
                ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): search: ")+ symbolName );
                if ( ptrSymbol->getArrayDimension() == 1 ){
                    ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): found dim1: ")+ symbolName );
                    for( int i = 0; i < newSize; ++i ){
                        if ( i < oldSize ) { m_model.setRawData( tableData[i], page, offset + i, STDataModelBase::USER_FORCE_WRITE ); }
                        else { m_model.setRawData( tableData[oldSize - 1], page, offset + i, STDataModelBase::USER_FORCE_WRITE ); }
                    }
                }else if ( ptrSymbol->getArrayDimension() == 2 ){
                    ///Logger::logf( LOG_MSG, _T("FirmwareManager::updateModel(): found dim2: ")+ symbolName );
                    int oldIdx = 0;
                    int newIdx = 0;
                    for( int i = 0; i < oldRowSize; ++i ){
                        for ( int j = 0; j < oldColSize; ++j ){
                            setRawData( iniSymbol, tableData[oldIdx], page, offset + newIdx);
                            ++oldIdx;
                            ++newIdx;
                        }
                        newIdx += newColSize - oldColSize;
                    }
                }
            }else{
                if ( ptrSymbol->getArrayDimension() == 1 ){
                    double delta = (double)oldColSize/(double)newColSize;
                    double dIdx = 0;
                    for( int i = 0; i < newColSize; ++i ){
                        int deltaInt = dIdx;
                        m_model.setRawData( tableData[deltaInt], page, offset + i, STDataModelBase::USER_FORCE_WRITE );
                        dIdx += delta;
                    }
                    m_model.setRawData( tableData[oldColSize-1], page, offset + newColSize-1, STDataModelBase::USER_FORCE_WRITE );
                }else if ( ptrSymbol->getArrayDimension() == 2 ){
                    double deltaCol = (double)oldColSize/(double)newColSize;
                    double deltaRow = (double)oldRowSize/(double)newRowSize;
                    int newIdx = 0;
                    for( int i = 0; i < newRowSize; ++i )
                    {
                        double rowIdx = i* deltaRow;

                        for ( int j = 0; j < newColSize; ++j )
                        {
                            double colIdx = j* deltaCol;
                            int oldIdx = (int)rowIdx*oldColSize + (int)colIdx;

                            setRawData( iniSymbol, tableData[oldIdx], page, offset + newIdx);
                            ++newIdx;
                        }
                        setRawData( iniSymbol, tableData[(int)rowIdx*oldColSize + oldColSize-1], page, offset + newIdx -1);
                    }

                    /// last row
                    int oldLastRowIdx = oldColSize * (oldRowSize -1);
                    int newLastRowIdx = newColSize * (newRowSize -1);
                    for ( int j = 0; j < iniSymbol->getArrayColSize(); ++j )
                    {
                        double oldIdx = (double)oldLastRowIdx + j * deltaCol;
                        setRawData( iniSymbol, tableData[(int)oldIdx], page, offset + newLastRowIdx+j);
                    }
                    setRawData( iniSymbol, tableData[oldColSize*oldRowSize -1], page, offset + newColSize*newRowSize -1);

                }
            }
            ptrSymbol->setArrayColSize(newColSize);
            ptrSymbol->setArrayRowSize(newRowSize);
        }
    }
*/
}

/*
void FWConfigUploadManager::createHeaderSymbol( const SymbolTable& symbolTable, const wxString& headerName )
{
    String2CharRefHash& tableDataH    = m_fcDataFile.getTableDataHNoConst();
    String2IntHash& tableSizeH      = m_fcDataFile.getTableSizeHNoConst();

    /// if no r or k table in cfg then there is no need to map
    if(m_cfgSymbolTable.find(headerName) == m_cfgSymbolTable.end() ) return;

    /// if found r or k table in ini then there is no need to map
    if(symbolTable.find(headerName) != symbolTable.end() ) return;
    ///Logger::logf( LOG_MSG, _T("FirmwareManager::createHeaderSymbol(): need map: ") );

    Symbol* headerSymbol = m_cfgSymbolTable[headerName];

    SymbolTable::const_iterator IT = m_cfgSymbolTable.begin();
    SymbolTable tmpSt;
    for ( ; IT != m_cfgSymbolTable.end(); ++IT ){
        Symbol* ptrSymbol = IT->second;
        if( ! ptrSymbol->isArray() ) continue;
        ///Logger::logf( LOG_MSG, _T("FirmwareManager::createHeaderSymbol(): array: ")+ ptrSymbol->getName() );

        wxString symbolName = ptrSymbol->getName() + _T("_") +headerName;
        Logger::logf( LOG_MSG, _T("FirmwareManager::createHeaderSymbol(): search: ")+ symbolName );
        if(symbolTable.find(symbolName) == symbolTable.end() ) continue;
        Symbol* iniSymbol = symbolTable.find(symbolName)->second;
        if( ! iniSymbol->isArray() ) continue;

        Symbol* newSymbol = new Symbol(*headerSymbol);
        newSymbol->setName(symbolName);
        tmpSt[symbolName] = newSymbol;
        ///Logger::logf( LOG_MSG, _T("FirmwareManager::createHeaderSymbol(): found: ")+ symbolName );

        String2CharRefHash::const_iterator tableDataIT = tableDataH.find( headerName );
        if( tableDataIT == tableDataH.end() ) { continue; }
        tableDataH[symbolName] = tableDataH[headerName];

        tableSizeH[symbolName] = headerSymbol->getArraySize();
        ///Logger::logf( LOG_MSG, _T("FirmwareManager::createHeaderSymbol(): map: ")+ headerName +_T("->")+ symbolName);
    }

    SymbolTable::const_iterator IT2 = tmpSt.begin();
    for ( ; IT2 != tmpSt.end(); ++IT2){
        wxString symbolName = IT2->first;
        Symbol* newSymbol = IT2->second;
        m_cfgSymbolTable[symbolName] = newSymbol;
    }
}
*/

void FWConfigUploadManager::setRawData(Symbol* ptrSymbol, uint8_t data, int page, int offset)
{
    if ( ptrSymbol->getName() == _T("l") )
    {
        double mtValue = 65535.0 / ( (double)data + 200.0 ) - 73.0 + 0.55;
        m_model.setRawData( (uint8_t)mtValue, page, offset, STDataModelBase::USER_FORCE_WRITE );
    }else{
        m_model.setRawData(data, page, offset, STDataModelBase::USER_FORCE_WRITE );
    }
}


//void FWConfigUploadManager::writeAllDirty( const STDataModelBase& model, CommandRequestBase* controller )
//{
//    for ( int i = 1; i < model.getPageCount(); ++i )
//    {
//        if ( !model.isPageValid( i ) || !model.isPageDirty( i ) ) { continue; }
//
//        for ( int j = 0; j < model.getPageSize( i ); ++j )
//        {
//            if ( !model.isDataDirty( i, j ) ) { continue; }
//
//            //Logger::logf( LOG_MSG, _T("FWConfigUploadManager::writeAllDirty()\n") );
//            controller->request( new WritePageCommand( i, j , model.getRawData( i, j ) ) );
//        }
//    }
//}

//size_t FWConfigUploadManager::getDirtyCount()
//{
//    return m_model.getDirtyCount();
//}


void FWConfigUploadManager::setVariableData(VariableDataDH& vd, const FirmwareConfigData& fcd)
{
    //create and fill VariableDataDH with scalar values
    const String2IntHash& scalarH = fcd.getScalarValuesH();
    String2IntHash::const_iterator IT = scalarH.begin();
    for ( ; IT != scalarH.end(); ++IT )
    {
        vd.addValue( IT->first, IT->second);
    }

    const wxArrayString&  tablesA = fcd.getTableOrderA();
    for(size_t idx = 0; idx < tablesA.GetCount(); ++idx){
          wxString tableName = tablesA.Item(idx);
          VemsMatrix<uint8_t>* matrix = fcd.getTable(tableName);
          for ( size_t i = 0; i < matrix->size(); ++i )
        {
            wxString name = tableName + wxString::Format( _T("[%d]"), i );
            vd.addValue(name, matrix->get(i) );
        }

    }
}


void  FWConfigUploadManager::verifyConfig()
{
    if ( m_pConfigParser ) { delete m_pConfigParser; }

    m_pConfigParser = new FirmwareConfigParser( &m_fcDataECU );

    m_iState = UMS_WAIT_DATA;
}


void FWConfigUploadManager::processConfigLine( const wxString& line )
{
    if ( m_iState != UMS_WAIT_DATA ) { return ; }

    m_pConfigParser->parseOneLine( line );
}


FirmwareConfigDifferences* FWConfigUploadManager::compare()
{
    return m_fcDataECU.compare( *m_fcDataFile );
}


void FWConfigUploadManager::applyFilter(    const StringSet& filter,
                                            wxArrayString& missing )
{
    /// remove filtered scalar keys
    const wxArrayString& scalars = m_fcDataFile->getScalarOrderA();
    wxArrayString remove;

    for ( size_t i = 0; i < scalars.GetCount(); ++i )
    {
        if ( filter.find( scalars[i] ) == filter.end() ) { remove.Add( scalars[i] ); }
    }

    for ( size_t i = 0; i < remove.GetCount(); ++i )
    {
        m_fcDataFile->removeScalar( remove[i] );
    }

    remove.Clear();

    /// remove filtered table keys
    const wxArrayString& tables = m_fcDataFile->getTableOrderA();

    for ( size_t i = 0; i < tables.GetCount(); ++i )
    {
        if ( filter.find( tables[i] ) == filter.end() ) { remove.Add( tables[i] ); }
    }

    for ( size_t i = 0; i < remove.GetCount(); ++i )
    {
        m_fcDataFile->removeTable( remove[i] );
    }

    /// find missing keys
    const String2IntHash& scalarHash = m_fcDataFile->getScalarValuesH();
    StringSet::const_iterator it = filter.begin();

    for ( ; it != filter.end(); ++it ){
        VemsMatrix<uint8_t>* matrix = m_fcDataFile->getTable(*it);
        if ( (scalarHash.find( *it ) != scalarHash.end()) || matrix  ) { continue; }

        missing.Add( *it );
    }
}

FirmwareConfigDifferences*  FWConfigUploadManager::compareConfig( STDataModelBase* model )
{
    bool isRawLambdaUsed = MTConfigHelper::getMTConfigData()->isRawLambdaUsed();
    SymbolMgr* iniSymbols = &MTConfigHelper::getMTConfigData()->getSymbolMgr();

    STDataIOVemsCfg cfg( model );
    cfg.saveTo( &m_fcDataECU, iniSymbols, isRawLambdaUsed );

    return m_fcDataECU.compare( *m_fcDataFile );
}

void FWConfigUploadManager::createUploadData( const STDataModelBase& deviceModel, bool removeSameBytes )
{
    m_uploadData.clear();
    m_uploadCount = 0;

    for ( int i = 1; i < m_model.getPageCount(); ++i )
    {
        if ( !m_model.isPageValid( i ) || !m_model.isPageDirty( i ) ) { continue; }

        for ( int j = 0; j < m_model.getPageSize( i ); ++j )
        {
            if ( !m_model.isDataDirty( i, j ) ) { continue; }

            if ( removeSameBytes && ( m_model.getRawData(i,j) == deviceModel.getRawData(i,j) ) ) { continue; }

            m_uploadData[i].insert(j);
            ++m_uploadCount;
        }
    }
}

void FWConfigUploadManager::writeUploadData( CommandRequestBase* controller )
{
    IntToIntSetHash::const_iterator pageIter = m_uploadData.begin();

    for ( ; pageIter != m_uploadData.end(); ++pageIter ) {
        int pageNo = pageIter->first;
        const IntSet& page = pageIter->second;

        IntSet::const_iterator offsetIter = page.begin();

        for ( ; offsetIter != page.end(); ++offsetIter ) {
            int pageOffset = *offsetIter;
            uint8_t value = m_model.getRawData( pageNo, pageOffset );

            controller->request( new WritePageCommand( pageNo, pageOffset , value ) );
        }
    }
}
