#include "GPSDetectorClient.h"

#include "../logger.h"
#include "../GPSNMEA/NMEAParse.h"
#include "GPSCommunicationEvent.h"

// m_maxbuffersize: waiting for 256 bytes data (if it is a gps data, it should be enough)
// m_maxtimeout: after 2500 ms if no data came in -> timeout
GPSDetectorClient::GPSDetectorClient( wxEvtHandler* evtHandler ) : m_maxBufferSize( 256 ), m_maxTimeOut( 2500 )
{
    // wait for data
    m_state = CHAR_IN;

    // create the buffer
    m_buffer = new unsigned char[m_maxBufferSize];
    m_bufferSize = 0;

    m_evtHandler = evtHandler;
    m_timeOutCounter = 0;

    m_sp.Start();
}

GPSDetectorClient::~GPSDetectorClient()
{
    delete[] m_buffer;
}

void GPSDetectorClient::write( uint8_t* buffer, int& size )
{
    /// nothing to write to the port
    // nop
}

void GPSDetectorClient::read( uint8_t* buffer, int size )
{
    switch ( m_state )
    {
        case CHAR_IN:
            readChar( buffer, size );
            break;

        case TIME_OUT:
        case BUFFER_FULL:
        default:
            break;
    }
}

void GPSDetectorClient::readChar( uint8_t* buffer, int size )
{
    if ( ( m_timeOutCounter >= m_maxTimeOut ) || ( m_sp.Time() > 2.5 * m_maxTimeOut ) )
    {
        m_state = TIME_OUT;
        return;
    }

    if ( size == 0 )
    {
        ++m_timeOutCounter;
        wxThread::Sleep( 1 );
        return;
    }

    // else copy data into the buffer, it it isnot full
    if ( m_bufferSize + size >= m_maxBufferSize )
    {
        m_state = BUFFER_FULL;
        return;
    }

    memcpy( m_buffer + m_bufferSize, buffer, size );

    m_bufferSize += size;
}

void GPSDetectorClient::finish()
{
    GPSCommunicationEvent event;

    if ( m_state == TIME_OUT )
    {
        event.state = GPSCommunicationEvent::CANNOT_OPEN_PORT;

        if ( m_evtHandler )
        {
            m_evtHandler->AddPendingEvent( event );
        }
    }
    if ( m_state == BUFFER_FULL )
    {
        wxString buffer = wxString::From8BitData( (char*)m_buffer, m_bufferSize );

        if ( buffer.Find(_T("$GPGGA")) == wxNOT_FOUND )
        {
            /// NOT VALID GPS DATA
            event.state = GPSCommunicationEvent::NOT_FOUND;

            if ( m_evtHandler )
            {
                m_evtHandler->AddPendingEvent( event );
            }
            //LOGP( _T("No valid GPS Data"))
        }
        else
        {
            NMEAParse gpsParser;
            // LOGP( buffer )
            int i = 0;
            while ( ( i < m_bufferSize ) && ( gpsParser.feedGPScharacter(buffer[i++] ) != END ) );

            GPSDataStruct data = gpsParser.gpsdata;

            event.state = GPSCommunicationEvent::GPS_FOUND;

            event.gpsdata = data;

            if ( m_evtHandler )
            {
                m_evtHandler->AddPendingEvent( event );
            }
        }
    }
}

bool GPSDetectorClient::isFinished()
{
    return ( m_state == TIME_OUT || m_state == BUFFER_FULL );
}
