#include "FirmwareDirector.h"
#include "../SimpleFwDlg.h"
#include "../FirmwareTypes.h"
#include "../firmwareconfigitem.h"
#include "../logger.h"
#include "../UserLog.h"
#include "../fileparser/IHexParser.h"
#include "../events/InputEvent.h"
#include "../events/CommunicationEvent.h"
#include "../events/FirmwareVerifyEvent.h"
#include "../events/FirmwareUploadEvent.h"
#include "../events/FirmwareUploadedEvent.h"
#include "../gui/FirmwareDirector.h"
#include "../vemsprotocol/VemsDeviceControllerBase.h"
#include "../vemsprotocol/FirmwareData.h"
#include "../vemsprotocol/FirmwareVerifyController.h"



FirmwareDirector::FirmwareDirector( wxWindow* parent, VemsDeviceControllerBase* deviceController )
{
    m_deviceController = deviceController;
    m_parent = parent;
    m_filepatches.clear();
}

FirmwareDirector::~FirmwareDirector()
{
    disconnectCommunicationEventHandler();
}

void FirmwareDirector::verifyFirmware( Project* proj )
{
    m_filepatches.clear();

    if ( !isEnabled() ) {
        showErrorMessage();
        return;
    }

    SimpleFwDlg dlg( m_parent, VERIFY );

    if ( proj ) {
        dlg.Init( proj );
    } else {
        FirmwareInfo fwInfo = m_deviceController->getFirmwareInfo();
        dlg.Init( &fwInfo );
    }

    FirmwareConfigItemArray configItemA;

    if (dlg.ShowModal() == wxID_OK){
        dlg.getFirmwareConfigItemArray( configItemA );
        wxString hexFilePath = getHexFile(&configItemA);
        IHexParser ihexp( m_deviceController->getDevice() );
        if( ihexp.parse(hexFilePath) != 0){
             LOGP( _T("Error: hex File does not exist!") );
            return;
        }

        FirmwareData* fwData = ihexp.getFirmwareData();
        addPatchFiles( &configItemA, fwData );
        startVerifyFirmware( fwData );
    }
}

void FirmwareDirector::uploadFirmware( Project* proj )
{
//    if ( !isEnabled() ) {
//        showErrorMessage();
//        return;
//    }

    SimpleFwDlg dlg( m_parent, UPLOAD );

    if ( proj ) {
        dlg.Init( proj );
    } else {
        FirmwareInfo fwInfo = m_deviceController->getFirmwareInfo();
        dlg.Init( &fwInfo );
    }

    FirmwareConfigItemArray configItemA;

    if ( dlg.ShowModal() == wxID_OK ) {
        dlg.getFirmwareConfigItemArray( configItemA );
        wxString hexFilePath = getHexFile(&configItemA);
        IHexParser ihexp( m_deviceController->getDevice() );
        if( ihexp.parse(hexFilePath) != 0){
            LOGP( _T("Error: hex File does not exist!") );
            return;
        }

        FirmwareData* fwData = ihexp.getFirmwareData();
        addPatchFiles( &configItemA, fwData );
        startUploadFirmware( fwData );
    }
}

wxString FirmwareDirector::getHexFile(FirmwareConfigItemArray* configItemA)
{
    FirmwareConfigItem* configItem;
    for(uint16_t i = 0; i <configItemA->Count(); i++ ){
        configItem = configItemA->Item(i);

        if( configItem->type == _T("firmware") ){
            return configItem->getAbsolutePath();
        }
    }
    return wxEmptyString;
}

bool FirmwareDirector::isEnabled() const
{
    return ( m_deviceController->isConnected() && ( m_deviceController->isV3() || m_deviceController->isRound() ) );
}

void FirmwareDirector::showErrorMessage() const
{
        UserLog::vLog( _T("ecu_not_found_1") );
        wxString msg = _T("ECU is not detected!");
        wxMessageBox(msg);
}

void FirmwareDirector::startVerifyFirmware( FirmwareData* firmwareData )
{
    m_deviceController->stop( VDC_FLAG_NOEVENT );

    m_firmwareVerifyController = new FirmwareVerifyController( m_deviceController, this );

    connectCommunicationEventHandler();
    connectFirmwareVerifyEventHandlers();

    createProgressBar( VERIFY, firmwareData->getPages() );

    m_firmwareVerifyController->verify( firmwareData );
}

void FirmwareDirector::startUploadFirmware( FirmwareData* firmwareData )
{
    m_deviceController->stop( VDC_FLAG_NOEVENT );

    m_firmwareVerifyController = new FirmwareVerifyController( m_deviceController, this );

    connectCommunicationEventHandler();
    connectFirmwareUploadEventHandlers();

    createProgressBar( UPLOAD, firmwareData->getPages() );

    m_firmwareVerifyController->upload( firmwareData );
}

void FirmwareDirector::connectCommunicationEventHandler()
{
    Connect(    wxID_ANY,
                wxEVT_COMMUNICATION,
                CommunicationEventHandler( FirmwareDirector::onCommunicationEvent ),
                NULL, this );
}

void FirmwareDirector::disconnectCommunicationEventHandler()
{
    Disconnect( wxID_ANY,
                wxEVT_COMMUNICATION,
                CommunicationEventHandler( FirmwareDirector::onCommunicationEvent ),
                NULL, this );
}

void FirmwareDirector::connectFirmwareVerifyEventHandlers()
{
    Connect(    FirmwareVerifyEvent::PAGE_VERIFIED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwarePageVerified ),
                NULL, this );

    Connect(    FirmwareVerifyEvent::PAGE_VERIFIED_AND_PATCHED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwarePageVerified ),
                NULL, this );

    Connect(    FirmwareVerifyEvent::SUCCEEDED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwareVerifingSucceeded ),
                NULL, this );

    Connect(    FirmwareVerifyEvent::FAILED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwareVerifingFailed ),
                NULL, this );
}

void FirmwareDirector::disconnectFirmwareVerifyEventHandlers()
{
    Disconnect( FirmwareVerifyEvent::PAGE_VERIFIED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwarePageVerified ),
                NULL, this );

    Disconnect( FirmwareVerifyEvent::PAGE_VERIFIED_AND_PATCHED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwarePageVerified ),
                NULL, this );

    Disconnect( FirmwareVerifyEvent::SUCCEEDED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwareVerifingSucceeded ),
                NULL, this );

    Disconnect( FirmwareVerifyEvent::FAILED, wxEVT_FIRMWARE_VERIFY,
                FirmwareVerifyEventHandler( FirmwareDirector::onFirmwareVerifingFailed ),
                NULL, this );
}

void FirmwareDirector::connectFirmwareUploadEventHandlers()
{
    Connect(    FirmwareUploadEvent::ERASING_PAGES, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwarePagesErasing ),
                NULL, this );

    Connect(    FirmwareUploadEvent::PAGE_UPLOADED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwarePageUploaded ),
                NULL, this );

    Connect(    FirmwareUploadEvent::UPLOADING_PAGES_SUCCEEDED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwareUploadingSucceeded ),
                NULL, this );

    Connect(    FirmwareUploadEvent::UPLOADING_PAGES_FAILED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwareUploadingFailed ),
                NULL, this );
}

void FirmwareDirector::disconnectFirmwareUploadEventHandlers()
{
    Disconnect( FirmwareUploadEvent::ERASING_PAGES, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwarePagesErasing ),
                NULL, this );

    Disconnect( FirmwareUploadEvent::PAGE_UPLOADED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwarePageUploaded ),
                NULL, this );

    Disconnect( FirmwareUploadEvent::UPLOADING_PAGES_SUCCEEDED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwareUploadingSucceeded ),
                NULL, this );

    Disconnect( FirmwareUploadEvent::UPLOADING_PAGES_FAILED, wxEVT_FIRMWARE_UPLOAD,
                FirmwareUploadEventHandler( FirmwareDirector::onFirmwareUploadingFailed ),
                NULL, this );
}

void FirmwareDirector::createProgressBar( int processType, int max )
{
    wxString deviceStr = m_deviceController->getDeviceAsString();
    wxString title = _T("ECU ( ") + deviceStr;
    title += ( processType == VERIFY ? _T(" ) flash verify progress!") : _T(" ) flash upload progress!") );

    m_progressDialog = new wxProgressDialog(    title,
                                                _T("Page index: "),
                                                max,
                                                m_parent,
                                                wxPD_APP_MODAL |
                                                wxPD_ELAPSED_TIME |
                                                wxPD_SMOOTH |
                                                wxPD_ESTIMATED_TIME |
                                                wxPD_REMAINING_TIME );
}

void FirmwareDirector::cleanupVerifyFirmware()
{
    disconnectCommunicationEventHandler();
    disconnectFirmwareVerifyEventHandlers();
}

void FirmwareDirector::cleanupUploadFirmware()
{
    disconnectCommunicationEventHandler();
    disconnectFirmwareUploadEventHandlers();
}

void FirmwareDirector::onCommunicationEvent( CommunicationEvent& event )
{
    disconnectCommunicationEventHandler();

    if ( m_progressDialog ) {
        m_progressDialog->Destroy();
        m_progressDialog = NULL;
    }

    if ( event.GetId() == CommunicationEvent::TIMEOUT ) {

        m_deviceController->timeout();

    }

}

/// FirmwareVerifyEvent handlers
void FirmwareDirector::onFirmwarePageVerified( FirmwareVerifyEvent& event )
{
    size_t pageCount = event.getFirmwarePageCount();
    size_t pageIdx = event.getFirmwarePageIdx();

    wxString message = wxString::Format( _T("Page index: %d / %d"), pageCount, pageIdx );

    if ( event.GetId() == FirmwareVerifyEvent::PAGE_VERIFIED_AND_PATCHED )
    {
        m_filepatches.insert( event.getPatchFileName() );
    }

    m_progressDialog->Update( pageIdx, message );
}
#include <algorithm>
#include <set>
void FirmwareDirector::onFirmwareVerifingSucceeded( FirmwareVerifyEvent& event )
{
    LOG
    m_progressDialog->Destroy();
    m_progressDialog = NULL;

    wxString message = _T("Firmware verify was successful!");

    if ( m_filepatches.size() > 0 )
    {
        message += _T("\n\nFound patches:\n");
        StringSet::const_iterator it = m_filepatches.begin();

        for ( ; it != m_filepatches.end(); ++it )
        {
            message += (*it) + _T("\n");
        }
    }

    wxMessageDialog dialog( m_parent, message,
                            _T("Firmware verify"), wxOK | wxICON_INFORMATION, wxDefaultPosition );
    if ( dialog.ShowModal() == wxID_OK ) {
        cleanupVerifyFirmware();
        m_deviceController->resume();
//        m_deviceController->start();
    }
}

void FirmwareDirector::onFirmwareVerifingFailed( FirmwareVerifyEvent& event )
{
    m_progressDialog->Destroy();
    m_progressDialog = NULL;

    wxMessageDialog dialog( m_parent, _T("Firmware verify was failed!"),
                            _T("Firmware verify"), wxOK | wxICON_ERROR, wxDefaultPosition );
    if ( dialog.ShowModal() == wxID_OK ) {
        cleanupVerifyFirmware();
        m_deviceController->resume();
    }
}

/// FirmwareUploadEvent handlers
void FirmwareDirector::onFirmwareUploadingStarted( FirmwareUploadEvent& event )
{
}

void FirmwareDirector::onFirmwarePagesErasing( FirmwareUploadEvent& event )
{
    if ( m_progressDialog ) {
        m_progressDialog->Pulse( _T("Erasing pages... ") );
    }
}

void FirmwareDirector::onFirmwarePagesErasingSucceeded( FirmwareUploadEvent& event )
{

}

void FirmwareDirector::onFirmwareUploadingPagesStarted( FirmwareUploadEvent& event )
{
}

void FirmwareDirector::onFirmwarePageUploaded( FirmwareUploadEvent& event )
{
    size_t pageCount = event.getFirmwarePageCount();
    size_t pageIdx = event.getFirmwarePageIdx();

    wxString message = wxString::Format( _T("Page index: %d / %d"), pageIdx, pageCount );
    m_progressDialog->Update( pageIdx, message );
}

void FirmwareDirector::onFirmwareUploadingSucceeded( FirmwareUploadEvent& event )
{
    m_progressDialog->Destroy();
    m_progressDialog = NULL;

    wxMessageDialog dialog( m_parent, _T("Firmware upload was successful!"),
                            _T("Firmware upload"), wxOK | wxICON_INFORMATION, wxDefaultPosition );

    if ( dialog.ShowModal() == wxID_OK ) {
        cleanupUploadFirmware();
        m_deviceController->start( VDC_FLAG_AUTO_LEAVE_BOOT_MODE | VDC_FLAG_NO_RECONNECT );

        FirmwareUploadedEvent evt;
        m_parent->AddPendingEvent( evt );
    }
}

void FirmwareDirector::onFirmwareUploadingFailed( FirmwareUploadEvent& event )
{
}

int FirmwareDirector::addPatchFiles( FirmwareConfigItemArray* configItemA, FirmwareData*  fwData )
{
     wxArrayString* patchFileA = getPatchFiles(configItemA);
     for(uint16_t i = 0; i <patchFileA->Count(); i++ ){
           wxString pathFilePath = patchFileA->Item(i);
           IHexParser ihexPatch( m_deviceController->getDevice() );
            if( ihexPatch.parse(pathFilePath) != 0){
//                observerNotify(_T("Warning: Patch File not found:")+ pathFilePath + _T("!") );
            }else{
//                 Logger::logf(LOG_MSG, _T("FirmwareMgr::addPatchFiles(): ")+pathFilePath);
                FirmwareData *patchData = ihexPatch.getFirmwareData();
                fwData->addPatch(patchData);
                delete patchData;
            }
     }
     return 0;
}

wxArrayString* FirmwareDirector::getPatchFiles(FirmwareConfigItemArray* configItemA)
{
    wxArrayString * A = new wxArrayString();

    FirmwareConfigItem* configItem = new FirmwareConfigItem();
    for(uint16_t i = 0; i <configItemA->Count(); i++ ){
        configItem = configItemA->Item(i);

        if( configItem->type == _T("firmware_patch") ){
            if(configItem->active){
                A->Add(configItem->getAbsolutePath());
            }
        }
    }
    return A;
}
