#include "CalcModelAnalyzer.h"

#include <wx/arrimpl.cpp>
#include <wx/textfile.h>
#include <typeinfo>
#include <wx/wxsqlite3.h>

#include "LogHistoryMgrHelper.h"
#include "../MTConfigHelper.h"

#include "../DirHelper.h"
#include "../logger.h"


WX_DEFINE_OBJARRAY( CalcSymbolA )
WX_DEFINE_OBJARRAY( DiffTA );

/// the analyzer

CalcModelAnalyzer::CalcModelAnalyzer( SymbolTable* symTable ) : RTDataAnalyzerBase()
{
    wxSQLite3Database db;
    wxString dbName = DirHelper::GetPath( dCONFIG ) + _T("calcmodel.db");
    db.Open( dbName );

    wxArrayString reqDH;

    try {
        wxSQLite3ResultSet res = db.ExecuteQuery( _T("SELECT * FROM calc_symbols") );

        while ( res.NextRow() )
        {
            CalcSymbol sym;
            sym.name = res.GetString( _T("symbol_name") );

            if ( symTable->find( sym.name ) == symTable->end() )
            {
                continue;
            }
            LOGP( _T("found:") + sym.name );

            sym.minlimit = res.GetDouble( _T("diff_limit_min") );
            sym.maxlimit = res.GetDouble( _T("diff_limit_max") );

            if ( reqDH.Index( sym.name ) == wxNOT_FOUND )
                reqDH.Add( sym.name );
            m_symbols.Add( sym );
        }
    }
    catch ( wxSQLite3Exception e )
    {
        //
    }

    db.Close();

    setRequiredSymbols( reqDH );

}

CalcModelAnalyzer::~CalcModelAnalyzer()
{
    m_symbols.Clear();
}

void CalcModelAnalyzer::processData()
{
    size_t syms = m_symbols.GetCount();

    for ( size_t i = 0; i < syms; ++i )
    {
        wxString& name = m_symbols[i].name;

        double minLimit = m_symbols[i].minlimit, maxLimit = m_symbols[i].maxlimit;

        DoubleHash::iterator it = m_data->find( name );
        if ( it == m_data->end() )
        {
            continue;
        }

        double val = (*m_data)[name], absval = fabs(val);

        if ( absval <= minLimit || absval >= maxLimit )   // ok
        {
            continue;
        }

        DiffT d;
        d.rtidx = (int)((*m_data)[_T("rtidx")]);
        d.value = (*m_data)[name];

        m_diffs[name].Add( d );
    }
}

/// the process

CalcModelAnalyzerProcess::CalcModelAnalyzerProcess()
{
    m_mgr = new LogHistoryMgr( *LogHistoryMgrHelper::getInstance() );

    m_analyzer = new CalcModelAnalyzer( &MTConfigHelper::getMTConfigData()->getSymbolTable() );
    m_mgr->getRTDataMgr()->RegisterViewer( m_analyzer );
    m_analyzer->SetDataMgr( m_mgr->getRTDataMgr() );

    m_mgr->registerObserver(this);

    m_isrunning = false; m_isDisabled = false;
}

CalcModelAnalyzerProcess::~CalcModelAnalyzerProcess()
{
    delete m_mgr;
    delete m_analyzer;
}

bool CalcModelAnalyzerProcess::process()
{
    if ( m_isrunning || m_isDisabled )
        return false;

    observerNotify( m_mgr->getHistoryCount() );

    m_isrunning = true;

    m_mgr->analyzeHistory();

    return false;
}

void CalcModelAnalyzerProcess::notify( int i, void* classType )
{
    LogHistoryMgr* cm = (LogHistoryMgr*)(classType);
    if ( cm == m_mgr )
    {
        observerNotify( i, 0 );
    }
}

void CalcModelAnalyzerProcess::start()
{
    setDisabled( false );
    ProcessBase::start( (WXTHREAD_MIN_PRIORITY + WXTHREAD_DEFAULT_PRIORITY) / 2 );
}

void CalcModelAnalyzerProcess::finished()
{
    m_isrunning = false;
    setDisabled( true );

    CalcModelAnalyzer* a = dynamic_cast<CalcModelAnalyzer*>(m_analyzer);

    if ( a )
        m_diffs = a->getResult();

    observerNotify( _T("calcModelAnalyzisP") );
}

wxString CalcModelAnalyzerProcess::getLogFileName()
{
    return ( m_mgr->getLogFileName() );
}

/// the viewer

int idCalcModelSymbol       = wxNewId();
int idCalcModelSaveTable    = wxNewId();

BEGIN_EVENT_TABLE( CalcModelAnalyzerViewer, wxFrame )
    EVT_COMBOBOX( idCalcModelSymbol, CalcModelAnalyzerViewer::OnSelectSymbol )
END_EVENT_TABLE()

CalcModelAnalyzerViewer::CalcModelAnalyzerViewer( wxWindow *parent, const wxString& caption ) :
    wxFrame( parent, wxID_ANY, caption, wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE | wxFRAME_FLOAT_ON_PARENT )
{

    m_popupMenu = NULL;

    SetBackgroundColour( wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE) );
    ClearBackground();

    wxBoxSizer* sizer = new wxBoxSizer( wxVERTICAL );

    wxArrayString nullArr;
    m_symbolnames = new wxComboBox( this, idCalcModelSymbol, _T(""), wxDefaultPosition, wxDefaultSize, 0, NULL, wxCB_DROPDOWN | wxCB_READONLY | wxCB_SORT );

    sizer->Add( m_symbolnames, 0, wxEXPAND | wxALL, 5 );

    wxStaticText* txt = new wxStaticText( this, -1, _T("Right-click on grid: save to csv") );
    sizer->Add( txt, 0, wxEXPAND | wxALL, 5 );

    m_table = new wxGrid( this, wxID_ANY );
    m_table->CreateGrid( 1, 2 );
    m_table->SetColLabelValue( 0, _T("RT Index") ); m_table->SetColLabelValue( 1, _T("Diff") );

    m_table->EnableEditing( false );

    m_table->Connect( wxEVT_GRID_CELL_RIGHT_CLICK, wxMouseEventHandler( CalcModelAnalyzerViewer::OnRightClick ), 0, this );
    m_table->Connect( wxEVT_GRID_LABEL_RIGHT_CLICK, wxMouseEventHandler( CalcModelAnalyzerViewer::OnRightClick ), 0, this );
    m_table->Connect( wxEVT_GRID_CELL_LEFT_DCLICK, wxGridEventHandler( CalcModelAnalyzerViewer::OnDoubleClickCell ), 0, this );


    sizer->Add( m_table, 1, wxEXPAND | wxTOP | wxLEFT | wxRIGHT, 5 );

    SetSizer( sizer );
    sizer->SetSizeHints( this );
    sizer->Fit( this );
}

CalcModelAnalyzerViewer::~CalcModelAnalyzerViewer()
{
    delete m_process;
}

void CalcModelAnalyzerViewer::setLogHistory( LogHistoryBase* lh )
{
    m_lh = lh;
}

void CalcModelAnalyzerViewer::notify( int i )
{
    m_dlg = new wxProgressDialog( _T("Please wait"), _T(""), i, NULL,
        wxPD_APP_MODAL | wxPD_SMOOTH | wxPD_AUTO_HIDE );
    m_dlg->Show();
    wxSafeYield( m_dlg,true );
}

void CalcModelAnalyzerViewer::notify( int i, int j )
{
    m_dlg->Update( i );
}

void CalcModelAnalyzerViewer::calculate()
{
    m_process = new CalcModelAnalyzerProcess();
    m_process->registerObserver( this );
    m_process->start();
}

void CalcModelAnalyzerViewer::notify( const wxString& str )
{
    if ( str == _T("calcModelAnalyzisP") )
    {
        m_dlg->Destroy();
        m_result = m_process->getResult();

        wxArrayString choices;
        String2DiffTAHash::const_iterator it = m_result.begin();
        for ( ; it != m_result.end(); ++it )
        {
            choices.Add( it->first );
        }

        m_symbolnames->Append( choices );

        Show();
    }
}

void CalcModelAnalyzerViewer::OnSelectSymbol( wxCommandEvent& event )
{
    wxString str = m_symbolnames->GetValue();

    DiffTA& table = m_result[str];

    unsigned int count = table.GetCount();
    if ( count < 1 )
    {
        event.Skip();
        return;
    }

    Freeze();
    m_table->ClearGrid();

    // "resize"
    if ( m_table->GetNumberRows() < (int)count )
        m_table->AppendRows( count - m_table->GetNumberRows() );
    else if ( m_table->GetNumberRows() > (int)count )
        m_table->DeleteRows( count - 1, m_table->GetNumberRows() - count - 1 );


    for ( size_t i = 0; i < count; ++i )
    {
        DiffT t = table[i];
        wxString rtIdxS = wxString::Format( _T("%d"), t.rtidx );
        wxString valueS = wxString::Format( _T("%.2f"), t.value );
        m_table->SetCellValue( i, 0, rtIdxS );
        m_table->SetCellValue( i, 1, valueS );
    }
    m_table->Fit();
    GetSizer()->RecalcSizes();
//    GetSizer()->Fit(this);
//    GetSizer()->SetSizeHints(this);

    Thaw();
}

void CalcModelAnalyzerViewer::OnRightClick( wxMouseEvent& event )
{
    if ( m_popupMenu == NULL )
    {
        m_popupMenu = new wxMenu();
        m_popupMenu->Append( idCalcModelSaveTable, _T("Save to file") );
        Connect( idCalcModelSaveTable, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( CalcModelAnalyzerViewer::Save) );
    }

    PopupMenu( m_popupMenu );
}

void CalcModelAnalyzerViewer::OnDoubleClickCell( wxGridEvent& event )
{
    int row = event.GetRow();

    wxString posStr = m_table->GetCellValue( row, 0 );

    if ( posStr == wxEmptyString )
    {
        event.Skip();
        return;
    }

    long pos;
    if ( posStr.ToLong( &pos ) == false )
    {
        event.Skip();
        return;
    }

    m_lh->refresh( pos, true );
}

void CalcModelAnalyzerViewer::Save( wxCommandEvent& event )
{
    wxString symbolname = m_symbolnames->GetValue();

    wxString fname = m_process->getLogFileName();

    if ( fname == wxEmptyString )
    {
        event.Skip();
        return;
    }

    if ( fname.EndsWith( _T("vemslog") ) )
        fname.RemoveLast(7);
    fname += symbolname + _T(".csv");

    wxTextFile f( fname );
    if ( f.Exists() )
    {
        f.Open();
        f.Clear();
    }
    else
        f.Create();

    f.AddLine( symbolname );

    size_t nr = m_table->GetNumberRows();
    for ( size_t i = 0; i < nr; ++i )
    {
        wxString str = m_table->GetCellValue( i, 0 ) + _T(";") + m_table->GetCellValue( i, 1 );
        f.AddLine( str );
    }
    f.Write();
    f.Close();

    event.Skip();
}
