#include "RT2ViewerConverter.h"

#include "../logger.h"
#include "../Math.h"
#include "../SymbolMgr.h"

#include "../vemsprotocol/SIPRPacket.h"
#include "../vemsprotocol/SIPRTypes.h"
#include "../common/FunctionMeter.h"
#include "../common/types/DoubleHash.h"

RT2ViewerConverter::RT2ViewerConverter(SymbolMgr* mt, DoubleHash* viewerModel )
{
    //ctor
    m_symbolMgr = mt;
    m_viewerModel = viewerModel;
    m_buffer = NULL;
    m_size = 0;
    m_dependencyA = NULL;
    m_address = 0;


    if( m_viewerModel ){
        m_hasInternalModel = false;
    }else{
        m_viewerModel = new DoubleHash();
        m_hasInternalModel = true;
    }

    m_variableData = VariableDataDH( m_viewerModel );
    m_rpn.setVariableData( &m_variableData );
}

RT2ViewerConverter::~RT2ViewerConverter()
{
    //dtor
    if( m_hasInternalModel ) { delete m_viewerModel; }
	if(m_dependencyA) { purge(m_dependencyA); }
}

void RT2ViewerConverter::resetHash()
{
    m_viewerModel->clear();
    m_rpn.reset();
}

void RT2ViewerConverter::refreshDependency(const DoubleHash& requiredData)
{
    GraphNodeSet nodes;
    fillGraphNodeSet(requiredData, nodes);

    const Graph& dag = m_symbolMgr->getDAG();

    if(m_dependencyA) purge(m_dependencyA);
    m_dependencyA = dag.getNodeDedepending( nodes );

    //    dag.printDepending(*m_dependencyA);
}

void RT2ViewerConverter::convert(uint8_t* buff, int size, DoubleHash& requiredData, int subType)
{

    m_buffer = buff;
    m_size =size;

    // time shift
    DoubleHash::const_iterator it = requiredData.begin(), endit = requiredData.end();

    for ( ; it != endit; ++it )
    {
        if ( it->first.Find(_T("[")) != wxNOT_FOUND )
            (*m_viewerModel)[it->first] = it->second;
        if ( it->first == _T("time") )
            (*m_viewerModel)[it->first] = it->second;
    }

    //
//    Logger::logf( LOG_MSG, _T("------------------------") );
//    for ( size_t i = 0; i < m_dependencyA->GetCount(); ++i )
//    {
//        GraphNodeSet* ns = m_dependencyA->Item(i);
//        GraphNodeSet::iterator it = ns->begin(), e = ns->end();
//        for ( ; it!= e; ++it )
//        {
//            GraphNode* node = *it;
//            Logger::logf( LOG_MSG, _T("node name = ") + node->getName() );
//        }
//    }
//

    fillScalarValues(subType);

    fillDependentValues();
    //evaluate all symbol requested by viewers

    fillRegisteredValues(requiredData);

}

void RT2ViewerConverter::convertAIM(uint8_t* buff, DoubleHash& requiredData)
{
    SymbolTable& st = m_symbolMgr->getSymbolTable();

    if(st.size() == 0) {
        return;
}

    /*DoubleHash::const_iterator it = requiredData.begin(), endit = requiredData.end();

    for( ; it != endit; ++it ) {
        if ( it->first.Find(_T("[")) != wxNOT_FOUND )
            (*m_viewerModel)[it->first] = it->second;
        if ( it->first == _T("time") )
            (*m_viewerModel)[it->first] = it->second;
    }*/
    for(SymbolTable::const_iterator it = st.begin(); it != st.end(); it++) {

        Symbol* s = it->second;

        //value = getValue(s->getSize(), offset, s->getScale(), s->getTranslation());

        if(s->getOffset() == buff[0]) {
            double value = getAIMValue( buff, s->getScale(), s->getTranslation() );
            (*m_viewerModel)[s->getName()] = value;
        }
    }

    fillDependentValues();
    /// evaluate all symbol requested by viewers
    fillRegisteredValues(requiredData);
}

double RT2ViewerConverter::getAIMValue( uint8_t* buffer, double scale, double translation )
{
    double val = buffer[1] * 256 + buffer[2];
    return ( val * scale + translation );
}

void RT2ViewerConverter::convert(const SIPRPacket& packet, DoubleHash& requiredData)
{
    if( packet.type() == SIPR_READ_MEMORY_32)
    {
        m_address = packet.data()[0] << 24;
        m_address += packet.data()[1] << 16;
        m_address += packet.data()[2] << 8;
        m_address += packet.data()[3];

        m_buffer = &packet.data()[4];
        m_size = packet.size() - 4;
        /*if ( m_address == 0x140 )
            Logger::logf(LOG_MSG, _T("TPS=%d"), m_buffer[15]);*/
        convert(m_buffer, m_size, requiredData, SiprSensorType::SIPR_READ_MEMORY_32);
        return;
    }

    if( packet.type() == SIPR_AIM)
    {
        m_buffer = packet.data();
        m_size = 3;

        convertAIM(m_buffer, requiredData);
        return;
    }


    if ( (packet.type() != SIPR_READ_SENSORS)  ) { return; }
    //selectInput( INPUT_TYPE_SIPR );
    int subType = packet.getSubType();

    switch(subType){
        case SiprSensorType::SIPR_RT_MT_NORMAL:
        case SiprSensorType::SIPR_RT_MT_EXT:
        case SiprSensorType::SIPR_RT_ROUND:
            m_buffer = &packet.data()[3];
            m_size = packet.size() - 3;
            break;
        case SiprSensorType::GPS:
        default:
            m_buffer = packet.data();
            m_size = packet.size();
    }
    convert(m_buffer, m_size, requiredData, subType);
}

/// The raw buffer contains these data and therefore do not have dependency
void RT2ViewerConverter::fillScalarValues(int rtDataType)
{
    const SymbolTable* st = m_symbolMgr->getScalarTable(rtDataType);

    //Logger::logf(LOG_MSG, _T("RTDataRawInputBase::fillScalarValues() - %p"), st );

    if(!st) {
        return;
    }

    //Logger::logf(LOG_MSG, _T("RTDataRawInputBase::fillScalarValues()") - 10 );

    double  value;
    for(SymbolTable::const_iterator it = st->begin(); it != st->end(); it++){
        Symbol* s = it->second;
        if(!s) { continue ; } //is it possible?
        if ( s->getName() == _T("time")) continue;
        if(rtDataType == SiprSensorType::SIPR_READ_MEMORY_32){
            int offset = s->getOffset() - m_address;
            /*if ( s->getName() == _T("engine_t_batt") )
                Logger::logf(LOG_MSG, s->getName()+_T(", offset=%d, address=%x"),  s->getOffset(), m_address);*/
            if( (offset < 0) || (offset >= 32)) { continue;} /// the packet not contains the given variable
            //if ( s->getName() == _T("engine_t_batt") )
            {
            value = getValue(s->getSize(), offset, s->getScale(), s->getTranslation());
            }

            //Logger::logf(LOG_MSG, s->getName() + _T(", value=%f"), value);

        }else{
            value = getValue(s->getSize(), s->getOffset(), s->getScale(), s->getTranslation());
        }
        (*m_viewerModel)[s->getName()] = value;
        //Logger::logf(LOG_MSG, _T("RTDataRawInputBase::fillScalarValues() o:%i,v:%f name:") + s->getName(), s->getOffset(), value);
    }
}

void RT2ViewerConverter::fillScalarValues()
{
    SymbolTable& st = m_symbolMgr->getSymbolTable();
    for(SymbolTable::iterator it = st.begin(); it != st.end(); it++){
        Symbol* s = it->second;
        if(!s) { continue ; } //is it possible?
        if ( s->getName() == _T("time")) continue;
        if(!s->isConst() && s->isScalar()){
            double value = getValue(s->getSize(), s->getOffset()
                                            , s->getScale(), s->getTranslation());
            (*m_viewerModel)[s->getName()] = value;
             //Logger::log(LOG_MSG, wxString::Format(_T("RTDataMgrSerial::fillScalarValues() Symbol: ") + s->getName()+ _T(", Value: %f"), value));
            //m_rtRawValues[s->getName()] = m_RealTimeDesc.getValue(s->getSize(), s->getOffset()
            //,1, 0);
        }
    }
}

/** @brief getValue
*
*   @todo: document this function
* copied from RealTimeDesc
*/
double RT2ViewerConverter::getValue(int size, int offset, double scale, double translate)
{
   int value = getValue(size, offset);

    //Logger::log(LOG_MSG, wxString::Format(_T("Offset: %d, Size: %d, Scale: %f Tr: %f"), offset, size, scale, translate));
   return scale * ((double)value + translate);
}

int RT2ViewerConverter::getValue(int size, int offset)
{
    if ( ( offset + size ) > m_size ) { return 0; }
    if(size < 1 || size > 4) return 0;
    return Math::getNumber((uint8_t*)&(m_buffer[offset]),size);
}


void RT2ViewerConverter::fillGraphNodeSet(const DoubleHash& requiredData, GraphNodeSet& nodes)
{
    const SymbolTable& symbolTable = m_symbolMgr->getExpTable();
    DoubleHash::const_iterator IT = requiredData.begin();
    for(; IT != requiredData.end(); ++IT )
    {
        wxString symbolName = IT->first;
        //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillGraphNodeSet(): m_RegisteredSymbolTable: ")+ symbolName);

        SymbolTable::const_iterator it = symbolTable.find(symbolName);
        if(it == symbolTable.end() ) continue;
        Symbol* aSymbol = it->second;
        if( aSymbol->hasDepending() ){
            nodes.insert( aSymbol );
            //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillGraphNodeSet(): depending: ")+ symbolName);
        }
    }
}

void RT2ViewerConverter::purge( GraphNodeSetArray* m_nodeArray)
{
    //delete containers( array, set )
    for( size_t i = 0; i < m_nodeArray->GetCount(); i++ )
    {
        GraphNodeSet* levelSet = m_nodeArray->Item( i );
        delete levelSet;
    }
    delete m_nodeArray;
}

void RT2ViewerConverter::fillDependentValues()
{
    //LOG
    if(!m_dependencyA) return;
    size_t i;
    //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() START Dag level: %i"),m_dependencyA->GetCount() );
    for( i = 0; i < m_dependencyA->GetCount(); i++ )
    {
        GraphNodeSet* levelSet = m_dependencyA->Item( i );

        //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() - level : %i"), i );

        //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() - level size: %i"), levelSet->size() );

        GraphNodeSet::const_iterator itSet = levelSet->begin();
        for( ; itSet != levelSet->end(); ++itSet )
        {
            //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() - node addres : %p"), *itSet );
            Symbol* eSymbol = ( Symbol* )(*itSet);
            //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() - node address: %p"), eSymbol );
            //m_rtRawValues[ eSymbol->getName() ] = getValue( eSymbol->getName() );
            wxString symbolName = eSymbol->getName();
            if ( symbolName == _T("time") ) continue;
            //Logger::logf(LOG_MSG,_T("m_viewerModel=%p, e=%i, symbolName=")+symbolName,m_viewerModel,m_viewerModel->find(symbolName) == m_viewerModel->end() ? 0 : 1 );
            bool res; res = getValue( symbolName,  (*m_viewerModel)[ symbolName ] );
            //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() name: ")+ symbolName );
            //if(!res) (*m_viewerModel)[symbolName] = -1;

            //double value = getValue( symbolName );
            //(*m_viewerModel)[ symbolName ] = value;
            //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::fillDependentValues() - expression : ")
            //           + symbolName + _T(" = %f"), value );
        }
    }
}

void RT2ViewerConverter::fillRegisteredValues(DoubleHash& requiredData)
{
    //Logger::logf( LOG_MSG, _T("RT2ViewerConverter::fillRegistered()size:%i"),  (*m_RegisteredSymbolTable).size());

    DoubleHash::iterator it;
    for( it = requiredData.begin(); it != requiredData.end(); ++it )
    {
        wxString symbolName = it->first;
        if ( symbolName == _T("time") ) continue;

        //if values already evaluated no need evalute ( use evaluated data )
        /*
        if( m_viewerModel->find( symbolName ) != m_viewerModel->end() )
        {
            (*m_RegisteredSymbolTable)[symbolName] = (*m_viewerModel)[ symbolName ];
            Logger::logf( LOG_MSG, _T("RT2ViewerConverter::fillRegisteredSymbolTable() 1 - %f symbol: ") + symbolName,  (*m_RegisteredSymbolTable)[symbolName]);
        }
        else
        {*/

            bool res; res = getValue(symbolName, requiredData[symbolName]);
            //if(!res) requiredData[symbolName] = -1;

            //double value = getValue(symbolName);
            //requiredData[symbolName] = value;
            //(*m_viewerModel)[ symbolName ] = value;
            //Logger::logf( LOG_MSG, _T("RT2ViewerConverter::fillRegistered() %f symbol: ") + symbolName,  (requiredData[symbolName]));
        //}

       // Logger::logf( LOG_MSG, _T("RT2ViewerConverter::fillSymbolValues() - symbol : ") + symbolName );
    }

}

/** @brief getValue
  *
  * @todo: document this function
  */
bool RT2ViewerConverter::getValue(const wxString& symbolname, double& value)
{
    //return 10 +rand()/1000;
    //Logger::log(LOG_MSG,_T("----DataMgrSerial::getValue()---- Symbol: ") + symbolname  );

    if(!m_symbolMgr) { return false; } //exception needed ?!

    Symbol* s = m_symbolMgr->getSymbol(symbolname);

	if(!s){ return false; } // exception needed ?!

    if( /*m_rtRawValues.find(symbolname) != m_rtRawValues.end() */ s->isScalar() ){
//	    Logger::log(LOG_MSG, wxString::Format(_T("Symbol (scalar): ") + symbolname + _T(", Value: %f !!!!!!!!!!!!!"), m_rtRawValues[symbolname] ));
		//Logger::logf( LOG_MSG, _T("RTDataRawInputBase::getValue() - ") + s->getName() + _T(" = %f"), m_rtRawValues[symbolname] );
		value = (*m_viewerModel)[symbolname];
        return true;

	}else if(s->isExpr()){
	    wxString expr = s->getExpr();
	    //double value = m_eval.calculate(expr);
        //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::getValue() - symbol : ") + s->getName() );

        value = m_rpn.calculate( expr );

        //Logger::logf( LOG_MSG, _T("RTDataRawInputBase::getValue() - ") + s->getName() + _T(" = %f"), value );
        //Logger::logf(LOG_MSG, _T("Symbol: ") + symbolname + _T(" Expr: ") + expr + _T(", Value: %f"), value );
        /*if( symbolname == _T("longitude") )
        {
            Logger::logf( LOG_MSG, _T("RTDataRawInputBase::getValue() - longitude = %f"), value );
        }*/

	  	return true;
    }else{
        return false; // exception needed ?!
    }

}

void RT2ViewerConverter::printViewerModel()
{
    size_t s = m_viewerModel->size();
    Logger::logf(LOG_MSG, _T("RT2ViewerConverter::printViewerModell() size: %d"), s);
    DoubleHash::const_iterator IT =  m_viewerModel->begin();
    for(; IT != m_viewerModel->end();++IT ){
        wxString key = IT->first;
        double value = IT->second;
        Logger::logf(LOG_MSG, _T("RT2ViewerConverter::printViewerModell(): ")+key+_T("=%f"), value);
    }
}

