#include "RTLiveAnalyzer.h"

#include "TablePanel.h"
#include "AccSetMatrix.h"

#include "../expression/VariableData.h"
#include "../loghistory/ExpressionSQL.h"

#include "../rtdatamanager/RTDataMgrBase.h"
#include "../stdatamanager/STDataMgrBase.h"
#include "../vemsgauge/VemsTableDesc.h"

#include "../logger.h"
#include "../SymbolMgr.h"



BEGIN_EVENT_TABLE( RTLiveAnalyzer, wxFrame )
    /// TODO
END_EVENT_TABLE()

RTLiveAnalyzer::RTLiveAnalyzer( wxWindow* parent, RTDataMgrBase* rtMgrBase, STDataMgrBase* stMgrBase, SymbolMgr* symbolMgr, VemsTableDesc* tableDesc )
    : wxFrame( parent, -1, _T("VE Live Analyze"), wxDefaultPosition, wxDefaultSize )
{
    SetIcon( wxIcon( _T("vemsLogoIcon2") ) );

    SetDataMgr( rtMgrBase );
    m_rtDataMgr->RegisterViewer( this );

    m_symbolMgr = symbolMgr;

    m_tableDesc = tableDesc;
    m_xBinSymbol = m_symbolMgr->getSymbol( m_tableDesc->xSTSymbol );
    m_yBinSymbol = m_symbolMgr->getSymbol( m_tableDesc->ySTSymbol );

    m_xRTSymbol = m_symbolMgr->getSymbol( m_tableDesc->xRTSymbol );
    m_yRTSymbol = m_symbolMgr->getSymbol( m_tableDesc->yRTSymbol );

    m_cols = (unsigned int) m_xBinSymbol->getShape0();
    m_rows = (unsigned int) m_yBinSymbol->getShape0();

    m_timeMatrix = new VemsMatrix<double>( m_rows, m_cols );
    m_accSetMatrix = new AccSetMatrix( m_rows, m_cols );
    m_dataMatrix = new VemsMatrix<double>( m_rows, m_cols );

    m_stDataMgr = stMgrBase;

    m_defaultExpr = ExpressionSQL::getExpressionByName( _T("default_expression") );

    RPNParser rpn;
    // get required symbols from expression
    wxArrayString syms = rpn.getVariablesArray( m_defaultExpr );

    // get required symbols from filter
    m_filter = _T("((pulseWidth > 0) && (lambda < 2) && (lambda > 0.1) && (coolant > 70))");
    wxArrayString exprSyms = rpn.getVariablesArray( m_filter );

    for ( size_t i = 0; i < exprSyms.GetCount(); ++i )
    {
        LOGP( exprSyms[i] )
        syms.Add( exprSyms[i] );
    }

    // add  x and y symbols
    wxString xRTName = m_xRTSymbol->getName();
    wxString yRTName = m_yRTSymbol->getName();
    syms.Add( xRTName );
    syms.Add( yRTName );

    setRequiredSymbols( syms );

    buildGUI();
}

RTLiveAnalyzer::~RTLiveAnalyzer()
{
    //dtor
    m_rtDataMgr->UnRegisterViewer( this );
    delete m_dataMatrix;
    delete m_accSetMatrix;
    delete m_timeMatrix;
}

void RTLiveAnalyzer::processData()
{
    /// TODO
    RPNParser rpn;

    VariableDataDH vd( m_data );
    rpn.setVariableData( &vd );

    DoubleHash::iterator it = m_data->begin();

    double filterVal = rpn.calculate( m_filter );

    if ( filterVal == 0 )
    {
        ++m_dropped;
        return;
    }

    double expr_val = rpn.calculate( m_defaultExpr );

    wxString xRTName = m_xRTSymbol->getName();
    wxString yRTName = m_yRTSymbol->getName();

    int xvalue = 0;
    int yvalue = 0;

    it = m_data->find( xRTName );
    if ( it != m_data->end() )
    {
        xvalue = (int)( it->second );
    }

    it = m_data->find( yRTName );
    if ( it != m_data->end() )
    {
        yvalue = (int)( it->second );
    }

    updateMatrix( xvalue, yvalue, expr_val );
}

void RTLiveAnalyzer::buildGUI()
{
    SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );
    ClearBackground();

    wxSizer* sizer = new wxBoxSizer( wxVERTICAL );

    wxSizer* hsizer = new wxBoxSizer( wxHORIZONTAL );
    wxStaticText* txt = new wxStaticText( this, -1, _T("Set minimal amount of data for a cell") );
    hsizer->Add( txt, 0, wxEXPAND | wxALL, 3 );
    m_minData = new wxTextCtrl( this, -1, _T("5"), wxDefaultPosition, wxSize( 40, -1 ) );
    hsizer->Add( m_minData, 0, wxEXPAND | wxALL, 3 );
    txt = new wxStaticText( this, -1, _T("secs") );
    hsizer->Add( txt, 0, wxEXPAND | wxALL, 3 );

    sizer->Add( hsizer, 0, wxALL, 3 );

    wxArrayString xbin, ybin;
    getBins( xbin, ybin );

    m_table = new TablePanel( this, 600, 300, xbin, ybin, true );
    m_table->setBgColor(wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE));
    sizer->Add( m_table, 1, wxEXPAND | wxALL, 3 );

    m_table->setMatrix( m_dataMatrix );
    m_table->setCountMatrix( m_timeMatrix );

    SetSizer( sizer );
    sizer->Fit( this );
}

void RTLiveAnalyzer::getBins( wxArrayString& xbins, wxArrayString& ybins )
{
    // row names
    int i;
    short prev_value = 0;
    for ( i = (int)(m_rows-1); i >=0; --i )
    {
        short value = (short) m_stDataMgr->getDataFromViewerModel(*m_yBinSymbol, 0, i);
        if ( i < (int)(m_rows-1) )
        {
            short limit = (value + prev_value) / 2;
            m_HorizLimits.Add(limit);
        }

        if ( i == (int)(m_rows-2) )
            m_ymax = prev_value + ( prev_value - value );

        if ( i == 0 )
            m_ymin = value - (prev_value - value);

        prev_value = value;

        ybins.Add(wxString::Format( _T("%d"), value ));

    }

    // col names
    for ( i = 0; i < (int)m_cols; ++i )
    {
        short value = (short) m_stDataMgr->getDataFromViewerModel(*m_xBinSymbol,0,i);

        if ( (int)i > 0 )
        {
            short limit = (value + prev_value) / 2;
            m_VertLimits.Add(limit);
        }

        if ( i == (int)(m_cols-1) )
            m_xmax = value + ( value - prev_value );

        if ( i == 1 )
            m_xmin = prev_value - (value - prev_value);

        prev_value = value;

        xbins.Add(wxString::Format( _T("%d"), value ));
    }
}

void RTLiveAnalyzer::updateMatrix( int xvalue, int yvalue, double expr_value )
{
    unsigned int i;
    int matcol = -1;
    int matrow = -1;

    if ( xvalue > m_xmax || xvalue < m_xmin || yvalue > m_ymax || yvalue < m_ymin )
    {
        ++m_outofrange;
        return;
    }

    for ( i = 0; i < m_cols - 1; ++i )
    {
        if ( xvalue < m_VertLimits[i] )
        {
            matcol = i;
            break;
        }
    }

    for ( i = 0; i < m_rows - 1; ++i )
    {
        if ( yvalue > m_HorizLimits[i])
        {
            matrow = i;
            break;
        }
    }

    matcol = ( matcol != -1 ) ? matcol : m_cols - 1;
    matrow = ( matrow != -1 ) ? matrow : m_rows - 1;

    m_timeMatrix->set( matrow, matcol, ( m_timeMatrix->get( matrow, matcol ) + 1.0 ) );

    AccSetT* cell = m_accSetMatrix->getData( matrow, matcol );
    (*cell)( expr_value, weight = 1.0 );

    double val = mean(*cell);

    Logger::logf(LOG_MSG, _T("row=%d, col=%d mean=%f"), matrow, matcol, val);

    m_dataMatrix->set( matrow, matcol, val );

    m_table->Refresh(false);
}
