#include "TablePanel.h"

#include <wx/dcbuffer.h>
#include "../logger.h"

BEGIN_EVENT_TABLE( TablePanelBase, wxPanel )
    EVT_PAINT( TablePanelBase::onPaint )
    EVT_SIZE( TablePanelBase::onSize )
END_EVENT_TABLE()

TablePanelBase::TablePanelBase( wxWindow* parent, int width, int height, const wxArrayString& xbins, const wxArrayString& ybins, bool showHeader )
{
    m_width     = width;
    m_height    = height;
    m_xbins     = xbins;
    m_ybins     = ybins;
    m_showHeader = showHeader;

    m_showValues = true;

    m_numformat = _T("%.0f");

    m_mincount = 1;

    m_rows = m_ybins.GetCount();
    m_cols = m_xbins.GetCount();

    m_matrix = NULL;

    Create( parent, wxID_ANY, wxDefaultPosition, wxSize( width, height ) );

    int totalcols = ( showHeader ) ? ( m_cols + 1 ) : m_cols;
    int totalrows = ( showHeader ) ? ( m_rows + 1 ) : m_rows;
    m_cellw = (int) ((width - totalcols * 2) / totalcols );
    m_cellh = (int) ( (height - totalrows *2) / totalrows );

    m_cellsize.SetWidth( m_cellw );
    m_cellsize.SetHeight( m_cellh );

    m_calculated = false;

    m_labelBkColor = wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE);

    resetColorGradients();
}

TablePanelBase::~TablePanelBase()
{
    if ( m_matrix )
    {
        delete m_matrix;
        m_matrix = NULL;
    }
    if ( m_countmatrix )
    {
        delete m_countmatrix;
        m_countmatrix = NULL;
    }
}

void TablePanelBase::setMatrix( VemsMatrix<double>* matrix )
{
    if ( matrix->cols() != m_cols || matrix->rows() != m_rows )
    {
        return;
    }

    if ( m_matrix == NULL )
    {
        m_matrix = new VemsMatrix<double>( m_rows, m_cols );
    }

    memcpy( m_matrix->getData(), matrix->getData(), m_cols * m_rows * sizeof(double) );

    if ( m_countmatrix == NULL )
    {
        m_countmatrix = new VemsMatrix<double>( m_rows, m_cols );
        m_countmatrix->fill( 1.0 );
    }

    m_calculated = false;

    Refresh(false);
}

void TablePanelBase::setCountMatrix( VemsMatrix<double>* matrix )
{
    if ( matrix->cols() != m_cols || matrix->rows() != m_rows )
    {
        return;
    }

    memcpy( m_countmatrix->getData(), matrix->getData(), m_cols * m_rows * sizeof(double) );
}

VemsMatrix<double>* TablePanelBase::getMatrix()
{
    return m_matrix;
}

VemsMatrix<double>* TablePanelBase::getCountMatrix()
{
    return m_countmatrix;
}

void TablePanelBase::setCustomColors( DoubleArray limits, DoubleArray colors )
{
    m_colorlimits = limits;
    m_customcolorsA = colors;

    m_customcolors = true;
    Refresh();
}

void TablePanelBase::resetColorGradients()
{
    m_gradientColorArray.Clear();

    m_gradientColorArray.Add(wxColor(32,64,255));
    m_gradientColorArray.Add(wxColor(32, 128, 32));

    m_gradientColorArray.Add(wxColor(224, 224, 32));

    m_gradientColorArray.Add(wxColor(128,64,0));
    m_gradientColorArray.Add(wxColor(224, 128, 64));

    m_customcolors = false;

    if ( m_matrix != NULL )
        Refresh(false);
}

void TablePanelBase::onPaint( wxPaintEvent& event )
{
    wxBufferedPaintDC dc( this );

    dc.SetBackground( m_bgColor );
    dc.Clear();

    dc.SetBrush(*wxWHITE_BRUSH);
    dc.SetPen( *wxWHITE_PEN );

    int totalcols = ( m_showHeader ) ? ( m_cols + 1 ) : m_cols;
    int totalrows = ( m_showHeader ) ? ( m_rows + 1 ) : m_rows;

    Draw3DBorder( dc, wxRect(0,0,(m_cellw + 2 ) * totalcols, ( m_cellh + 2 ) * totalrows ) );

    dc.DrawRectangle(m_cellw + 3, m_cellh+3, (m_cellw+2) * m_cols - 2, (m_cellh+2) * m_rows -2 );

    Draw3DCells( dc );

    if ( m_matrix != NULL )
    {
        DrawData( dc );
    }
}

void TablePanelBase::Draw3DCells( wxDC& dc )
{
    if ( m_showHeader == false )
    {
        return;
    }

    wxBrush oldbrush = dc.GetBrush();
    dc.SetPen( wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE) );
    dc.SetBrush( m_labelBkColor );

    // get text point size
    wxFont font( 1, wxFONTFAMILY_SWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD);
    int w1 = CalcFontSize(dc,wxRect(0,0,m_cellw,m_cellh), m_xbins[m_xbins.GetCount()-1],font); /// need to discuss
    int w2 = CalcFontSize(dc,wxRect(0,0,m_cellw,m_cellh), m_ybins[0],font);
    int fontsize = wxMin(w1,w2);
    font.SetPointSize(fontsize);
    dc.SetFont(font);
    int i = 0, xpos = 0;
    while ( i < m_cols+1 )
    {
        Draw3DBorder(dc,wxRect(xpos, 0, m_cellw+2, m_cellh+2));

        // draw
        dc.SetTextForeground(wxColor(0,0,0));
        if ( i > 0)
        {
            wxSize s = dc.GetTextExtent(m_xbins[i-1]);
            dc.DrawText( m_xbins[i-1], xpos + ( m_cellw - s.x)/2, (m_cellh - s.y)/2 );
        }


        xpos += ( m_cellw + 2);
        i++;
        if ( i > 1 && i < m_cols+1)
            dc.DrawLine( xpos, m_cellh+2, xpos, (m_cellh+2)* ( m_rows+1) -1 );
    }
    i = 1; size_t ypos = m_cellh+2;
    while ( i < m_rows+1 )
    {
        Draw3DBorder(dc, wxRect(0, ypos, m_cellw+2, m_cellh+2));

        dc.SetTextForeground(wxColor(0,0,0));
        wxSize s = dc.GetTextExtent(m_ybins[i-1]);
        dc.DrawText( m_ybins[i-1], ( m_cellw - s.x)/2, ypos + (m_cellh - s.y)/2 );

        ypos += ( m_cellh + 2);
        i++;
        if ( i > 1 && i < m_rows+1)
            dc.DrawLine( m_cellw+2, ypos, (m_cellw+2)* ( m_cols+1) -1, ypos );
    }
}

void TablePanelBase::resizeTo( int width, int height )
{
    m_width = width;
    m_height = height;

    int totalcols = ( m_showHeader ) ? ( m_cols + 1 ) : m_cols;
    int totalrows = ( m_showHeader ) ? ( m_rows + 1 ) : m_rows;

    m_cellw = (int) ((width - totalcols * 2) / totalcols );
    m_cellh = (int) ( (height - totalrows *2) / totalrows );

    m_cellsize.SetWidth(m_cellw);
    m_cellsize.SetHeight(m_cellh);
    Refresh(false);
}

void TablePanelBase::Draw3DBorder( wxDC& dc, const wxRect& rc )
{
	wxPen oldPen = dc.GetPen();

	dc.SetPen( wxPen(wxSystemSettings::GetColour(wxSYS_COLOUR_3DSHADOW), 1, wxSOLID));
	dc.DrawLine(rc.x, rc.y, rc.GetRight() + 1, rc.y);
	dc.DrawLine(rc.x, rc.y, rc.x, rc.GetBottom() + 1);
	dc.DrawLine(rc.GetRight(), rc.GetBottom(), rc.GetRight(), rc.y);
	dc.DrawLine(rc.GetRight(), rc.GetBottom(), rc.x, rc.GetBottom());

	dc.SetPen(*wxWHITE_PEN);
	dc.DrawLine(rc.x + 1, rc.y + 1, rc.GetRight(), rc.y + 1);
	dc.DrawLine(rc.x + 1, rc.y + 1, rc.x + 1, rc.GetBottom());

	dc.SetPen(oldPen);
}

void TablePanelBase::onSize( wxSizeEvent& event )
{
     unsigned int w = event.GetSize().x;
     unsigned int h = event.GetSize().y;

     resizeTo( w, h );

     event.Skip();
}

int TablePanelBase::CalcFontSize( wxDC& dc, wxRect rect, const wxString& str, wxFont font )
{
    int w = rect.width, h = rect.height;
    wxCoord w1, h1;

    int pointsize = 1;
    do
    {
        font.SetPointSize( pointsize );
        dc.GetTextExtent( str, &w1, &h1, NULL, NULL, &font );

        if ( (w1>=w) || (h1>=h) )
        {
            break;
        }

        else ++pointsize;

    } while ( 1 );

    return ( pointsize - 1 );
}

int TablePanelBase::CalcCellFontSize( wxDC& dc )
{
    wxFont font( 1, wxFONTFAMILY_SWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL );

    int fontsize = CalcFontSize( dc, m_cellsize, _T("-999.99"), font );
    return fontsize;
}

void TablePanelBase::GetTblMinMax(wxDC& dc)
{
    if ( m_matrix == NULL ) return;

    // initialization

    m_tblmin =  9E9;
    m_tblmax = -9E9;

    for ( size_t i = 0; i < m_matrix->rows(); ++i )
    {
        for ( size_t j = 0; j < m_matrix->cols(); ++j )
        {
            double value = m_matrix->get( i, j );

            if ( m_tblmin > value )
            {
                m_tblmin = value;
            }
            if (m_tblmax < value)
            {
                m_tblmax = value;
            }
        }
    }
    m_calculated = true;
}

wxColor TablePanelBase::dbl2Color(double rgb)
{
    unsigned char r,g,b;
    unsigned int irgb = (unsigned)rgb;
    r = ( irgb / (256*256) ); irgb = ( irgb % (256*256) );
    g = ( irgb / 256 ); b = irgb % 256;

    return wxColor( r, g, b );
}

wxColor TablePanelBase::calcCustomColorCellVal(double value)
{
    if ( m_customcolorsA.GetCount() < 1 || m_colorlimits.GetCount() < 1 )
        return wxColor(255,255,255);

    size_t i;

    int maxindex = m_colorlimits.GetCount()-1;

    if ( value <= m_colorlimits[0] )
        return dbl2Color( m_customcolorsA[0] );

    if ( value >= m_colorlimits[maxindex])
        return dbl2Color( m_customcolorsA[maxindex] );

    for ( i = 0; i < m_colorlimits.GetCount()-1; ++i )
    {
        double limitval = m_colorlimits[i+1];
        if ( value < limitval )
        {
            break;
        }
    }
    // interpolation
    double minval = m_colorlimits[i], maxval = m_colorlimits[i+1];

    double corr = ( value - minval ) / ( maxval - minval );

    wxColor leftColor = dbl2Color(m_customcolorsA[i]);
    wxColor rightColor = dbl2Color(m_customcolorsA[i+1]);

    wxColor interpolatedCol;
    unsigned int red = leftColor.Red() - (int)((double)(leftColor.Red() - rightColor.Red()) * corr);
    unsigned int green = leftColor.Green() - (int)((double)(leftColor.Green() - rightColor.Green()) * corr);
    unsigned int blue = leftColor.Blue() - (int)((double)(leftColor.Blue() - rightColor.Blue()) * corr);
    return wxColor(red,green,blue);
}

inline wxRect TablePanelBase::GetCellRect( int row, int col )
{
    if ( m_showHeader )
    {
        wxRect r( (m_cellw+2)*(col+1)+1, (m_cellh+2)*(row+1)+1, m_cellw, m_cellh );
        return r;
    }
    else
    {
        wxRect r( (m_cellw+2)*(col) + 1, (m_cellh+2)*(row) + 1, m_cellw, m_cellh );
        return r;
    }
}

void TablePanelBase::DrawData(wxDC& dc)
{
    if ( m_matrix == NULL ) return;

    if ( !m_calculated )
        GetTblMinMax(dc);

    int fontsize = CalcCellFontSize(dc);

    wxFont font( 1, wxFONTFAMILY_SWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL);
    font.SetPointSize( fontsize-1 );
    dc.SetFont( font );

    wxString str;


    for ( size_t i = 0; i < m_matrix->rows(); ++i )
    {
        for ( size_t j = 0; j < m_matrix->cols(); ++j )
        {
            double cellcount = m_countmatrix->get( i, j );

            double value = m_matrix->get( i, j );

            ///
            if ( cellcount < (double)m_mincount )
            {
                dc.SetTextForeground(wxColor(160,160,160));

                dc.SetBrush( *wxWHITE );
                dc.SetPen(*wxWHITE);
                wxRect r = GetCellRect( i, j );
                dc.DrawRectangle( r.x, r.y, r.width, r.height );

                if ( m_showValues == false )
                {
                    continue;
                }

                wxSize s = dc.GetTextExtent( _T("N/A") );

                int x;
                int y;

                if ( m_showHeader )
                {
                    x = (m_cellw+2)*(j+1)+1 + ( m_cellw - s.x)/2;
                    y = (m_cellh+2)*(i+1)+1 + (m_cellh - s.y)/2;
                }
                else
                {
                    x = (m_cellw+2)*(j)+1 + ( m_cellw - s.x)/2;
                    y = (m_cellh+2)*(i)+1 + (m_cellh - s.y)/2;
                }

                dc.DrawText( _T("N/A") , x, y );
                dc.SetTextForeground(wxColor(0,0,0));

                continue;

            }
            else
            {
                wxColor color = GetCellColor( value );

                str = wxString::Format( m_numformat, value);

                wxRect r = GetCellRect( i, j);
                dc.SetBrush(color); dc.SetPen(color);
                dc.DrawRectangle( r.x, r.y, r.width, r.height );
            }

            if ( m_showValues == false )
            {
                continue;
            }


            wxSize s = dc.GetTextExtent( str );

            int x, y;

            if ( m_showHeader )
            {
                x = (m_cellw+2)*(j+1)+1 + ( m_cellw - s.x)/2;
                y = (m_cellh+2)*(i+1)+1 + (m_cellh - s.y)/2;
            }
            else
            {
                x = (m_cellw+2)*(j)+1 + ( m_cellw - s.x)/2;
                y = (m_cellh+2)*(i)+1 + (m_cellh - s.y)/2;
            }

            dc.DrawText( str, x, y );
            dc.SetTextForeground(wxColor(0,0,0));
        }
    }
}

wxColor TablePanelBase::GetCellColor( double value )
{
    if ( m_customcolors )
        return calcCustomColorCellVal(value);


    if( m_gradientColorArray.IsEmpty() ) { return *wxRED; }

    int maxIndex = m_gradientColorArray.Count()-1;

    const float clr_factor = ( value - m_tblmin) / (m_tblmax - m_tblmin);

    if( clr_factor < 0 ) { return *wxRED; }

    float baseIndex = clr_factor * maxIndex;
    int startIndex = (int)baseIndex;
    int endIndex = startIndex + 1;
    if(endIndex >= maxIndex ){  endIndex = maxIndex;}

    wxColor startColor = m_gradientColorArray.Item(startIndex);
    wxColor endColor = m_gradientColorArray.Item(endIndex);

    float reminder = baseIndex - startIndex;

    const int r = startColor.Red() - int((int(startColor.Red()) - int(endColor.Red())) * reminder);
    const int g = startColor.Green() - int((int(startColor.Green()) - int(endColor.Green())) * reminder);
    const int b = startColor.Blue() - int((int(startColor.Blue()) - int(endColor.Blue())) * reminder);

    return wxColor(r,g,b);
}

///------------------------------------------------------------------------------
///------------------------------------------------------------------------------
// TablePanel implementation

BEGIN_EVENT_TABLE( TablePanel, TablePanelBase )
    EVT_RIGHT_DOWN ( TablePanel::onRightClick )
    EVT_LEFT_DOWN ( TablePanel::onLeftClick )
    EVT_MOUSE_EVENTS( TablePanel::onMouse )

    EVT_KEY_DOWN ( TablePanel::onKeyDown )
END_EVENT_TABLE();


TablePanel::TablePanel( wxWindow* parent, int width, int height, const wxArrayString& xbins, const wxArrayString& ybins, bool showHeader )
    : TablePanelBase( parent, width, height, xbins, ybins, showHeader )
{
}

TablePanel::~TablePanel()
{
}


void TablePanel::onLeftClick( wxMouseEvent& event )
{
    event.Skip();
}

void TablePanel::onRightClick( wxMouseEvent& event )
{
    event.Skip();
}

void TablePanel::onMouse( wxMouseEvent& event )
{
    event.Skip();
}

void TablePanel::onKeyDown( wxKeyEvent& event )
{
    event.Skip();
}
