#include <iostream>
#include <stdlib.h>
#include "arg_parser.h"
#include "signalgen.h"

using namespace std;

void usage(char *progname) {       
  cout <<  "Usage:" << progname << " [options] \n";
  cout << "Options: " << endl;
  cout << " -a <amplitude> " << endl;
  cout << "        Amplitude in 10ths of a percent. This can also be specified " << endl;
  cout << "        as a range that is linear through the duration of the signal" << endl;
  cout << "        for example 1000-500 would reduce the amplitude from 100% to" << endl;
  cout << "        50% in a linear fashion. 500-1000-500 would start at 50%, " << endl;
  cout << "        increase to 100% then decrease back to 50%. Default is 100%" << endl;
  cout << " -c <count> " << endl;
  cout << "        Tooth count on the trigger" << endl;
  cout << " -f <filename> / --file <filename> " << endl;
  cout << "        Output wavefile as <filename>. Default is output.wav " << endl;
  cout << " -h / --help " << endl;
  cout << "        Display usage options " << endl;
  cout << " -l <seconds> / --length <seconds> " << endl;
  cout << "        Length of generated .wav file in seconds. Default is 30 " << endl;
  cout << " -m <count> " << endl;
  cout << "        Missing tooth count on the trigger" << endl;
  cout << " -r <rpm> " << endl;
  cout << "        RPM at which to generate the signal. This can also be specified " << endl;
  cout << "        as a range that is linear through the duration of the signal" << endl;
  cout << "        for example 50-5000 would start at 50 RPM and increase to 5000." << endl;
  cout << "        50-200-200-500 would start at 50, rise to 200, hold there for " << endl;
  cout << "        1/4 of the duration then rise to 500 RPM. Default is 1000." << endl;
  cout << " -t <coiltype> / --type <coiltype>" << endl;
  cout << "        Type of trigger to emulate " << endl;
  cout << "          " << (char)COIL << ": Simple coil type" << endl;
  cout << "          " << (char)MULTITOOTH << ": multitooth \n";
  cout << "          " << (char)FIAT << ": Fiat Stilo\n";
  cout << "          " << (char)LOTUS << ": Lotus Elise\n";
  cout << "          " << (char)SUBARU << ": Subaru EJ series\n";
  cout << "          " << (char)COSWORTH << ": Lancia/Cosworth 4 crank + 2 cam\n";
  cout << "          " << (char)MAZDA << ": Mazda BP05 4 + 1 style\n";
}


// parses a string consisting of numbers with dashes into a vector of numbers
vector<uint16_t> parseRange(std::string r) {
  string str=r;
  vector<uint16_t> ret;
  // while there are any '-' chars then there must be more tokens
  string::size_type loc;
  while ( (loc = str.find('-')) != string::npos ) {
    ret.push_back(strtol(str.substr(0,loc).c_str(),NULL,10));
    str.erase(0,loc+1); 
  }

  // there must be a final number in there so extract it and push on the vector
  ret.push_back(strtol(str.c_str(),NULL,10));
  return ret;
}


int main(int argc, char *argv[]) {
  SignalGen* sg = new SignalGen();

  // define and process the options
  const Arg_parser::Option options[] =
    {
    { 'a', "",         Arg_parser::yes   },
    { 'c', "",         Arg_parser::yes   },
    { 'h', "help",     Arg_parser::no    },
    { 'f', "file",     Arg_parser::yes   },
    { 'm', "",         Arg_parser::yes   },
    { 'l', "length",   Arg_parser::yes   },
    { 't', "type",     Arg_parser::yes   },
    { 'r', "rpm",      Arg_parser::yes   },
    {   0, 0,          Arg_parser::no    } };

  Arg_parser parser( argc, argv, options );
  if( parser.error().size() ) {                         // bad option
    cerr << parser.error().c_str() << endl; 
    return 1; 
  }

  for( int argind = 0; argind < parser.arguments(); ++argind ) {
    const int code = parser.code( argind );
    if( !code ) break;                                  // no more options
    switch( code ) {
      case 'a':
        sg->setAmps(parseRange(parser.argument( argind )));
        break;
      case 'c':
        sg->setWheelToothCnt(strtol(parser.argument( argind ).c_str(),NULL,10));
        break;
      case 'f': 
        sg->setFilename(parser.argument( argind ).c_str());
        break;
      case 'h': 
        usage(argv[0] ); 
        return 0;
      case 'r':
        sg->setRPMs(parseRange(parser.argument( argind )));
        break;
      case 't': { 
        std::string typestr=parser.argument(argind);
        sg->setPrimaryTrig(typestr[0]);
        }
        break;
      case 'l':
        sg->setDuration(strtol(parser.argument( argind ).c_str(),NULL,10));
        break;
      case 'm':
        sg->setWheelMissing(strtol(parser.argument( argind ).c_str(),NULL,10));
        break;
    }
  } // end process options

  // should probably have more error checking
  if( sg->openWavFile() ){
    sg->generate_wav_output();
    sg->closeWavFile();
  }

  return 0;
}


