#include "STDataIOVemsCfg.h"


#include "STDataModelBase.h"
#include "../Math.h"
#include "../Symbol.h"
#include "../SymbolMgr.h"
#include "../logger.h"
#include "../version.h"
#include "../firmwareconfig/FirmwareConfigData.h"


#include <wx/textfile.h>
#include <wx/sstream.h>
#include <wx/wfstream.h>


void STDataIOVemsCfg::saveTo(   const wxString& header,
                                const SymbolArray& headerSymbols,
                                SymbolMgr* iniSymbols,
                                const wxString& filename,
                                bool isRawLambdaUsed )
{
    LOG
    m_isRawLambdaUsed = isRawLambdaUsed;
    wxFile file ( filename, wxFile::write );

    if ( !file.IsOpened() ) { return; }

    /// write header
    file.Write( header.mbc_str(), header.Length() );
    file.Write( _T("\n") );

    writeScalarSymbolsToFile( file, headerSymbols );
    writeArraySymbolsToFile( file, iniSymbols->getConfigTableSymbolsInOrder(), iniSymbols->getConfigTableSymbols() );

    file.Close();
}

void STDataIOVemsCfg::saveTo( FirmwareConfigData* fcData, SymbolMgr* iniSymbols, bool isRawLambdaUsed )
{
    m_isRawLambdaUsed = isRawLambdaUsed;
    saveScalarsTo( fcData, iniSymbols->getCfgSymbolTable() );
    saveArraysTo( fcData, iniSymbols->getConfigTableSymbols() );
}

void STDataIOVemsCfg::saveScalarsTo( FirmwareConfigData* fcData, const SymbolTable& headerSymbols )
{
    SymbolTable::const_iterator iter = headerSymbols.begin();

    for ( ; iter != headerSymbols.end(); ++iter )
    {
        Symbol* s = iter->second;

        uint8_t data = m_model->getRawData( *s );

        wxString symName = s->getName();

        if ( symName.IsEmpty() ) continue;

        fcData->addScalar( symName, data, 0 );
    }
}

void STDataIOVemsCfg::saveArraysTo( FirmwareConfigData* fcData, const SymbolTable& configTables )
{
    SymbolTable::const_iterator iter = configTables.begin();

    for ( ; iter != configTables.end(); ++iter )
    {
        Symbol* configArraySymbol = iter->second;
        saveArrayTo( fcData, configArraySymbol );
    }
}

void STDataIOVemsCfg::saveArrayTo( FirmwareConfigData* fcData, const Symbol* symbol )
{
    wxString name = symbol->getName();
    int page = symbol->getPage();
    int offset = symbol->getOffset();
    int index = 0;
    size_t rows = symbol->getArrayRowSize();
    size_t cols = symbol->getArrayColSize();

    for ( size_t row = 0; row < rows; ++row )
    {
        wxArrayString rowData;
        for ( size_t col = 0; col < cols; ++col )
        {
            index = row * cols + col;

            uint8_t data = m_model->getRawData( page, offset + index );

            /// lambda hack
            if ( name.IsSameAs( _T("l") ) && !m_isRawLambdaUsed )
            {
                double d = ( 65535.0 - ( (double)data + 73.0 ) * 200 ) / ( double(data) + 73.0 );

                data = (uint8_t)d;
            }

            rowData.Add( wxString::Format( _T("%02X"), data ) );
        }

        fcData->addTableLine( name, row, rowData );
    }
}

void STDataIOVemsCfg::writeScalarSymbolsToFile( wxFile& file, const SymbolArray& headerSymbols )
{
    wxString line;

    for ( size_t i = 0; i < headerSymbols.GetCount(); ++i )
    {
        Symbol* s = headerSymbols[i];
        uint8_t data = m_model->getRawData( *s );

        wxString name = s->getName();

        long idx = Math::hex2int( name.AfterFirst( _T('[') ).BeforeLast( _T(']') ) );

        if (  idx > 15 )
        {
            char idxChar = 'F' + idx - 15;
            wxString prefix = name.BeforeFirst( _T('[') );
            name = prefix + wxString::Format( _T("[%c]"), idxChar );
        }

        line = name + wxString::Format( _T("=%02X\n"), data );
        file.Write( line );
    }
}


void STDataIOVemsCfg::writeArraySymbolsToFile( wxFile& file, const SymbolArray& orderedConfigTables, const SymbolTable& configTables )
{
    wxString line;

    for ( size_t  i = 0; i < orderedConfigTables.GetCount(); ++i )
    {
        Symbol* configArraySymbol = orderedConfigTables[i];

        if ( configArraySymbol->isVemsCfgArrayAxis() ) { continue; }
        //if ( isArrayHeaderSymbol( configArraySymbol ) ) { continue; }
        writeArraySymbolToFile( file, configArraySymbol );
        writeArraySymbolHeaderToFile( file, configArraySymbol, configTables );
    }
}

/*
bool STDataIOVemsCfg::isArrayHeaderSymbol( const Symbol* symbol )
{
    return ( symbol->getName().Length() > 1 );
}
*/

void STDataIOVemsCfg::writeArraySymbolToFile( wxFile& file, const Symbol* symbol )
{
    wxString line;

    wxString name = symbol->getName();
    int page = symbol->getPage();
    int offset = symbol->getOffset();
    int index = 0;
    size_t rows = symbol->getArrayRowSize();
    size_t cols = symbol->getArrayColSize();

    for ( size_t row = 0; row < rows; ++row )
    {
        wxString specName;
        long specRow;
        if(symbol->isPecialArray(name, specName, specRow) ){ /// h0 -> h[0]
            line = wxString::Format( specName + _T("[%1X]="), specRow );
        }else{
            line = wxString::Format( name + _T("[%1X]="), row );
        }

        for ( size_t col = 0; col < cols; ++col )
        {
            index = row * cols + col;

            uint8_t data = m_model->getRawData( page, offset + index );

            /// lambda hack
            if ( name.IsSameAs( _T("l") ) && !m_isRawLambdaUsed )
            {
                double d = ( 65535.0 - ( (double)data + 73.0 ) * 200 ) / ( double(data) + 73.0 );

                data = (uint8_t)d;
            }

            line += wxString::Format( _T("%02X"), data );
            if ( col < ( cols - 1 ) )
            {
                line += _T(" ");
            }
        }

        file.Write( line + _T("\n") );
    }
}


void STDataIOVemsCfg::writeArraySymbolHeaderToFile( wxFile& file, const Symbol* symbol, const SymbolTable& configTables )
{
    wxString line;
    wxString name = symbol->getName();

    SymbolTable::const_iterator head_it = configTables.find( name + _T("_r") );

    if ( head_it != configTables.end() )
    {
        line = getTableHeaderLine( name, head_it->second );
        file.Write( line + _T("\n") );
    }

    head_it = configTables.find( name + _T("_k") );
    if ( head_it != configTables.end() )
    {
        line = getTableHeaderLine( name, head_it->second );
        file.Write( line + _T("\n") );
    }
}


wxString STDataIOVemsCfg::getTableHeaderLine( const wxString& tableName, const Symbol* headerSymbol )
{
    wxString headerName = headerSymbol->getName().AfterLast( _T('_') );
    wxString line = tableName + _T("[") + headerName + _T("]=");

    int page = headerSymbol->getPage();
    int offset = headerSymbol->getOffset();

    size_t cols = headerSymbol->getArrayColSize();

    for ( size_t col = 0; col < cols; ++col )
    {
        uint8_t data = m_model->getRawData( page, offset + col );
        line += wxString::Format( _T("%02X"), data );
        if ( col < ( cols - 1 ) )
        {
            line += _T(" ");
        }
    }

    return line;
}

