#include "VemsToolBarToolBase.h"

#include "../VemsAuiManager.h"

#include "../logger.h"


/** \class VemsToolBarToolBase
  *
  * \~english
  * \brief ToolBar icon base class
  * \details
  * This provides the root class for every ToolBar icon handler class. \n\n
  * \ref tbtover
  *
  *
  * \~hungarian
  * \brief ToolBar ikon ősosztály
  * \details
  * Ez az ősosztálya minden, ToolBar ikont kezelő osztálynak. \n\n
  * \ref tbtover
  *
  *
  */

IMPLEMENT_ABSTRACT_CLASS(VemsToolBarToolBase, wxEvtHandler);

VemsToolBarToolBase::VemsToolBarToolBase(int id)
                    : m_itemId (-1)
                    , m_toolID (0)
                    , m_toolBar (NULL)
                    , m_defaultBitmap (NULL)
                    , m_inactiveBitmap (NULL)
                    , m_shortHelpString (wxEmptyString)
                    , m_longHelpString  (wxEmptyString)
                    , m_AUImanager (NULL)
{
    //ctor

    m_toolID = id;
}


VemsToolBarToolBase::~VemsToolBarToolBase()
{
    //dtor
}

/** \~english
  * Set the tool bmp from outer file. png format only
  * \deprecated Old function. Need not use.
  *
  * \~hungarian
  * A Tool bitmap állítható be egy külső fájlból. csak png formátum
  * \deprecated Régi függvény. Nem használatos.
  *
  */
void VemsToolBarToolBase::SetToolBmp(const wxString &filename, bool active)
{
    wxBitmap *bmp = new wxBitmap(filename, wxBITMAP_TYPE_PNG);

    SetToolBmp(bmp,active);
}

/** \~english
  * Set the tool bitmap from wxBitmap class.
  * \param bmp the bitmap data on \c wxBitmap class
  * \param active the type of the tool (that can be disabled or enabled)
  *
  * \~hungarian
  * A tool bitamp állítható be wxBitmap osztályból.
  * \param bmp bitmap adat a \c wxBitmap osztályból
  * \param active a tool tipusa (ami lehet inaktív vagy akítv)
  *
  */
void VemsToolBarToolBase::SetToolBmp(wxBitmap *bmp, bool active)
{
    if(bmp == NULL) return;

    if(m_toolBar == NULL) return;

    if(!bmp->IsOk()) return;

    //Logger::logf(LOG_MSG,_T("VemsToolBarToolBase::SetToolBmp ok"));

    if(active)
        m_toolBar->SetToolNormalBitmap(m_toolID, *bmp);
    else
        m_toolBar->SetToolDisabledBitmap(m_toolID, *bmp);

    m_AUImanager->Update();
    //m_handler->GetAuiMgr()->Update();
    // ^without this, the toolbar is flicking after changing button bitmap

}

/** \~english
  * Set the tool bitmap with \link #ToolBitmapIds
  *
  * \~hungarian
  * A tool bitamp állítható be \link #ToolBitmapIds \endlink segítségével
  *
  */
void VemsToolBarToolBase::SetToolBmp(int id, bool active)
{

    if(m_bmpHash.find(id) == m_bmpHash.end()) return;

    SetToolBmp(m_bmpHash[id],active);

    m_ToolBmpId = id;

}

/** \~english
  * It can initialize the help strings which can be set in the child class. \c m_longHelpString \c m_shortHelpString
  *
  * \~hungarian
  * A gyermek osztályból beállítható a tool-hoz tartozó felbukkanó szövegek \c m_longHelpString \c m_shortHelpString. Ezek inicializálására szolgál a függvény.
  *
  */
void VemsToolBarToolBase::InitHelpStrings()
{
    if(m_toolBar == NULL) return;

    m_toolBar->SetToolLongHelp(m_toolID,m_longHelpString);
    m_toolBar->SetToolShortHelp(m_toolID,m_shortHelpString);
}

/** \~english
  * This function loads the files which added in the child class constructor to the \link #m_bmpKeyListArray \endlink and stores in the \link #ToolBitmapIds \endlink-key based \link #m_bmpHash \endlink .
  *
  * \~hungarian
  * A függvény betölti a gyerekosztály konstruktorában a \link #m_bmpKeyListArray \endlink-hez hozzáadott fájlokat és elraktározza a \link #m_bmpHash \endlink-ben ahol \link #ToolBitmapIds \endlink kulcscsal érhető el.
  *
  */
void VemsToolBarToolBase::InitBmpRes()
{

    if(m_toolBar == NULL) return;

    for(IntArrStrHash::iterator it = m_bmpKeyListArray.begin(); it != m_bmpKeyListArray.end(); it++)   {
        if( it->second.GetCount() == 2) {
            if( m_bmpHash[it->first] != NULL ) delete m_bmpHash[it->first];
            m_bmpHash[it->first] = LoadAndResample( it->second[0] , m_toolBar->GetToolBitmapSize(), true, it->second[1] );
        }
        if( it->second.GetCount() == 1) {
            if( m_bmpHash[it->first] != NULL ) delete m_bmpHash[it->first];
            m_bmpHash[it->first] = LoadAndResample( it->second[0] , m_toolBar->GetToolBitmapSize());
        }
    }

}

/** \~english
  * Add request which are fill it int the \link #FrameLayoutHandler \endlink class.
  *
  * \~hungarian
  * Kérést lehet hozzáadni amit a \link #FrameLayoutHandler \endlink osztályon keresztül lehet teljesíteni.
  * \~ \n \ref tbtsender
  */
void VemsToolBarToolBase::AddRequest(ToolBarToolsNeedId tbnid)
{
    m_needObjectIds.Add(tbnid);
}

/** \~english
  * Add requestarray which are fill it int the \link #FrameLayoutHandler \endlink class.
  *
  * \~hungarian
  * Kéréseket lehet hozzáadni melyeket a \link #FrameLayoutHandler \endlink osztályon keresztül lehet teljesíteni.
  * \~ \n \ref tbtsender
  */
void VemsToolBarToolBase::AddRequest(const TbNeedArray &tbnidarray)
{

    WX_APPEND_ARRAY(m_needObjectIds, tbnidarray);

}


/** \~english
  * It loads outer file to wxBitmap class and resample on \c size parameter or if smaller the \c lowlimit parameter and \c islowlimit is true use \c alternatefile parameter.
  *
  * \~hungarian
  * Betölti a fájlt egy wxBitmap osztályba és átméretezi a \c size -ra vagy ha kisebb mint \c lowlimit paraméter és \c islowlimit igaz \c alternatefile -t használja.
  *
  */

wxBitmap* VemsToolBarToolBase::LoadAndResample(const wxString &filename, const wxSize &size, bool islowlimit, const wxString &alternatefile, uint16_t lowlimit)
{
    wxBitmap *bmp;

    if( islowlimit && size.x <= lowlimit ) {
        bmp = new wxBitmap();
        if( !bmp->LoadFile(alternatefile, wxBITMAP_TYPE_PNG) ) return NULL;
        return bmp;
    }

    wxImage image = wxImage();


    if(!image.LoadFile(filename, wxBITMAP_TYPE_PNG)) return NULL;

    if( size.x != image.GetWidth() )
        bmp = new wxBitmap(image.ResampleBicubic(size.x,size.y));
        //bmp = new wxBitmap(image.Rescale( size.x, size.y, wxIMAGE_QUALITY_HIGH ));
        //bmp = new wxBitmap(image.ShrinkBy ( image.GetWidth() / size.x , image.GetHeight() / size.y ));
    else
        bmp = new wxBitmap(image);

    if(!bmp->IsOk()) return NULL;

    return bmp;
}


/** \~english
  * \page tbtover VemsToolBarTool overview
  * The VemsToolBarTool class-system is capable to put an icon to the main frame toolbar. \n \n
  * So have to create a new class what is inherited from the VemsToolBarToolBase class. Creating a new icon:
  * \li In the new class constructor add paths for \link VemsToolBarToolBase::m_bmpKeyListArray m_bmpKeyListArray \endlink
  * with \link #ToolBitmapIds \endlink . There is an opportunity to can set more than one file for one ID, because the
  * wxWidgets image-resize algorithm is not so good when resizing images which is contains text. So can add a big (example 48x48)
  * and a smaller (ex 16x16) image. There is not load the pictures into the memory.
  * \li Implement the \link #VemsToolBarToolBase::Init Init \endlink function. You must call the \link #VemsToolBarToolBase::InitBmpRes InitBmpRes \endlink
  * in this function. In this place can set the tool active and inactive image. \link VemsToolBarToolBase::m_defaultBitmap m_defaultBitmap \endlink and
  * \link #VemsToolBarToolBase::m_inactiveBitmap m_inactiveBitmap \endlink. Now the pictures which added in the constructor is available, so can set the
  * active-inactive bmps with \link #VemsToolBarToolBase::m_bmpHash m_bmpHash \endlink with \link #ToolBitmapIds \endlink type key. \n
  * Add request to the object-sender systems. \ref tbtsender \n
  * Set the help strings. Initializing later.
  * \li Implement the neccesary event-handler functions.
  * \li Implement \link #VemsToolBarToolBase::SendObject SendObject \endlink function. \ref tbtsender
  *
  * \~hungarian
  * \page tbtover VemsToolBarTool körültekintés
  * A VemsToolBarTool osztályrendszer alkalmas egy ikont kirakni a fő ablak toolbarjára. \n \n
  * Ehhez létre kell hozni egy osztályt mely a VemsToolBarToolBase osztályból öröklődik. Az új ikon létrehozásának ajánlott menete: \n
  * \li Az új osztály konstruktorában ajánlott a \link VemsToolBarToolBase::m_bmpKeyListArray m_bmpKeyListArray \endlink asszociatív tömbhöz
  * \link #ToolBitmapIds \endlink kulcsokkal kell a kívánt képek elérési útját megadni. Lehetőség van egy azonosítóhoz több képet
  * megadni. Erre azért van szükség, mert a wxWidgets képátméretező algoritmusa nem ad teljesen olvasható képet ha szöveget kicsinyítünk.
  * Így megadható egy nagy (pl 48x48-as) és egy direkt kicsinek tervezett képet (pl 16x16). A képek betöltése a memóriába nem itt történik.
  * \li Az \link #VemsToolBarToolBase::Init Init \endlink függvény megvalósítása. Ebben a függvéynben érdemes az \link #VemsToolBarToolBase::InitBmpRes InitBmpRes \endlink
  * függvényt meghívni ami a konstruktorban megadott fájlneveket tölti be. Ezután megadható a tool aktív és inaktiv képe. Ezt a
  * \link VemsToolBarToolBase::m_defaultBitmap m_defaultBitmap \endlink és \link #VemsToolBarToolBase::m_inactiveBitmap m_inactiveBitmap \endlink
  * megadásával lehet. Az \link #VemsToolBarToolBase::m_bmpHash m_bmpHash \endlink-ből \link #ToolBitmapIds \endlink kulcsokkal már elérhetőek a
  * konstruktorban megadott képek. \n
  * Az objektum küldő rendszerhez itt érdemes hozzáadni a kéréseket, melyeket a FrameLayoutHandler osztályon keresztül lehet teljesíteni. \ref tbtsender \n
  * A toolhoz tartozó help-stringeket is itt kell megadni. Az inicializálás később történik.
  * \li A szükséges eseménykezelők megvalósítása.
  * \li A \link #VemsToolBarToolBase::SendObject SendObject \endlink megvalósítása. \ref tbtsender
  *
  */

/** \~english
  * \page tbtsender SendObject overview
  * If added a request with \link VemsToolBarToolBase::AddRequest AddRequest \endlink function then can fill with FrameLayoutHandler class.
  * This class is created in the MainFrame class. Create a new-type-request and usage:
  * \li Add new key to \link #ToolBarToolsNeedId \endlink
  * \li Handle the "event" when arriving a request in the neccessary \link #VemsToolBarToolBase::SendObject SendObject \endlink function implementations.
  * \li Fill the request with the FrameLayoutHandler. Use \link #FrameLayoutHandler::SendObject \endlink function.
  * \~hungarian
  * \page tbtsender SendObject körültekintés
  * Ha lett kérés hozzáadva az \link VemsToolBarToolBase::AddRequest AddRequest \endlink fügvénnyel, azokat a FrameLayoutHandler osztályon keresztül
  * lehet teljesíteni. Ez az osztály jelenleg a MainFrame osztályban van létrhozva. Új tipusú kérés létrehozása és használata:
  * \li Új kulcs hozzáadása \link #ToolBarToolsNeedId \endlink-hoz.
  * \li A kérés megérkezésének lekezelése a \link #VemsToolBarToolBase::SendObject SendObject \endlink függvény adott implementációiban. pl: TbTbPlay::SendObject
  * \li A kérés teljesítése, tehát objektum dobása a FrameLayoutHandler osztályon keresztül, a \link #FrameLayoutHandler::SendObject \endlink függvénnyel.
  */
