#include "TriggerLog.h"
#include <wx/file.h>

#include "../logger.h"
#include "../Math.h"
#include "Unescaper.h"
/*

If there is a sec trigger than:
N*prim trigger(2byte) TS, Sec trig TS(2 byte), CRC(2byte), 0x7e,

Escaping:
    0x7e -> 0x7d 0x5e
    0x7d -> 0x7d 0x5d

*/

/// -----------------------------------------------------

TriggerLog::TriggerLog()
{
    clearData();
}

TriggerLog::~TriggerLog()
{
    //dtor
}

void TriggerLog::clearData()
{
    m_prevvalue =  0;
    m_threshold = 1.5;

    m_prevdiff = 0;
    m_rawTriggerData.Clear();

    m_rawTriggerFrameA.Clear();

    m_timelines.clear();

    /// DEPRECATED: only prim&sectrig
    /*
    m_primaryTrTime.Clear();
    m_secondraryTrTime.Clear();
    */
    ///-------------------

    m_resultH.clear();
}

void TriggerLog::setFirmwareInfo( FirmwareInfo& fwInfo )
{
    m_fwinfo = fwInfo;

    m_fwVersion = m_fwinfo.getFirmwareVersionAsString();
    m_serialNum = m_fwinfo.getSerialNumber();
}

void TriggerLog::getHeaderInfo( wxString& fwVer, wxString& createTime, wxString& serialnum )
{
    fwVer = m_fwVersion;
    createTime = m_creationTime;
    serialnum = m_serialNum;
}

int TriggerLog::unescape( const IntArray& src, IntArray& out )
{
    bool res = true;
    int escCount = 0;

    Unescaper escaper;

    for ( size_t i = 0; i < src.GetCount(); ++i )
    {
        int v = src.Item( i );

        uint8_t escapedData;

        if ( escaper.parse( v, escapedData ) )
        {
            out.Add( escapedData );
        }

        escCount += escaper.m_escCount;
    }

    if ( res )
    {
        return escCount;
    }
    else
    {
        return -1;
    }
}

/// src: raw triggerlog with sec trigger frames
/// idA: index of frames in raw triggerlog
///idx: requested frame idx
///
IntArray TriggerLog::getFrame( const IntArray& src, const IntArray& idxA, size_t idx )
{
    size_t endIdx = idxA.Item( idx );

    size_t startIdx = 0;

    if ( idx > 0 )
    {
        startIdx = idxA.Item( idx - 1 ) + 1;

        if( startIdx >= src.GetCount() )
        {
            if( src.GetCount() > 0 )
            {
                startIdx = src.GetCount() - 1;
            }
            else
            {
                startIdx = 0;
            }
        }
    }

    IntArray frame;

    for ( size_t i = startIdx; i < endIdx; ++i )
    {
        frame.Add(src.Item(i));
    }
    return frame;
}

IntArray TriggerLog::getFramesIdx( const IntArray& src )
{
    IntArray frameIdx;

    for ( size_t i = 0; i < src.GetCount(); ++i )
    {
        int v = src[i];

        if ( v == 0x7e )
        {
            frameIdx.Add( i );
        }
    }

    return frameIdx;
}

void TriggerLog::fillTRDataArray( const IntArray& src, TriggerFrameArray& dest )
{
    IntArray out;

    int res = unescape( src, out );

    if( res == 0 )
    {
        // ESC sequence not found, it can be any type
        Logger::logf(LOG_MSG, _T("TriggerLog::fillTRDataArray() unescape UNKNOWN!") );
    }
    else if ( res > 0 )
    {
        // maybe NEW type!
        //Logger::logf(LOG_MSG, _T("TriggerLog::fillTRDataArray() unescape OK!: %i"),res );
    }
    else
    {
        // OLD type!
        Logger::logf(LOG_MSG, _T("TriggerLog::fillTRDataArray() unescape ERROR!") );
        return;
    }

    IntArray framesIdx = getFramesIdx( m_rawTriggerData );

    if ( framesIdx.GetCount() == 0 )
    {
        if( res != -1 )
        {
            // new type but frame is not found
            Logger::logf(LOG_MSG, _T("TriggerLog::fillTRDataArray() unescape OK!: %i"),res );
        }
        else
        {
            // OLD log!
            Logger::logf(LOG_MSG, _T("TriggerLog::fillTRDataArray() unescape ERROR!") );
        }
        return;
    }

    for ( size_t i = 0; i < framesIdx.GetCount(); ++i )
    {
        IntArray frame = getFrame( m_rawTriggerData, framesIdx, i );

        //Logger::logf(LOG_MSG, _T("TriggerLog::fillTriggerData() frameIdx:%i size:%i"),i, frame.GetCount() );

        IntArray out;

        unescape( frame, out );

        TriggerFrame* trFrame = new TriggerFrame();

        int res = trFrame->addFrame( out );

        // empty frames do not count
        if ( res == TriggerFrame::TRFRAME_EMPTY_FRAME )
        {
            continue;
        }

        if ( m_resultH.find( res ) == m_resultH.end() )
        {
            m_resultH[res] = 1;
        }
        else
        {
            m_resultH[res]++;
        }

        if ( res < 256 )
        {
            dest.Add( trFrame );
        }

    }

     Logger::logf(LOG_MSG, _T("TriggerLog::fillTriggerData() END"));
}

void TriggerLog::_fillTimeLine( const TriggerFrameArray& src )
{
    // multiple of 65536. Add this to all timestamps
    int clockbase = 0;

    int prevTime = 0;

    for ( size_t i =0; i < src.GetCount(); ++i )
    {
        TriggerFrame* trFrame = src.Item( i );

        // Skip bad-CRC frames, but omit CRC-check for the first frame, because it is
        // likely bad anyway, especially if primtrig-only sequence which is a big frame without CRC
        if ( (i != 0) && !trFrame->checkCRC() )
        {
            continue;
        }

        trFrame->convertFrame2Time( m_timelines, clockbase, prevTime );
        // Logger::logf( LOG_MSG, _T("ftline=%d"), m_timelines[TriggerFrame::CRC_ONLY].GetCount() );
    }
}

void TriggerLog::fillTriggerData()
{
    fillTRDataArray(m_rawTriggerData, m_rawTriggerFrameA);

    _fillTimeLine(m_rawTriggerFrameA);
}

DoubleArray& TriggerLog::getNthTimeLine( unsigned int n )
{
    Int2DoubleArrayHash::iterator it = m_timelines.find( n );
    if ( it != m_timelines.end() )
    {
        return it->second;
    }
    else
    {
        return *new DoubleArray();
    }
}

/*--------------------------*/
/* NOT NEEDED               */
/*--------------------------*/
/*void TriggerLog::hexLog( const IntArray& src, int col )
{
    wxString line;
    int reminder = 0;

    size_t i = 0;

    for ( ; i < src.GetCount(); ++i )
    {
        wxString hex = wxString::Format(_T(" 0x%02X"),src[i] );

        line += hex;

        if ( reminder == col - 1 )
        {
            Logger::logf(LOG_MSG,_T("idx:(%i) ")+line, i);
            line.Empty();
            reminder = 0;
        }
        else
        {
            ++reminder;
        }
    }

    if ( reminder )
    {
        Logger::logf(LOG_MSG,_T("idx:(%i) ")+line, i);
    }
}*/

/// DEPRECATED CODE - only prim&sectrig
/*
void TriggerLog::fillTimeLine(const TriggerFrameArray& src)
{
    /// TODO cut this
    int clockbase = 0; // multiple of 65536. Add this to all timestamps
    int prevTime = 0;
    for(size_t i =0; i < src.GetCount(); ++i){
        TriggerFrame* trFrame = src.Item(i);
        // Skip bad-CRC frames, but omit CRC-check for the first frame, because it is
        // likely bad anyway, especially if primtrig-only sequence which is a big frame without CRC
        if( (i != 0) && !trFrame->checkCRC() )continue;
        //Logger::logf(LOG_MSG, _T("TriggerLog::fillTRTime() CRC OK idx: %i"), i);
        LOG
        trFrame->convertFrame2Time(m_primaryTrTime, m_secondraryTrTime, clockbase, prevTime);
        Logger::logf( LOG_MSG, _T("ftline=%d"), m_primaryTrTime.GetCount() );
    }
}
*/
