#include "TriggerLogManager.h"

#include <wx/datetime.h>

#include "../DirHelper.h"
#include "../RawLogHelper.h"
#include "../logger.h"

#include "TriggerLogV3.h"
#include "TriggerLogImg.h"
#include "TriggerLogTail.h"
#include "TriggerLogController.h"

#include "../vemsprotocol/VemsDeviceControllerBase.h"

#include "../UserLog.h"
#include "TriggerLogGraph.h"
#include "TriggerLogSelector.h"
#include "../vemsprotocol/ECUInfo.h"

BEGIN_EVENT_TABLE( TriggerLogManager, wxEvtHandler )

    EVT_BUTTON( wxID_ANY, TriggerLogManager::OnStartStop )

END_EVENT_TABLE()

TriggerLogManager::TriggerLogManager( VemsDeviceControllerBase * vdc, const wxString& fileName, wxEvtHandler* evtHandler, int tltype )

{
    // initialization
    m_vdc = vdc;
    m_tlcontroller = NULL;

    m_wasECUOnline = false;

    if ( vdc )
        m_wasECUOnline = true;

    m_resultFrame = NULL;

    // log from V3 (online)
    if ( fileName.IsEmpty() )
    {
        m_type = 0;

        // get the fwinfo
        FirmwareInfo finfo = vdc->getFirmwareInfo();

        // create the triggerlog
        m_triggerLog = new TriggerLogV3( evtHandler );
        m_evtHandler = evtHandler;

        m_triggerLog->setFirmwareInfo( finfo );

        m_iscontinuous = true;

        // stop thread
        m_vdc->stop( VDC_FLAG_NOEVENT );
    }

    // load from file
    else
    {
        m_type = 1;

        m_fileName = fileName;

        m_iscontinuous = false;

        if ( tltype == TriggerLogSelector::TLOG_FILE )
        {
            m_triggerLog = new TriggerLogImg( fileName );
            }
        // not supported/used
        /*else
        {
            LOG
            m_triggerLog = new TriggerLogTail( fileName, evtHandler );

            //showResults();
        }
        */
    }
}

TriggerLogManager::~TriggerLogManager()
{
    delete m_triggerLog;

    if ( m_tlcontroller )
    {
        m_tlcontroller->stop();
        m_tlcontroller->Disconnect( wxEVT_RAW_TRIGGER_LOG,RawTriggerLogEventHandler( TriggerLogManager::OnRawTriggerLogEvent ), NULL, this );
        delete m_tlcontroller;
    }

    // ECU was online
	if ( m_vdc ) {
		m_vdc->resume();
	}
    //EndSession();
    //LOGP( _T("Destructor was called..."))
}

void TriggerLogManager::handleButtons( int type )
{
    if ( type == 1 ) // start
    {
        m_triggerLog->clearData();

        if ( m_tlcontroller )
        {
            m_tlcontroller->stop();

            m_tlcontroller->Disconnect( wxEVT_RAW_TRIGGER_LOG,RawTriggerLogEventHandler( TriggerLogManager::OnRawTriggerLogEvent ), NULL, this );

            delete m_tlcontroller;
        }

        m_tlcontroller = new TriggerLogController( m_vdc );

        m_tlcontroller->start();

        m_tlcontroller->Connect( wxEVT_RAW_TRIGGER_LOG,RawTriggerLogEventHandler( TriggerLogManager::OnRawTriggerLogEvent ), NULL, this );

        // if a prev. logfile is opened -> close it
        if ( m_triggerlogfile.IsOpened() )
            m_triggerlogfile.Close();

        // then create a new logfile
        wxDateTime dt = wxDateTime::Now();
        wxString fname = m_vdc->getFirmwareInfo().getSerialNumber() + _T("-") + dt.FormatISODate() + _T("-") + dt.FormatISOTime() + _T(".triggerlog");
        fname.Replace( _T(":"), _T(".") );
        m_triggerLog->m_creationTime = dt.FormatISODate() + _T(" ") + dt.FormatISOTime();
        m_triggerlogfile.Create( DirHelper::GetPath( dLOG ) + fname );

        // set CSV filename
        m_fileName = fname;

        // write VT header to the beginning of the logfile
        writeToLogFile( NULL, true );
    }

    if ( type == 2 )
    {
        if ( m_tlcontroller )
        {
            m_tlcontroller->stop();
            m_tlcontroller->Disconnect( wxEVT_RAW_TRIGGER_LOG,RawTriggerLogEventHandler( TriggerLogManager::OnRawTriggerLogEvent ), NULL, this );
        }

        // close the opened logfile
        if ( m_triggerlogfile.IsOpened() )
            m_triggerlogfile.Close();

    }

    if ( type == 3 )
    {
        showResults();
    }

    if ( type == 4 ) // graph closed
    {
        wxWindow* resultFrame = wxFindWindowByName( _T("Trigger log results"), m_winparent );

        if ( resultFrame != NULL )
            resultFrame->Destroy();

        if ( m_tlcontroller )
        {
            m_tlcontroller->stop();
            m_tlcontroller->Disconnect( wxEVT_RAW_TRIGGER_LOG,RawTriggerLogEventHandler( TriggerLogManager::OnRawTriggerLogEvent ), NULL, this );
        }

        m_unescapedFrame.Clear();
    }
}


// TriggerLogGraph: start / stop v3 button handler
void TriggerLogManager::OnStartStop( wxCommandEvent& event )
{
    event.Skip();
}

void TriggerLogManager::OnRawTriggerLogEvent( RawTriggerLogEvent& event )
{
    wxMemoryBuffer& rawData = event.GetDataBuffer();

    size_t len = rawData.GetDataLen();

    uint8_t* ptr = (uint8_t*)rawData.GetData();

    writeToLogFile( &rawData );

    for( size_t i = 0; i < len; i++, ptr++)
    {
        m_triggerLog->m_rawTriggerData.Add( *ptr );

        if( *ptr == 0x7e )
        {
            TriggerLogEvent trEvent;

            int res = trEvent.m_frame.addFrame( m_unescapedFrame );

            // update statistics
            if ( res != TriggerFrame::TRFRAME_EMPTY_FRAME )
            {
                if( m_triggerLog->m_resultH.find( res ) == m_triggerLog->m_resultH.end()  )
                {
                    m_triggerLog->m_resultH[res] = 1;
                }
                else
                {
                    m_triggerLog->m_resultH[res] ++;
                }
            }

            // send only valid frame
            if( res < 256 )
            {
                // Send Event
                if ( m_evtHandler )
                    m_evtHandler->AddPendingEvent(trEvent);
            }
            else
            {
//                LOG
            }

            m_unescapedFrame.Clear();
            m_unescaper.init();
        }
        else
        {
            uint8_t data;

            if( m_unescaper.parse( *ptr, data ) )
            {
                m_unescapedFrame.Add(data);
            }
        }
    }
}

void TriggerLogManager::writeToLogFile( wxMemoryBuffer* buffer, bool start )
{
    if ( start )
    {
        char header[512];
        memset( header, 0, 512 );

        wxString hdr = RawLogHelper::getVemslogHeader( m_vdc->getFirmwareInfo(), 2, true );

        for ( size_t i = 0; i < hdr.Length(); ++i )
            header[i] = hdr[i];

        m_triggerlogfile.Write( header, 512 );
    }
    if ( buffer != NULL )
    {
        size_t len = buffer->GetDataLen();

        if ( len > 0 )
        {
            unsigned char* buf = new unsigned char[len];
            memcpy( buf, (unsigned char*) buffer->GetData(), len );

            m_triggerlogfile.Write( buf, len );

            delete[] buf;
        }
    }
}
/*
bool TriggerLogManager::StartSession()
{
    if( m_vemsTuneController->getDevice() != ECUInfo::DEVICE_V3)
    {
        return false;
    }

    m_vemsTuneController->setECUOnline( false, true );

//    m_vemsTuneController->getEcuStateMgr()->setState( ECUInfo::MAN );

    return true;
}

bool TriggerLogManager::EndSession()
{
    // restart ECU if it was online before
    if ( !m_wasECUOnline || m_type != 0 )
        return false;

    m_commBase->clear();
    m_vemsTuneController->setECUOnline( true, false );

    return true;
}
*/
void TriggerLogManager::showResults()
{
    wxWindow* win = wxFindWindowByName( _T("Trigger log results"), m_winparent );
    if ( win != NULL )
    {
        win->Show();
        win->SetFocus();
    }
    else
    {
        m_resultFrame = new TriggerLogResults( m_winparent, m_triggerLog );
        m_resultFrame->setTriggerLogFileName( m_fileName );
        m_resultFrame->Show();
        m_resultFrame->SetFocus();
    }
}
