#include "TriggerLogResults.h"
#include "TriggerLogHeader.h"
#include "../DirHelper.h"
#include "../common/typedef.h"
#include "../common/types/IntArray.h"
#include <wx/textfile.h>
#include <wx/filename.h>

#include "../logger.h"
#include "TriggerLogGraph.h"

int TLOG_RES_SAVE2CSV = wxNewId();

BEGIN_EVENT_TABLE( TriggerLogResults, wxFrame )

    EVT_BUTTON( TLOG_RES_SAVE2CSV, TriggerLogResults::OnSaveCSV )

END_EVENT_TABLE()

TriggerLogResults::TriggerLogResults( wxWindow* parent, TriggerLog* triggerLog )
{
    m_parent = parent;

    m_triggerLog = triggerLog;

    m_fileName = _T("");
    buildGUI();
}

void TriggerLogResults::buildGUI()
{
    Logger::logf(LOG_MSG, _T("TriggerLogResults::buildGUI() START") );

    Create( m_parent, -1, _T("Trigger log results") , wxDefaultPosition, wxDefaultSize );

    SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );

    ClearBackground();

    wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );

    wxScrolledWindow* panel = new wxScrolledWindow( this, -1 );

    panel->SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_3DFACE ) );

    panel->ClearBackground();

    // add scrollbars
    panel->SetScrollRate( 5, 10 );

    wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);

    /// create & fill the grid
    m_grid = new wxGrid( panel, -1, wxDefaultPosition, wxSize(480,180) );

    DoubleArray& primaryTrTime = m_triggerLog->getNthTimeLine(TriggerFrame::CRC_ONLY);
    DoubleArray& secondraryTrTime = m_triggerLog->getNthTimeLine(TriggerFrame::SEC_TRIG);


    size_t primSize = primaryTrTime.GetCount();
    size_t secSize = secondraryTrTime.GetCount();
    size_t sumCount = primSize + secSize;

    Logger::logf(LOG_MSG, _T("TriggerLogResults::addTimeLineGrid(): prCount: %i"), primSize );
    Logger::logf(LOG_MSG, _T("TriggerLogResults::addTimeLineGrid(): secCount: %i"), secSize );

    m_grid->CreateGrid( sumCount, 6 );
    wxArrayString colnames;
    colnames.Add( _T("Primary Time (ms)") );
    colnames.Add( _T("Primary Diff (ms)") );
    colnames.Add( _T("Primary Diff Ratio") );
    colnames.Add( _T("Secondary Time (ms)") );
    colnames.Add( _T("Secondary Diff (ms)") );
    colnames.Add( _T("Comment") );

    for ( uint8_t i = 0; i < colnames.GetCount(); ++i )
        m_grid->SetColLabelValue( i, colnames[i] );

    sizer->Add( m_grid, 1, wxEXPAND | wxALL, 5 );

    m_grid->SetDefaultCellAlignment( wxALIGN_RIGHT, wxALIGN_CENTRE );
    m_grid->AutoSizeColumns( true );

    size_t primIdx = 0;
    size_t secIdx = 0;
    size_t idx = 0;

    double threshold = 1.5;
    double currPrim = 0;
    double currSec = 0;

    double prev = 0;
    double prevdiff = 0;
    size_t missingidx = 0; // idx of the last missing tooth

    while ( idx < sumCount )
    {
        if ( primIdx < primSize )
        {
            currPrim = primaryTrTime.Item( primIdx );
        }
        if( secIdx < secSize )
        {
            currSec = secondraryTrTime.Item( secIdx );
        }

        if( ( ( secSize == 0 ) && ( primSize != 0 ) ) || ( ( currPrim < currSec ) && ( primIdx < primSize ) ) )
        {
            m_grid->SetCellValue( idx, 0, wxString::Format( _T("%.3f, (%i)"), currPrim, primIdx ) );
            double diff = currPrim - prev;
            m_grid->SetCellValue( idx, 1, wxString::Format( _T("%.3f"), diff ) );

            double ratio = ( prevdiff == 0 ) ? 1.0 : ( (double)diff / (double) prevdiff );
            m_grid->SetCellValue( idx, 2, wxString::Format( _T("%.2f"), ratio ) );

            if ( ratio > threshold )
            {
                int t = (int)floor( ratio + 0.5 ) - 1;

                m_grid->SetCellValue( idx, 5, wxString::Format( _T("MISSING TOOTH: %d, cycle=%d"),
                    t, idx - missingidx ) );

                m_grid->SetCellFont( idx, 5, m_grid->GetLabelFont() );

                missingidx = idx;
            }

            primIdx++;

            prev = currPrim;

            prevdiff = diff;
            //Logger::logf(LOG_MSG, _T("TriggerLogResults::addTimeLineGrid(): pidx: %i, v: %f"),primIdx, currPrim );
        }
        else if( ( (secSize != 0 ) && (primSize == 0)) || (secIdx < secSize ) )
        {
            m_grid->SetCellValue( idx, 3, wxString::Format( _T("%.3f, (%i)"), currSec, secIdx ) );

            double diff = currSec - prev;
            m_grid->SetCellValue( idx, 4, wxString::Format( _T("%.3f"), diff ) );

            secIdx++;
            prev = currSec;
            //Logger::logf(LOG_MSG, _T("TriggerLogResults::addTimeLineGrid(): sidx: %i, v: %f"),secIdx, currSec );
        }
        else
        {
            //Logger::logf(LOG_MSG, _T("TriggerLogResults::addTimeLineGrid(): pidx: %i, secidx: %i"),primIdx,  secIdx );
        }

        idx++;
    }

    m_grid->AutoSizeColumn( 2 );
    m_grid->AutoSizeColumn( 5 );

    panel->SetSizer( sizer );
    sizer->Fit( panel );
    sizer->SetSizeHints( panel );

    wxBoxSizer* bsizer = new wxBoxSizer( wxHORIZONTAL );
    wxButton* saveButton = new wxButton( this, TLOG_RES_SAVE2CSV, wxT("Save table to CSV") );

    int cId = wxNewId();
    wxButton* closeButton = new wxButton( this, cId, _T("Close") );
    Connect( cId, wxEVT_COMMAND_BUTTON_CLICKED, wxCommandEventHandler( TriggerLogResults::OnClose ) );

    bsizer->Add( saveButton, 0, wxALL, 3 );
    bsizer->Add( closeButton, 0, wxALL, 3 );

    mainsizer->Add( bsizer, 0, wxALL, 5 );
    mainsizer->Add( panel, 1, wxEXPAND | wxALL, 5 );

    SetSizer( mainsizer );
    mainsizer->Fit( this );
    mainsizer->SetSizeHints( this );
}


///--------------------------------------
/// TODO: for more than 2 trigger types
///--------------------------------------

void TriggerLogResults::OnSaveCSV( wxCommandEvent& event )
{
    if ( event.GetId() != TLOG_RES_SAVE2CSV )
    {
        event.Skip();
        return;
    }

    wxFileName fn( m_fileName );
    wxString defName = fn.GetName() + _T(".csv");

    wxFileDialog dlg( this, _T("Save file"), DirHelper::GetPath( dLOG ), defName, _T("Comma separated values (*.csv)|*.csv"),
        wxFD_SAVE | wxFD_OVERWRITE_PROMPT );

    if ( dlg.ShowModal() == wxID_OK )
    {
        wxString fName = dlg.GetPath();
        wxTextFile file( fName );

        // it file exists -> clear its content
        if ( file.Exists() )
        {
            file.Open();
            file.Clear();
        }
        // or create it
        else
        {
            file.Create();
        }

        // write comment
        file.AddLine( _T("; This is a csv-formatted version of a vems trigger log file") );

        wxString fwver, sernum, crtime;
        m_triggerLog->getHeaderInfo( fwver, crtime, sernum );
        Logger::logf(LOG_MSG, fwver + crtime + sernum);

        wxString hdrinfo = _T("; Created on: ");
        if ( crtime.IsEmpty() ) hdrinfo += _T("(unknown)");
        else hdrinfo += crtime;

        hdrinfo += _T(" - Serial number: ");
        if ( sernum.IsEmpty() ) hdrinfo += _T("(unknown)");
        else hdrinfo += sernum;

        hdrinfo += _T(" - Firmware version: ");
        if ( fwver.IsEmpty() ) hdrinfo += _T("(unknown)");
        else hdrinfo += fwver;

        file.AddLine( hdrinfo );

        /// TODO this

        file.AddLine( _T("id,Primary Time (ms),Primary Diff (ms),Primary Diff Ratio,Secondary Time (ms),Secondary Diff (ms),Comment") );

        for ( int i = 0; i < m_grid->GetNumberRows(); ++i )
        {
            wxString str = wxString::Format( _T("%d,"), i );
            for ( int j = 0; j < m_grid->GetNumberCols(); ++j )
            {
                wxString cellval = m_grid->GetCellValue( i, j );
                if ( cellval.Find(_T(",")) != wxNOT_FOUND )
                    cellval = _T("\"") + cellval + _T("\"");
                str += cellval;
                str += _T(",");
            }
            str = str.RemoveLast();

            file.AddLine( str );
        }
        file.Write();

        file.Close();

        wxMessageBox( _T("CSV saved"), _T(""), wxOK | wxICON_INFORMATION );
    }
}

void TriggerLogResults::OnClose( wxCommandEvent& event )
{
    Destroy();
}
