#include "ConfigletPanel.h"


#include <wx/dir.h>
#include <wx/filename.h>


#include "../stdatamanager/STDataModelBase.h"
#include "../stdatamanager/STDataMgrBase.h"
#include "../SymbolMgr.h"
#include "../firmwareconfig/FirmwareConfigData.h"
#include "../fileparser/FirmwareConfigParser.h"
#include "../logger.h"
#include "../FirmwareInfo.h"
#include "../common/VemsContext.h"


struct ChangedSymbol
{
    Symbol symbol;
    size_t row;
    size_t col;

    ChangedSymbol( const Symbol& s , size_t r = 0, size_t c = 0 )
    {
        symbol  = s;
        row     = r;
        col     = c;
    }
};


int idConfigletComboBox = wxNewId();


BEGIN_EVENT_TABLE( ConfigletPanel, wxPanel )
    EVT_COMBOBOX( idConfigletComboBox, ConfigletPanel::OnSelect )
END_EVENT_TABLE()


ConfigletPanel::ConfigletPanel( wxWindow* parent, const wxString& dirName, SymbolMgr* symbolmgr)
    : wxPanel( parent )
{
    m_symbolMgr = symbolmgr;
    init( dirName );

}


ConfigletPanel::~ConfigletPanel()
{
    unregisterViewer();

    /*
    if ( m_bConfigletSelected && !m_pSTCtrl->isWrited() )
    {
        m_pSTCtrl->requestReadAllDirty();
    }
    */
}


void ConfigletPanel::init( const wxString& dirName )
{
    m_dirName = dirName;
    wxString title = _T("Defaults");

    //create sizer
    wxStaticBox* pStaticBox =  new wxStaticBox( this, wxID_ANY, title );
    wxFont font( pStaticBox->GetFont() );
    font.SetWeight( wxFONTWEIGHT_BOLD );
    pStaticBox->SetOwnFont( font );
    wxStaticBoxSizer* pMainSizer = new wxStaticBoxSizer( pStaticBox, wxHORIZONTAL );
    SetSizer( pMainSizer );

    SymbolTable& symbols = m_symbolMgr->getCfgSymbolTable();

    if (  symbols.empty() )
    {
        wxObject* obj = VemsContext::getObject(_T("FirmwareInfo"));

        wxString fwVer;

        if(obj){
            fwVer = ((FirmwareInfo*) obj)->getFirmwareVersionAsString();
        }

        //wxString msg = _T("Descriptor file has been not found for current firmware\n");
        wxString msg = _T("Firmware ") + fwVer + _T(" not installed!\n");
        msg += _T("Firmware can be downloaded with Tools/Firmware/Firmware Web Tool.");

        wxStaticText* text = new wxStaticText( this, wxID_ANY, msg );
        text->SetForegroundColour( *wxRED );
        GetSizer()->Add( text, 1, wxEXPAND | wxALL, 5 );
        return;
    }

    //create file list
    wxDir dir( m_dirName );
    wxString filename;
    wxArrayString files;
    files.Add( _T("default") );

    if ( dir.IsOpened() )
    {

        bool cont = dir.GetFirst( &filename, _T("*.vemscfg"), wxDIR_FILES );
        while ( cont )
        {
            add(filename, files);
            cont = dir.GetNext( &filename );
        }

        cont = dir.GetFirst( &filename, _T("*.txt"), wxDIR_FILES );
        while ( cont )
        {
            add(filename, files);
            cont = dir.GetNext( &filename );
        }

    }

    //else?

    wxComboBox* combo = new wxComboBox( this,
                                        idConfigletComboBox,
                                        _T("default"),
                                        wxDefaultPosition,
                                        wxDefaultSize,
                                        files,
                                        wxCB_READONLY );

    GetSizer()->Add( combo, 1, wxEXPAND | wxALL, 5 );
}

void ConfigletPanel::OnSelect( wxCommandEvent& event )
{
    refreshChangedSymbolsFromDeviceModel();
    clearChangedSymbols();

    wxString comboValue = ( (wxComboBox*)event.GetEventObject() )->GetValue();
    if ( comboValue.IsSameAs( _T("default") ) )
    {
        m_stDataMgr->updateRegisteredViewers();
        return;
    }

    /// parse selected configlet
    FirmwareConfigData      fcData;
    FirmwareConfigParser    configParser( &fcData );
    wxString strFileName = m_dirName +  m_configName2FileName[comboValue];

    configParser.parse( strFileName );
    Logger::logf( LOG_MSG, fcData.scalarDataToString() );


    /// load configlet data to model
    const SymbolTable& symbols = m_symbolMgr->getCfgSymbolTable();
    const String2IntHash& scalars =  fcData.getScalarValuesH();

    for( String2IntHash::const_iterator it = scalars.begin(); it != scalars.end(); ++it )
    {
        //Logger::logf( LOG_MSG, _T("ConfigletPanel::OnSelect()" + it->first) );
        SymbolTable::const_iterator it_symbol = symbols.find( it->first );

        if( it_symbol == symbols.end() )
        {
            Logger::logf( LOG_MSG, _T("Symbol ") + it->first + _T(" not found in global.h!!!") );
            continue;
        }

        uint8_t value   = (uint8_t)it->second;

        ChangedSymbol* changed = new ChangedSymbol( *(it_symbol->second) );

        m_stDataMgr->setRawModelData( *(it_symbol->second), 0, 0, value );
        m_changedSymbols.Add( changed  );
    }


    /// load h-table rows
    for(size_t i = 0; i < 6; i++){
        wxString name = wxString::Format(_T("h%i"), i);

        loadHTable(fcData, name);
    }
    m_stDataMgr->updateRegisteredViewers();
}

void ConfigletPanel::loadHTable(const FirmwareConfigData& fcData, const wxString name)
{
    VemsMatrix<uint8_t>* pTable = fcData.getTable( name );
    if ( pTable )
    {
        const SymbolTable& symbols = m_symbolMgr->getCfgSymbolTable();
        const SymbolTable& tables = m_symbolMgr->getConfigTableSymbols();
        SymbolTable::const_iterator it_symbol = tables.find( name );

        if ( it_symbol != symbols.end() )
        {
            size_t size    = it_symbol->second->getArraySize();

            for ( size_t i = 0; i < size; ++i )
            {
                if ( !pTable->dirty( i ) ) { continue; }

                uint8_t value = pTable->get( i );

                m_stDataMgr->setRawModelData( *(it_symbol->second), 0, i, value );
                m_changedSymbols.Add( new ChangedSymbol( *(it_symbol->second), 0, i ) );
            }
        }
    }
}

void ConfigletPanel::requestWrite()
{

    Logger::logf( LOG_MSG, _T("ConfigletPanel::requestWriteData()") );

    for ( size_t i = 0; i < m_changedSymbols.GetCount(); ++i )
    {
        ChangedSymbol* changed = m_changedSymbols[i];

        m_stDataMgr->requestWrite( changed->symbol, changed->row, changed->col );
    }

}


void ConfigletPanel::refreshChangedSymbolsFromDeviceModel()
{
    for ( size_t i = 0; i < m_changedSymbols.GetCount(); ++i )
    {
        ChangedSymbol* changed = m_changedSymbols[i];

        m_stDataMgr->updateRawFromDeviceModel( changed->symbol, changed->row, changed->col );
    }
}

void ConfigletPanel::clearChangedSymbols()
{
    WX_CLEAR_ARRAY( m_changedSymbols );
}

wxString ConfigletPanel::toString()
{
    return _T("ConfigletPanel");
}

void ConfigletPanel::add(const wxString& fileName, wxArrayString& files)
{
    wxFileName fn(fileName);
    wxString confName = fn.GetName();
    m_configName2FileName[confName] = fileName;
    files.Add( confName );
}
