#ifndef USERDIALOGTEMPLATE_H
#define USERDIALOGTEMPLATE_H

#include <wx/wx.h>
#include <wx/gbsizer.h>

#include "StaticViewersController.h"
#include "../Symbol.h"
#include "../stdatamanager/STDataViewerBase.h"
#include "../firmwareconfig/ConfigValidatorViewerBase.h"

class UserDialog;
class UserDialogManager;
class STDataModelBase;
class ConfigValidatorBase;
class FirmwareInfo;

/**********************************************************
*   StaticLabel
**********************************************************/
class StaticLabel : public wxWindow
{
public:
	StaticLabel(wxWindow *parent, const wxString& label, bool vertical);
	virtual ~StaticLabel();

private:

	bool m_isVertical;
    void OnPaint(wxPaintEvent&);

    DECLARE_EVENT_TABLE()
};


class UserDialogTemplate;

/**********************************************************
*   STViewerTextCtrl
**********************************************************/
class STViewerTextCtrl : public wxTextCtrl, public STDataViewerBase
{
    public:
        STViewerTextCtrl( wxWindow* parent, wxWindowID id, const wxString& value = _T(""),
                          const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize,
                          long style = 0, const wxValidator& validator = wxDefaultValidator, const wxString& name = wxTextCtrlNameStr);

        void update();
        void requestRead();
        void updateModel();
        void requestWrite();

        void setFormat( int digits_in );

        void setSymbol( const Symbol& symbol_in ) { m_symbol = symbol_in; }
        void setExpression( const wxString& expr ) { m_expr = expr; }
        wxString toString();
        SymbolArray getSymbols();

    private:
        UserDialogTemplate* m_parent;

        wxString    m_format;
        wxString    m_expr;
        Symbol      m_symbol;

        wxWindow*   getNextViewerWidget();
        bool        isEnabledViewerWidget( wxWindow* widget );
        void        OnEnter( wxCommandEvent& event );
        void        OnKeyDown( wxKeyEvent& event );
        void        OnKillFocus( wxFocusEvent& event );

        void        ReValidate();

    DECLARE_EVENT_TABLE()
};

/**********************************************************
*   STViewerComboBox
**********************************************************/
class STViewerComboBox : public wxComboBox, public STDataViewerBase
{
    public:
//        STViewerComboBox(   wxWindow* parent, wxWindowID id, const wxString& value,
//                            const wxPoint& pos, const wxSize& size, const wxArrayString& choices , long style );
        STViewerComboBox( wxWindow* parent, const Symbol& sym, const wxString& expr );

        void update();
        void requestRead();
        void updateModel();
        void requestWrite();

        void setSymbol( const Symbol& symbol_in ) { m_symbol = symbol_in; }
        void setExpression( const wxString& expr ) { m_expr = expr; }
        wxString toString();
        SymbolArray getSymbols();

    private:
        UserDialogTemplate* m_parent;
        wxString            m_expr;
        Symbol              m_symbol;
        wxArrayString       m_labels;
        int                 m_value;

        void OnTextUpdated( wxCommandEvent& event );
        void    ReValidate();
        DECLARE_EVENT_TABLE();
};

WX_DECLARE_HASH_MAP( int, wxString, wxIntegerHash, wxIntegerEqual, ID_TO_SYMBOL);

WX_DECLARE_STRING_HASH_MAP( wxStaticBitmap*, symbol2IconHash );
WX_DECLARE_STRING_HASH_MAP( wxStaticText*, symbol2LabelHash );

/**********************************************************
*   @name   UserDialogTemplate

    @brief  create a UserDialog from UserDialogDescriptor
            ud parameter must be valid ( not NULL )
**********************************************************/
class UserDialogTemplate : public wxScrolledWindow, public StaticViewersController, public ConfigValidatorViewerBase
{
    public:
        UserDialogTemplate();
        UserDialogTemplate( wxWindow* parent,
                            const UserDialog& desc,
                            SymbolMgr* symbolmgr,
                            ConfigValidatorBase* validator,
                            wxWindowID id = wxID_ANY,
                            const wxPoint& pos = wxDefaultPosition,
                            const wxSize& size = wxDefaultSize,
                            long style = wxTAB_TRAVERSAL | wxVSCROLL );

        virtual ~UserDialogTemplate();

        bool Init();
        bool Create( wxWindow* parent,
                     const UserDialog& desc,
                     SymbolMgr* symbolmgr,
                     ConfigValidatorBase* validator,
                     wxWindowID id = wxID_ANY,
                     const wxPoint& pos = wxDefaultPosition,
                     const wxSize& size = wxDefaultSize,
                     long style = wxTAB_TRAVERSAL | wxVSCROLL );

        // gauge group functions
        void setTablePopupMenuTitle( const wxString& title) { m_tablePopupMenuTitle = title; }
        bool onDropFiles( const wxArrayString& filenames );
        void importFromFile(  const wxString& file );

        static StaticViewersController* newInstance(wxWindow* parent, const UserDialog& desc,
                    UserDialogBase* udMgr, SymbolMgr* symbolMgr, STDataMgrBase* stMgr, ConfigValidatorBase* validator, wxArrayString symbolsToMark = wxArrayString() );

        void setConfigValidator( ConfigValidatorBase* validator ) { m_validator = validator; }

        void notify( const StringHash& errors_and_warnings );
        void setSymbolsToMark( wxArrayString& symbols ) { m_SymbolMarked = symbols; }

        ConfigValidatorBase*    getConfigValidator() { return m_validator; }

    private:
        ID_TO_SYMBOL            m_idToSymbol;
        wxString                m_tablePopupMenuTitle;
        wxGridBagSizer*         m_mainSizer;
        wxArrayString           m_symbolsOnDialog;

        ConfigValidatorBase*    m_validator;
        symbol2IconHash         m_symbol2icon;
        symbol2IconHash         m_markedsymbolIcon;
        symbol2LabelHash        m_symbol2label;
        wxArrayString           m_SymbolMarked;

        static wxBitmap         m_warningIcon;
        static wxBitmap         m_errorIcon;
        static wxBitmap         m_markedIcon;

        bool CreateDialog( const UserDialog& desc, SymbolMgr* symbolmgr );
        void addMissingWarning( const wxString& label, const wxString& variable );
        void addLabel( const wxString& label );
        void addEditField( STDataMgrBase* controller, const wxString& label, const Symbol& symbol_in, const wxString& expr );
        void addComboBox( STDataMgrBase* controller, const wxString& label, const Symbol& symbol_in, const wxString& expr);
        void addConfigletPanel( const UserDialog& desc, SymbolMgr* symbolmgr );

        void loadModelFromDialog( STDataModelBase& model );
        void updateDialogFromModel( STDataModelBase& model );
        void showErrorDialog(const FirmwareInfo& fwInfo);

        void OnLeftUpTextField( wxMouseEvent& event );
        void OnClose( wxCloseEvent& event );

        DECLARE_EVENT_TABLE();
};


#endif // USERDIALOGTEMPLATE_H
