//
// VEMSTableEditor.h - universal table editor for VEMS
//

#ifndef __HAVE_VEMSTABLEEDITOR_H
#define __HAVE_VEMSTABLEEDITOR_H

class wxScrolledWindow;

#include "wx/scrolwin.h"
#include "wx/generic/scrolwin.h"
#include "../common/types/String2IntHash.h"
#include "../common/types/PointArray.h"
#include "../common/types/IntArray.h"
#include "../common/types/Int2StringHash.h"
#include "../common/VemsMatrix.h"
#include "GroupMember.h"

class SelectableWidgetBase;
class FocusParentSinkerBase;

//
// VEMSTableEditor - table editor for VEMS
//

// Constants
const int VEMS_TABLE_EDITOR_DEFAULT_COL_WIDTH = 50;					// Default width of cells in the table
const int VEMS_TABLE_EDITOR_MAX_COLS = 16;							// Maximum number of columns in the grid
const int VEMS_TABLE_EDITOR_MAX_ROWS = 14;							// Maximum number of rows in the grid

/*
const int VEMS_TABLE_EDITOR_DEFAULT_CELL_FONT_SIZE = 16;			// Default font size for the cells (px)
const int VEMS_TABLE_EDITOR_DEFAULT_LABEL_FONT_SIZE = 11;			// Default font size for labels (px)
const int VEMS_TABLE_EDITOR_ROW_COL_CELL_EXTRA_WIDTH = 16;			// Extra width of cells (improve readability) (px)
*/

// Key codes control
enum
{
	VEMS_TABLE_EDITOR_KEY_INC_SMALL = 0,							// Increment cell value (small step)
	VEMS_TABLE_EDITOR_KEY_INC_LARGE,								// Increment cell value (large step)
	VEMS_TABLE_EDITOR_KEY_DEC_SMALL,								// Decrement cell value (small step)
	VEMS_TABLE_EDITOR_KEY_DEC_LARGE,								// Decrement cell value (large step)

	VEMS_TABLE_EDITOR_KEY_NAVIGATE_LEFT,							// Move cursor left
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_RIGHT,							// Move cursor right
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_UP,								// Move cursor up
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_DOWN,							// Move cursor down

	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_COPY_LEFT,					// Move cursor left and copy value of previous cell
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_COPY_RIGHT,					// Move cursor right and copy value of previous cell
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_COPY_UP,						// Move cursor up and copy value of previous cell
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_COPY_DOWN,					// Move cursor down and copy value of previous cell

	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_HIGHLIGHT_LEFT,				// Move cursor left and highlight rectangular area of cells
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_HIGHLIGHT_RIGHT,				// Move cursor right and highlight rectangular area of cells
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_HIGHLIGHT_UP,				// Move cursor up and highlight rectangular area of cells
	VEMS_TABLE_EDITOR_KEY_NAVIGATE_AND_HIGHLIGHT_DOWN,				// Move cursor down and highlight rectangular area of cells

	VEMS_TABLE_EDITOR_KEY_NAVIGATE_GRID_LEFTTOP_CORNER,				// Move cursor to the upper left cell (grid home)
    VEMS_TABLE_EDITOR_KEY_NAVIGATE_GRID_RIGHTBOTTOM_CORNER,			// Move cursor to the lower right cell (grid end)

	VEMS_TABLE_EDITOR_KEY_LOCK_CELLS,								// Lock current cells or highlighed cells

	VEMS_TABLE_EDITOR_KEY_LOCK_ALL_CELLS,							// Lock all cells

	VEMS_TABLE_EDITOR_KEY_UNLOCK_ALL_CELLS,							// Unlock all cells

	VEMS_TABLE_EDITOR_KEY_INTERPOLATE_CELLS,						// Interpolate cell(s)

	VEMS_TABLE_EDITOR_KEY_EDIT_VALUE,								// Edit value under the cursor

	VEMS_TABLE_EDITOR_KEY_HIGHLIGHT_ALL_CELLS,						// Highlight all cells

	VEMS_TABLE_EDITOR_KEY_UNDO_CELL_MODIFICATION,					// Undo cell(s) modification (restore previous cell)
	VEMS_TABLE_EDITOR_KEY_UNDO_ALL_CELLS_MODIFICATION,				// Unde modification of all cells in the grid

	VEMS_TABLE_EDITOR_KEY_MULTIPLY_CELLS,                           // Multiply selected cell(s)
	VEMS_TABLE_EDITOR_KEY_DIVIDE_CELLS,                             // Divide selected cell(s)

	VEMS_TABLE_EDITOR_KEY_SELECT_LOAD,

	VEMS_TABLE_EDITOR_CHANGE_TABLE,
	VEMS_TABLE_EDITOR_WRITE_FW_CONFIG,
	VEMS_TABLE_EDITOR_BURN_FW_CONFIG,
	VEMS_TABLE_EDITOR_REFRESH_FW_CONFIG,

	VEMS_TABLE_EDITOR_KEY_AUTO_FOLLOW_LOAD,
	VEMS_TABLE_EDITOR_KEY_DISPLAY_HELP,

	VEMS_TABLE_EDITOR_KEY_MAX_INDEX
};

// Event ids of VEMSTableEditor popup context-sencetive menu
enum
{
	VEMS_TABLE_EDITOR_PM_INC_CELL_LARGE = 2000,
	VEMS_TABLE_EDITOR_PM_DEC_CELL_LARGE,
	VEMS_TABLE_EDITOR_PM_INC_CELL_SMALL,
	VEMS_TABLE_EDITOR_PM_DEC_CELL_SMALL,
	VEMS_TABLE_EDITOR_PM_LOCK_CELL,
	VEMS_TABLE_EDITOR_PM_UNLOCK_CELL,
	VEMS_TABLE_EDITOR_PM_UNDO_CELL_MODIFICATION,
	VEMS_TABLE_EDITOR_PM_UNDO_ALL_CELLS_MODIFICATION,
	VEMS_TABLE_EDITOR_PM_LOCK_ALL_CELLS,
	VEMS_TABLE_EDITOR_PM_UNLOCK_ALL_CELLS,
	VEMS_TABLE_EDITOR_PM_ACCEPT_ALL_CHANGES,
	VEMS_TABLE_EDITOR_PM_EDIT_HEADER,
	VEMS_TABLE_EDITOR_PM_MOVE_HEADER,
	VEMS_TABLE_EDITOR_PM_INSERT_HEADER,
	VEMS_TABLE_EDITOR_PM_MULTIPLY_CELLS,
	VEMS_TABLE_EDITOR_PM_DIVIDE_CELLS,
	VEMS_TABLE_EDITOR_PM_IMPORT,
	VEMS_TABLE_EDITOR_PM_REMAP,
};

// Platform dependent stuff
#define vemsABS(a)    (((a) < 0) ? (-(a)) : (a))
#define vemsMemset  memset

//
// VEMSTableEditor class definiton
//

WX_DECLARE_HASH_MAP(int, wxBitmap, wxIntegerHash, wxIntegerEqual, IntToBitmapHash);


WX_DECLARE_OBJARRAY(wxColor, GradientColorArray);

class GaugeDescManager;
class GaugeProperties;
class wxHotKeyManager;

class VEMSTableEditor : public wxControl, public GroupMember
{
    DECLARE_DYNAMIC_CLASS(VEMSTableEditor);

    public:
        VEMSTableEditor() {};
        VEMSTableEditor(wxWindow *parent, GaugeDescManager * dmgr, GaugeProperties &prop);
        //, const wxPoint &point = wxDefaultPosition, const wxSize &size = wxDefaultSize,GaugeGroupManager* groupManager = NULL, const wxString& groupName = wxEmptyString );
        virtual ~VEMSTableEditor();

        // set tables from ini
        void SetTables( const wxArrayString& tables ) { m_tables = tables; }

        // VEMSTableEditor API
        void DeleteAllCells();										// Delete ALL cells (cols & rows) in the table
        bool CreateCells(int numrows, int numcols);					// Create grid

        bool SetColLabelValue(int col, double);						// Change caption of the specified column
        void AcceptModifiedColLabelValue( int col );
        int GetColLabelValue(int col);								// Return caption of the specified column

        bool SetRowLabelValue(int col, double);						// Change caption of the specified row
        void AcceptModifiedRowLabelValue( int row );
        int GetRowLabelValue(int row);								// Return caption of the specified row

        void SetCellsMinValue(float v);								// Set minimum value for cells
        float GetCellsMinValue() const;								// Get minimum value for cells

        void SetCellsMaxValue(float v);								// Set maximum value for cells
        float GetCellsMaxValue() const;								// Get maximum value for cells

        ///row header
        void    SetRowHeaderMinValue( double v ) { m_minRowHeaderValue = v; }   // Set minimum value for row header
        double  GetRowHeaderMinValue() const { return m_minRowHeaderValue; }    // Get minimum value for row header
        void    SetRowHeaderMaxValue( double v ) { m_maxRowHeaderValue = v; }   // Set maximum value for row header
        double  GetRowHeaderMaxValue() const { return m_maxRowHeaderValue; }    // Get maximum value for row header
        bool    WasRowLabelModified( int row );

        ///col header
        void    SetColHeaderMinValue( double v ) { m_minColHeaderValue = v; }   // Set minimum value for col header
        double  GetColHeaderMinValue() const { return m_minColHeaderValue; }    // Get minimum value for col header
        void    SetColHeaderMaxValue( double v ) { m_maxColHeaderValue = v; }   // Set maximum value for col header
        double  GetColHeaderMaxValue() const { return m_maxColHeaderValue; }    // Get maximum value for col header
        bool    WasColLabelModified( int col );

        ///load from VemsMatrix<double>
        void    loadFromVemsMatrix( const VemsMatrix<double>& m );
        void    loadRowHeaderFromVemsMatrix( const VemsMatrix<double>& m );
        void    loadColHeaderFromVemsMatrix( const VemsMatrix<double>& m );

        float GetCellValue(int row, int col);						// Modify value of cell specified
        bool SetCellValue(int row, int col, float v);				// Return current value of cell specified

        bool WasCellModified(int row, int col) const;				// Retuns status of the cell specified - was cell modified or not
        bool AreCellsModified() const;
        void AcceptAllModifiedCells();								// Accept all cell modifications
        void AcceptSingleModifiedCell( int row, int col );          // Accept a cell modification

        void SetCurrentLoad(float load, float rpm);					// Set current load marker
        void ClearCurrentLoad();									// Removed current load marker from grid

        int GetNumberRows() const;									// Return number of rows
        int GetNumberCols() const;									// Return number of colums

        // Precision control
        bool SetValuePrecision(int width, int precision);			// Set precision of values in the cells
        void GetValuePrecision(int *width, int *precision);			// Retrive current precision of values in the cells

        bool SetSmallStepValue(float small_step);					// Set small inrement value
        float GetSmallStepValue() const;							// Get small inrement value

        bool SetLargeStepValue(float large_step);					// Set large increment value
        float GetLargeStepValue() const;							// Get large increment value

        // Key map control
        bool SetKey(unsigned int func, int keycode, int modifiers = 0);		// Change key code and key modifiers for specified functionality
        int  GetKeycode(unsigned int func) const;
        int  GetKeyModifiers(unsigned int func) const;

        // Look & feel control
        bool SetCellFont(const wxFont& font);
        wxFont GetCellFont() const;

        void SetCellTextColor(const wxColor& clr);
        wxColor GetCellTextColor() const;

        bool SetLabelFont(const wxFont& font);
        wxFont GetLabelFont() const;

        void SetLabelTextColor(const wxColor& clr);
        wxColor GetLabelTextColor() const;

        bool SetZoomboxFont(const wxFont& font);
        wxFont GetZoomboxFont() const;

        void SetZoomboxTextColor(const wxColor& clr);
        wxColor GetZoomboxTextColor() const;

        void SetCellBackgroundColor(const wxColor& clr);
        wxColor GetCellBackgroundColor() const;

        void AddCellBackgroundColorGradient(const wxColor& clr1);
        void SetCellBackgroundColorGradient(const wxColor& clr1, const wxColor& clr2);
        bool GetCellBackgroundColorGradient(wxColor* clr1, wxColor* clr2) const;

        void SetLockedCellBackgroundColor(const wxColor& clr);
        wxColor GetLockedCellBackgroundColor() const;

        void SetModifiedCellTextColor(const wxColor& clr);
        wxColor GetModifiedCellTextColor() const;

        void SetGridLinesColor(const wxColor& clr);
        wxColor GetGridLinesColor() const;

        void SetCurrentLoadMarkerColor(const wxColor& clr);
        wxColor GetCurrentLoadMarkerColor() const;

        void SetWidgetBackgroundColor(const wxColor& clr);
        wxColor GetWidgetBackgroundColor() const;

        void SetCursorColor(const wxColor& clr);
        wxColor GetCursorColor() const;

        void SetLabel3DColors(const wxColor& inner, const wxColor& outer);
        void GetLabel3DColors(wxColor* inner, wxColor* outer);

        void SetLoadMarkerRadius(double r);  /// temporarly radius <= cell_height/2 !!!! (tip. is equal)
        double GetLoadMarkerRadius() { return m_loadMarkerRadius; };

        // Helper API
        wxSize GetBestSize();

        void CalcWidgetSize(const wxSize &size);

        //gauge group function
        void SetPopupMenuTitle( const wxString& title ) { m_popupMenuTitle = title; }
        void AppendPopupMenuItem( int id
                                , const wxString&  item = _T("")
                                , const wxString& helpString = _T("")
                                , wxItemKind kind = wxITEM_NORMAL )
                                { m_popupMenu.Append( id, item, helpString, kind ); }


        void UpdateSize();
        void moveCursorToPosition( const wxPoint& position );

        VemsMatrix<double>*  GetRowLabels();
        VemsMatrix<double>*  GetColLabels();

        void AutoFollowLoad();

        void SetFocusParentSinker(FocusParentSinkerBase* fps) { m_focusSinker = fps; }

    protected:
        //virtual void OnSingleCellChanged( int row, int col, double d ){}

        double GetSideRatio();

        SelectableWidgetBase* GetSelectableWidgetBase();

        void OnSetFocus(wxFocusEvent &evt);
        void OnKillFocus(wxFocusEvent &evt);

    private:
        String2IntHash m_hotkeySymbolToEnum;

        wxHotKeyManager * m_hotkeyMgr;

        int m_cellFontSize;
        int m_labelFontSize;
        double m_rowColCellExtraWidth;

        // Internal common variables
        int m_numRows, m_numCols;																	// Number of rows & cols
        int m_cursorRow, m_cursorCol;																// Current cursor position
        int m_selectedRowLabel;
        int m_selectedColLabel;

        bool m_showLoadMarker;
        double m_currentMarkerLoad, m_currentMarkerRpm;												// Current load marker values

        int m_drawOriginRow, m_drawOriginCol;														// Cell to start widget drawing from

        int m_highlightOriginRow, m_highlightOriginCol;												// The begining of rectangular area of highlighted cells (when using CTRL + arrow keys)

        float m_minCellValue, m_maxCellValue;														// Minimum and maximum values of cells
        double m_minRowHeaderValue, m_maxRowHeaderValue;											// Minimum and maximum values of row header
        double m_minColHeaderValue, m_maxColHeaderValue;											// Minimum and maximum values of col header

        float m_smallStepValue, m_largeStepValue;													// Small / large step values

        VemsMatrix<int> m_rowLabels;												                // Row labels
        VemsMatrix<int> m_rowLabelsBak;												                // Row labels backup

        VemsMatrix<int> m_colLabels;												                // Column labels
        VemsMatrix<int> m_colLabelsBak;												                // Column labels backup

        VemsMatrix<double> m_cellValues;					                                        // Cell values
        VemsMatrix<double> m_cellBakValues;				                                            // Backup cell values to maintain accept/undo functionality

        unsigned char m_cellLockFlag[VEMS_TABLE_EDITOR_MAX_ROWS][VEMS_TABLE_EDITOR_MAX_COLS];		// Cell lock flags
        unsigned char m_cellHighlightFlag[VEMS_TABLE_EDITOR_MAX_ROWS][VEMS_TABLE_EDITOR_MAX_COLS];	// Cell selection ("highlight") flags

        int m_valueWidth;																			// Cell value precision - integer part
        int m_valuePrecision;																		// Cell value precision - fractional part

        // Key codes
        struct VEMSTableKeyPair
        {
            int keycode;
            int modifiers;
        };

        VEMSTableKeyPair m_keyMap[VEMS_TABLE_EDITOR_KEY_MAX_INDEX];

        // Look & feel control variables
        wxFont m_cellFont;						// Font of regular cell
        wxFont m_labelFont;						// Font for labels
        wxFont m_zoomboxFont;					// Font for cell's editor input field

        wxColor m_cellTextColor;				// Color of the cell text
        //wxColor m_cellBkColor;					// Color of cell background
        //wxColor m_cellBkColorGradient;			// Color of cell background #2 (Gradient mode)
        wxColor m_cellHighlightBkColor;			// Background color for selected cell
        GradientColorArray m_gradientColorArray;

        wxColor m_labelTextColor;				// Color of the label text
        wxColor m_labelBkColor;					// Background color of the labels

        wxColor m_zoomboxTextColor;				// Color of the cell's editor text

        wxColor m_modifiedCellTextColor;		// Color for text in modified (changed) cell
        wxColor m_lockedCellBkColor;			// Background color of locked cell
        wxColor m_gridLinesColor;				// Color for grid separator lines
        wxColor m_cellHighlightTextColor;       // Text color for highlighted cell

        wxColor m_currentLoadMarkerColor;		// Color of current load marker

        wxColor m_bkColor;                      // Background color of the widget
        wxColor m_cursorColor;                  // Color of the cursor

        wxColor m_inner3DColor, m_outer3DColor; // Colors for 3-dimension borders on labels.

        // Internal layout variables
        //wxSize m_cellSize;
        wxRealPoint m_cellSize;
        wxRect m_gridRect;
        wxRect m_gridWarningRect;
        double m_rowLabelWidth, m_colLabelHeight;
        wxSize m_widgetSize;

        // Internal methods
        wxString GetValueFormatMask() const;

        static bool IsKeycodesEqual(int keycode1, int keycode2);
        static bool IsKeyModifiersEqual(int wx_modifiers, int vems_table_modifiers);
        int GetFunctionByKey(int keycode, int modifiers);
        wxString GetKeyTextualForFunction(int func) const;

        void SetDefaults();
        void RecalcSizes();                 // Recalculate best row / column / cell size

        bool IsScrollNeededToViewCell(int row, int col, int *scrollRow = 0, int *scrollCol = 0) const;
        bool EnsureCellVisible(int row, int col);

        void LockCell(int row, int col);
        void UnlockCell(int row, int col);
        bool IsCellLocked(int row, int col) const;
        bool IsCellHighlighted(int row, int col) const;
        void LockHighlightedCells();
        void LockAllCells();
        void UnlockAllCells();

        wxRect GetCellRectPix(int row, int col) const;            // In logical coords
        wxRect GetCellRectVal(int row, int col) const;            // In values (rpm, load)
        wxRect GetLabelRowRect(int row) const;
        wxRect GetLabelColRect(int col) const;

        bool HitTestCell(const wxPoint& pos, int *row = 0, int *col = 0) const;
        bool HitTestLabelRow(const wxPoint& pos, int *row = 0) const;
        bool HitTestLabelCol(const wxPoint& pos, int *col = 0) const;
        bool HitTestLeftTopCorner(const wxPoint& pos) const;

        void HighlightRow(int row);
        void UnhighlightRow(int row);
        bool IsRowHighlighted(int row) const;

        void HighlightColumn(int col);
        void UnhighlightColumn(int col);
        bool IsColumnHighlighted(int col) const;

        void HighlightCell(int row, int col);
        void UnhighlightCell(int row, int col);

        void HighlightAllCells();
        void UnhighlightAllCells();
        bool AreAllCellsHighlighted() const;
        bool AreCellsHighlighted() const;

        bool HighlightRectangularAreaOfCells(int start_row, int start_col, int end_row, int end_col);
        bool IsRectangularAreaOfCellsHighlighted() const;

        void ResetHighlightOrigin();

        void InterpolateCell(int row, int col);
        void InterpolateSelection();

        void UndoCellModification(int row, int col);
        void UndoHighlightedCellsModification();
        void UndoAllCellsModification();

        void IncCellLarge();
        void DecCellLarge();
        void IncCellSmall();
        void DecCellSmall();
        void LockUnlockCells();

        void RequestValueForCursorCell( const wxString& initialValue = _T("") );

        void DrawCurrentLoadMarker(wxDC& dc, const wxRect& clientRc);
        void Draw3DBorder(wxDC& dc, const wxRect& rc);

        void Send( int type, int row = -1, int col = -1, double v = 0.0 );
        void ShowColHeaderDialog( int type );
        void ShowRowHeaderDialog( int type );

        void MultiplyCell( int row, int col, double d );
        void MultiplySelection( double d );                 // multiple a cell with d
        void ShowMultiplyDialog();                          // multiple highlighted cells with d

        void DivideCell( int row, int col, double d );      // divide a cell with d
        void DivideSelection( double d );                   // divide highlighted cells with d
        void ShowDivideDialog();

        void SelectLoad();                                  // select load cell
        bool isCursorPositionValid() const;
        void showChangeValueDialog( const wxString& initialValue );
        void changeSelection( double d );

        // Utilities
        enum MoveDirection { LEFT, RIGHT, UP, DOWN };
        enum MoveStyle { MOVE, COPY, HIGHLIGHT };

        void CopyFrom( int row, int col );
        void HighlightFrom( int row, int col );
        void EnsureCellVisible( int dir );
        bool MoveCursorIsPossible( int dir );
        void MoveCursor( int dir, int style = MOVE );
        bool ChangeCellValue( int row, int col, double delta );
        void ChangeSelectedCells( double delta, bool readAfterWrite = false );
        void ChangeCells( int start_row, int start_col, int end_row, int end_col, double d );
        void SendCursorCellChangedEvent();
        void AppendPopupMenu( wxMenu& menu, int popupMenuID, const wxString& eventSymbol, const wxString& miscStr = _T(" ") );

        // Event handlers
        void OnPaint(wxPaintEvent&);								// Painting of the widget
        void OnEraseBackground(wxEraseEvent&);
        void OnKeyDown(wxKeyEvent&);                                // Keyboard key presses
        void OnKeyUp( wxKeyEvent& event );
        void OnMouseLeftClick(wxMouseEvent&);						// Sigle clock of left mouse button
        void OnMouseLeftDoubleClick(wxMouseEvent&);					// Double click of left mouse button
        void OnMouseRightClick(wxMouseEvent&);						// Single click of right mouse button
        void OnSize(wxSizeEvent &evt);
        void OnMouseMove(wxMouseEvent& evt);
        void OnMouseLeftUp(wxMouseEvent& evt);

        // Event handlers for popup menu
        void OnPopupMenuMultiply( wxCommandEvent& event );
        void OnPopupMenuDivide( wxCommandEvent& event );
        void OnPopupMenuIncCellLarge(wxCommandEvent& event);
        void OnPopupMenuDecCellLarge(wxCommandEvent& event);
        void OnPopupMenuIncCellSmall(wxCommandEvent& event);
        void OnPopupMenuDecCellSmall(wxCommandEvent& event);
        void OnPopupMenuLockCell(wxCommandEvent& event);
        void OnPopupMenuUnlockCell(wxCommandEvent& event);
        void OnPopupMenuUndoCellModification(wxCommandEvent& event);
        void OnPopupMenuUndoAllCellsModification(wxCommandEvent& event);
        void OnPopupMenuLockAllCells(wxCommandEvent& event);
        void OnPopupMenuUnlockAllCells(wxCommandEvent& event);
        void OnPopupMenuAcceptAllChanges(wxCommandEvent& event);
        void OnPopupMenuEditHeader(wxCommandEvent& event);
        void OnPopupMenuChangeTable(wxCommandEvent& event);
        void OnPopupMenuImport( wxCommandEvent& event );
        void OnPopupMenuRemap( wxCommandEvent& event );

        DECLARE_EVENT_TABLE()


        //IntToBitmapHash m_tableCache;
        PointArray      m_currentRefreshCells;
        PointArray      m_oldRefreshCells;

        PointArray      m_currentRefreshCellsforLoad;
        PointArray      m_oldRefreshCellsforLoad;

        PointArray      m_refreshedCells;


        bool            m_refreshAllCells;
        bool            m_isChange;
        bool            m_isLoadChange;
        bool            m_isSizeChanging;

        wxBitmap        *m_face;

        wxPoint         m_currentLoadCoordPix;
        wxPoint         m_currentLoadCoord;

        //wxPoint         m_max

        double m_minGridRpm;
        double m_maxGridRpm;

        double m_minGridLoad;
        double m_maxGridLoad;

        double m_loadMarkerRadius;

        double m_clientAR;
        double m_parentAR;

        IntArray    m_currentRefresRowLabel;
        IntArray    m_oldRefresRowLabel;

        IntArray    m_currentRefresColLabel;
        IntArray    m_oldRefresColLabel;

        //gauge group variables
        wxString        m_popupMenuTitle;
        wxMenu          m_popupMenu;
        wxArrayString   m_tables;           // available tables from ini
        Int2StringHash  m_menuIDtoTableName;

        wxWindow   *m_parent;
        wxSize      m_lastSize;

        wxPoint     m_delta;

        bool        m_isAutoFollowLoad;

        wxBitmap*   m_helpPanelBmp;
        bool        m_isShowHelp;
        bool        m_isResized;

        FocusParentSinkerBase*  m_focusSinker;

        void drawHelpPanel( wxDC& dc );
        double getHelpTextHeight( int clientHeight, size_t lineCount, double margin, double rowScale );
        void createHelpPanel( int clientHeight );

    protected:

        wxWindow* GetSelectFaceParent();

        void NormalizeFont(wxFont &font, int limit1 = 10, int limit2 = 20);

        void DrawRefreshLabels(wxDC &dc);

        void DrawRowLabel(wxDC &dc, int row);
        void DrawColLabel(wxDC &dc, int col);

        void CalculateGridValueLimits();

        wxColor CalculateCellBgColor(const wxPoint &point);


        void SetFullRefresh(bool isRefresh = true) { m_refreshAllCells = isRefresh; };
        void SetLoadChange(bool isChange = true) { m_isLoadChange = isChange; };
        void SetCellChange(bool isChange = true) { m_isChange = isChange; };

        void ClearRefreshArrays();
        bool SearchInPointArray(const PointArray &parr, const wxPoint &point);

        bool AddNewRefreshCell(short x, short y);
        bool AddNewRefreshCell(const wxPoint& point);

        bool AddOldRefreshCell(short x, short y);
        bool AddOldRefreshCell(const wxPoint& point);

        bool AddNewRefreshCellforLoad(short x, short y);
        bool AddNewRefreshCellforLoad(const wxPoint& point);

        bool AddNewRefreshColLabelIndex(short idx);
        bool AddNewRefreshRowLabelIndex(short idx);

        void RefreshBackgroundCustomCells(wxDC &dc);
        void RefreshCellLabels(wxDC &dc);

        void CalculateLoadCellPoints();
        void DrawBackgroundCustomCells(wxDC &dc, const wxPoint &point);
        void DrawCellLabel(wxDC &dc, const wxPoint &point);

};


//
// VEMSCellEditorDialog class definition
//

class VEMSCellEditorDialog : public wxDialog
{
public:
	VEMSCellEditorDialog(wxWindow *parent, float cur_val, float min_value, float max_value, const wxRect& cellRect, const wxFont& zoomboxFont, const wxColor& zoomboxTextColor, const wxString& formatMask, const wxString& intialValue);
	virtual ~VEMSCellEditorDialog();

	bool IsValueAccepted() const;
	float GetResultValue() const;

private:
	bool m_isValAccepted;
	float m_resultVal;
	float m_minVal;
	float m_maxVal;

	wxTextCtrl *m_valueText;
	wxButton   *m_okButton;
	wxButton   *m_cancelButton;

	// Event handlers
	void OnButton(wxCommandEvent& event);

	DECLARE_EVENT_TABLE()
};

#endif     // !__HAVE_VEMSTABLEEDITOR_H



