#include "BootModeBase.h"

#include "../CommunicationBase.h"
#include "../SerialException.h"
#include "../pageexception.h"
#include "../logger.h"
#include "../Math.h"

BootModeBase::BootModeBase(CommunicationBase* commBase)
{
    //ctor
    m_commBase = commBase;
}

BootModeBase::~BootModeBase()
{

}

bool BootModeBase::checkFirmwareType()
{
    if( m_firmwareData->getType() == m_type) return true;

    Logger::logf(LOG_MSG, _T("BootModeBase::checkFirmwareType() invalid firmware type: ")+ m_firmwareData->getTypeAsString() );
    return false;
}

bool BootModeBase::failSafePageHandle(uint32_t idx, pageHandleFV p)
{
    bool tryAgain = false;
    int tryCount = 0;
    int PAGE_MAX_TRY_COUNT = 5;
    wxString tmp;
    do{
        tryAgain = false;
        try
        {
            (this->*p)(idx);
            tryAgain = false;
        }
        catch (PageException& pe)
        {
            tmp = pe.toString();
            tryAgain = true;
            tryCount++;
            //TODO: invoke a visitor that collect the pages NTCFinder and a GUI should use it.
        }catch (SerialException& se){
            //Logger::logf(LOG_MSG, _T("SERIAL EXCEPTION occured-----------------------------"));
            Logger::logf(LOG_MSG, _T("SERIAL EXCEPTION: ") + se.toString() );
            tryCount++;
        }catch (...){
            Logger::logf(LOG_MSG, _T("UNKNOWN EXCEPTION occured-----------------------------"));
            tryCount++;
        }

    }while(tryAgain && (tryCount < PAGE_MAX_TRY_COUNT) );
    if(tryAgain){
            observerNotify(tmp);
    }
    return !tryAgain;
}

void BootModeBase::observerVerifyInit()
{
    size_t pagesNum = m_firmwareData->getPages();
    Logger::logf(LOG_MSG, _T("BootModeBase::observerVerifyInit(): pages:%d"), pagesNum);
    wxString deviceStr = ECUInfo::getDeviceAsString(m_device);
    observerInit(_T("ECU (") +deviceStr+ _T(") flash Verify Progress"), _T("Page index:"), pagesNum );
}

void BootModeBase::observerUploadInit()
{
    size_t pagesNum = m_firmwareData->getPages();
    wxString deviceStr = ECUInfo::getDeviceAsString(m_device);
    observerInit(_T("ECU (") +deviceStr+ _T(") flash Upload Progress"), _T("Page index:"), pagesNum );
}


int BootModeBase::selectPage(const char* pageCmd, uint32_t pageIdx )
{
    uint32_t pageID = getPageID(pageIdx);

    m_commBase->SendString((uint8_t*)pageCmd);
    int result = m_commBase->sendNumber( pageID );
    if ( result == -1 )
    {
        throw PageException(PAGE_ADDRESS_SYNC_ERROR, pageIdx);
        //return PAGE_ADDRESS_SYNC;
    }
#ifdef SFILE_LIB
	wxMilliSleep((5+50)*10*1000/(m_commBase->getBaudRate()));
#endif
    result = m_commBase->ReadString(m_buff, 1, 64); // we except 1 byte in fact
    if(result == 1){
        if( m_buff[0] == '\r' ){ // OK
            //Logger::logf(LOG_MSG, _T("BootModeV3::selectPage() address in SYNC OK! idx: %i"), idx );
        }else{ // typically 'N' for bad checksum or nothing for serious out of sync error
            Logger::logf(LOG_MSG, _T("BootModeBase::selectPage() address in SYNC failed! idx: %i"), pageIdx);
            throw PageException(PAGE_ADDRESS_SYNC_ERROR, pageIdx);
        }
    }else{
        Logger::logf(LOG_MSG, _T("BootModeBase::selectPage() address in SYNC failed, received bytecount: %d, idx: %i"), result, pageIdx );
         throw PageException(PAGE_ADDRESS_SYNC_ERROR, pageIdx);
    }
    return 0;
}

int BootModeBase::verifyPageCommon(const char* cmd, uint32_t pageIdx )
{
  m_commBase->SendString((uint8_t*)cmd);
    //int len =
#ifdef SFILE_LIB
	wxMilliSleep((4+40)*10*1000/(m_commBase->getBaudRate()));
#endif
	int result = m_commBase->ReadString(m_buff,4, 16);   // we except 4 byte (cheksum) in fact
    if(result != 4){
            Logger::logf(LOG_MSG, _T("BootModeBase::verifyPage() getChecksum result: %i"), result);
    }

    uint32_t filePageCheckSum = m_firmwareData->getPageCheckSum(pageIdx);
    uint32_t ecuPageCheckSum	= Math::getNumber( m_buff);

    if( !checksumVerify(ecuPageCheckSum, filePageCheckSum ) )
    //if (filePageCheckSum != ecuPageCheckSum)
    {
        PageException pe = PageException(PAGE_VERIFY_CHECKSUM_ERROR, pageIdx);
        pe.m_pageCheckSum = filePageCheckSum;
        pe.m_ecuCheckSum = ecuPageCheckSum;

        Logger::logf(LOG_MSG, _T("BootModeBase::verifyPage() different checksum! idx: %i,%08X,%08X"), pageIdx, filePageCheckSum, ecuPageCheckSum);
        throw pe;
    }
    return ecuPageCheckSum;
}

int BootModeBase::uploadPageCommon(const char* cmd, uint32_t pageIdx)
{

    m_commBase->SendString((uint8_t*)cmd);
    bool found;
    int result;
    uint8_t *buf = m_firmwareData->getPageData(pageIdx, found);
    uint32_t pageSize = m_firmwareData->getPageSize();
    result = m_commBase->sendStringBlock(buf, pageSize, 1);
    if (result != (int)pageSize)
    {
        throw PageException(PAGE_WRITE_ERROR, pageIdx);
    }

    Math::getBytes(m_firmwareData->getPageCheckSum(pageIdx), m_buff);

    result = m_commBase->SendString(m_buff, 4); // send checksum
    if(result != 4){
        Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon() send checksum byte: Failed") );
    }
#ifdef SFILE_LIB
	wxMilliSleep((1+10)*10*1000/(m_commBase->getBaudRate()));
#endif
    result = m_commBase->ReadString(m_buff,1, 64); // we except 1 byte in fact
    if(result == 1){
        if( m_buff[0] == '\r' ){ // OK
            //Logger::logf(LOG_MSG, _T("Firmware::uploadPage() after checksum read in SYNC OK!") );
        }else{ // typically 'N' for bad checksum or nothing for serious out of sync error
            Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon() page write checksum failed! Received Byte: %c"), m_buff[0] );
            throw PageException(PAGE_WRITE_CHECKSUM_ERROR, pageIdx);
        }
    }else{
        Logger::logf(LOG_MSG, _T("BootModeBase::uploadPageCommon failed, received bytecount: %d"), result );
         throw PageException(PAGE_WRITE_SYNC_ERROR, pageIdx);
    }

    return PAGE_OK;
}

int BootModeBase::verifyFlashCommon()
{
    if( !checkFirmwareType() ) return -1;

    uint32_t i;
    int error_cnt = 0;
    int MAX_PAGE_ERROR_COUNT=15;
//    uint32_t ecuPageCs = 0;

    observerVerifyInit();
    size_t pagesNum = m_firmwareData->getPages();

    for(i = 0; i < pagesNum && error_cnt < MAX_PAGE_ERROR_COUNT; i++)
    {
        observerNotify(i, i); //not a page index;
        uint32_t idx = m_firmwareData->getPageIndexbyOrderIdx(i);
        if( !failSafeVerifyPage(idx) ){
            error_cnt++;
        }
    }
    observerNotify(pagesNum, pagesNum);
    return error_cnt;

}

int BootModeBase::uploadFlashCommon(uint32_t maxPageIdx)
{
    if( !checkFirmwareType() ) return -1;

    uint32_t i;
    int error_cnt = 0;
    int MAX_PAGE_ERROR_COUNT=15;

    observerUploadInit();
    size_t pagesNum = m_firmwareData->getPages();
    for(i = 0; i < pagesNum && error_cnt < MAX_PAGE_ERROR_COUNT; i++)
    {
        observerNotify(i, i); //not a page index;
        uint32_t idx = m_firmwareData->getPageIndexbyOrderIdx(i);
        if( (maxPageIdx > 0) && (idx > maxPageIdx) ){
            wxString msg = wxString::Format(
                _T("Page write error in page: %i. (Page write to idx > %i is not allowed!)"), idx, maxPageIdx);
            observerNotify(msg);
            continue;
        }
        if( !failSafeUploadPage(idx) ) error_cnt++;
    }

    observerNotify(pagesNum, pagesNum);
    return error_cnt;
}

bool BootModeBase::checksumVerify(uint32_t ecu_checksum, uint32_t file_checksum )
{
    Logger::logf(LOG_MSG, _T("BootModeBase::checksumVerify() START"));
    return (ecu_checksum == file_checksum);
}
