#include "ManProtocol.h"


#include "MegatuneProtocol.h"
#include "SIPRTypes.h"
#include "SIPRPacket.h"
#include "../events/InputEvent.h"
#include "../logger.h"
#include "../Math.h"
#include "../FirmwareInfo.h"

#include <wx/listimpl.cpp>


WX_DEFINE_LIST( ManCommandList );


///----------------------------------------------------------------------------------------ManCommand
ManCommand::ManCommand( int             type,
                        const uint8_t*  data,
                        uint8_t         size,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             sleepAfterSend )

{
    init( type, data, size, expectedRxSize, timeoutCount, clearCount, sleepAfterSend  );
}

ManCommand::ManCommand( int             type,
                        const wxString& command,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             sleepAfterSend )
{
    init( type, command, expectedRxSize, timeoutCount, clearCount, sleepAfterSend );
}

ManCommand::ManCommand( int             type,
                        const wxString& command )
{
    init( type, command, 0, 0, 0, 0 );

}

void ManCommand::init(  int             type,
                        const uint8_t*  data,
                        uint8_t         size,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             sleepAfterSend )
{
    m_pData = new uint8_t[ size ];
    memcpy( m_pData, data, size );
    m_size  = size;
    m_iExpectedRxSize   = expectedRxSize;
    m_iTimeoutCount     = timeoutCount;
    m_iClearCount       = clearCount;
    m_iType             = type;
    m_iSleepAfterSend   = sleepAfterSend;
}

void ManCommand::init(  int             type,
                        const wxString& command,
                        int             expectedRxSize,
                        int             timeoutCount,
                        int             clearCount,
                        int             sleepAfterSend )
{
#if defined (WIN32) || (__WINCE_STANDARDSDK__)
    char* c = _strdup( command.mb_str() );
#else
    char* c = strdup( command.mb_str() );
#endif

    init( type, (uint8_t*)c, strlen( c ), expectedRxSize, timeoutCount, clearCount, sleepAfterSend );
}


///-----------------------------------------------------------------------------------------------ManProtocolMode
bool ManProtocolMode::isInState( int stateType )
{
    return m_pCurrentState->getType() == stateType;
}

void ManProtocolMode::changeState( ManProtocolState* state )
{
    m_pCurrentState = state;
}


///--------------------------------------------------------------------------------------ManProtocolModeNormal
ManProtocolModeNormal::ManProtocolModeNormal()
{
    Logger::logf( LOG_MSG, _T("ManProtocolModeNormal::ManProtocolModeNormal()") );

    setType( ManProtocolMode::NORMAL );
    changeState( ManProtocolStateIdle::instance() );
}

bool ManProtocolModeNormal::isInTxState()
{
    return ( m_pCurrentState->isSameTypeAs( ManProtocolState::IDLE ) );
//                m_pCurrentState->isSameTypeAs( ManProtocolState::TIMEOUT ) );
}

bool ManProtocolModeNormal::isInRxState()
{
    return (    m_pCurrentState->isSameTypeAs( ManProtocolState::WAIT ) ||
                m_pCurrentState->isSameTypeAs( ManProtocolState::CLEAR ) );
}

bool ManProtocolModeNormal::isIdle()
{
    return isInState( ManProtocolState::IDLE );
}

void ManProtocolModeNormal::execute( ManProtocol* p )
{
    m_pCurrentState->execute( p );
}


///---------------------------------------------------------------------------------------ManProtocolModeDownload
ManProtocolModeDownload::ManProtocolModeDownload()
{
    Logger::logf( LOG_MSG, _T("ManProtocolModeDownload::ManProtocolModeDownload()") );
    setType( ManProtocolMode::DOWNLOAD );
    changeState( ManProtocolStateDownloadIdle::instance() );
}

bool ManProtocolModeDownload::isInTxState()
{
    return m_pCurrentState->isSameTypeAs( ManProtocolState::DOWNLOAD_IDLE );
}

bool ManProtocolModeDownload::isInRxState()
{
    return m_pCurrentState->isSameTypeAs( ManProtocolState::DOWNLOAD_WAIT );
}

bool ManProtocolModeDownload::isIdle()
{
    return isInState( ManProtocolState::DOWNLOAD_IDLE );
}

void ManProtocolModeDownload::execute( ManProtocol* p )
{
    m_pCurrentState->execute( p );
}


///--------------------------------------------------------------------------------------ManProtocolModeSaveSD
ManProtocolModeSaveSD::ManProtocolModeSaveSD()
{
    Logger::logf( LOG_MSG, _T("ManProtocolModeSaveSD::ManProtocolModeSaveSD()") );

    setType( ManProtocolMode::SAVE_SD );
    changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
}

bool ManProtocolModeSaveSD::isInTxState()
{
    return  m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_NEED_RT_SIZE ) ||
            m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_NEED_DATA );

}

bool ManProtocolModeSaveSD::isInRxState()
{
    return  m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_WAIT_RT_SIZE ) ||
            m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_WAIT_DATA ) ||
            m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_CLEAR_AFTER_RT_SIZE ) ||
            m_pCurrentState->isSameTypeAs( ManProtocolState::SAVE_SD_CLEAR_AFTER_DATA );

}

bool ManProtocolModeSaveSD::isIdle()
{
    return isInState( ManProtocolState::SAVE_SD_NEED_RT_SIZE );
}

void ManProtocolModeSaveSD::execute( ManProtocol* p )
{
    m_pCurrentState->execute( p );
}


///----------------------------------------------------------------------------------------ManProtocolModeNormal States

///------------------------------------------------------------------------------------------------ManProtocolStateIdle
ManProtocolStateIdle* ManProtocolStateIdle::s_pInstance = NULL;

ManProtocolStateIdle* ManProtocolStateIdle::instance()
{
    //Logger::logf( LOG_MSG, _T("ManProtocolStateIdle::instance()") );

    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateIdle(); }

    return s_pInstance;
}

void ManProtocolStateIdle::execute( ManProtocol* p )
{
    if ( !p->hasNextCommand() && p->needDisable() ) { p->setDisabled(); }

    if ( p->isDisabled() ) { return; }

    if ( !p->hasNextCommand() ) { return; }

    p->sendCommand();

    p->resetRxProperties();

    if ( p->isWaitForResponse() )
    {
        p->changeState( ManProtocolStateWait::instance() );
    }
    else if ( p->isNeedClear() )
    {
        p->changeState( ManProtocolStateClear::instance() );
    }
}

///------------------------------------------------------------------------------------------------ManProtocolStateWait
ManProtocolStateWait* ManProtocolStateWait::s_pInstance = NULL;

ManProtocolStateWait* ManProtocolStateWait::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateWait(); }

    return s_pInstance;
}

void ManProtocolStateWait::execute( ManProtocol* p )
{
    if ( p->isDataReceived() )
    {
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( p->isExpectedDataReceived() )
        {
            p->notify();
            p->changeState( ManProtocolStateClear::instance() );
        }
    }
    else
    {
        p->increaseTimeoutCount();

        if ( p->isTimedOut() )
        {
            p->notifyTimeout();
            p->changeState( ManProtocolStateIdle::instance() );
        }
    }
}

///---------------------------------------------------------------------------------------------ManProtocolStateClear
ManProtocolStateClear* ManProtocolStateClear::s_pInstance = NULL;

ManProtocolStateClear* ManProtocolStateClear::instance()
{
    Logger::logf( LOG_MSG, _T("-----ManProtocolStateClear::instance()") );

    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateClear(); }

    return s_pInstance;
}


void ManProtocolStateClear::execute( ManProtocol* p )
{
    p->increaseClearCount();

    if ( p->isCleared() ) {
        p->sleep();
        p->changeState( ManProtocolStateIdle::instance() );
}
}

/*
///--------------------------------------------------------------------------------------------ManProtocolStateTimeout
ManProtocolStateTimeout* ManProtocolStateTimeout::s_pInstance = NULL;

ManProtocolStateTimeout* ManProtocolStateTimeout::instance()
{
    Logger::logf( LOG_MSG, _T("ManProtocolStateTimeout::instance()") );

    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateTimeout(); }

    return s_pInstance;
}

void ManProtocolStateTimeout::execute( ManProtocol* p )
{
    Logger::logf( LOG_MSG, _T("ManProtocolStateTimeout::execute()") );
    p->changeState( ManProtocolStateIdle::instance() );
}
*/

///---------------------------------------------------------------------------------------ManProtocolModeDownload States

///-----------------------------------------------------------------------------------------ManProtocolStateDownloadIdle
ManProtocolStateDownloadIdle* ManProtocolStateDownloadIdle::s_pInstance = NULL;

ManProtocolStateDownloadIdle* ManProtocolStateDownloadIdle::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateDownloadIdle(); }

    return s_pInstance;
}


void ManProtocolStateDownloadIdle::execute( ManProtocol* p )
{
    if ( p->needDisable() ) { p->setDisabled(); }

    if ( p->isDisabled() ) { return; }

    if ( !p->hasNextCommand() ) { return; }

    p->sendCommand();

    p->changeState( ManProtocolStateDownloadWait::instance() );
}

///-------------------------------------------------------------------------------------------ManProtocolStateDownloadWait

ManProtocolStateDownloadWait* ManProtocolStateDownloadWait::s_pInstance = NULL;

ManProtocolStateDownloadWait* ManProtocolStateDownloadWait::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateDownloadWait(); }

    return s_pInstance;
}

void ManProtocolStateDownloadWait::execute( ManProtocol* p )
{
    //Logger::logf( LOG_MSG, _T("ManProtocolStateDownloadWait::execute()") );
    if ( p->isDataReceived() )
    {
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( !p->isEOL() ) { return; }

        //Logger::logf( LOG_MSG, _T("ManProtocolStateDownloadWait::execute() - EOL") );
        //p->printConfigLine();

        if ( p->isEmptyLine() )
        {
            p->increaseEmptyLineCount();

            if ( p->hasTooMuchEmptyLines() )
            {
                Logger::logf( LOG_MSG, _T("ManProtocolStateDownloadWait::execute() - Too much empty lines!") );
                p->sendEvent( SIPR_CONFIG_ERROR );
            }
        }
        else
        {
            p->sendConfigLine();
            p->increaseLineCount();
        }

        p->resetRxProperties();
        // p->changeState( ManProtocolStateDownloadIdle::instance() );

    }
    else
    {
        p->increaseTimeoutCount();

        if ( !p->isTimedOut() ) { return; }

        if( p->isNotEnoughtLines() )
        {
            // really timeout
            Logger::logf(   LOG_MSG,
                            _T("ManProtocolStateDownloadWait::execute() - Not enought lines!") );
            p->sendEvent( SIPR_CONFIG_ERROR );

        }
        else
        {
            //TEST
            wxString msg;

            if ( p->isCurrentCommandSameAs( VemsCommand::DOWNLOAD_CONFIG ) )
            {
                msg = _T("Download config");
                p->sendEvent( SIPR_CONFIG_END );
            }
            else if ( p->isCurrentCommandSameAs( VemsCommand::DOWNLOAD_TABLES ) )
            {
                msg = _T("Download table");
                p->sendEvent( SIPR_TABLES_END );
            }

            Logger::logf(   LOG_MSG,
                            _T("ManProtocolStateDownloadWait::execute() - ") + msg +
                            _T(" - line count : %d"), p->getLineCount() );

            // response finished
            // ...
            //p->sendEvent( SIPR_CONFIG_END );
        }

        p->resetDownloadProperties();
        p->resetRxProperties();
        p->changeState( ManProtocolStateDownloadIdle::instance() );

    }

/*
        if ( p->isExpectedDataReceived() )
        {
            p->notify();
            p->changeState( ManProtocolStateClear::instance() );
        }
    }
    else
    {
        p->increaseTimeoutCount();

        if ( p->isTimedOut() ) { p->changeState( ManProtocolStateTimeout::instance() ); }
    }
    */
}

///--------------------------------------------------------------------------------------ManProtocolModeSaveSD States
///-----------------------------SAVE_SD_NEED_RT_SIZE, SAVE_SD_WAIT_RT_SIZE, SAVE_SD_NEED_DATA, SAVE_SD_SAVE_WAIT_DATA

///----------------------------------------------------------------------------------ManProtocolStateSaveSDNeedRTSize
ManProtocolStateSaveSDNeedRTSize* ManProtocolStateSaveSDNeedRTSize::s_pInstance = NULL;

ManProtocolStateSaveSDNeedRTSize* ManProtocolStateSaveSDNeedRTSize::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDNeedRTSize(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDNeedRTSize::execute( ManProtocol* p )
{
    //Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDNeedRTSize::execute() - START") );
    if ( !p->hasNextCommand() ) { return; }

    p->sendCommand();
    //Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDNeedRTSize::execute() - SEND COMMAND") );
    p->resetRxProperties();

    if ( p->isWaitForResponse() )
    {
        p->changeState( ManProtocolStateSaveSDWaitRTSize::instance() );
    }
}


///----------------------------------------------------------------------------------ManProtocolStateSaveSDWaitRTSize
ManProtocolStateSaveSDWaitRTSize* ManProtocolStateSaveSDWaitRTSize::s_pInstance = NULL;

ManProtocolStateSaveSDWaitRTSize* ManProtocolStateSaveSDWaitRTSize::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDWaitRTSize(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDWaitRTSize::execute( ManProtocol* p )
{
    if ( p->isDataReceived() )
    {
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( !p->isExpectedDataReceived() ) { return; }

        if ( !p->isCRCValid() )
        {
            /// TODO
            Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitRTSize::execute() - CRC ERROR!!!") );
            p->resetSDVariables();
            p->sendEvent( SIPR_SD_CRC_ERROR );
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            return;
        }

        if ( p->isSDCardEmpty() )
        {
            Logger::logf(   LOG_MSG, _T("ManProtocolStateSaveSDWaitRTSize::execute() - SDCard is empty!!!") );
            if ( p->isCheckCardEmpty() ) { p->resetCheckCardEmpty(); }
            p->resetSDVariables();
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            p->sendEvent( SIPR_SD_EMPTY );
            return;
        }

        //Logger::logf(   LOG_MSG, _T("ManProtocolStateSaveSDWaitRTSize::execute() - check sd empty: %d"), p->isCheckCardEmpty() );

        if ( p->isCheckCardEmpty() )
        {
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            p->resetSDVariables();
            p->resetCheckCardEmpty();
            p->sendEvent( SIPR_SD_NOT_EMPTY );
            return;
        }

            // p->sendStartSaveLogEvent();

            //  p->setRTSize();
        //Logger::logf(   LOG_MSG, _T("ManProtocolStateSaveSDWaitRTSize::execute() - Block received\n") + p->toString() );

        //p->changeState( ManProtocolStateSaveSDNeedData::instance() );
        p->changeState( ManProtocolStateSaveSDClearAfterRTSize::instance() );
            //p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            //p->notify();
            //p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );

    }
    else
    {
        p->increaseTimeoutCount();

        if ( !p->isTimedOut() ) { return; }

        Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitRTSize::execute() - TIMEOUT\n"));

        p->resetSDVariables();
        p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
    }
}

///----------------------------------------------------------------------------------ManProtocolStateSaveSDClearAfterRTSize
ManProtocolStateSaveSDClearAfterRTSize* ManProtocolStateSaveSDClearAfterRTSize::s_pInstance = NULL;

ManProtocolStateSaveSDClearAfterRTSize* ManProtocolStateSaveSDClearAfterRTSize::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDClearAfterRTSize(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDClearAfterRTSize::execute( ManProtocol* p )
{
    if ( p->isClearedAfterRTSize() )
    {
        if ( p->isSearchStartIndex() ) { p->setSearchIndex(); }
        p->changeState( ManProtocolStateSaveSDNeedData::instance() );
    }

    /*
    if ( size ) { Logger::logf( LOG_MSG, _T("MegatuneProtocol::rx() - CLEAR - %d"), m_iClearCount ); }

    if ( ++m_iClearCount >= m_pCurrentCommand->getClearCount() )
            {
                m_state = IDLE;
            }
        */
}

///----------------------------------------------------------------------------------ManProtocolStateSaveSDNeedData
ManProtocolStateSaveSDNeedData* ManProtocolStateSaveSDNeedData::s_pInstance = NULL;

ManProtocolStateSaveSDNeedData* ManProtocolStateSaveSDNeedData::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDNeedData(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDNeedData::execute( ManProtocol* p )
{
    p->getNextBlocks();

    if ( !p->hasNextCommand() ) { return; }

    p->sendCommand();

    p->resetRxProperties();

    if ( p->isWaitForResponse() )
    {
        p->changeState( ManProtocolStateSaveSDWaitData::instance() );
    }
}


///----------------------------------------------------------------------------------ManProtocolStateSaveSDWaitData
ManProtocolStateSaveSDWaitData* ManProtocolStateSaveSDWaitData::s_pInstance = NULL;

ManProtocolStateSaveSDWaitData* ManProtocolStateSaveSDWaitData::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDWaitData(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDWaitData::execute( ManProtocol* p )
{
    if ( p->isDataReceived() )
    {
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( !p->isExpectedDataReceived() ) { return; }

        if ( !p->isCRCValid() )
        {
            /// TODO
            Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitData::execute() - CRC ERROR!!!") );
            p->resetSDVariables();
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            p->sendEvent( SIPR_SD_CRC_ERROR );
            return;
        }

        p->verifyBlocks();

        // blocks are invalid
        if ( !p->isValid() && !p->isSearchStartIndex() )
        {
            Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitData::execute() - NOT ALL BLOCKS ARE VALID!!!") );
            p->sendValidBlocks();
            p->resetSDVariables();
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            return;
        }

        // blocks are valid
        if ( !p->isSearchStartIndex() )
        {
            p->sendValidBlocks();
            p->increaseBlockIndex();
        }
        else { p->setSearchIndex(); }


        p->changeState( ManProtocolStateSaveSDNeedData::instance() );
    }
    else
    {
        p->increaseTimeoutCount();

        if ( !p->isTimedOut() ) { return; }

        Logger::logf( LOG_MSG, _T("ManProtocolStateSDWaitData::execute() - TIMEOUT") );
        p->resetSDVariables();
        p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
        p->sendEvent( SIPR_SD_TIMEOUT );
    }
}

///----------------------------------------------------------------------------------MPSSearchLastWaitData
MPSSearchLastWaitData* MPSSearchLastWaitData::s_pInstance = NULL;

MPSSearchLastWaitData* MPSSearchLastWaitData::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  MPSSearchLastWaitData(); }

    return s_pInstance;
}

void MPSSearchLastWaitData::execute( ManProtocol* p )
{
    if ( p->isDataReceived() )
    {
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( !p->isExpectedDataReceived() ) { return; }
/*
        if ( !p->isCRCValid() )
        {
            /// TODO
            Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitData::execute() - CRC ERROR!!!") );
            p->resetSDVariables();
            p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
            p->sendEvent( SIPR_SD_CRC_ERROR );
            return;
        }
*/
        p->verifyBlocks();

        //if ( !p->isValid() )

        if ( p->isValid() )
        {
            p->changeState( ManProtocolStateSaveSDNeedData::instance() );
        }


/*
        if ( !p->isValid() )
        {
            if ( !p->isSearchStartIndex() )
            {

                p->sendValidBlocks();
                /// TODO
                Logger::logf( LOG_MSG, _T("ManProtocolStateSaveSDWaitData::execute() - NOT ALL BLOCKS ARE VALID!!!") );
                p->resetSDVariables();
                p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );

                //p->sendEvent( SIPR_SD_LOG_STOP );
                // clear sd variables
                return;
            }

            // increaseBlockIndex();
            // setSearchDown();
        }

        // valid
        if ( !p->isSearchStartIndex() )
        {
            p->sendValidBlocks();
            //if ( p->isSearchStartIndex() ) { p->setBlockIndexForSearch(); }
            p->increaseBlockIndex();
        }

        p->changeState( ManProtocolStateSaveSDNeedData::instance() ); */
    }
    else
    {
        p->increaseTimeoutCount();

        if ( !p->isTimedOut() ) { return; }

        Logger::logf( LOG_MSG, _T("ManProtocolStateSDWaitData::execute() - TIMEOUT") );
        p->resetSDVariables();
        p->changeState( ManProtocolStateSaveSDNeedRTSize::instance() );
        p->sendEvent( SIPR_SD_TIMEOUT );
    }


}

///----------------------------------------------------------------------------------ManProtocolStateSaveSDClearAfterData
ManProtocolStateSaveSDClearAfterData* ManProtocolStateSaveSDClearAfterData::s_pInstance = NULL;

ManProtocolStateSaveSDClearAfterData* ManProtocolStateSaveSDClearAfterData::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDClearAfterData(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDClearAfterData::execute( ManProtocol* p )
{

}

/*
///----------------------------------------------------------------------------------ManProtocolStateSDWait
ManProtocolStateSDWait* ManProtocolStateSDWait::s_pInstance = NULL;

ManProtocolStateSDWait* ManProtocolStateSDWait::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSDWait(); }

    return s_pInstance;
}

void ManProtocolStateSDWait::execute( ManProtocol* p )
{
    if ( p->isDataReceived() )
    {
        Logger::logf( LOG_MSG, _T("ManProtocolStateSDWait::execute() - DATA RECEIVED") );
        p->clearTimeoutCount();

        p->addReceivedData();

        if ( p->isExpectedDataReceived() )
        {
            //if ( p->isCRCValid() )

            Logger::logf(   LOG_MSG, _T("ManProtocolStateSDWait::execute() - Block received , crc: %i"), p->isCRCValid() );

            //p->notify();
            p->changeState( ManProtocolStateSDIdle::instance() );
        }
    }
    else
    {
        p->increaseTimeoutCount();

        if ( p->isTimedOut() )
        {
            Logger::logf( LOG_MSG, _T("ManProtocolStateSDWait::execute() - TIMEOUT") );
            //p->changeState( ManProtocolStateTimeout::instance() );
            p->changeState( ManProtocolStateSDIdle::instance() );
        }
    }
}


//enum SyncState  { NEED_PAGE_SIZE, GET_PAGE_SIZE, NEED_TEST_PAGE_SIZE, TEST_PAGE_SIZE, HAVE_PAGE_SIZE };
//  1. NEED_RT_SIZE
        1.1 send READ_A_BLOCK - > WAIT_RT_SIZE
            1.1.1   EXPECTED DATA RECEIVED
                        CRC OK
                            -> NEED_DATA
                        CRC BAD
                    EXPECTED DATA NOT RECEIVED

    2. NEED_DATA
        send READ_N_BLOCK - WAIT_DATA
            EXPECTED DATA RECEIVED
                CRC OK
                VALID
                    NEXT LOG -> send event
                    NEED DATA


*/
/*

///--------------------------------------------------------------------------ManState - IDLE

ManProtocolStateIdle* ManProtocolStateIdle::s_pInstance = NULL;

ManProtocolStateIdle* ManProtocolStateIdle::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateIdle(); }

    return s_pInstance;
}


void ManProtocolStateIdle::execute( ManProtocol* p )
{
    if ( !p->hasNextCommand() ) { return; }

    p->sendCommand();

    p->resetRxProperties();

    if ( p->isWaitForResponse() )
    {
        p->changeState( ManProtocolStateWait::instance() );
    }
}





///--------------------------------------------------------------------------ManState - SD_IDLE
ManProtocolStateSDIdle* ManProtocolStateSDIdle::s_pInstance = NULL;

ManProtocolStateSDIdle* ManProtocolStateSDIdle::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSDIdle(); }

    return s_pInstance;
}


void ManProtocolStateSDIdle::execute( ManProtocol* p )
{
    // SaveSD - GET_RT / WAIT_RT
    //
    //request ( )

    1.  m_pCurrentCommand = "msg" // get rt data length
    2.  m_pCurrentCommand = hexIdx, len // get rt data length

        //changeState( SAVE_SD_LOG ) wait data size

    3.  m_pCurrentCommand = "msG" , until  END
         changeState( SAVE_SD_LOG_RX ) wait log size




 WAIT_RT
    if ( received )
        1.  checkCRC()
        2. if ( crcOK )

           else
            getDataSize()
            if ( dataSize <= 0 || dataSize ==255) // CARD_EMPTY
            else DATA_SIZE_OK changeState( SAVE_SD_LOG_TX )


           else { COMM_ERROR; }

    else { timeout - COMM_ERROR  }

}

///--------------------------------------------------------------------------ManState - SAVE_SD_WAIT
ManProtocolStateSaveSDWait* ManProtocolStateSaveSDWait::s_pInstance = NULL;

ManProtocolStateSaveSDWait* ManProtocolStateSaveSDWait::instance()
{
    if ( !s_pInstance ) { s_pInstance = new  ManProtocolStateSaveSDWait(); }

    return s_pInstance;
}

void ManProtocolStateSaveSDWait::execute( ManProtocol* p )
{
}
*/

///-----------------------------------------------------------------------------------------------------ManCommandFactory
int ManCommandFactory::s_timeout = 0;

ManCommand* ManCommandFactory::create( VemsCommand* command )
{
    switch ( command->getType() )
    {
		case VemsCommand::DOWNLOAD_TABLES:
		    return new ManCommand( VemsCommand::DOWNLOAD_TABLES, _T("mct"), 0, 10, 0 );
		case VemsCommand::DOWNLOAD_CONFIG:
            return new ManCommand( VemsCommand::DOWNLOAD_CONFIG, _T("mcd"), 0, 10, 0 );
        case VemsCommand::SD_LOG_START :
            return new ManCommand( VemsCommand::SD_LOG_START, _T("mse") );
        case VemsCommand::SD_LOG_STOP :
            return new ManCommand( VemsCommand::SD_LOG_STOP, _T("msd") );
        case VemsCommand::SD_LOG_ERASE :
            return new ManCommand( VemsCommand::SD_LOG_ERASE, _T("msc") );
        case VemsCommand::SD_LOG_GET_STATUS:
            // expected size?
            return new ManCommand( VemsCommand::SD_LOG_GET_STATUS, _T("mss") );
        case VemsCommand::PING:
            return new ManCommand( VemsCommand::PING, _T("mxc"), 23, s_timeout, 2 );
        case VemsCommand::GET_COMPILE_TIME:
            return new ManCommand( VemsCommand::GET_COMPILE_TIME, _T("mdv"), 16, s_timeout, 2 );
        case VemsCommand::GET_FIRMWARE_VERSION:
            return new ManCommand( VemsCommand::GET_FIRMWARE_VERSION, _T("mdV"), 8, s_timeout, 2 );
        case VemsCommand::GET_HWSIG:
            return new ManCommand( VemsCommand::GET_HWSIG, _T("mci"), 12, s_timeout, 2 );
        case VemsCommand::SET_STATE:
            switch ( command->getState() )
            {
                case ECUInfo::MEGATUNE:
                    //Logger::logf( LOG_MSG, _T("ManCommandFactory::create() - VemsCommand::SET_STATE - ECUInfo::MEGATUNE") );
                    return new ManCommand( VemsCommand::SET_STATE, _T("mmbye"), 5, s_timeout, 10 );
                default:
                    return NULL;
            }
        case VemsCommand::SD_LOG_SAVE_A_BLOCK :
        {
		    int startIndex = command->getStartIndex();
		    startIndex &= 0xFFFFFF;
		    wxString strCmd = wxString::Format( _T("msg%06X"), startIndex );
		    Logger::logf( LOG_MSG, _T("ManCommandFactory::create() - SAVE SD : ") + strCmd );
		    return new ManCommand( VemsCommand::SD_LOG_SAVE_A_BLOCK, strCmd, ManProtocol::s_iSD_BLOCK_SIZE, 200, 20 );
        } break;
        case VemsCommand::SD_LOG_SAVE_N_BLOCKS :
        {
		    int startIndex = command->getStartIndex();
		    startIndex &= 0xFFFFFF;
		    wxString strCmd = wxString::Format( _T("msG%06X"), startIndex );
		    Logger::logf( LOG_MSG, _T("ManCommandFactory::create() - SAVE SD : ") + strCmd );
		    int expectedSize = ManProtocol::s_iSD_BLOCK_SIZE * ManProtocol::s_iSD_BLOCK_COUNT;
		    return new ManCommand(  VemsCommand::SD_LOG_SAVE_N_BLOCKS, strCmd, expectedSize, 200, 20 );
        } break;
        case VemsCommand::SET_BAUDRATE:
        {
            int baudCommand = 0;

            switch ( command->getBaudRate() )
            {
                case 115200:
                    baudCommand = 8;
                    break;
                default:
                    baudCommand = 1000000 / command->getBaudRate() - 1;
            }

            wxString strCmd = wxString::Format( _T("mcB%02X"), baudCommand );
            Logger::logf( LOG_MSG, _T("ManCommandFactory::create() - SET_BAUDRATE : ") + strCmd );

            return new ManCommand(  VemsCommand::SET_BAUDRATE, strCmd, 4, 100, 0 );
        } break;
        case VemsCommand::WBO2:
        {
            bool isEnable = command->getWBO2();
            wxString strCmd;
            if(isEnable){
                strCmd = _T("mde02");
            }else{
                strCmd = _T("mde00");
            }
             Logger::logf( LOG_MSG, _T("ManCommandFactory::create()(VemsCommand::WBO2)-sendcmd: ") + strCmd );
            return new ManCommand(  VemsCommand::WBO2, strCmd, 0, 0, 0 );
        }
            break;

        case VemsCommand::SAVE_CONFIG:
        {
            SaveConfigCmd* cmd = dynamic_cast<SaveConfigCmd*>(command);
            if ( cmd ) {
                wxString strCmd ;

                if ( cmd->configType == 0 ) {
                    strCmd = _T("mcS0");
                } else if ( cmd->configType == 1 ) {
                    strCmd = _T("mcS1");
                } else {
                    LOGP( wxString::Format( _T("Invalid config type! - %d"), cmd->configType ) );
                    return NULL;
    }

                return new ManCommand(  VemsCommand::SAVE_CONFIG, strCmd, 0, 0, 100, 10000 );
            }
        } break;

        case VemsCommand::LOAD_CONFIG:
        {
            LoadConfigCmd* cmd = dynamic_cast<LoadConfigCmd*>(command);
            if ( cmd ) {
                wxString strCmd ;

                if ( cmd->configType == 0 ) {
                    strCmd = _T("mcL0");
                } else if ( cmd->configType == 1 ) {
                    strCmd = _T("mcL1");
                } else {
                    LOGP( wxString::Format( _T("Invalid config type! - %d"), cmd->configType ) );
    return NULL;
}

                return new ManCommand(  VemsCommand::LOAD_CONFIG, strCmd, 0, 0, 100, 10000 );
            }
        } break;

    }

    return NULL;
}

///-------------------------------------------------------------------------------------------ManProtocol
ManProtocol::ManProtocol( wxEvtHandler* eventHandler, int timeout )
{
    m_pEventHandler     = eventHandler;
    m_pCurrentMode      = new ManProtocolModeNormal();
    m_pCurrentCommand   = NULL;

    // rx variables
    resetRxProperties();
    m_pRxBuffer         = NULL;
    m_pTxBuffer         = NULL;
    m_pTxBufferSize     = NULL;
    m_iRxBufferSize     = 0;
    //m_rxBuffer[ s_iRX_BUFFER_SIZE ] fill 0 ?

    // download variables
    resetDownloadProperties();

    // sd variables
    m_bIsNextLog            = false;
    m_bCheckCardEmpty       = false;
    m_iBlockIndex           = 0;
    m_iBlockCount           = 1;
    m_iCRCErrorCount        = 0;
    m_iLastSecCountInBlock  = -255;
    m_iInvalidBlockIndex    = -1;
    m_iRTDataSize           = 0;
    m_iNextLogIndex         = -1;

    // sd search variables for SD Card save log last T minute
    m_bSearchStartIndex     = false;
    m_bSearchStartIndexUp   = true;
    m_iSearchCount          = 0;
    m_iSearchMask           = 1;
    m_iBackTime             = 0;

    // disable variables
    m_bNeedDisable          = false;
    m_bIsDisabled           = false;

    m_iTimeout                      = timeout;
    ManCommandFactory::s_timeout    = timeout;
}

ManProtocol::~ManProtocol() {}

int ManProtocol::rx( uint8_t* rxBuffer, int size )
{
    // set rx buffer
    m_pRxBuffer     = rxBuffer;
    m_iRxBufferSize = size;

    if ( m_pCurrentMode->isInRxState() ) { m_pCurrentMode->execute( this ); }

    //if ( size ) { Logger::logf( LOG_MSG, _T("ManProtocol::rx() - %02X"), m_pRxBuffer[0] ); }

    return 0;
}

int ManProtocol::tx( uint8_t* txBuffer, int& size )
{
    if ( m_pCurrentMode->isIdle() && m_txQueue.IsEmpty() )
    {
        processNextCommand();
    }

    // set tx buffer
    m_pTxBuffer     = txBuffer;
    m_pTxBufferSize = &size;

    if ( m_pCurrentMode->isInTxState() ) { m_pCurrentMode->execute( this ); }

    return 0;
}

void ManProtocol::request( VemsCommand* command )
{
    m_RequestQueue.push_back( command );
}

void ManProtocol::processNextCommand()
{
    // Logger::logf( LOG_MSG, _T("ManProtocol::processNextCommand() - BEGIN") );

    if ( m_RequestQueue.IsEmpty() ) { return; }

    VemsCommand* command = m_RequestQueue.front();
    m_RequestQueue.pop_front();

    if ( !command ) { return; }

    changeMode( command );

    switch( command->getType() )
    {
        case VemsCommand::DOWNLOAD_ALL :
        {
            pushCommand( new DownloadConfigCommand() );
            pushCommand( new DownloadTablesCommand() );
        } break;
        case VemsCommand::SD_LOG_CHECK_EMPTY:
        {
            m_bCheckCardEmpty = true;
            m_iBlockCount = 1;
            pushCommand( new SDLogSaveABlockCommand( 0 ) );
        } break;
       case VemsCommand::SD_LOG_SAVE :
       {
            if( command->getBackTime() > 0 )
            {
                initSearchMode( command->getBackTime() );
            }

            m_iBlockCount = 1;
            pushCommand( new SDLogSaveABlockCommand( 0 ) );
        } break;
    }

    pushCommand( command );
}


void ManProtocol::changeMode( VemsCommand* command )
{
    switch( command->getType() )
    {
        // mormal mode commands
        case VemsCommand::SD_LOG_START :
        case VemsCommand::SD_LOG_STOP :
        case VemsCommand::SD_LOG_ERASE :
        case VemsCommand::SD_LOG_GET_STATUS:
        case VemsCommand::SET_BAUDRATE:
        case VemsCommand::PING :
        case VemsCommand::GET_FIRMWARE_VERSION :
        case VemsCommand::GET_COMPILE_TIME:
        case VemsCommand::GET_HWSIG:
        case VemsCommand::WBO2:
        case VemsCommand::SET_STATE:
        case VemsCommand::SAVE_CONFIG:
        case VemsCommand::LOAD_CONFIG:
            if ( m_pCurrentMode->getType() != ManProtocolModeNormal::NORMAL )
            {
                changeMode( new ManProtocolModeNormal() );
            } break;
        // download mode commands
        case VemsCommand::DOWNLOAD_ALL :
        case VemsCommand::DOWNLOAD_TABLES :
        case VemsCommand::DOWNLOAD_CONFIG :
            if ( m_pCurrentMode->getType() != ManProtocolMode::DOWNLOAD )
            {
                changeMode( new ManProtocolModeDownload() );
            } break;
        // save sd mode commads
        case VemsCommand::SD_LOG_SAVE:
        case VemsCommand::SD_LOG_CHECK_EMPTY:
            if ( m_pCurrentMode->getType() != ManProtocolMode::SAVE_SD )
            {
                changeMode( new ManProtocolModeSaveSD() );
            } break;
    }
}

void ManProtocol::pushCommand( VemsCommand* command )
{
    ManCommand* manCmd = ManCommandFactory::create( command );
    if ( manCmd ) { m_txQueue.push_back( manCmd ); }
    delete command;
}

bool ManProtocol::isInState( int stateType )
{
    return m_pCurrentMode->isInState( stateType );
}

void ManProtocol::changeState( ManProtocolState* state )
{
    m_pCurrentMode->changeState( state );
}

void ManProtocol::changeMode( ManProtocolMode* mode )
{
    m_pCurrentMode = mode;
}

bool ManProtocol::isCleared()
{
    if ( m_iClearCount >= m_pCurrentCommand->getClearCount() )
    {
        Logger::logf( LOG_MSG, _T("ManProtocolStateClear::execute() - CLEARED : %d"), m_iClearCount );
    }

    return ( m_iClearCount >= m_pCurrentCommand->getClearCount() );
}

bool ManProtocol::isDataReceived()
{
    return m_iRxBufferSize!=0?true:false;
}

void ManProtocol::clearTimeoutCount()
{
    m_iTimeoutCount = 0;
}

void ManProtocol::addReceivedData()
{
    // assert !m_pRxBuffer
    // need to document always recived 0 or 1 byte
    if ( m_iRxDataLength >= s_iRX_BUFFER_SIZE )
    {
        Logger::logf( LOG_MSG, _T("ManProtocol::addReceivedData() - BUFFER OVERFLOW %d >= %d"),
                        m_iRxDataLength, s_iRX_BUFFER_SIZE );
    }

    m_rxBuffer[ m_iRxDataLength++ ] = m_pRxBuffer[0];

    printf( "ManProtocol::addReceivedData() - %d \r", m_iRxDataLength );
}

bool ManProtocol::isExpectedDataReceived()
{
    // assert  if ( !m_pCurrentCommand ) { return true; }
    if ( m_pCurrentCommand->getExpectedRxSize() == m_iRxDataLength )
    {
        //Logger::logf(   LOG_MSG, _T("ManProtocol::isExpectedDataReceived() - size: %d\n") + toString(),
                        //m_iRxDataLength );

    }

    return ( m_pCurrentCommand->getExpectedRxSize() == m_iRxDataLength );
}


bool ManProtocol::isTimedOut()
{
    // assert  if ( !m_pCurrentCommand ) { return true; }
    return ( m_iTimeoutCount >= m_pCurrentCommand->getTimeoutCount() );
}

void ManProtocol::sendCommand()
{
    // if m_pCurrentCommand->isMacro()
        // if getNextCommand
    if ( m_pCurrentCommand ) { delete m_pCurrentCommand; }

    // get command from txQueue
    m_pCurrentCommand = m_txQueue.front();
    m_txQueue.pop_front();

    if ( !m_pCurrentCommand ) return;

//    Logger::logf( LOG_MSG, _T("ManProtocol::sendCommand() - m_pCurrentCommand: ") +
//        m_pCurrentCommand->getType() );
    //assert if ( !m_pCurrentCommand ) { return 0; }

    // if is macro command
    // set
    //

    //
    int size = m_pCurrentCommand->getSize();
    const uint8_t*  data = m_pCurrentCommand->getData();

    memcpy( m_pTxBuffer, data, size );

    *m_pTxBufferSize = size;
}

void ManProtocol::resetRxProperties()
{
    // Logger::logf( LOG_MSG, _T("ManProtocol::resetRxProperties()") );
    m_iTimeoutCount = 0;
    m_iClearCount   = 0;
    m_iRxDataLength = 0;
}

bool ManProtocol::isWaitForResponse()
{
    // assert  if ( !m_pCurrentCommand ) { return false; }
    return m_pCurrentCommand->isWaitForResponse();
}

bool ManProtocol::isNeedClear()
{
    // assert  if ( !m_pCurrentCommand ) { return false; }
    return m_pCurrentCommand->isNeedClear()!=0?true:false;
}


bool ManProtocol::isEOL() const
{
    if ( m_iRxDataLength > 1 )
    {
        if (    m_rxBuffer[ m_iRxDataLength - 2 ] == '\r' &&
                m_rxBuffer[ m_iRxDataLength - 1 ] == '\n' ) { return true; }
    }

    return false;
}

//
bool ManProtocol::isEmptyLine()
{
    return m_iRxDataLength == 2;
}

void ManProtocol::resetDownloadProperties()
{
    m_iLineCount        = 0;
    m_iEmptyLineCount   = 0;
}

//

void ManProtocol::printConfigLine()
{
    if ( m_iRxDataLength > 1 )
    {
        m_rxBuffer[ m_iRxDataLength - 2 ] = '\0';
        wxString line( (char*)m_rxBuffer, wxConvUTF8 );
        Logger::logf( LOG_MSG, _T("ManProtocol::printConfigLine() - ") + line );
    }
}

void ManProtocol::sendConfigLine()
{
    //Logger::logf( LOG_MSG, _T("ManProtocol::sendConfigLine()") );
    if ( m_iRxDataLength <= 1 || m_iRxDataLength > 254 )
    {
        // send message
        return;
    }

    m_rxBuffer[ m_iRxDataLength - 2 ] = '\0';
    SIPRPacket p( 0, SIPR_CONFIG_LINE, m_iRxDataLength - 1, m_rxBuffer );
    sendEvent( p );
}

void ManProtocol::increaseTimeoutCount()
{
    //Logger::logf( LOG_MSG, _T("ManProtocol::increaseTimeoutCount() - count: %d"), ++m_iTimeoutCount );
    ++m_iTimeoutCount;
}

/*
bool ManProtocol::checkCRC()
{

}
*/

bool ManProtocol::isCRCValid( uint8_t* data, int len)
{

    uint16_t crccalc=getCRC( data, len-2 );
    uint16_t crcrecv=data[len-2] << 8; // BIG endian
    crcrecv += data[len-1];

    if( crccalc == crcrecv){
        return true;
    }

    Logger::logf(LOG_MSG,_T("ManProtocol::isCRCValid() - invalid, calc=%04x, recv=%04x"), crccalc, crcrecv);
    return false;
}

uint16_t ManProtocol::getCRC(uint8_t* data, int len)
{
    uint16_t crc = 0;
    int i;
    for(i = 0; i < len; i++){
        crc = Math::crc16(crc, data[i]);
    }
    return crc;
}

bool ManProtocol::isSDCardEmpty()
{
    m_iRTDataSize = m_rxBuffer[ s_iSD_RT_SIZE_BLOCK_INDEX ];

    Logger::logf( LOG_MSG, _T("ManProtocol::isSDCardEmpty() - m_iRTDataSize = %d"), m_iRTDataSize );

    if ( m_iRTDataSize <= 0 || m_iRTDataSize == 255 ) { return true; }

    return false;
}

bool ManProtocol::isClearedAfterRTSize()
{
    return m_iClearCount++ >= m_pCurrentCommand->getClearCount();
    /*
    if ( m_iClearCount >= m_pCurrentCommand->getClearCount() )
    {
        Logger::logf( LOG_MSG, _T("ManProtocolStateClear::execute() - CLEARED : %d"), m_iClearCount );
    }

    return ( m_iClearCount >= m_pCurrentCommand->getClearCount() );
    */
}
/*
bool ManProtocol::getRTDataSize()
{
    m_rxBuffer[ ]

    uint8_t data[512+2+2];
    if( getMMCBlock(0, data, 514) ){
        bool iscrcOK = isCRCValid(data, 514);
        if(!iscrcOK) return -1; //COMM_ERROR

        int dataSize = data[510];
        if(dataSize <= 0 || dataSize == 255) return -2; //CARD EMPTY
        return dataSize;
    }
    return -1; //COMM_ERROR

}
*/

bool ManProtocol::isCRCValid()
{
    /*
    if ( ( m_iRxDataLength > 0 ) && ( m_iRxDataLength < s_iRX_BUFFER_SIZE ) )
    {
        return isCRCValid( m_rxBuffer, m_iRxDataLength );
    }

    return false;
    */

    bool isCRC_OK = false;

    m_iCRCErrorCount = 0;

    for ( int i = 0; i < m_iBlockCount; ++i )
    {
        // need assert or verify
        isCRC_OK = isCRCValid( &m_rxBuffer[ i * s_iSD_BLOCK_SIZE ], s_iSD_BLOCK_SIZE );
        //isCRC_OK = isCRCValid(&data[i*blockSize], blockSize);
        if ( !isCRC_OK )
        {
            Logger::logf(LOG_MSG,_T("ManProtocol::isCRCValid() - CRC error !!! blockIdx: %i"), i);
            ++m_iCRCErrorCount;
            continue;
        }
    }

    return m_iCRCErrorCount < s_iMAX_CRC_ERROR_COUNT;
}


/*
        /// CRC for all block
        for(i = 0; i< blockCount; i++){
            iscrcOK = isCRCValid(&data[i*blockSize], blockSize);
            if( !iscrcOK){
             Logger::logf(LOG_MSG,_T("MMCLogger save() CRC16 failed! idx: %i"), i);
             crcErrorCount++;
             continue; // Even for a full 0xFF block we require valid CRC.
            }else{
                crcErrorCount = 0;
            }
        }
*/
void ManProtocol::getNextBlocks()
{
    if ( isSearchStartIndex() )
    {
        //request( new SDLogSaveABlockCommand( m_iBlockIndex ) );
        pushCommand( new SDLogSaveABlockCommand( m_iBlockIndex ) );
    }
    else
    {
        m_iBlockCount = s_iSD_BLOCK_COUNT;
        //request( new SDLogSaveNBlocksCommand( m_iBlockIndex ) );
        pushCommand( new SDLogSaveNBlocksCommand( m_iBlockIndex ) );
    }
}




bool ManProtocol::isNextLog(uint8_t * data)
{
    int secOffset = 0;
    int firstSecCountInBlock = data[secOffset];

    if(firstSecCountInBlock == m_iLastSecCountInBlock){
        m_iLastSecCountInBlock = firstSecCountInBlock;
        return false;
    }
    if( (firstSecCountInBlock - m_iLastSecCountInBlock) == 1 ){
        m_iLastSecCountInBlock = firstSecCountInBlock;
        return false;
    }
    if( (firstSecCountInBlock == 0) && (m_iLastSecCountInBlock == 255) ){
        m_iLastSecCountInBlock = firstSecCountInBlock;
        return false;
    }
    m_iLastSecCountInBlock = firstSecCountInBlock;

    Logger::logf(LOG_MSG,_T("MMCLogger isNextLog() blockIdx = TRUE"));

    return true;
}


bool ManProtocol::isDataValid( uint8_t* data, int len )
{
    int data_or=0;
    int data_and=0xff;
    int i;
    for(i = 0; i < len; i++){
    data_or|=data[i];
    data_and&=data[i];
    }
    if ((data_or==0) || (data_and==0xff)) return false;
    else return true;

}

/*
bool ManProtocol::isValid()
{

}

sendEvents
{

}



if ( isValid )

else
*/
/// return all read blocks are valid or not
/// set m_iInvalidBlockIndex, m_iNextLogIndex
/// need rename
void ManProtocol::verifyBlocks()
{
    bool    bIsValid    = false;
    int     iBlockIndex = 0;

    m_iInvalidBlockIndex    = -1;
    m_iNextLogIndex         = -1;

    for ( int i = 0; i < m_iBlockCount; ++i )
    {
        iBlockIndex = i * s_iSD_BLOCK_SIZE;
        bIsValid = isDataValid( &m_rxBuffer[ iBlockIndex ], s_iSD_BLOCK_SIZE - 18 );

        if ( bIsValid )
        {
            if ( isNextLog( &m_rxBuffer[ iBlockIndex ] ) ) { m_iNextLogIndex = iBlockIndex; }
        }
        else
        {
            m_iInvalidBlockIndex = iBlockIndex;
            break;
        }
    }
}

bool ManProtocol::isValid() { return m_iInvalidBlockIndex == -1; }

void ManProtocol::sendValidBlocks()
{
    int iBlockIndex = 0;

    for ( int i = 0; i < m_iBlockCount; ++i )
    {
        iBlockIndex = i * s_iSD_BLOCK_SIZE;

        if ( iBlockIndex == m_iInvalidBlockIndex )
        {
            sendEvent( SIPR_SD_LOG_STOP );
            return;
        }

        if ( iBlockIndex == m_iNextLogIndex ) { sendEvent( SIPR_SD_LOG_START ); }

        notify( iBlockIndex );
    }

    sendEvent( SIPR_SD_LOG_NEXT_BLOCK );
}


/*
///TODO - rename
bool ManProtocol::isValid()
{
    bool isValid        = false;
    bool foundInvalid   = false;
    m_bIsNextLog        = false;

    for ( int i = 0; i < m_iBlockCount; ++i )
    {
        // need assert or verify
        isValid = isDataValid( &m_rxBuffer[ i * s_iSD_BLOCK_SIZE ], s_iSD_BLOCK_SIZE - 18 );
        //isCRC_OK = isCRCValid(&data[i*blockSize], blockSize);
        if ( isValid )
        {
            // Logger::logf(LOG_MSG,_T("ManProtocol::isCRCValid() - CRC error !!! blockIdx: %i"), i);
            // ++m_iCRCErrorCount;
            // continue;
            if ( isNextLog( &m_rxBuffer[ i * s_iSD_BLOCK_SIZE ] ) )
            {
                m_bIsNextLog = true;
                Logger::logf(LOG_MSG,_T("ManProtocol::isValid()- NextLog!!! blockIdx: %i"), i);
                //send start log event
                SIPRPacket p( 0, SIPR_SD_LOG_START );
                sendEvent( p );
            }


            notify( i * s_iSD_BLOCK_SIZE );
        }
        else
        {
            if ( !foundInvalid )
            {
                foundInvalid = true;
                m_iInvalidBlockIndex = i;
                Logger::logf( LOG_MSG, _T("ManProtocol::isValid() - FALSE, blockIdx: %i"), i );
            }
        }


        //m_iCRCErrorCount = 0;
    }

    if ( !isValid ) { sendEvent( SIPR_SD_LOG_STOP ); }
    else
    {
        SIPRPacket p( 0, SIPR_SD_LOG_NEXT_BLOCK );
        sendEvent( p );
    }

    //m_iBlockIndex += s_iSD_BLOCK_COUNT;

    return isValid;
}
*/

void ManProtocol::increaseBlockIndex()
{
    m_iBlockIndex += s_iSD_BLOCK_COUNT;
}

void ManProtocol::resetSDVariables()
{
    m_bIsNextLog            = false;
    m_bCheckCardEmpty       = false;
    m_iBlockIndex           = 0;
    m_iBlockCount           = 1;
    m_iCRCErrorCount        = 0;
    m_iLastSecCountInBlock  = -255;
    m_iInvalidBlockIndex    = -1;
    m_iRTDataSize           = 0;
}

bool ManProtocol::isNextLog()
{
    return m_bIsNextLog;
}

void ManProtocol::notify( int blockIndex )
{
    //Logger::logf( LOG_MSG, _T("ManProtocol::notify() - blockIndex: %d"), blockIndex );

    uint8_t m_buffer[ 512 ];
    m_buffer[0] = 0;
    m_buffer[1] = 0;
    m_buffer[2] = SiprSensorType::SIPR_RT_MT_NORMAL;

    for ( int i = 0; i < s_iSD_BLOCK_SIZE - 2 - m_iRTDataSize; i += m_iRTDataSize )
    {
        //Logger::logf( LOG_MSG, _T("ManProtocol::notify() - i: %d, tps: %u"), i, m_rxBuffer[ blockIndex + i + 8 ] );

        memcpy( &m_buffer[3], &m_rxBuffer[ blockIndex + i ], m_iRTDataSize );

        SIPRPacket p( 0, SIPR_READ_SENSORS, m_iRTDataSize + 3, m_buffer );

        sendEvent( p );
    }

    /*
    if ( !m_pEventHandler ) { return; }

    uint8_t m_buffer[ 512 ];
    m_buffer[0] = 0;
    m_buffer[1] = 0;
    m_buffer[2] = SiprSensorType::MEGATUNE;

    for ( int i = 0; i < s_iSD_BLOCK_SIZE; i += m_iRTDataSize )
    {
        memcpy( &m_buffer[3], &m_rxBuffer[ blockIndex + i ], m_iRTDataSize );

        SIPRPacket p( 0, SIPR_READ_SENSORS, m_iRTDataSize + 3, m_buffer );

        sendEvent( p );
    }
    */
}



void ManProtocol::sendEvent( const SIPRPacket& packet )
{
    InputEvent e;
    e.setPacket( packet );
    if ( m_pEventHandler ) { m_pEventHandler->AddPendingEvent( e ); }
}

void ManProtocol::sendEvent( int type )
{
    SIPRPacket p( 0, type );
    sendEvent( p );
}

/*
void ManProtocol::sendSdCardNotEmptyEvent()
{
    SIPRPacket p( 0, SIPR_SD_NOT_EMPTY );
    sendEvent( p );
}

void ManProtocol::sendStartSaveLogEvent()
{
    SIPRPacket p( 0, SIPR_SD_LOG_START );
    sendEvent( p );
}

void ManProtocol::sendStopSaveLogEvent()
{
    SIPRPacket p( 0, SIPR_SD_LOG_STOP );
    sendEvent( p );
}

void ManProtocol::sendSdCardEmptyEvent()
{
    SIPRPacket p( 0, SIPR_SD_EMPTY );
    sendEvent( p );
}
*/

bool ManProtocol::isCurrentCommandSameAs( int vemsCommandType )
{
    if ( !m_pCurrentCommand ) { return false; }

    return m_pCurrentCommand->getType() == vemsCommandType;
}

bool ManProtocol::isCheckCardEmpty()
{
    return m_bCheckCardEmpty;
}

void ManProtocol::resetCheckCardEmpty()
{
    m_bCheckCardEmpty = false;
}

void ManProtocol::notify()
{
    switch( m_pCurrentCommand->getType() )
    {
        case VemsCommand::PING:
        {
            wxString s = wxString::From8BitData( (const char *)m_rxBuffer, m_iRxDataLength );

            Logger::logf( LOG_MSG, _T("-----ManProtocol::notify() - VemsCommand::PING - len: %d\n"), m_iRxDataLength );
            Logger::logf( LOG_MSG, _T("-----ManProtocol::notify() - VemsCommand::PING ->") + s + _T("<-\n") );

            SIPRPacket p( 0, SIPR_ACK );
            sendEvent( p );
        } break;
        case VemsCommand::SET_BAUDRATE:
        {
            Logger::logf( LOG_MSG, _T("ManProtocol::notify() - send VemsCommand::SET_BAUDRATE - BEGIN") );
            SIPRPacket p( 0, SIPR_SET_BAUD );
            sendEvent( p );
        } break;
        case VemsCommand::GET_COMPILE_TIME :
        {
            SIPRPacket p( 0, SIPR_GET_COMPILE_TIME, m_iRxDataLength, m_rxBuffer );
            sendEvent( p );
        } break;
        case VemsCommand::GET_FIRMWARE_VERSION :
        {
            //const wxCharBuffer buffer = str.To8BitData();
            wxString s = wxString::From8BitData( (const char*)m_rxBuffer, m_iRxDataLength );
            FirmwareVersion ver;

            FirmwareInfo::extractFirmwareVersion( s, ver );

            const uint8_t   size = 3;
            uint8_t         data[size] = { ver.major, ver.minor, ver.build };
            //
            SIPRPacket p( 0, SIPR_GET_VERSION, size, data ); // createFromString( type, string ); | createFrom8BitData( uint8_t* data, int length )
            sendEvent( p );
        } break;
        case VemsCommand::GET_HWSIG:
        {
            SIPRPacket p( 0, SIPR_GET_HWSIG, m_iRxDataLength, m_rxBuffer );
            sendEvent( p );
        } break;
    }
}

void ManProtocol::notifyTimeout()
{
//    Logger::logf( LOG_MSG, _T("ManProtocol::notifyTimeout() - VemsCommand::TIMEOUT") );
    SIPRPacket p( 0, SIPR_TIMEOUT );
    sendEvent( p );
}

void ManProtocol::disable()
{
//    Logger::logf( LOG_MSG, _T("-----ManProtocol::disable() - %d"), m_bIsDisabled );
    if ( !m_bIsDisabled ) { m_bNeedDisable = true; }
}

void ManProtocol::enable()
{
    m_bIsDisabled = false;
}

bool ManProtocol::isDisabled()
{
    return m_bIsDisabled;
}

void ManProtocol::setDisabled()
{
    m_bIsDisabled   = true;
    m_bNeedDisable  = false;
}

void ManProtocol::sleep()
{
    if ( m_pCurrentCommand->getSleepAfterWrite() > 0 ) {
        LOGP( wxString::Format( _T("m_pCurrentCommand->getSleepAfterWrite(): %d"), m_pCurrentCommand->getSleepAfterWrite() ) );
        wxThread::Sleep( m_pCurrentCommand->getSleepAfterWrite() );
    }
}

void ManProtocol::initSearchMode( int iBackTime )
{
    m_bSearchStartIndex     = true;
    m_bSearchStartIndexUp   = true;
    m_iSearchCount          = 0;
    m_iSearchMask           = 1;
    m_iBackTime             = iBackTime;
    // m_iInvalidBlockIndex    = -1;
    // m_iBlockIndex           = 0;
}

void ManProtocol::setSearchIndex()
{
    Logger::logf( LOG_MSG, _T("ManProtocol::setSearchIndex() - BEGIN\n")
                    _T("\t\tm_bSearchStartIndexUp: %d\n")
                    _T("\t\tisValid(): %d\n" )
                    _T("\t\tm_iBlockIndex: %d\n" )
                    _T("\t\tm_iSearchMask: %d\n" )
                    _T("\t\tm_iSearchCount: %d\n" ),
                    m_bSearchStartIndexUp,
                    isValid(),
                    m_iBlockIndex,
                    m_iSearchMask,
                    m_iSearchCount );



    if ( m_bSearchStartIndexUp )
    {
        if ( isValid() )
        {
            m_iBlockIndex |= m_iSearchMask;
            m_iSearchMask <<= 1;
            ++m_iSearchCount;
        }
        else
        {
            ++m_iBlockIndex;
            m_bSearchStartIndexUp = false;
        }
    }
    else
    {
        if ( !isValid() ) { m_iBlockIndex &= ~m_iSearchMask; }
        m_iSearchMask >>= 1;
        m_iBlockIndex |= m_iSearchMask;
        --m_iSearchCount;

        // finish search
        if ( m_iSearchCount < 0 )
        {
            int iBackBlockIndex = (int)(m_iBackTime * 60 * m_iRTDataSize / 51.2);
            if ( iBackBlockIndex < m_iBlockIndex ) { m_iBlockIndex -= iBackBlockIndex; }
            else{ m_iBlockIndex = 0; }

            m_bSearchStartIndex = false;
        }
    }
}

// TEST
wxString ManProtocol::toString()
{
    wxString s;

    // header
    s.Append( _T("ManProtocol::toString()\n") );

    // state variables
    wxString sState;

    sState = wxString::Format( _T("\tState variables\n")
                            _T("\t\tm_pCurrentMode: %p \n")
                            _T("\t\tm_txQueue.Size: %d\n")
                            _T("\t\tm_pCurrentCommand: %p\n")
                            _T("\t\tm_pEventHandler addr: %p\n"),
                            m_pCurrentMode,
                            m_txQueue.size(),
                            m_pCurrentCommand,
                            m_pEventHandler );
    s.Append( sState );

    // rx variables
    wxString sRx;

    sRx = wxString::Format( _T("\tRx variables\n")
                            _T("\t\tm_iClearCount: %d\n")
                            _T("\t\tm_iTimeoutCount: %d\n")
                            _T("\t\tm_iRxDataLength: %d\n")
                            _T("\t\tm_iRxBufferSize: %d\n"),
                            m_iClearCount,
                            m_iTimeoutCount,
                            m_iRxDataLength,
                            m_iRxBufferSize );
    s.Append( sRx );

    // sd variables
    wxString sSD;

    sSD = wxString::Format( _T("\tSD variables\n")
                            _T("\t\tm_bIsNextLog: %d\n")
                            _T("\t\tm_bCheckCardEmpty: %d\n")
                            _T("\t\tm_iBlockIndex: %d\n")
                            _T("\t\tm_iBlockCount: %d\n")
                            _T("\t\tm_iCRCErrorCount: %d\n")
                            _T("\t\tm_iLastSecCountInBlock: %d\n")
                            _T("\t\tm_iInvalidBlockIndex: %d\n")
                            _T("\t\tm_iRTDataSize: %d\n"),
                            m_bIsNextLog,
                            m_bCheckCardEmpty,
                            m_iBlockIndex,
                            m_iBlockCount,
                            m_iCRCErrorCount,
                            m_iLastSecCountInBlock,
                            m_iInvalidBlockIndex,
                            m_iRTDataSize );
    s.Append( sSD );

    /*
        ManProtocolMode*    m_pCurrentMode;
        ManCommandList      m_txQueue;
        ManCommand*         m_pCurrentCommand;
        wxEvtHandler*       m_pEventHandler;



        int                 m_iClearCount;
        int                 m_iTimeoutCount;
        int                 m_iRxDataLength;
        uint8_t*            m_pRxBuffer;
        uint8_t*            m_pTxBuffer;
        int*                m_pTxBufferSize;
        int                 m_iRxBufferSize;
        uint8_t             m_rxBuffer[ s_iRX_BUFFER_SIZE ];

        // download variables
        int                 m_iLineCount;
        int                 m_iEmptyLineCount;

        // under test
        bool                m_bNeedDisable;
        bool                m_bIsDisabled;
   */

   return s;
}

/*
    if command backTime > 0
        get rt size
        search end
            send msg Index

        wait data
            isValid
                ...
                send msg Index

            isNotValid

                send msg Index


*/

/*
uint32_t MMCLogger::getStartIdx(uint16_t back_time)//min
{

    uint8_t tmp[32];
    m_commBase->ReadString(tmp, 16);

    Logger::logf(LOG_MSG,_T("MMCLogger getStartIdx time=%i"), back_time);

    int dataSize;
    int errorCount = 0;

    do{
        dataSize = getDataSize();
        Logger::logf(LOG_MSG,_T("MMCLogger getstartidx0 RTdataRecord Size=%i"), dataSize);
        if(dataSize < 0) { errorCount++; }
    }while( (dataSize < 0) &&  errorCount < 4);

    if (dataSize == -2) return 0;

    int8_t count=0;
    uint32_t blockIdx=0x0;
    uint32_t mask=0x01;
    do{
        Logger::logf(LOG_MSG,_T("MMCLogger getstartidx%d()up blockidx=%i maszk=%d"), count,blockIdx,mask);
        wxString message = wxString::Format(_T("Searcing: %d"), blockIdx );
        observerNotify(message);

        blockIdx |=mask;
        mask <<=1;
        count++;
    }while( getBlockState(blockIdx) != -2 );
    blockIdx++;
    do{
        Logger::logf(LOG_MSG,_T("MMCLogger getstartidx%d()down blockidx=%i maszk=%d"), count,blockIdx,mask);
        wxString message = wxString::Format(_T("Searching: %d"), blockIdx );
        observerNotify(message);

        if ( getBlockState(blockIdx) == -2 ) blockIdx &= ~mask;
        mask >>= 1;
        blockIdx |= mask;
        count--;
    }while(count>=0);

    uint32_t back_block=(uint16_t)(back_time*60*dataSize/51.2);
    back_block < blockIdx ? blockIdx = blockIdx-back_block : blockIdx = 0;
    Logger::logf(LOG_MSG,_T("MMCLogger getstartidx final blockidx=%i ,back_block:%d"),blockIdx, back_block);

    return blockIdx;


}

int MMCLogger::getBlockState(uint32_t blockIdx)
{
   uint8_t data[512+2+2];

    if( getMMCBlock(blockIdx, data, 514) ){
        bool iscrcOK = isCRCValid(data, 514);
        if(!iscrcOK) return -1; //COMM_ERROR

        int dataSize = data[510];
        if(dataSize == 0 || dataSize == 255) return -2; //CARD EMPTY
        return dataSize;
    }
    return -2; //COMM_timed_out

}*/


void ManProtocol::reset()
{
    resetRxProperties();
}
