#include "PasswordHelper.h"
#include "../DirHelper.h"
#include "../logger.h"
#include "../Math.h"

#include <wx/textfile.h>
#include <wx/regex.h>
#include <wx/tokenzr.h>

int PasswordHelper::checkPassword( const wxString& serialNumber, wxString* crcHex  )
{
    wxTextFile passwordsFile( DirHelper::GetPath(dCONFIG) + _T("password.txt") );

    if ( !passwordsFile.Open() ) {
        return PASSWORD_FILE_NOT_FOUND;
    }

    wxString line = wxEmptyString;

    for ( line = passwordsFile.GetFirstLine(); !passwordsFile.Eof();  line = passwordsFile.GetNextLine() )
    {
        /// ignore comments
        if ( line.StartsWith(_T(";") ) || line.StartsWith(_T("#") ) ) {
            continue;
        }

        wxStringTokenizer tokenizer( line, _T("=") );

        if ( tokenizer.CountTokens() != 2 ) {
            continue;
        }

        wxString serialToken = tokenizer.GetNextToken();
        serialToken.Trim().Trim(false);

        if ( !serialToken.IsSameAs(serialNumber) ) {
            continue;
        }

        wxString password = tokenizer.GetNextToken();
        password.Trim().Trim(false);

        passwordsFile.Close();

        if ( isPasswordAccepted( password ) ) {
            if ( crcHex ) {
                uint32_t magic = Math::hex2int( password );
                uint16_t crc = resolveMagic( magic );
                *crcHex = wxString::Format( _T("%04X"), crc );
            }
            return PASSWORD_ACCEPTED;
        } else {
            return INVALID_PASSWORD;
        }

    }

    passwordsFile.Close();
    return PASSWORD_NOT_FOUND;
}

bool PasswordHelper::isPasswordFormatAccepted( const wxString& password )
{
    wxRegEx regExp( _T("(^[0-9a-fA-F]{8}$)"), wxRE_ADVANCED );
    return regExp.Matches( password );
}

bool PasswordHelper::isPasswordAccepted( const wxString& password )
{
    uint32_t magic = Math::hex2int( password );

    return checkMagic( magic );
}

bool PasswordHelper::checkMagic( uint32_t data )
{
    int i;

    for ( i = 0; i < 16; ++i )
    {
        uint8_t evenIdx = 2 * i;
        uint8_t oddIdx = 31 - evenIdx;

        uint8_t evenBit = ( data & ( 1 << evenIdx ) ) >> evenIdx;
        uint8_t oddBit = ( data & ( 1 << ( oddIdx ) ) ) >> ( oddIdx );

        if ( evenBit != oddBit ) {
            return false;
        }
    }

    return true;
}

uint16_t PasswordHelper::resolveMagic( uint32_t data )
{
    int i = 0;

    uint16_t result = 0;

    for ( i = 0; i < 16; ++i )
    {
        uint8_t evenBit = ( data & ( 1 << ( 2*i ) ) ) >> ( 2*i );
        result |= ( evenBit << i );
    }

    return result;
}
